/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/ansi/AnsiFormat.java#1 $
 *
 * Copyright (c) 2011 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.ansi;

import java.util.Arrays;

import com.google.common.base.Joiner;
import com.google.common.base.Predicates;
import com.google.common.collect.Iterables;
import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

/**
 * An ANSI text format encapsulating the following attributes:
 * <ul>
 * <li>Foreground {@link AnsiColor}
 * <li>Background {@link AnsiColor}
 * <li>{@link AnsiStyle}
 * </ul>
 *
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2011/07/06 $
 */
public class AnsiFormat {

    private static final char ESC = '\u001b';

    private final AnsiColor foreground;
    private final AnsiColor background;
    private final AnsiStyle style;

    private final String prefix;
    private final String suffix;

    public AnsiFormat(AnsiColor foreground, AnsiColor background, AnsiStyle style) {
        this.foreground = foreground;
        this.background = background;
        this.style = style;

        String attrs = Joiner.on(';').join(
            Iterables.filter(
                Arrays.asList(new Integer[] {
                    this.foreground == null ? null : this.foreground.foregroundCode(),
                    this.background == null ? null : this.background.backgroundCode(),
                    this.style == null ? null : this.style.code()
                }),
                Predicates.not(Predicates.isNull())
            )
        );

        prefix = ESC + "[" + attrs + "m";
        suffix = ESC + "[m";
    }

    /**
     * Apply this <tt>AnsiFormat</tt> to the given text.
     * @param text Plain text to format
     * @return The text bracketed within ANSI escape sequences that apply this {@link AnsiFormat}.
     */
    public String apply(String text) {
        return prefix + text + suffix;
    }

    public AnsiColor getForeground() {
        return this.foreground;
    }

    public AnsiColor getBackground() {
        return this.background;
    }

    public AnsiStyle getStyle() {
        return this.style;
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder()
            .append(this.foreground)
            .append(this.background)
            .append(this.style)
            .toHashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof AnsiFormat)) {
            return false;
        }

        AnsiFormat that = (AnsiFormat) obj;
        return new EqualsBuilder()
            .append(this.foreground, that.foreground)
            .append(this.background, that.background)
            .append(this.style, that.style)
            .isEquals();
    }

//    public static void main(String[] args) {
//        System.out.println(new AnsiFormat(AnsiColor.CYAN, null, null).apply("foo"));
//        System.out.println(new AnsiFormat(AnsiColor.RED, AnsiColor.WHITE, AnsiStyle.BOLD).apply("bar"));
//        System.out.println(new AnsiFormat(null, null, null).apply("baz"));
//    }
}
