/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/concurrent/AdjustableSemaphore.java#1 $
 *
 * Copyright (c) 2010 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.concurrent;

import java.util.concurrent.Semaphore;

import net.jcip.annotations.GuardedBy;

/**
 * A simple semaphore implementation that allows the
 * maximum number of permits to be adjusted.
 * <p/>
 * {@link Semaphore} does not allow the maximum number
 * of permits to be adjusted after construction, so
 * we needed to create AdjustableSemaphore.
 * <p/>
 * If you don't need to adjust the maximum number of permits
 * after construction, you should probably just use {@link Semaphore}.
 */
public final class AdjustableSemaphore extends Semaphore {

    /**
     * The current maximum number of permits.
     */
    @GuardedBy("this")
    private int maxPermits;

    public AdjustableSemaphore(int initialMaxPermits, boolean fair) {
        super(initialMaxPermits, fair);
        setMaxPermitsInternal(initialMaxPermits);
    }

    public AdjustableSemaphore(int initialMaxPermits) {
        super(initialMaxPermits);
        setMaxPermitsInternal(initialMaxPermits);
    }

    public synchronized int getMaxPermits() {
        return this.maxPermits;
    }

    /**
     * Sets the new maximum number of permits to be controlled
     * by this semaphore.
     * <p/>
     * If the new maximum is less than the current maximum,
     * then any threads currently waiting on a permit will
     * continue to wait until enough permits are released
     * to meet the new maximum.
     *
     * @param newMax new maximum number of permits (must be > 0)
     * @throws IllegalArgumentException if newMax <= 0
     */
    public void setMaxPermits(int newMax) {
        if (newMax <= 0) {
            throw new IllegalArgumentException(String.format("maxPermits must be > 0, %d is not a valid value", newMax));
        }

        /*
         * Must synchronize r/w access to this.maxPermits
         */
        synchronized (this) {
            int oldMax = getMaxPermits();

            if (newMax == oldMax) {
                /*
                 * Nothing changed.
                 *
                 * No need to adjust.
                 */
                return;
            } else if (newMax > oldMax) {
                /*
                 * Maximum value has increased.
                 *
                 * Calling release with the delta will allow more acquires to proceed.
                 */
                release(newMax - oldMax);
            } else {
                /*
                 * Maximum value has decreased.
                 *
                 * Calling reducePermits with the (positive) delta will allow less acquires to proceed.
                 *
                 * Any threads currently waiting in acquire will continue to wait until
                 * enough releases have been performed to be below the newMax
                 */
                reducePermits(oldMax - newMax);
            }
            setMaxPermitsInternal(newMax);
        }

    }

    private synchronized void setMaxPermitsInternal(int newMax) {
        this.maxPermits = newMax;
    }
}
