/*
 * $Id: //depot/prod/newport_news/main/modules/nn-platform/src/main/java/com/netapp/common/execution/PasswordMaskingShellCommand.java#3 $
 *
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.execution;

import java.util.Iterator;
import java.util.List;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

/**
 * A ShellCommand that masks passwords in the user facing and log facing commands.
 *
 * This provides a limited amount of security because of the following problems:
 * <ul>
 * <li>If a password is empty or null, then no mask is displayed (allowing an astute user to deduce that the password is empty)</li>
 * <li>If other text identical to the password appears anywhere else in the command, then that text is masked as well.</li>
 * <li>A user can still monitor the process listing to find the password when the shell command is executed.</li>
 * </ul>
 *
 * @author claird
 * @version $Revision: #3 $ $Date: 2011/06/24 $
 */
public class PasswordMaskingShellCommand extends ShellCommand {

    public static final String DEFAULT_PASSWORD_MASK = "^^^^^";

    private final String password;

    private String mask = DEFAULT_PASSWORD_MASK;

    /**
     * @param password password to be masked
     * @param command command arguments with unmasked password
     */
    public PasswordMaskingShellCommand(String password, List<String> command) {
        super(command);
        this.password = password;
    }

    /**
     * @param password password to be masked
     * @param command command arguments with unmasked password
     */
    public PasswordMaskingShellCommand(String password, String... command) {
        super(command);
        this.password = password;
    }

    @Override
    public String getLogFacingCommand() {
        if (this.password == null || this.password.length() == 0) {
            return super.getLogFacingCommand();
        }

        StringBuilder logFacingCommand = new StringBuilder();
        for (Iterator<String> argumentListIter = getArgumentList().iterator(); argumentListIter.hasNext();) {
            String argument = argumentListIter.next();

            String maskedArgument = argument == null ? null : argument.replace(this.password, getMask());

            logFacingCommand.append(maskedArgument);

            if (argumentListIter.hasNext()) {
                logFacingCommand.append(" ");
            }

        }
        return logFacingCommand.toString();
    }


    @Override
    public ShellRunnable transformInto(String... argumentList) {
        return new PasswordMaskingShellCommand(this.password, argumentList);
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder()
            .appendSuper(super.hashCode())
            .append(this.password)
            .toHashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj instanceof PasswordMaskingShellCommand) {
            if (this == obj) {
                return true;
            }

            PasswordMaskingShellCommand rhs = (PasswordMaskingShellCommand) obj;
            return new EqualsBuilder()
                .appendSuper(super.equals(obj))
                .append(this.password, rhs.password)
                .isEquals();
        } else {
            return false;
        }
    }
    public void setMask(String mask) {
        this.mask = mask;
    }

    public String getMask() {
        return this.mask;
    }

}
