/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/graph/GraphvizEncoder.java#4 $
 *
 * Copyright (c) 2011 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.graph;

import java.io.PrintWriter;
import java.io.Writer;
import java.util.Map;

import com.google.common.collect.Maps;

import com.netapp.common.graph.Graph.Edge;

/**
 * Encode the {@link DisplayGraph} in <a href="http://www.graphviz.org/">Graphviz DOT</a> syntax.
 *
 * @author sirianni
 * @version $Revision: #4 $ $Date: 2012/04/26 $
 */
public class GraphvizEncoder {

    /**
     * Direction of the graph layout.
     * See <a href="http://www.graphviz.org/content/attrs#drankdir">Graphviz documentation</a>.
     */
    public enum RankDir {
        TOP_TO_BOTTOM("TB"),
        LEFT_TO_RIGHT("LR"),
        BOTTOM_TO_TOP("BT"),
        RIGHT_TO_LEFT("RL");

        private String keyword;

        RankDir(String keyword) {
            this.keyword = keyword;
        }
    }

    private final PrintWriter out;

    public GraphvizEncoder(Writer writer) {
        this.out = new PrintWriter(writer);
    }

    public void encode(DisplayGraph graph) {
        this.encode(graph, RankDir.TOP_TO_BOTTOM);
    }

    public void encode(DisplayGraph graph, RankDir rankDir) {
        String openBrace = rankDir == RankDir.TOP_TO_BOTTOM || rankDir == RankDir.BOTTOM_TO_TOP ? "{" : "";
        String closeBrace = rankDir == RankDir.TOP_TO_BOTTOM || rankDir == RankDir.BOTTOM_TO_TOP ? "}" : "";

        out.println("digraph G {");
        out.println();

        out.println("compound=true");
        out.println("rankdir=" + rankDir.keyword);
        out.println("bgcolor=white; fontcolor=black; fontname=Helvetica; fontsize=10.0");
        out.printf("label=\"%s\"%n", graph.getName());
        out.println();

        out.println("edge [color=black, fontcolor=black, fontname=Helvetica, fontsize=10.0]");
        out.println();

        out.println("node [color=black, fontcolor=black, fontname=\"Helvetica-Bold\", fontsize=10.0, shape=record, style=\"solid,filled\"]");
        out.println();

        Map<DisplayNode, String> nodeNameMap = Maps.newHashMap();
        int nodeId = 0;
        for (DisplayNode node : graph.getGraph().getNodes()) {
            String nodeName = "node_" + nodeId++;
            nodeNameMap.put(node, nodeName);

            out.printf("%s [%n", nodeName);

            String contentString = getContentString(node);

            out.printf("  label=\"%s%s%s\"%n",
                openBrace,
                contentString,
                closeBrace);
            out.printf("  fillcolor=%s%n", node.getColor());
            out.println("]");
            out.println();
        }

        for (Edge<DisplayNode> edge : graph.getGraph().getEdges()) {
            String sourceNodeName = nodeNameMap.get(edge.getSource());
            String destNodeName = nodeNameMap.get(edge.getDest());

            out.printf("%s -> %s%n", sourceNodeName, destNodeName);
        }

        out.println("}");

        out.flush();
    }

    /**
     * Escape special characters in DOT labels
     * The characters escaped are ' ', '<', and '>'
     * @param input string to escape characters in
     * @return escaped string
     */
    private String escapeSpecialCharacters(String input) {
        return input.replaceAll(" ", "\\ ").replaceAll("<", "\\\\<").replaceAll(">", "\\\\>");
    }

    private String getContentString(DisplayNode node) {
        StringBuilder builder = new StringBuilder();

        if (node.getTopContent() != null) {
            builder.append(escapeSpecialCharacters(node.getTopContent()));
        }

        if (node.getMiddleContent() != null) {
            if (node.getTopContent() != null) {
                builder.append("\\n|");
            }
            builder.append(escapeSpecialCharacters(node.getMiddleContent()));
        }

        if (node.getBottomContent() != null) {
            if (node.getTopContent() != null || node.getMiddleContent() != null) {
                builder.append("\\l|");
            }

            builder.append(escapeSpecialCharacters(node.getBottomContent()));
            builder.append("\\n");
        }
        return builder.toString();
    }

}
