/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/logging/NALogger.java#8 $
 *
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.logging;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.netapp.common.util.ListenerProxy;

/**
 * Log messages through the logging framework.
 * @author sunday
 * @version $Revision: #8 $ $Date: 2011/11/14 $
 */

@edu.umd.cs.findbugs.annotations.SuppressWarnings("LI")
public class NALogger {
    private static final Map<String, NALogger> loggers = new ConcurrentHashMap<String, NALogger>();

    /*
     * Lazy initialization holder class idiom for static fields ("Effective Java" p. 283)
     *
     * Initializing the listeners is somewhat expensive for CLIs (around 80ms) mainly due to proxy initialization.
     * So we avoid the initialization until a listener is added.
     * And since listeners are not very common for CLIs, we can
     * avoid the initialization altogether.
     */
    private static class ListenersHolder {
        private static final ListenerProxy<LogListener> LISTENERS = new ListenerProxy<LogListener>(LogListener.class);
    }

    private static LogAdapterFactory factory;
    static {
        // TODO: Add ability to specify a different factory
        // via a system property or file in the classpath
        try {
            factory = new SLF4JAdapterFactory();
        } catch (Throwable t) { }

        if (factory == null) {
            factory = new JdkLoggerAdapterFactory();
        }
    }

    private LogAdapter adapter;

    /** Name of this logger. */
    private String name;

    private NALogger(String name) {
        this.name = name;
        this.adapter = factory.createLogAdapter(name);
    }

    /**
     * Get a logger for the supplied class
     * @param clazz
     * @return logger for the supplied class
     */
    public static NALogger getLogger(Class<?> clazz) {
        return getLogger(clazz.getName());
    }

    public static NALogger getLogger(String name) {
        NALogger logger = loggers.get(name);
        if (logger == null) {
            logger = new NALogger(name);
            loggers.put(name, logger);
        }
        return logger;
    }

    /**
     * Get a logger for the calling context automatically.
     * @return Logger associated to the calling class
     */
    public static NALogger getLogger() {
        Throwable t = new Throwable();
        String caller = t.getStackTrace()[1].getClassName();
        return getLogger(caller);
    }

    /**
     * Add a new listener to be notified when a message is logged.
     * Listeners are weakly referenced and will not be called if the listener
     * is garbage collected.
     * @param listener
     */
    public static void addListener(LogListener listener) {
        ListenersHolder.LISTENERS.addListener(listener);
    }

    protected static void notifyListeners(LogMessage message) {
        ListenersHolder.LISTENERS.proxy().append(message);
    }

    public String debug(String msg) {
        return internal(msg);
    }

    public String debug(String msg, Object... params) {
        return internal(msg, params);
    }

    public String debug(String msg, Throwable t, Object... params) {
        return internal(msg, t, params);
    }

    public String internal(String msg) {
        adapter.log(LogLevel.DEBUG, msg);
        return msg;
    }

    public String internal(String msg, Object... params) {
        String formattedMsg = String.format(msg, params);
        adapter.log(LogLevel.DEBUG, formattedMsg);
        return formattedMsg;
    }

    public String internal(String msg, Throwable t, Object... params) {
        String formattedMsg = String.format(msg, params);
        adapter.log(LogLevel.DEBUG, formattedMsg, t);
        return formattedMsg;
    }

    public String internalTrace(String msg) {
        adapter.log(LogLevel.TRACE, msg);
        return msg;
    }

    public String internalTrace(String msg, Object... params) {
        String formattedMsg = String.format(msg, params);
        adapter.log(LogLevel.TRACE, formattedMsg);
        return formattedMsg;
    }

    public String internalTrace(String msg, Throwable t, Object... params) {
        String formattedMsg = String.format(msg, params);
        adapter.log(LogLevel.TRACE, formattedMsg, t);
        return formattedMsg;
    }

    String log(LogMessage msg) {
        adapter.log(msg);
        notifyListeners(msg);
        return msg.localizedMessage();
    }

    String log(LogMessage msg, Throwable t) {
        adapter.log(msg, t);
        notifyListeners(msg);
        return msg.localizedMessage();
    }

    /*
     * standard logging methods
     */

    public String log(LogLevel level, MessageKey key) {
        LogMessage msg = new LogMessage(level, this.name, key, (Object[])null);
        return log(msg);
    }

    public String log(LogLevel level, MessageKey key, Throwable t) {
        LogMessage msg = new LogMessage(level, this.name, key, (Object[])null);
        return log(msg, t);
    }

    public String log(LogLevel level, MessageKey key, Object... params) {
        LogMessage msg = new LogMessage(level, this.name, key, params);
        return log(msg);
    }

    public String log(LogLevel level, MessageKey key, Throwable t, Object... params) {
        LogMessage msg = new LogMessage(level, this.name, key, params);
        return log(msg, t);
    }

    public String trace(MessageKey key) {
        return log(LogLevel.TRACE, key);
    }

    public String trace(MessageKey key, Throwable t) {
        return log(LogLevel.TRACE, key, t);
    }

    public String trace(MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.TRACE, key, t, params);
    }

    public String trace(MessageKey key, Object... params) {
        return log(LogLevel.TRACE, key, params);
    }

    public String debug(MessageKey key) {
        return log(LogLevel.DEBUG, key);
    }

    public String debug(MessageKey key, Throwable t) {
        return log(LogLevel.DEBUG, key, t);
    }

    public String debug(MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.DEBUG, key, t, params);
    }

    public String debug(MessageKey key, Object... params) {
        return log(LogLevel.DEBUG, key, params);
    }

    public String info(MessageKey key) {
        return log(LogLevel.INFO, key);
    }

    public String info(MessageKey key, Throwable t) {
        return log(LogLevel.INFO, key, t);
    }

    public String info(MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.INFO, key, t, params);
    }

    public String info(MessageKey key, Object... params) {
        return log(LogLevel.INFO, key, params);
    }

    public String warn(MessageKey key) {
        return log(LogLevel.WARN, key);
    }

    public String warn(MessageKey key, Throwable t) {
        return log(LogLevel.WARN, key, t);
    }

    public String warn(MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.WARN, key, t, params);
    }

    public String warn(MessageKey key, Object... params) {
        return log(LogLevel.WARN, key, params);
    }

    public String error(MessageKey key) {
        return log(LogLevel.ERROR, key);
    }

    public String error(MessageKey key, Throwable t) {
        return log(LogLevel.ERROR, key, t);
    }

    public String error(MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.ERROR, key, t, params);
    }

    public String error(MessageKey key, Object... params) {
        return log(LogLevel.ERROR, key, params);
    }

    public String log(String level, MessageKey key) {
        return log(LogLevel.parseLevel(level), key);
    }

    public String log(String level, MessageKey key, Throwable t) {
        return log(LogLevel.parseLevel(level), key, t);
    }

    public String log(String level, MessageKey key, Object... params) {
        return log(LogLevel.parseLevel(level), key, params);
    }

    public String log(String level, MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.parseLevel(level), key, t, params);
    }

    /***
     * Check if logger is enabled for specified log level.
     * @param level
     * @return True if this Logger is enabled for the LogLevel level,
     * false otherwise.
     */
    public boolean isLogLevelEnabled(LogLevel level) {
        return this.adapter.isLogLevelEnabled(level);
    }
}
