/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/logging/NDC.java#4 $
 *
 * Copyright (c) 2010 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.common.logging;

import static com.netapp.common.ansi.AnsiColor.*;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import com.google.common.base.Function;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.Collections2;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;

import com.netapp.common.ansi.AnsiColor;
import com.netapp.common.ansi.AnsiFormat;


/**
 * Implementation of the <b>Nested Diagnostic Context</b>.  It provides a nested (stack based) context for logging purposes.
 * This implementation does not propagate to spawned threads.
 * Note that this implementation is dependent on synchronous logging (i.e. the logging must happen in the same thread
 * the NDC is set in).  If this requirement becomes a problem we can re-implement this class on top of the
 * {@link org.slf4j.MDC}, albeit in a less efficient way.
 *
 * @author thomaswb
 * @version $Revision: #4 $
 *          Date: $Date: 2013/08/14 $
 */
public class NDC {

    private static final ThreadLocal<Deque<NDCElement>> STACKS = new ThreadLocal<Deque<NDCElement>>() {
        @Override
        protected Deque<NDCElement> initialValue() {
            return new ArrayDeque<NDCElement>();
        }
    };

    private NDC() {
        //Prevent instantiation of this class
    }

    public static void push(String s) {
        Deque<NDCElement> stack = STACKS.get();
        stack.push(new NDCElement(stack, s));
    }

    public static void push(String... strings) {
        for (String s : strings) {
            push(s);
        }
    }

    public static String pop() {
        try {
            return STACKS.get().pop().getText();
        } catch (NoSuchElementException e) {
            return null;
        }

    }

    public static void pop(int n) {
        Deque<NDCElement> stack = STACKS.get();
        try {
            for (int i = 0; i < n; i++) {
                stack.pop();
            }
        } catch (NoSuchElementException e) {
            // Ignore
        }
    }

    public static void clear() {
        STACKS.get().clear();
    }

    public static String[] getElements() {
        List<String> result = Lists.newArrayList();
        for (Iterator<NDCElement> i = get().descendingIterator(); i.hasNext(); ) {
            result.add(i.next().text);
        }
        return result.toArray(new String[result.size()]);
    }

    static Deque<NDCElement> get() {
        return STACKS.get();
    }

    static class NDCElement {

        private static final Function<NDCElement, AnsiColor> COLOR_SELECTOR = new Function<NDC.NDCElement, AnsiColor>() {
            @Override
            public AnsiColor apply(NDCElement input) {
                return input.getColor();
            }
        };

        private final AnsiFormat format;
        private final String text;

        NDCElement(Deque<NDCElement> parents, String text) {
            this.text = text;
            this.format = new AnsiFormat(ColorPicker.pickColor(parents, text), null, null);
        }

        String getText() {
            return this.text;
        }

        String getColoredText() {
            return this.format.apply(text);
        }

        AnsiColor getColor() {
            return this.format.getForeground();
        }

        @Override
        public int hashCode() {
            return new HashCodeBuilder()
                .append(this.format)
                .append(this.text)
                .toHashCode();
        }

        @Override
        public boolean equals(Object obj) {
            if (!(obj instanceof NDCElement)) {
                return false;
            }

            NDCElement that = (NDCElement) obj;
            return new EqualsBuilder()
                .append(this.format, that.format)
                .append(this.text, that.text)
                .isEquals();
        }

        @Override
        public String toString() {
            return this.text;
        }
    }

    /**
     * An intelligent color picker for NDC elements.
     * Three main features:
     * <ol>
     * <li>Assigns a child {@link NDCElement} a different color than its ancestors.
     * <li>Tries to assign two distinct {@link NDCElement} instances with the same text <i>the same color</i>.
     * <blockquote>
     * Bard needs this capability since it frequently pushes the same NDC text into different <tt>NDCElement</tt> instances
     * (e.g. the same job name is pushed onto the NDC in different spots and often in different threads).
     * </blockquote>
     * <li>Tries to assign two sibling {@link NDCElement} instances different colors.
     * <blockquote>
     * This provides better visual distinction between adjacent log messages.
     * </blockquote>
     * </ol>
     *
     * Colors are assigned in a round-robin fashion to ensure maximum variation in colors for adjacent messages.
     */
    private static class ColorPicker {

        /**
         * The set of candidate {@link AnsiColor}s that we use for NDC coloring.
         */
        private static final Set<AnsiColor> COLORS = EnumSet.of(
            RED,
            GREEN,
            YELLOW,
            BLUE,
            MAGENTA,
            CYAN
        );

        /**
         * The counter for root {@link NDCElement}s.  Used to round-robin among the candidate colors for root elements.
         */
        private static final AtomicInteger ROOT_COUNTER = new AtomicInteger();

        /**
         * Counters for children of a given {@link NDCElement} stack.  Use to round-robin among the candidate colors for sibling {@link NDCElement}s.
         */
        private static final LoadingCache<List<NDCElement>, AtomicInteger> CHILD_COUNTS = CacheBuilder.newBuilder()
            .expireAfterAccess(10, TimeUnit.MINUTES)
            .maximumSize(500)
            .build(new CacheLoader<List<NDCElement>, AtomicInteger>() {
                @Override
                public AtomicInteger load(List<NDCElement> input) {
                    return new AtomicInteger();
                }
            });

        /**
         * A cache of colors by {@link NDCPath}.
         * Used to assign two distinct {@link NDCElement} instances with the same {@link NDCPath} the same color.
         */
        private static final LoadingCache<NDCPath, AnsiColor> CACHE = CacheBuilder.newBuilder()
            .expireAfterAccess(10, TimeUnit.MINUTES)
            .maximumSize(500)
            .build(new CacheLoader<NDCPath, AnsiColor>() {
                @Override
                public AnsiColor load(NDCPath key) {
                    Set<AnsiColor> colors = EnumSet.copyOf(COLORS);

                    // Remove all the ancestor colors so that we assign a different color to ourselves
                    colors.removeAll(Collections2.transform(key.parents, NDCElement.COLOR_SELECTOR));

                    /*
                     * If we've used all the colors reset the list
                     * TODO This is not really optimal coloring
                     */
                    if (colors.isEmpty()) {
                        colors = COLORS;
                    }

                    // Get the proper sibling count so that we assign a different color than our siblings
                    AtomicInteger counter = key.parents.isEmpty()
                            ? ROOT_COUNTER
                            : CHILD_COUNTS.getUnchecked(key.parents);

                    int colorIndex = counter.getAndIncrement() % colors.size();

                    return Iterables.get(colors, colorIndex);
                }
            });

        private static class NDCPath {
            private List<NDCElement> parents;
            private String text;

            NDCPath(Deque<NDCElement> parents, String text) {
                this.parents = ImmutableList.copyOf(parents);
                this.text = text;
            }

            @Override
            public int hashCode() {
                return new HashCodeBuilder()
                    .append(parents)
                    .append(text)
                    .toHashCode();
            }

            @Override
            public boolean equals(Object obj) {
                if (!(obj instanceof NDCPath)) {
                    return false;
                }

                NDCPath that = (NDCPath) obj;
                return new EqualsBuilder()
                    .append(this.parents, that.parents)
                    .append(this.text, that.text)
                    .isEquals();
            }
        }

        static AnsiColor pickColor(Deque<NDCElement> parents, String text) {
            return CACHE.getUnchecked(new NDCPath(parents, text));
        }
    }

}
