/*
 * $Id: //depot/prod/newport_news/main/modules/nn-platform/src/main/java/com/netapp/common/platform/unix/UnixAuthenticatorNativeImpl.java#8 $
 *
 * Copyright (c) 2008 Network Appliance, Inc.
 * All rights reserved.
 *
 */
package com.netapp.common.platform.unix;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

import com.netapp.common.platform.Authenticator;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.ConfigProperty;

/**
 * Authenticates using either
 *
 * A) passwd/shadow  (only on linux AND pam service name is null)
 *
 * OR
 *
 * B) modular authentication using
 *     1.) PAM under Linux, Solaris, HPUX and OSX
 *     2.) LAM under AIX
 *
 * @author rippy
 * @version $Revision: #8 $ $Date: 2012/11/13 $
 */
public class UnixAuthenticatorNativeImpl implements Authenticator {

    private static final ConfigProperty<String> PAM_SERVICE_NAME = ConfigProperty.stringProperty("pam.service.name", "snapmanager");

    private String errorMessage;

    /**
     * PAM service name to use when authenticating.
     * When null, don't use pam (and look at passwd/shadow directly)
     */
    private String serviceName;

    /**
     * Use this constructor if you want to authenticate using passwd directly (and not use pam/lam).
     * Works only on linux.
     */
    public UnixAuthenticatorNativeImpl() {
        this((String) null);
    }

    /**
     * Use this constructor if you want to authenticate using pam/lam.
     */
    public UnixAuthenticatorNativeImpl(String pamServiceName) {
        this.serviceName = pamServiceName;
    }

    /**
     * Use this constructor to use pam/lam, and use the configured "pam.service.name".
     */
    public UnixAuthenticatorNativeImpl(ConfigBundle configBundle) {
        this.serviceName = PAM_SERVICE_NAME.lookup(configBundle);
    }

    /**
     * Tie in to native authentication library.
     *
     * The native logic will set the error message via {@link #setErrorMessage(String)}
     * to <tt>null</tt> on successful authentication, or the error message string
     * on failed authentication.
     *
     * passwd/shadow authentication will be used if appname is null and we're running on linux.
     *
     * pam/lam authentication will be used if appname is non-null.
     */
    native void nativeAuthenticate(String appname, String username, String password);

    static {
        System.loadLibrary("com_netapp_common_platform_unix_UnixAuthenticatorNativeImpl");
    }

    /**
     * Gets the error message of the last authentication request.
     *
     * Note that external synchronization must be in place for this
     * to be completely accurate. (i.e. the authenticate call
     * and the getErrorMessage call must be within the same synchronized block)
     */
    @Override
    public String getErrorMessage() {
        return this.errorMessage;
    }

    /**
     * Will be called by native code to indicate result of authentication.
     */
    public void setErrorMessage(String msg) {
        if (msg == null) {
            this.errorMessage = null;
        } else {
            /*
             * Copy so that the msg allocated in native code can be freed by native code
             */
            this.errorMessage = new StringBuilder(msg).toString();
        }
    }

    @Override
    public synchronized boolean authenticate(String username, String password) {
        setErrorMessage(null);

        this.nativeAuthenticate(this.serviceName, username, password);
        return this.errorMessage == null;
    }

    public static void main(String[] args) throws IOException {
        String username = null;
        String password = null;
        if (args.length == 2) {
            username = args[0];
            password = args[1];
        } else if (args.length == 1) {
            username = args[0];
            password = prompt("Password: ");
        } else if (args.length == 0) {
            username = prompt("Username: ");
            password = prompt("Password: ");
        } else {
            System.out.println("UnixAuthenticatorNativeImpl syntax:\n\n\t<username> <password>");
            System.exit(1);
        }

        // note that this always uses modular (pam/lam) auth
        UnixAuthenticatorNativeImpl authenticator = new UnixAuthenticatorNativeImpl(new ConfigBundle());
        System.out.println("Authenticated ['" + authenticator.authenticate(username, password) + "']\terrorString['" + authenticator.getErrorMessage() + "']");
    }

    @edu.umd.cs.findbugs.annotations.SuppressWarnings("DM_DEFAULT_ENCODING")
    private static String prompt(String prompt) throws IOException {
        System.out.print(prompt);
        System.out.flush();

        return new BufferedReader(new InputStreamReader(System.in)).readLine();
    }
}
