/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/text/DynamicTableWriter.java#2 $
 *
 * Copyright (c) 2010 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.text;

import java.util.Arrays;
import java.util.Formatter;
import java.util.LinkedList;
import java.util.List;

import edu.umd.cs.findbugs.annotations.SuppressWarnings;


/**
 * Writes a {@link Table} to a {@link Appendable} with left-justified columns.
 * Column width is determined by the maximum width of the column header or
 * value within the column.
 * Columns are delimited by a single space character.
 *
 * @author philipc
 * @version $Revision: #2 $ $Date: 2011/06/24 $
 */
public class DynamicTableWriter {

    public void write(Appendable appendable, Table table) {
        Formatter formatter = new Formatter(appendable);

        Integer[] columnWidths = getColumnWidths(table);

        String formatString = this.createFormatStringForColumns(columnWidths);

        printHeading(table, formatter, formatString, columnWidths);

        printRows(table, formatter, formatString);
    }

    private void printRows(Table table, Formatter formatter, String formatString) {
        for (String[] row : table.getRows()) {
            formatter.format(formatString, (Object[]) row);
        }
    }

    private void printHeading(Table table, Formatter formatter, String formatString, Integer[] columnWidths) {
        if (table.getHeadings() != null && table.getHeadings().length > 0) {
            formatter.format(formatString, (Object[]) table.getHeadings());

            String[] headingSeparator = this.getHeadingSeparator(columnWidths);
            formatter.format(formatString, (Object[]) headingSeparator);
        }
    }

    private Integer[] getColumnWidths(Table table) {
        int numColumns = table.getColumnCount();

        Integer[] columnWidths = new Integer[numColumns];

        updateColumnWidths(columnWidths, table.getHeadings());

        for (String[] row : table.getRows()) {
            updateColumnWidths(columnWidths, row);
        }
        return columnWidths;
    }

    private void updateColumnWidths(Integer[] columnWidths, String[] values) {
        if (values != null) {
            for (int i = 0; i < values.length; i++) {
                int valueWidth = (values[i] == null) ? 0 : values[i].length();
                int currentWidth = (columnWidths[i] == null) ? 0 : columnWidths[i];
                columnWidths[i] = Math.max(currentWidth, valueWidth);
            }
        }
    }

    private String createFormatStringForColumns(Integer... columnWidths) {
        StringBuilder formatString = new StringBuilder(10 * columnWidths.length);
        String prefix = "";
        for (int width : columnWidths) {
            formatString.append(prefix).append("%-").append(width).append("s");
            prefix = " "; // All columns after the first get spaces between
        }
        formatString.append(System.getProperty("line.separator"));
        return formatString.toString();
    }


    private String[] getHeadingSeparator(Integer[] columnWidths) {
        String[] headingSeparator = new String[columnWidths.length];
        for (int i = 0; i < columnWidths.length; i++) {
            char[] columnHeading = new char[columnWidths[i]];
            Arrays.fill(columnHeading, '-');
            headingSeparator[i] = new String(columnHeading);
        }
        return headingSeparator;
    }

    public static class Table {

        private String[] headings;
        private List<String[]> rows;

        @SuppressWarnings("EI2")
        public Table(String... headings) {
            this.headings = headings;
            this.rows = new LinkedList<String[]>();
        }

        public void addRow(String... columns) {
            this.rows.add(columns);
        }

        public List<String[]> getRows() {
            return this.rows;
        }

        @SuppressWarnings("EI")
        public String[] getHeadings() {
            return this.headings;
        }

        public int getColumnCount() {
            int columnCount = 0;
            if (this.headings != null) {
                columnCount = this.headings.length;
            }

            for (String[] row : this.rows) {
                columnCount = Math.max(columnCount, row.length);
            }
            return columnCount;
        }
    }

}
