/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/timing/StopWatch.java#10 $
 *
 * Copyright (c) 2011 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.timing;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.concurrent.Callable;

import org.apache.commons.lang.time.DurationFormatUtils;

import com.netapp.common.logging.LogLevel;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.UncheckedCallable;

/**
 * The <tt>StopWatch</tt> is a lightweight way of timing code blocks.
 * It follows the same approach taken by in the perf4j library.
 * Basically, whenever you are inclined to use System.currentTimeMillis() for quick
 * and dirty performance testing, use <tt>StopWatch</tt> instead.
 *
 * <p>
 * The typical pattern for <tt>StopWatch</tt> usage is:
 * <pre>
 * {@code
 * StopWatch stopWatch = new StopWatch().withCategory("foo").withTag("foo.bar").start();
 * // Interesting code goes here
 * stopWatch.stopAndLog();
 * }
 * </pre>
 *
 * The setting of the tag can be deferred to log time as well:
 * <pre>
 * {@code
 * StopWatch stopWatch = new StopWatch().withCategory("foo").start();
 * try {
 *     // Interesting code goes here
 *     stopWatch.withTag("foo.success").log();
 * } catch (...) {
 *     stopWatch.withTag("foo.error").log();
 * }
 * </pre>
 *
 * It is a best practice to use dotted "." syntax for scoping tag names.
 * This will facilitate filtering <tt>StopWatch</tt> log files for tags of interest.
 * This practice is followed in AutoZAPI - for example:
 * <ul>
 * <li><tt>zapi.connect.$target-filer</tt>: for connection timing
 * <li><tt>zapi.api.$api-name.$target-filer</tt>: for API timing
 * <li>etc.
 * </ul>
 *
 * <p>
 * The <tt>StopWatch</tt> will log the timings at <tt>TRACE</tt> level to a logger named
 * <tt>stopwatch</tt> (or <tt>{@code stopwatch.<category>} see {@link #withCategory(String)}).
 * The timings are logged in a self-describing JSON format.
 *
 * To direct the <tt>StopWatch</tt> logs to a log file in logback, use a clause such as the
 * following example in your <tt>logback.xml</tt>:
 * <pre>
 * {@code
 * <logger name="stopwatch" additivity="false" level="TRACE">
 *     <appender name="stopwatch" class="ch.qos.logback.core.rolling.RollingFileAppender">
 *         <file>./log/stopwatch.log</file>
 *         <encoder>
 *             <pattern>%msg%n</pattern>
 *         </encoder>
 *     </appender>
 * </logger>
 * }
 * </pre>
 *
 * <p>
 * See also: <a href="http://wikid.netapp.com/w/MEI/CSI/Bard/Development/Developer_Guide/StopWatch">
 * StopWatch wiki page</a>
 *
 * @author sirianni
 * @version $Revision: #10 $ $Date: 2013/02/07 $
 */
public class StopWatch implements Serializable {

    private static final String ROOT_LOGGER_NAME = "stopwatch";

    private static final NALogger ROOT_LOGGER = NALogger.getLogger(ROOT_LOGGER_NAME);

    private transient NALogger logger;
    private String category;
    private String tagFormat;
    private Object[] tagArgs;

    private long startTime;
    private long endTime;

    public StopWatch() {
        this.logger = ROOT_LOGGER;
        this.startTime = -1;
        this.endTime = -1;
    }

    /**
     * To be used when an already known 'startTime' 'endTime' has to be logged into stopwatch.log
     * Calling stop() or stopAndLog() on a {@link StopWatch} constructed using this constructor will
     * throw {@link IllegalStateException}. To log, use {@link StopWatch#log()} method directly
     *
     * @param startTime
     * @param endTime
     *
     * @see #StopWatch()
     */
    public StopWatch(long startTime, long endTime) {
        this();
        this.startTime = startTime;
        this.endTime = endTime;
    }

    /**
     * Set the tag for this {@link StopWatch}.
     * It is a best practice to use dotted "." syntax for scoping tag names.
     * This will facilitate filtering <tt>StopWatch</tt> log files for tags of interest.
     * This practice is followed in AutoZAPI - for example:
     * <ul>
     * <li><tt>zapi.connect.$target-filer</tt>: for connection timing
     * <li><tt>zapi.api.$api-name.$target-filer</tt>: for API timing
     * <li>etc.
     * </ul>
     *
     * @param tagFormat Tag name (or {@link String#format(String, Object...) string}.
     * @param tagArgs Format arguments - <tt>String.format(tagFormat, tagArgs)</tt> will be called if args is not empty.
     */
    public StopWatch withTag(String tagFormat, Object... tagArgs) {
        /*
         * Avoid formatting until actually logged.
         * This will avoid the formatting costs unless the
         * stopwatch is enabled when log() is called.
         */
        this.tagFormat = tagFormat;
        this.tagArgs = tagArgs;
        return this;
    }

    /**
     * Set the category for this {@link StopWatch}.
     *
     * @param category The category to use for this {@link StopWatch}.
     * The category is appended to the String <tt>stopwatch.</tt> to generate
     * the name of the {@link NALogger} used for logging.  This fully qualified logger name
     * (e.g. <tt>stopwatch.myCategory</tt> is then what can be referenced in logging
     * configuration files.  If not specified, messages from this {@link StopWatch}
     * will be logged via the "root" {@link StopWatch} logger name of <tt>stopwatch</tt>.
     * Category names should follow java package naming conventions and typically include
     * product top-level names, for example:
     * <ul>
     * <li><tt>dfm.monitor</tt>: StopWatches in the DFM monitor subsystem
     * <li><tt>zapi</tt>: StopWatches in the AutoZAPI subsystem
     * </ul>
     */
    public StopWatch withCategory(String category) {
        this.category = category;
        this.logger = NALogger.getLogger(ROOT_LOGGER_NAME + "." + category);
        return this;
    }

    /**
     * Start the {@link StopWatch}.
     */
    public StopWatch start() {
        this.startTime = System.currentTimeMillis();
        this.endTime = -1;
        return this;
    }

    /**
     * Stop the {@link StopWatch}
     */
    public StopWatch stop() {
        if (this.startTime == -1) {
            throw new IllegalStateException("Not started");
        } else if (this.endTime != -1) {
            throw new IllegalStateException("Already stopped");
        }

        this.endTime = System.currentTimeMillis();
        return this;
    }

    /**
     * Stop the {@link StopWatch} and log the timing record.
     */
    public void stopAndLog() {
        this.stop();
        this.log();
    }

    /**
     * Log the timing record.
     * The {@link StopWatch} will continue to run.
     */
    public void log() {
        if (this.isEnabled()) {
            String json = String.format(
                  "timing{"
                    + "\"tag\":\"%s\","
                    + "\"start\":%d,"
                    + "\"end\":%d,"
                    + "\"elapsed\":%d,"
                    + "\"elapsedHMS\":\"%s\""
                + "}",
                getTag(), this.startTime, this.endTime, this.getElapsedTime(), this.getElapsedTimeHMS());
            this.logger.internalTrace(json);
        }
    }

    /**
     * Gets the fully formatted tag value.
     */
    private String getTag() {
        return this.tagArgs.length == 0
            ? this.tagFormat
            : String.format(this.tagFormat, this.tagArgs);
    }

    /**
     * Get the elapsed time of this {@link StopWatch} in milliseconds.
     * The {@link StopWatch} will continue to run.
     */
    public long getElapsedTime() {
        if (this.startTime == -1) {
            return 0;
        } else if (this.endTime == -1) {
            return System.currentTimeMillis() - this.startTime;
        } else {
            return this.endTime - this.startTime;
        }

    }

    /**
     * Get the elapsed time of this {@link StopWatch} as a duration formatted String (<tt>HH:mm:ss.SSSS</tt>).
     * @see DurationFormatUtils#formatDurationHMS(long)
     */
    public String getElapsedTimeHMS() {
        return DurationFormatUtils.formatDurationHMS(this.getElapsedTime());
    }

    /**
     * Call the provided {@link UncheckedCallable} and record the elapsed time via this {@link StopWatch}.
     */
    public <V> V trace(UncheckedCallable<V> callable) {
        this.start();
        try {
            return callable.call();
        } finally {
            this.stopAndLog();
        }
    }

    /**
     * Call the provided {@link Callable} and record the elapsed time via this {@link StopWatch}.
     */
    public <V> V trace(Callable<V> callable) throws Exception {
        this.start();
        try {
            return callable.call();
        } finally {
            this.stopAndLog();
        }
    }

    private void readObject(ObjectInputStream inputStream) throws IOException, ClassNotFoundException {
        inputStream.defaultReadObject();
        this.logger = this.category == null ? ROOT_LOGGER : NALogger.getLogger(ROOT_LOGGER_NAME + "." + this.category);
    }

    /**
     * Is the logger enabled for this category?
     * @see #withCategory(String)
     */
    public boolean isEnabled() {
        return this.logger.isLogLevelEnabled(LogLevel.TRACE);
    }
}
