/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/util/ConfigBundle.java#4 $
 *
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Collection;
import java.util.Properties;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * Maintain a list of application configuration values.
 * @author sunday
 * @version $Revision: #4 $ $Date: 2011/06/24 $
 */
public class ConfigBundle extends Properties {

    private static final long serialVersionUID = 1L;

    private static final NALogger logger = NALogger.getLogger(ConfigBundle.class);
    private static final ConfigProperty<File> INSTALL_DIR_PROPERTY = new ConfigProperty<File>(File.class, "install.dir");

    private File installDir;

    private boolean useSystemProperties;
    private String systemPropertiesPrefix = "";

    /**
     * Construct an empty ConfigBundle.
     */
    public ConfigBundle() {
        super();
    }

    /**
     * Construct a new Configuration Bundle from the supplied file names. All files are relative to the executing class path.
     * Later files will supercede definitions from earlier files.
     * @param configFiles
     */
    public ConfigBundle(String... configFiles) {
        this(Thread.currentThread().getContextClassLoader(), configFiles);
    }

    /**
     * Construct a new Configuration Bundle from the supplied file names. All files are relative to the executing class path.
     * Later files will supersede definitions from earlier files.
     * @param configFiles
     * @param classLoader
     */
    public ConfigBundle(ClassLoader classLoader, String... configFiles) {
        this(classLoader, Arrays.asList(configFiles));
    }

    /**
     * Construct a new Configuration Bundle from the supplied file names. All files are relative to the executing class path.
     * Later files will supersede definitions from earlier files.
     * @param configFiles
     */
    public ConfigBundle(Collection<String> configFiles) {
        this(Thread.currentThread().getContextClassLoader(), configFiles);
    }

    /**
     * Construct a new Configuration Bundle from the supplied file names. All files are relative to the executing class path.
     * Later files will supersede definitions from earlier files.
     * @param configFiles
     * @param classLoader
     */
    public ConfigBundle(ClassLoader classLoader, Collection<String> configFiles) {
        this();
        String file = null;
        try {
            for (String ifile : configFiles) {
                file = ifile;
                InputStream stream = classLoader.getResourceAsStream(file);

                if (stream != null) {
                    this.load(stream);
                    stream.close();
                }
            }
        } catch (IOException e) {
            throw new RuntimeException(logger.error(MsgKey.UTIL_CONFIG_EXCEPTION_LOADING_FILE, e, file, e.getLocalizedMessage()));
        }
    }

    public ConfigBundle(Properties properties) {
        this();
        this.putAll(properties);
    }

    /**
     * Construct a new Configuration Bundle by combining bundles.
     * Later bundles supersede definitions from earlier ones.
     * @param bundles
     */
    public ConfigBundle(ConfigBundle... bundles) {
        this();
        for (ConfigBundle bundle : bundles) {
            if (bundle != null) {
                this.putAll(bundle);
            }
        }
    }

    public void setUseSystemProperties(boolean useSystemProperties) {
        this.useSystemProperties = useSystemProperties;
    }

    public void setSystemPropertiesPrefix(String systemPropertiesPrefix) {
        this.systemPropertiesPrefix = systemPropertiesPrefix;
    }

    @Override
    public String getProperty(String key) {
        if (this.useSystemProperties) {
            String property = System.getProperty(this.systemPropertiesPrefix + key);
            if (property != null) {
                return property;
            }
        }
        return super.getProperty(key);
    }

    public File getInstallDir() {
        if (this.installDir == null) {
            this.installDir = INSTALL_DIR_PROPERTY.lookup(this);
        }

        if (this.installDir == null) {
            String env = System.getenv("INSTALL_DIR");
            if (env != null) {
                this.installDir = new File(env);
            }
        }

        if (this.installDir == null) {
            this.installDir = new File(".");
        }
        return installDir;
    }

    @Deprecated
    public Properties getProperties() {
        return this;
    }

    @Deprecated
    public static Properties load(ClassLoader loader, String[] configFiles) {
        String file = null;
        Properties returnProps = new Properties();
        try {
            for (String ifile : configFiles) {
                file = ifile;
                InputStream stream = loader.getResourceAsStream(file);

                if (stream != null) {
                    Properties props = new Properties(returnProps);
                    props.load(stream);
                    stream.close();
                    returnProps = props;
                }
            }
        } catch (IOException e) {
            throw new RuntimeException(logger.error(MsgKey.UTIL_CONFIG_EXCEPTION_LOADING_FILE, e, file, e.getLocalizedMessage()));
        }
        return returnProps;
    }

    @Override
    public synchronized int hashCode() {
        return super.hashCode();
    }

    @Override
    public synchronized boolean equals(Object o) {
        return super.equals(o);
    }
}
