/*
 * $Id: //depot/prod/newport_news/main/modules/nn-core/src/main/java/com/netapp/common/xml/XmlUtils.java#3 $
 *
 * Copyright (c) 2011 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.xml;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;

import javax.xml.stream.EventFilter;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import com.google.common.base.Predicate;
import com.google.common.base.Throwables;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Utility methods for working with XML.
 *
 * @author sirianni
 * @version $Revision: #3 $ $Date: 2012/08/16 $
 */
public class XmlUtils {

    private XmlUtils() { }

    /**
     * Transform the given xml {@link Source} to a String.
     */
    public static String toString(Source xmlSource) {
        try {
            StreamResult streamResult = new StreamResult(new StringWriter());
            TransformerFactory.newInstance().newTransformer().transform(xmlSource, streamResult);
            return streamResult.getWriter().toString();
        } catch (Exception e) {
            throw Throwables.propagate(e);
        }
    }

    /**
     * Transform the given DOM {@link Node} to a String.
     */
    public static String toString(Node node) {
        return toString(new DOMSource(node));
    }

    /**
     * Copy the contents of the given {@link XMLEventReader} into a {@link DocumentFragment} until
     * the given termination {@link Predicate} is satisfied <b>or</b> the end of the XML event stream is reached.
     *
     * The {@link XMLEvent} satisfying the termination <tt>Predicate</tt> is <b>not</b> included in
     * the document fragment.
     *
     * @param reader Source of the copy
     * @param documentFragment Destination of the copy
     * @param terminationPredicate Termination condition
     * @throws XMLStreamException
     */
    public static void copy(XMLEventReader reader, DocumentFragment documentFragment, Predicate<XMLEvent> terminationPredicate) throws XMLStreamException {
        /*
         * The built-in Java 6 SJSXP parser handles a DOMResult on a DocumentFragment correctly.
         * However, woodstox does not (see http://jira.codehaus.org/browse/WSTX-257).
         *
         * In order to support both parsers in AutoZAPI, we work around this by creating the DOMResult on
         * an Element and then re-parenting the Nodes to a DocumentFragment.
         *
         * We should remove this workaround once WSTX-257 is fixed.
         */
        Document document = documentFragment.getOwnerDocument();
        Element fragmentContainerElement = document.createElement("fragment-container");

        XmlUtils.copy(reader, fragmentContainerElement, terminationPredicate);

        /*
         * Now re-parent the DOM Elements we read from the StAX stream to a newly created DocumentFragment node
         */
        NodeList childNodes = fragmentContainerElement.getChildNodes();
        List<Element> childElements = new ArrayList<Element>(childNodes.getLength());
        for (int i = 0; i < childNodes.getLength(); i++) {
            Node node = childNodes.item(i);
            if (node instanceof Element) {
                childElements.add((Element) node);
            }
        }

        for (Element childElement : childElements) {
            documentFragment.appendChild(childElement);
        }
    }

    /**
     * Copy the contents of the given {@link XMLEventReader} into a {@link Node} until
     * the given termination {@link Predicate} is satisfied <b>or</b> the end of the XML event stream is reached.
     *
     * The {@link XMLEvent} satisfying the termination <tt>Predicate</tt> is <b>not</b> included in
     * the document fragment.
     *
     * @param reader Source of the copy
     * @param node Destination of the copy. Must be a {@link Document} or {@link Element}.
     * @param terminationPredicate Termination condition
     * @throws XMLStreamException
     */
    public static void copy(XMLEventReader reader, Node node, Predicate<XMLEvent> terminationPredicate) throws XMLStreamException {

        XMLEventWriter xmlEventWriter = XMLOutputFactory.newInstance().createXMLEventWriter(new DOMResult(node));
        try {
            for (XMLEvent event = reader.nextEvent(); !terminationPredicate.apply(event) && reader.hasNext(); event = reader.nextEvent()) {
                xmlEventWriter.add(event);
            }
            xmlEventWriter.flush();
        } finally {
            xmlEventWriter.close();
        }
    }

    public static class Stax {

        public static final EventFilter WHITESPACE_FILTER = new EventFilter() {
            @Override
            public boolean accept(XMLEvent event) {
                return !(event.isCharacters() && event.asCharacters().isWhiteSpace());
            }
        };

        public static final EventFilter COMMENT_AND_WHITESPACE_FILTER = new EventFilter() {
            @Override
            public boolean accept(XMLEvent event) {
                if (event.isCharacters() && event.asCharacters().isWhiteSpace()) {
                    return false;
                } else if (event.getEventType() == XMLEvent.COMMENT) {
                    return false;
                }
                return true;
            }
        };

        private Stax() { }

        /**
         * Consume events off the {@link XMLEventReader} until the predicate is satisfied.
         *
         * @return The first {@link XMLEvent} that satisfies the predicate
         */
        @SuppressWarnings("unchecked")
        public static <T extends XMLEvent> T consumeUntil(XMLEventReader reader, Predicate<XMLEvent> predicate) throws XMLStreamException {
            XMLEvent event = reader.nextEvent();
            while (!predicate.apply(event)) {
                event = reader.nextEvent();
            }
            return (T) event;
        }

        public static class Predicates {

            private Predicates() { }

            /**
             * Returns true when the {@link XMLEvent} represents the matching {@link javax.xml.stream.events.EndElement EndElement}
             * for the given {@link StartElement}.
             *
             * For example:
             * <pre>
             * {@code
             * <foo> <!-- This is the start element -->
             *   <bar>
             *     <baz>
             *       <foo>
             *       </foo> <!-- This is an end element with name 'foo' but not the *matching* end element -->
             *     </baz>
             *   </bar>
             * </foo> <!-- This is the *matching* end element -->
             */
            public static Predicate<XMLEvent> matchingEndElement(final StartElement startElement) {
                return new Predicate<XMLEvent>() {

                    /*
                     * We need to do some bookkeeping to handle the case where the nested XML fragment contains
                     * a tag that matches the tag of the enclosing XML field itself.
                     */

                    int endElementsOutstanding = 1;

                    @Override
                    public boolean apply(XMLEvent event) {
                        if (event.isStartElement() && startElement.getName().equals(event.asStartElement().getName())) {
                            endElementsOutstanding++;
                        } else if (event.isEndElement() && startElement.getName().equals(event.asEndElement().getName())) {
                            endElementsOutstanding--;
                        }
                        return endElementsOutstanding == 0;
                    }
                };

            }
        }

    }

    public static class Transform {

        private static final ThreadLocal<Transformer> IDENTITY_TRANSFORMER = new ThreadLocal<Transformer>() {
            @Override
            protected Transformer initialValue() {
                try {
                    return JaxpContext.transform().transformerFactory().newTransformer();
                } catch (TransformerConfigurationException e) {
                    throw Throwables.propagate(e);
                } catch (TransformerFactoryConfigurationError e) {
                    throw Throwables.propagate(e);
                }
            }
        };

        private Transform() { }

        /**
         * The identity XML transformer.
         * Useful for transforming XML from a {@link javax.xml.transform.Source} to a {@link javax.xml.transform.Result}
         * without changing the structure of the XML itself.
         */
        public static Transformer identityTransformer() {
            return IDENTITY_TRANSFORMER.get();
        }
    }

}
