package com.vmware.vcloud.sdk;

import com.vmware.vcloud.api.rest.schema.OperatingSystemFamilyInfoType;
import com.vmware.vcloud.api.rest.schema.OperatingSystemInfoType;
import com.vmware.vcloud.api.rest.schema.OrgListType;
import com.vmware.vcloud.api.rest.schema.QueryResultFileDescriptorRecordType;
import com.vmware.vcloud.api.rest.schema.ReferenceType;
import com.vmware.vcloud.api.rest.schema.SessionType;
import com.vmware.vcloud.api.rest.schema.SupportedOperatingSystemsInfoType;
import com.vmware.vcloud.api.rest.schema.versioning.SupportedVersionsType;
import com.vmware.vcloud.api.rest.schema.versioning.VersionInfoType;
import com.vmware.vcloud.sdk.admin.VcloudAdmin;
import com.vmware.vcloud.sdk.admin.extensions.VcloudAdminExtension;
import com.vmware.vcloud.sdk.constants.SdkMessage;
import com.vmware.vcloud.sdk.constants.Version;
import com.vmware.vcloud.sdk.constants.query.FormatType;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.GZIPOutputStream;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.apache.commons.codec.binary.Base64;
import org.apache.http.Header;
import org.apache.http.HttpHost;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.CookieStore;
import org.apache.http.client.CredentialsProvider;
import org.apache.http.client.HttpClient;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.impl.auth.BasicScheme;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.tsccm.ThreadSafeClientConnManager;
import org.apache.http.message.BasicHeader;
import org.apache.http.params.HttpConnectionParams;
import org.apache.http.params.HttpParams;
import org.w3c.dom.Element;

public class VcloudClient
{
  private static Logger logger = Logger.getLogger("com.vmware.vcloud.sdk");
  private String vCloudApiURL;
  private DefaultHttpClient client;
  private String userName;
  private String orgName;
  private HashMap<String, ReferenceType> orgRefsByName;
  private String vCloudTokenHeader;
  private String vCloudToken;
  HashMap<String, VersionInfoType> supportedApiVersions;
  static final HashMap<Version, String> SUPPORTED_SDK_VERSIONS = new HashMap();
  Version vcloudClientVersion;

  Version getVcloudClientVersion()
    throws VCloudException
  {
    return this.vcloudClientVersion;
  }

  String getVcloudTokenHeader()
    throws VCloudException
  {
    if (this.vCloudTokenHeader != null)
      return this.vCloudTokenHeader;
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.LOGIN_ERROR_MSG));
  }

  private void setVcloudTokenHeader(String vcloudTokenHeader)
  {
    this.vCloudTokenHeader = vcloudTokenHeader;
  }

  public String getVcloudToken()
    throws VCloudException
  {
    if (this.vCloudToken != null)
      return this.vCloudToken;
    throw new VCloudException(SdkUtil.getI18nString(SdkMessage.LOGIN_ERROR_MSG));
  }

  public void setVcloudToken(String vcloudToken)
  {
    this.vCloudToken = vcloudToken;
    this.vCloudTokenHeader = "x-vcloud-authorization";
  }

  public VcloudClient(String vCloudUrl, Version version)
  {
    try
    {
      if ((!vCloudUrl.endsWith("/")) && (!vCloudUrl.endsWith("\\"))) {
        vCloudUrl = vCloudUrl + "/";
      }
      new URL(vCloudUrl);
    } catch (MalformedURLException e) {
      throw new VCloudRuntimeException(e);
    }
    setVCloudApiURL(vCloudUrl + "api");
    this.vcloudClientVersion = version;
    ThreadSafeClientConnManager connectionManager = new ThreadSafeClientConnManager();
    this.client = new DefaultHttpClient(connectionManager);
  }

  public void setConnectionTimeout(int milliseconds)
  {
    HttpConnectionParams.setConnectionTimeout(this.client.getParams(), milliseconds);
  }

  public void setSocketTimeout(int milliseconds)
  {
    HttpConnectionParams.setSoTimeout(this.client.getParams(), milliseconds);
  }

  public HashMap<String, VersionInfoType> getSupportedVersions()
    throws VCloudException
  {
    if (this.supportedApiVersions == null) { this.supportedApiVersions = new HashMap();
      Response response;
      try {
        response = RestUtil.getSupportedVersions(this, getVcloudApiURL() + "/versions");
      }
      catch (VCloudRuntimeException e)
      {
        throw new VCloudRuntimeException(e);
      }
      SupportedVersionsType supportedVersionsType = null;
      if (response.isExpected(200)) {
        supportedVersionsType = (SupportedVersionsType)response.getResource();
      }
      else
      {
        response.handleUnExpectedResponse();
      }
      if (supportedVersionsType != null)
        for (VersionInfoType version : supportedVersionsType.getVersionInfo())
        {
          this.supportedApiVersions.put(version.getVersion(), version);
        } }
    return this.supportedApiVersions;
  }

  public void setMaxConnections(int maxConnections)
  {
    ((ThreadSafeClientConnManager)this.client.getConnectionManager()).setDefaultMaxPerRoute(maxConnections);

    if (maxConnections > 20)
      ((ThreadSafeClientConnManager)this.client.getConnectionManager()).setMaxTotal(maxConnections);
  }

  public void setProxy(String proxyHost, int port, String scheme)
  {
    HttpHost proxy = new HttpHost(proxyHost, port, scheme);
    this.client.getParams().setParameter("http.route.default-proxy", proxy);
  }

  public void setProxyCredentials(String username, String password)
  {
    HttpHost httpHost = (HttpHost)this.client.getParams().getParameter("http.route.default-proxy");

    if (httpHost != null)
      this.client.getCredentialsProvider().setCredentials(new AuthScope(httpHost.getHostName(), httpHost.getPort()), new UsernamePasswordCredentials(username, password));
  }

  public void login(String userName, String password)
    throws VCloudException
  {
    Header authHeader = BasicScheme.authenticate(new UsernamePasswordCredentials(userName, password), "UTF-8", false);

    VersionInfoType versionInfo = (VersionInfoType)getSupportedVersions().get(this.vcloudClientVersion.value());

    if (versionInfo == null) {
      throw new VCloudException(this.vcloudClientVersion.value() + " " + SdkUtil.getI18nString(SdkMessage.VERSION_NOT_SUPPORTED));
    }

    Response response = RestUtil.login(this, versionInfo.getLoginUrl(), authHeader);

    this.client.getCookieStore().clear();

    if (!response.isExpected(200))
    {
      response.handleUnExpectedResponse();
    }
    SessionType session = (SessionType)response.getResource();
    setOrgName(session.getOrg());
    setUserName(session.getUser());
  }

  public void ssoLogin(Element tokenElement, String orgName)
    throws VCloudException
  {
    TransformerFactory transFactory = TransformerFactory.newInstance();
    Transformer transformer;
    try
    {
      transformer = transFactory.newTransformer();
    } catch (TransformerConfigurationException e) {
      throw new VCloudRuntimeException(e);
    }
    StringWriter buffer = new StringWriter();
    try {
      transformer.transform(new DOMSource(tokenElement), new StreamResult(buffer));
    }
    catch (TransformerException e) {
      throw new VCloudRuntimeException(e);
    }
    String samlAssertionXML = buffer.toString();
    ssoLogin(samlAssertionXML, orgName);
  }

  public void ssoLogin(String samlAssertionXML, String orgName)
    throws VCloudException
  {
    ByteArrayOutputStream out = new ByteArrayOutputStream();
    try {
      GZIPOutputStream gout = new GZIPOutputStream(out);
      gout.write(samlAssertionXML.getBytes());
      gout.close();
    } catch (IOException e) {
      throw new VCloudRuntimeException(e);
    }
    String output = new Base64().encodeToString(out.toByteArray());
    String compressedEncodedSAMLAssertionXML = output.replaceAll(System.getProperty("line.separator"), "");
    Header samlAuthHeader;
    Header samlAuthHeader;
    if (orgName.equalsIgnoreCase("System")) {
      samlAuthHeader = new BasicHeader("Authorization", "SIGN token=\"" + compressedEncodedSAMLAssertionXML + "\"");
    }
    else
    {
      samlAuthHeader = new BasicHeader("Authorization", "SIGN token=\"" + compressedEncodedSAMLAssertionXML + "\"," + "org" + "=" + "\"" + orgName + "\"");
    }

    VersionInfoType versionInfo = (VersionInfoType)getSupportedVersions().get(this.vcloudClientVersion.value());

    if (versionInfo == null) {
      throw new VCloudException(this.vcloudClientVersion.value() + " " + SdkUtil.getI18nString(SdkMessage.VERSION_NOT_SUPPORTED));
    }

    Response response = RestUtil.login(this, versionInfo.getLoginUrl(), samlAuthHeader);

    this.client.getCookieStore().clear();
    if (!response.isExpected(200))
    {
      response.handleUnExpectedResponse();
    }
    SessionType session = (SessionType)response.getResource();
    setOrgName(session.getOrg());
    setUserName(session.getUser());
  }

  public void logout()
    throws VCloudException
  {
    SdkUtil.delete(this, getVcloudApiURL() + "/session", 204);

    setVcloudTokenHeader(null);
    setVcloudToken(null);
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.LOGOUT_INFO_MSG));
  }

  public HashMap<String, ReferenceType> getOrgRefsByName()
    throws VCloudException
  {
    return this.orgRefsByName == null ? getOrgReferences() : this.orgRefsByName;
  }

  public HashMap<String, ReferenceType> getUpdatedOrgList()
    throws VCloudException
  {
    return getOrgReferences();
  }

  public ReferenceType getOrgRefByName(String orgName)
    throws VCloudException
  {
    return this.orgRefsByName == null ? (ReferenceType)getOrgReferences().get(orgName) : (ReferenceType)this.orgRefsByName.get(orgName);
  }

  public Collection<ReferenceType> getOrgRefs()
    throws VCloudException
  {
    return this.orgRefsByName == null ? getOrgReferences().values() : this.orgRefsByName.values();
  }

  private HashMap<String, ReferenceType> getOrgReferences()
    throws VCloudException
  {
    String orgListUrl = getVcloudApiURL() + "/org/";

    logger.log(Level.INFO, orgListUrl);
    OrgListType orgListType_v1_5 = (OrgListType)SdkUtil.get(this, orgListUrl, 200);

    this.orgRefsByName = new HashMap();
    if ((orgListType_v1_5 == null) || (orgListType_v1_5.getOrg().size() == 0)) {
      logger.log(Level.WARNING, SdkUtil.getI18nString(SdkMessage.ORGS_EMPTY_INFO_MSG));
    }

    for (ReferenceType refType : orgListType_v1_5.getOrg()) {
      this.orgRefsByName.put(refType.getName(), refType);
    }
    return this.orgRefsByName;
  }

  public String getVcloudApiURL()
  {
    return this.vCloudApiURL;
  }

  void setVCloudApiURL(String cloudApiURL)
  {
    this.vCloudApiURL = cloudApiURL;
  }

  HttpClient getClient()
  {
    return this.client;
  }

  public VcloudAdmin getVcloudAdmin()
    throws VCloudException
  {
    return new VcloudAdmin(this);
  }

  public VcloudAdminExtension getVcloudAdminExtension()
    throws VCloudException
  {
    return new VcloudAdminExtension(this);
  }

  public static void setLogLevel(Level level)
  {
    logger.setLevel(level);
  }

  public void registerScheme(String name, Integer port, SSLSocketFactory factory)
  {
    Scheme scheme = new Scheme(name, port.intValue(), factory);
    ((ThreadSafeClientConnManager)this.client.getConnectionManager()).getSchemeRegistry().register(scheme);
  }

  public QueryService getQueryService()
  {
    return new QueryService(this);
  }

  public String getUserName()
  {
    return this.userName;
  }

  private void setUserName(String userName)
  {
    this.userName = userName;
  }

  public String getOrgName()
  {
    return this.orgName;
  }

  private void setOrgName(String orgName)
  {
    this.orgName = orgName;
  }

  public Boolean extendSession()
  {
    try
    {
      SdkUtil.get(this, getVcloudApiURL() + "/session", 200);

      logger.log(Level.INFO, getVcloudApiURL() + "/session");
    }
    catch (VCloudException e) {
      logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.SESSION_EXTENSION_FAILED));

      return Boolean.valueOf(false);
    }
    return Boolean.valueOf(true);
  }

  public HashMap<String, OperatingSystemFamilyInfoType> getOperatingSystemFamiliesByName()
    throws VCloudException
  {
    HashMap osFamilyInfos = new HashMap();
    SupportedOperatingSystemsInfoType supportedOperatingSystems = (SupportedOperatingSystemsInfoType)SdkUtil.get(this, getVcloudApiURL() + "/supportedSystemsInfo", 200);

    for (OperatingSystemFamilyInfoType osFamilyInfo : supportedOperatingSystems.getOperatingSystemFamilyInfo())
    {
      osFamilyInfos.put(osFamilyInfo.getName(), osFamilyInfo);
    }
    return osFamilyInfos;
  }

  public HashMap<String, OperatingSystemInfoType> getOperatingSystemsByName()
    throws VCloudException
  {
    HashMap osInfos = new HashMap();
    SupportedOperatingSystemsInfoType supportedOses = (SupportedOperatingSystemsInfoType)SdkUtil.get(this, getVcloudApiURL() + "/supportedSystemsInfo", 200);

    for (OperatingSystemFamilyInfoType osFamilyInfo : supportedOses.getOperatingSystemFamilyInfo())
    {
      for (OperatingSystemInfoType os : osFamilyInfo.getOperatingSystem()) {
        osInfos.put(os.getName(), os);
      }
    }
    return osInfos;
  }

  public ReferenceResult getServiceRefs()
    throws VCloudException
  {
    String query = getVcloudApiURL() + "/service" + "?" + getQueryService().buildQuery(FormatType.REFERENCE_VIEW);

    return (ReferenceResult)getQueryService().executeQuery(query);
  }

  public RecordResult<QueryResultFileDescriptorRecordType> getFileDescriptorRecords()
    throws VCloudException
  {
    String query = getVcloudApiURL() + "/files" + "?" + getQueryService().buildQuery(FormatType.RECORD_VIEW);

    return (RecordResult)getQueryService().executeQuery(query);
  }

  public InputStream getSchemaDefinition(String schemaFileName)
    throws VCloudException
  {
    logger.log(Level.INFO, SdkUtil.getI18nString(SdkMessage.GET_SCHEMA_DEFINITION) + " - " + schemaFileName);

    String url = getVcloudApiURL() + "/" + "v1.5" + "/schema/" + schemaFileName;

    return RestUtil.downloadFile(this, url);
  }

  static
  {
    SUPPORTED_SDK_VERSIONS.put(Version.V1_5, "application/*+xml;version=1.5");

    SUPPORTED_SDK_VERSIONS.put(Version.V5_1, "application/*+xml;version=5.1");
  }
}