package com.vmware.vcloud.sdk.admin;

import com.vmware.vcloud.api.rest.schema.AdminOrgType;
import com.vmware.vcloud.api.rest.schema.LinkType;
import com.vmware.vcloud.api.rest.schema.NetworksType;
import com.vmware.vcloud.api.rest.schema.ObjectFactory;
import com.vmware.vcloud.api.rest.schema.OrganizationReferenceType;
import com.vmware.vcloud.api.rest.schema.OrganizationReferencesType;
import com.vmware.vcloud.api.rest.schema.ProviderVdcReferencesType;
import com.vmware.vcloud.api.rest.schema.ReferenceType;
import com.vmware.vcloud.api.rest.schema.RightReferencesType;
import com.vmware.vcloud.api.rest.schema.RoleReferencesType;
import com.vmware.vcloud.api.rest.schema.RoleType;
import com.vmware.vcloud.api.rest.schema.VCloudType;
import com.vmware.vcloud.sdk.JAXBUtil;
import com.vmware.vcloud.sdk.SdkUtil;
import com.vmware.vcloud.sdk.VCloudException;
import com.vmware.vcloud.sdk.VcloudClient;
import com.vmware.vcloud.sdk.constants.SdkMessage;
import java.util.Collection;
import java.util.HashMap;
import java.util.logging.Logger;
import javax.xml.bind.JAXBElement;

public class VcloudAdmin
{
  private static Logger logger = Logger.getLogger("com.vmware.vcloud.sdk");
  private VCloudType vcloudResource_v1_5;
  private VcloudClient client;
  private ReferenceType systemAdminOrgReference;
  private HashMap<String, ReferenceType> adminOrganizationRefsByName;
  private HashMap<String, ReferenceType> providerVdcRefsByName;
  private HashMap<String, ReferenceType> rightRefsByName;
  private HashMap<String, ReferenceType> roleRefsByName;
  private HashMap<String, ReferenceType> externalNetworkRefsByName;

  public VcloudAdmin(VcloudClient client)
    throws VCloudException
  {
    this.client = client;

    this.vcloudResource_v1_5 = ((VCloudType)SdkUtil.get(client, client.getVcloudApiURL() + "/admin/", 200));

    sortAdminRefs_v1_5();
  }

  public AdminQueryService getAdminQueryService()
  {
    return new AdminQueryService(this.client);
  }

  public VCloudType getResource()
  {
    return this.vcloudResource_v1_5;
  }

  private void sortAdminRefs_v1_5()
  {
    this.adminOrganizationRefsByName = new HashMap();
    this.providerVdcRefsByName = new HashMap();
    this.rightRefsByName = new HashMap();
    this.roleRefsByName = new HashMap();
    this.externalNetworkRefsByName = new HashMap();

    if (getResource().getOrganizationReferences() != null) {
      for (OrganizationReferenceType orgRef : getResource().getOrganizationReferences().getOrganizationReference())
      {
        this.adminOrganizationRefsByName.put(orgRef.getName(), orgRef);
      }
    }
    if (getResource().getProviderVdcReferences() != null) {
      for (ReferenceType providerVdcRef : getResource().getProviderVdcReferences().getProviderVdcReference())
      {
        this.providerVdcRefsByName.put(providerVdcRef.getName(), providerVdcRef);
      }
    }

    if (getResource().getRightReferences() != null) {
      for (ReferenceType rightRef : getResource().getRightReferences().getRightReference())
      {
        this.rightRefsByName.put(rightRef.getName(), rightRef);
      }
    }
    if (getResource().getRoleReferences() != null) {
      for (ReferenceType roleRef : getResource().getRoleReferences().getRoleReference())
      {
        this.roleRefsByName.put(roleRef.getName(), roleRef);
      }
    }
    if (getResource().getNetworks() != null) {
      for (ReferenceType networkRef : getResource().getNetworks().getNetwork())
      {
        this.externalNetworkRefsByName.put(networkRef.getName(), networkRef);
      }
    }
    for (LinkType link : getResource().getLink())
      if ((link.getRel().equals("down")) && (link.getType().equals("application/vnd.vmware.admin.systemOrganization+xml")))
      {
        this.systemAdminOrgReference = new ReferenceType();
        this.systemAdminOrgReference = link;
        break;
      }
  }

  public AdminOrganization getSystemAdminOrg()
    throws VCloudException
  {
    if (this.systemAdminOrgReference == null) {
      throw new VCloudException(SdkUtil.getI18nString(SdkMessage.REFERENCE_NOT_FOUND_MSG));
    }

    return AdminOrganization.getAdminOrgByReference(this.client, this.systemAdminOrgReference);
  }

  public Role createRole(RoleType roleType)
    throws VCloudException
  {
    ObjectFactory of = new ObjectFactory();
    JAXBElement roleParams = of.createRole(roleType);
    String requestBodyXml = JAXBUtil.marshal(roleParams);
    String postURL = getResource().getHref() + "roles";

    return new Role(this.client, (RoleType)SdkUtil.post(this.client, postURL, requestBodyXml, "application/vnd.vmware.admin.role+xml", 201));
  }

  public AdminOrganization createAdminOrg(AdminOrgType adminOrgType)
    throws VCloudException
  {
    ObjectFactory of = new ObjectFactory();
    JAXBElement adminOrgParams = of.createAdminOrg(adminOrgType);

    String requestBodyXml = JAXBUtil.marshal(adminOrgParams);
    String postURL = getResource().getHref() + "orgs";

    return new AdminOrganization(this.client, (AdminOrgType)SdkUtil.post(this.client, postURL, requestBodyXml, "application/vnd.vmware.admin.organization+xml", 201));
  }

  public HashMap<String, ReferenceType> getAdminOrgRefsByName()
  {
    return this.adminOrganizationRefsByName;
  }

  public Collection<ReferenceType> getAdminOrgRefs()
  {
    return this.adminOrganizationRefsByName.values();
  }

  public ReferenceType getAdminOrgRefByName(String name)
  {
    return (ReferenceType)this.adminOrganizationRefsByName.get(name);
  }

  public HashMap<String, ReferenceType> getProviderVdcRefsByName()
  {
    return this.providerVdcRefsByName;
  }

  public Collection<ReferenceType> getProviderVdcRefs()
  {
    return this.providerVdcRefsByName.values();
  }

  public ReferenceType getProviderVdcRefByName(String name)
  {
    return (ReferenceType)this.providerVdcRefsByName.get(name);
  }

  public HashMap<String, ReferenceType> getRightRefsByName()
  {
    return this.rightRefsByName;
  }

  public Collection<ReferenceType> getRightRefs()
  {
    return this.rightRefsByName.values();
  }

  public ReferenceType getRightRefByName(String name)
  {
    return (ReferenceType)this.rightRefsByName.get(name);
  }

  public HashMap<String, ReferenceType> getRoleRefsByName()
  {
    return this.roleRefsByName;
  }

  public Collection<ReferenceType> getRoleRefs()
  {
    return this.roleRefsByName.values();
  }

  public ReferenceType getRoleRefByName(String name)
  {
    return (ReferenceType)this.roleRefsByName.get(name);
  }

  public HashMap<String, ReferenceType> getExternalNetworkRefsByName()
  {
    return this.externalNetworkRefsByName;
  }

  public Collection<ReferenceType> getExternalNetworkRefs()
  {
    return this.externalNetworkRefsByName.values();
  }

  public ReferenceType getExternalNetworkRefByName(String name)
  {
    return (ReferenceType)this.externalNetworkRefsByName.get(name);
  }
}