package com.vmware.vcloud.sdk_v1_0;

import com.vmware.vcloud.api.rest.schema.versioning.SupportedVersionsType;
import com.vmware.vcloud.api.rest.schema.versioning.VersionInfoType;
import com.vmware.vcloud.api.rest.schema_v1_0.OrgListType;
import com.vmware.vcloud.api.rest.schema_v1_0.ReferenceType;
import com.vmware.vcloud.sdk_v1_0.admin.VcloudAdmin;
import com.vmware.vcloud.sdk_v1_0.admin.extensions.VcloudAdminExtension;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.httpclient.HostConfiguration;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpConnectionManager;
import org.apache.commons.httpclient.HttpState;
import org.apache.commons.httpclient.MultiThreadedHttpConnectionManager;
import org.apache.commons.httpclient.UsernamePasswordCredentials;
import org.apache.commons.httpclient.auth.AuthScope;
import org.apache.commons.httpclient.auth.BasicScheme;
import org.apache.commons.httpclient.params.HttpConnectionManagerParams;

public class VcloudClient
{
  private static Logger logger = Logger.getLogger(VcloudClient.class.getSimpleName());
  private String apiVersionsUrl;
  private String vCloudApiURL;
  private HttpClient client;
  private OrgListType orgListType;
  private HashMap<String, VersionInfoType> supportedVersions;
  private HashMap<String, ReferenceType> orgRefsByName;
  private String vCloudAuthenticationHeaderName;
  private String vCloudAuthenticationHeaderValue;

  String getVCloudAuthenticationHeaderName()
    throws VCloudException
  {
    if (this.vCloudAuthenticationHeaderName != null)
      return this.vCloudAuthenticationHeaderName;
    throw new VCloudException("Login and try the operation");
  }

  void setVCloudAuthenticationHeaderName(String cloudAuthenticationHeaderName)
  {
    this.vCloudAuthenticationHeaderName = cloudAuthenticationHeaderName;
  }

  String getVCloudAuthenticationHeaderValue()
    throws VCloudException
  {
    if (this.vCloudAuthenticationHeaderValue != null)
      return this.vCloudAuthenticationHeaderValue;
    throw new VCloudException("Login and try the operation");
  }

  void setVCloudAuthenticationHeaderValue(String cloudAuthenticationHeaderValue)
  {
    this.vCloudAuthenticationHeaderValue = cloudAuthenticationHeaderValue;
  }

  public VcloudClient(String apiVersionsUrl)
  {
    try
    {
      new URL(apiVersionsUrl);
    } catch (MalformedURLException e) {
      throw new VCloudRuntimeException(e);
    }
    this.apiVersionsUrl = apiVersionsUrl;
    MultiThreadedHttpConnectionManager connectionManager = new MultiThreadedHttpConnectionManager();
    this.client = new HttpClient(connectionManager);
  }

  public void setConnectionTimeout(int milliseconds)
  {
    this.client.getHttpConnectionManager().getParams().setConnectionTimeout(milliseconds);
  }

  public void setSocketTimeout(int milliseconds)
  {
    this.client.getHttpConnectionManager().getParams().setSoTimeout(milliseconds);
  }

  public HashMap<String, VersionInfoType> getSupportedVersions()
    throws VCloudException
  {
    this.supportedVersions = new HashMap();

    Response response = RestUtil.getSupportedVersions(this, this.apiVersionsUrl);

    SupportedVersionsType supportedVersionsType = null;
    if (response.isExpected(200)) {
      supportedVersionsType = (SupportedVersionsType)response.getResource();
    }
    else
    {
      response.handleUnExpectedResponse();
    }
    if (supportedVersionsType != null)
      for (VersionInfoType version : supportedVersionsType.getVersionInfo())
      {
        this.supportedVersions.put(version.getVersion(), version);
      }
    return this.supportedVersions;
  }

  public void setMaxConnections(int maxConnections)
  {
    this.client.getHttpConnectionManager().getParams().setDefaultMaxConnectionsPerHost(maxConnections);

    if (maxConnections > 20)
      this.client.getHttpConnectionManager().getParams().setMaxTotalConnections(maxConnections);
  }

  public void setProxy(String proxyHost, int port)
  {
    this.client.getHostConfiguration().setProxy(proxyHost, port);
  }

  public void setProxyCredentails(String username, String password)
  {
    this.client.getState().setProxyCredentials(AuthScope.ANY, new UsernamePasswordCredentials(username, password));
  }

  public HashMap<String, ReferenceType> login(String userName, String password, VersionInfoType versionInfo)
    throws VCloudException
  {
    if (versionInfo == null) {
      throw new VCloudException("VersionInfo is Null");
    }

    setVCloudApiURL(versionInfo.getLoginUrl().replace("/login", ""));

    String authHeaderValue = BasicScheme.authenticate(new UsernamePasswordCredentials(userName, password), "UTF-8");

    Response response = RestUtil.login(this, versionInfo.getLoginUrl(), authHeaderValue);

    this.client.getState().clearCookies();

    if (response.isExpected(200)) {
      this.orgListType = ((OrgListType)response.getResource());
      parseOrgLists();
    }
    else {
      response.handleUnExpectedResponse();
    }

    return this.orgRefsByName;
  }

  public void logout()
    throws VCloudException
  {
    Response response = RestUtil.post(this, getVcloudApiURL() + "/logout", null, null, null);

    setVCloudAuthenticationHeaderName(null);
    setVCloudAuthenticationHeaderValue(null);
    if (!response.isExpected(200))
      response.handleUnExpectedResponse();
    logger.log(Level.INFO, "Successfully logged out");
  }

  public HashMap<String, ReferenceType> getUpdatedOrgList()
    throws VCloudException
  {
    String orgListUrl = getVcloudApiURL() + "/org/";

    logger.log(Level.INFO, orgListUrl);

    Response response = RestUtil.get(this, orgListUrl);

    if (response.isExpected(200)) {
      this.orgListType = ((OrgListType)response.getResource());
      parseOrgLists();
    }
    else {
      response.handleUnExpectedResponse();
    }
    return this.orgRefsByName;
  }

  private void parseOrgLists()
  {
    this.orgRefsByName = new HashMap();
    if ((this.orgListType == null) || (this.orgListType.getOrg().size() == 0)) {
      logger.log(Level.WARNING, "Nothing to parse the Org Lists");
      return;
    }
    for (ReferenceType ref : this.orgListType.getOrg())
      this.orgRefsByName.put(ref.getName(), ref);
  }

  public Organization getSystemOrg()
    throws VCloudException
  {
    ReferenceType systemAdminRef = new ReferenceType();
    systemAdminRef.setHref(getVcloudApiURL() + "/org/" + "1");

    return Organization.getOrganizationByReference(this, systemAdminRef);
  }

  public String getVcloudApiURL()
  {
    return this.vCloudApiURL;
  }

  void setVCloudApiURL(String cloudApiURL)
  {
    this.vCloudApiURL = cloudApiURL;
  }

  HttpClient getClient()
  {
    return this.client;
  }

  public Collection<ReferenceType> getOrgRefs()
  {
    return this.orgRefsByName.values();
  }

  public HashMap<String, ReferenceType> getOrgRefsByName()
  {
    return this.orgRefsByName;
  }

  public ReferenceType getOrgRefByName(String name)
  {
    return (ReferenceType)this.orgRefsByName.get(name);
  }

  public VcloudAdmin getVcloudAdmin()
    throws VCloudException
  {
    return new VcloudAdmin(this);
  }

  public VcloudAdminExtension getVcloudAdminExtension()
    throws VCloudException
  {
    return new VcloudAdminExtension(this);
  }
}