package org.eclipse.jdt.internal.compiler;

import java.util.ArrayList;
import java.util.HashMap;
import org.eclipse.jdt.core.compiler.CategorizedProblem;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.AllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.Annotation;
import org.eclipse.jdt.internal.compiler.ast.AnnotationMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.ast.ArrayAllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.ArrayInitializer;
import org.eclipse.jdt.internal.compiler.ast.ArrayQualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.ArrayReference;
import org.eclipse.jdt.internal.compiler.ast.ArrayTypeReference;
import org.eclipse.jdt.internal.compiler.ast.Assignment;
import org.eclipse.jdt.internal.compiler.ast.Block;
import org.eclipse.jdt.internal.compiler.ast.ClassLiteralAccess;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ConstructorDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ExplicitConstructorCall;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.FieldDeclaration;
import org.eclipse.jdt.internal.compiler.ast.FieldReference;
import org.eclipse.jdt.internal.compiler.ast.ImportReference;
import org.eclipse.jdt.internal.compiler.ast.Initializer;
import org.eclipse.jdt.internal.compiler.ast.Javadoc;
import org.eclipse.jdt.internal.compiler.ast.JavadocAllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.JavadocFieldReference;
import org.eclipse.jdt.internal.compiler.ast.JavadocMessageSend;
import org.eclipse.jdt.internal.compiler.ast.JavadocQualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.JavadocSingleTypeReference;
import org.eclipse.jdt.internal.compiler.ast.MemberValuePair;
import org.eclipse.jdt.internal.compiler.ast.MessageSend;
import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.NameReference;
import org.eclipse.jdt.internal.compiler.ast.ParameterizedQualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.ParameterizedSingleTypeReference;
import org.eclipse.jdt.internal.compiler.ast.QualifiedAllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.QualifiedNameReference;
import org.eclipse.jdt.internal.compiler.ast.QualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.SingleMemberAnnotation;
import org.eclipse.jdt.internal.compiler.ast.SingleNameReference;
import org.eclipse.jdt.internal.compiler.ast.SingleTypeReference;
import org.eclipse.jdt.internal.compiler.ast.Statement;
import org.eclipse.jdt.internal.compiler.ast.ThisReference;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeParameter;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.env.ICompilationUnit;
import org.eclipse.jdt.internal.compiler.env.ISourceType;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.parser.JavadocParser;
import org.eclipse.jdt.internal.compiler.parser.RecoveredElement;
import org.eclipse.jdt.internal.compiler.parser.Scanner;
import org.eclipse.jdt.internal.compiler.parser.SourceTypeConverter;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.HashtableOfObjectToInt;
import org.eclipse.jdt.internal.core.util.CommentRecorderParser;

public class SourceElementParser extends CommentRecorderParser
{
  public ISourceElementRequestor requestor;
  ISourceType sourceType;
  boolean reportReferenceInfo;
  char[][] typeNames;
  char[][] superTypeNames;
  int nestedTypeIndex;
  LocalDeclarationVisitor localDeclarationVisitor = null;
  CompilerOptions options;
  HashtableOfObjectToInt sourceEnds = new HashtableOfObjectToInt();
  HashMap nodesToCategories = new HashMap();
  boolean useSourceJavadocParser = true;

  public SourceElementParser(ISourceElementRequestor requestor, IProblemFactory problemFactory, CompilerOptions options, boolean reportLocalDeclarations, boolean optimizeStringLiterals)
  {
    this(requestor, problemFactory, options, reportLocalDeclarations, optimizeStringLiterals, true);
  }

  public SourceElementParser(ISourceElementRequestor requestor, IProblemFactory problemFactory, CompilerOptions options, boolean reportLocalDeclarations, boolean optimizeStringLiterals, boolean useSourceJavadocParser)
  {
    super(new ProblemReporter(
      DefaultErrorHandlingPolicies.exitAfterAllProblems(), 
      options, 
      problemFactory), 
      optimizeStringLiterals);

    this.problemReporter = new ProblemReporter(
      DefaultErrorHandlingPolicies.exitAfterAllProblems(), 
      options, 
      problemFactory) {
      public void record(CategorizedProblem problem, CompilationResult unitResult, ReferenceContext context) {
        unitResult.record(problem, context);
        SourceElementParser.this.requestor.acceptProblem(problem);
      }
    };
    this.requestor = requestor;
    this.typeNames = new char[4][];
    this.superTypeNames = new char[4][];
    this.nestedTypeIndex = 0;
    this.options = options;
    if (reportLocalDeclarations) {
      this.localDeclarationVisitor = new LocalDeclarationVisitor();
    }

    this.useSourceJavadocParser = useSourceJavadocParser;
    if (useSourceJavadocParser)
      this.javadocParser = new SourceJavadocParser(this);
  }

  private void acceptJavadocTypeReference(Expression expression) {
    if ((expression instanceof JavadocSingleTypeReference)) {
      JavadocSingleTypeReference singleRef = (JavadocSingleTypeReference)expression;
      this.requestor.acceptTypeReference(singleRef.token, singleRef.sourceStart);
    } else if ((expression instanceof JavadocQualifiedTypeReference)) {
      JavadocQualifiedTypeReference qualifiedRef = (JavadocQualifiedTypeReference)expression;
      this.requestor.acceptTypeReference(qualifiedRef.tokens, qualifiedRef.sourceStart, qualifiedRef.sourceEnd);
    }
  }

  public void addUnknownRef(NameReference nameRef)
  {
    if ((nameRef instanceof SingleNameReference)) {
      this.requestor.acceptUnknownReference(((SingleNameReference)nameRef).token, nameRef.sourceStart);
    }
    else
      this.requestor.acceptUnknownReference(((QualifiedNameReference)nameRef).tokens, nameRef.sourceStart, nameRef.sourceEnd);
  }

  public void checkComment()
  {
    if (((!this.diet) || (this.dietInt != 0)) && (this.scanner.commentPtr >= 0)) {
      flushCommentsDefinedPriorTo(this.endStatementPosition);
    }

    int lastComment = this.scanner.commentPtr;

    if (this.modifiersSourceStart >= 0)
    {
      while ((lastComment >= 0) && (Math.abs(this.scanner.commentStarts[lastComment]) > this.modifiersSourceStart)) lastComment--;
    }
    if (lastComment >= 0)
    {
      this.modifiersSourceStart = Math.abs(this.scanner.commentStarts[0]);

      while ((lastComment >= 0) && (this.scanner.commentStops[lastComment] < 0)) lastComment--;
      if ((lastComment >= 0) && (this.javadocParser != null)) {
        int commentEnd = this.scanner.commentStops[lastComment] - 1;

        this.javadocParser.reportProblems = ((this.currentElement == null) || (commentEnd > this.lastJavadocEnd));
        if (this.javadocParser.checkDeprecation(lastComment)) {
          checkAndSetModifiers(1048576);
        }
        this.javadoc = this.javadocParser.docComment;
        if (this.currentElement == null) this.lastJavadocEnd = commentEnd;
      }
    }

    if ((this.reportReferenceInfo) && (this.javadocParser.checkDocComment) && (this.javadoc != null))
    {
      TypeReference[] thrownExceptions = this.javadoc.exceptionReferences;
      if (thrownExceptions != null) {
        int i = 0; for (int max = thrownExceptions.length; i < max; i++) {
          TypeReference typeRef = thrownExceptions[i];
          if ((typeRef instanceof JavadocSingleTypeReference)) {
            JavadocSingleTypeReference singleRef = (JavadocSingleTypeReference)typeRef;
            this.requestor.acceptTypeReference(singleRef.token, singleRef.sourceStart);
          } else if ((typeRef instanceof JavadocQualifiedTypeReference)) {
            JavadocQualifiedTypeReference qualifiedRef = (JavadocQualifiedTypeReference)typeRef;
            this.requestor.acceptTypeReference(qualifiedRef.tokens, qualifiedRef.sourceStart, qualifiedRef.sourceEnd);
          }
        }

      }

      Expression[] references = this.javadoc.seeReferences;
      if (references != null) {
        int i = 0; for (int max = references.length; i < max; i++) {
          Expression reference = references[i];
          acceptJavadocTypeReference(reference);
          if ((reference instanceof JavadocFieldReference)) {
            JavadocFieldReference fieldRef = (JavadocFieldReference)reference;
            this.requestor.acceptFieldReference(fieldRef.token, fieldRef.sourceStart);
            if ((fieldRef.receiver != null) && (!fieldRef.receiver.isThis()))
              acceptJavadocTypeReference(fieldRef.receiver);
          }
          else if ((reference instanceof JavadocMessageSend)) {
            JavadocMessageSend messageSend = (JavadocMessageSend)reference;
            int argCount = messageSend.arguments == null ? 0 : messageSend.arguments.length;
            this.requestor.acceptMethodReference(messageSend.selector, argCount, messageSend.sourceStart);
            this.requestor.acceptConstructorReference(messageSend.selector, argCount, messageSend.sourceStart);
            if ((messageSend.receiver != null) && (!messageSend.receiver.isThis()))
              acceptJavadocTypeReference(messageSend.receiver);
          }
          else if ((reference instanceof JavadocAllocationExpression)) {
            JavadocAllocationExpression constructor = (JavadocAllocationExpression)reference;
            int argCount = constructor.arguments == null ? 0 : constructor.arguments.length;
            if (constructor.type != null) {
              char[][] compoundName = constructor.type.getParameterizedTypeName();
              this.requestor.acceptConstructorReference(compoundName[(compoundName.length - 1)], argCount, constructor.sourceStart);
              if (!constructor.type.isThis())
                acceptJavadocTypeReference(constructor.type);
            }
          }
        }
      }
    }
  }

  protected void classInstanceCreation(boolean alwaysQualified)
  {
    boolean previousFlag = this.reportReferenceInfo;
    this.reportReferenceInfo = false;
    super.classInstanceCreation(alwaysQualified);
    this.reportReferenceInfo = previousFlag;
    if (this.reportReferenceInfo) {
      AllocationExpression alloc = (AllocationExpression)this.expressionStack[this.expressionPtr];
      TypeReference typeRef = alloc.type;
      this.requestor.acceptConstructorReference(
        (typeRef instanceof SingleTypeReference) ? 
        ((SingleTypeReference)typeRef).token : 
        CharOperation.concatWith(alloc.type.getParameterizedTypeName(), '.'), 
        alloc.arguments == null ? 0 : alloc.arguments.length, 
        alloc.sourceStart);
    }
  }

  private long[] collectAnnotationPositions(Annotation[] annotations) { if (annotations == null) return null;
    int length = annotations.length;
    long[] result = new long[length];
    for (int i = 0; i < length; i++) {
      Annotation annotation = annotations[i];
      result[i] = ((annotation.sourceStart << 32) + annotation.declarationSourceEnd);
    }
    return result; }

  protected void consumeAnnotationAsModifier() {
    super.consumeAnnotationAsModifier();
    Annotation annotation = (Annotation)this.expressionStack[this.expressionPtr];
    if (this.reportReferenceInfo)
      this.requestor.acceptTypeReference(annotation.type.getTypeName(), annotation.sourceStart, annotation.sourceEnd);
  }

  protected void consumeClassInstanceCreationExpressionQualifiedWithTypeArguments() {
    boolean previousFlag = this.reportReferenceInfo;
    this.reportReferenceInfo = false;
    super.consumeClassInstanceCreationExpressionQualifiedWithTypeArguments();
    this.reportReferenceInfo = previousFlag;
    if (this.reportReferenceInfo) {
      AllocationExpression alloc = (AllocationExpression)this.expressionStack[this.expressionPtr];
      TypeReference typeRef = alloc.type;
      this.requestor.acceptConstructorReference(
        (typeRef instanceof SingleTypeReference) ? 
        ((SingleTypeReference)typeRef).token : 
        CharOperation.concatWith(alloc.type.getParameterizedTypeName(), '.'), 
        alloc.arguments == null ? 0 : alloc.arguments.length, 
        alloc.sourceStart);
    }
  }

  protected void consumeAnnotationTypeDeclarationHeaderName() { int currentAstPtr = this.astPtr;
    super.consumeAnnotationTypeDeclarationHeaderName();
    if (this.astPtr > currentAstPtr)
      rememberCategories(); }

  protected void consumeClassHeaderName1() {
    int currentAstPtr = this.astPtr;
    super.consumeClassHeaderName1();
    if (this.astPtr > currentAstPtr)
      rememberCategories(); 
  }

  protected void consumeClassInstanceCreationExpressionWithTypeArguments() { boolean previousFlag = this.reportReferenceInfo;
    this.reportReferenceInfo = false;
    super.consumeClassInstanceCreationExpressionWithTypeArguments();
    this.reportReferenceInfo = previousFlag;
    if (this.reportReferenceInfo) {
      AllocationExpression alloc = (AllocationExpression)this.expressionStack[this.expressionPtr];
      TypeReference typeRef = alloc.type;
      this.requestor.acceptConstructorReference(
        (typeRef instanceof SingleTypeReference) ? 
        ((SingleTypeReference)typeRef).token : 
        CharOperation.concatWith(alloc.type.getParameterizedTypeName(), '.'), 
        alloc.arguments == null ? 0 : alloc.arguments.length, 
        alloc.sourceStart);
    } }

  protected void consumeConstructorHeaderName() {
    long selectorSourcePositions = this.identifierPositionStack[this.identifierPtr];
    int selectorSourceEnd = (int)selectorSourcePositions;
    int currentAstPtr = this.astPtr;
    super.consumeConstructorHeaderName();
    if (this.astPtr > currentAstPtr) {
      this.sourceEnds.put(this.astStack[this.astPtr], selectorSourceEnd);
      rememberCategories();
    }
  }

  protected void consumeConstructorHeaderNameWithTypeParameters() { long selectorSourcePositions = this.identifierPositionStack[this.identifierPtr];
    int selectorSourceEnd = (int)selectorSourcePositions;
    int currentAstPtr = this.astPtr;
    super.consumeConstructorHeaderNameWithTypeParameters();
    if (this.astPtr > currentAstPtr) {
      this.sourceEnds.put(this.astStack[this.astPtr], selectorSourceEnd);
      rememberCategories();
    } }

  protected void consumeEnumConstantWithClassBody() {
    super.consumeEnumConstantWithClassBody();
    if (((this.currentToken == 30) || (this.currentToken == 27)) && 
      ((this.astStack[this.astPtr] instanceof FieldDeclaration))) {
      this.sourceEnds.put(this.astStack[this.astPtr], this.scanner.currentPosition - 1);
      rememberCategories();
    }
  }

  protected void consumeEnumConstantNoClassBody() { super.consumeEnumConstantNoClassBody();
    if (((this.currentToken == 30) || (this.currentToken == 27)) && 
      ((this.astStack[this.astPtr] instanceof FieldDeclaration))) {
      this.sourceEnds.put(this.astStack[this.astPtr], this.scanner.currentPosition - 1);
      rememberCategories();
    } }

  protected void consumeEnumHeaderName() {
    int currentAstPtr = this.astPtr;
    super.consumeEnumHeaderName();
    if (this.astPtr > currentAstPtr)
      rememberCategories();
  }

  protected void consumeExitVariableWithInitialization()
  {
    super.consumeExitVariableWithInitialization();
    if (((this.currentToken == 30) || (this.currentToken == 27)) && 
      ((this.astStack[this.astPtr] instanceof FieldDeclaration))) {
      this.sourceEnds.put(this.astStack[this.astPtr], this.scanner.currentPosition - 1);
      rememberCategories();
    }
  }

  protected void consumeExitVariableWithoutInitialization()
  {
    super.consumeExitVariableWithoutInitialization();
    if (((this.currentToken == 30) || (this.currentToken == 27)) && 
      ((this.astStack[this.astPtr] instanceof FieldDeclaration))) {
      this.sourceEnds.put(this.astStack[this.astPtr], this.scanner.currentPosition - 1);
      rememberCategories();
    }
  }

  protected void consumeFieldAccess(boolean isSuperAccess)
  {
    super.consumeFieldAccess(isSuperAccess);
    FieldReference fr = (FieldReference)this.expressionStack[this.expressionPtr];
    if (this.reportReferenceInfo)
      this.requestor.acceptFieldReference(fr.token, fr.sourceStart);
  }

  protected void consumeFormalParameter(boolean isVarArgs) {
    super.consumeFormalParameter(isVarArgs);

    flushCommentsDefinedPriorTo(this.scanner.currentPosition);
  }
  protected void consumeInterfaceHeaderName1() {
    int currentAstPtr = this.astPtr;
    super.consumeInterfaceHeaderName1();
    if (this.astPtr > currentAstPtr)
      rememberCategories(); 
  }

  protected void consumeMemberValuePair() { super.consumeMemberValuePair();
    MemberValuePair memberValuepair = (MemberValuePair)this.astStack[this.astPtr];
    if (this.reportReferenceInfo)
      this.requestor.acceptMethodReference(memberValuepair.name, 0, memberValuepair.sourceStart); }

  protected void consumeMarkerAnnotation()
  {
    super.consumeMarkerAnnotation();
    Annotation annotation = (Annotation)this.expressionStack[this.expressionPtr];
    if (this.reportReferenceInfo)
      this.requestor.acceptTypeReference(annotation.type.getTypeName(), annotation.sourceStart, annotation.sourceEnd);
  }

  protected void consumeMethodHeaderName(boolean isAnnotationMethod) {
    long selectorSourcePositions = this.identifierPositionStack[this.identifierPtr];
    int selectorSourceEnd = (int)selectorSourcePositions;
    int currentAstPtr = this.astPtr;
    super.consumeMethodHeaderName(isAnnotationMethod);
    if (this.astPtr > currentAstPtr) {
      this.sourceEnds.put(this.astStack[this.astPtr], selectorSourceEnd);
      rememberCategories();
    }
  }

  protected void consumeMethodHeaderNameWithTypeParameters(boolean isAnnotationMethod) {
    long selectorSourcePositions = this.identifierPositionStack[this.identifierPtr];
    int selectorSourceEnd = (int)selectorSourcePositions;
    int currentAstPtr = this.astPtr;
    super.consumeMethodHeaderNameWithTypeParameters(isAnnotationMethod);
    if (this.astPtr > currentAstPtr)
      this.sourceEnds.put(this.astStack[this.astPtr], selectorSourceEnd);
    rememberCategories();
  }

  protected void consumeMethodInvocationName()
  {
    super.consumeMethodInvocationName();

    MessageSend messageSend = (MessageSend)this.expressionStack[this.expressionPtr];
    Expression[] args = messageSend.arguments;
    if (this.reportReferenceInfo)
      this.requestor.acceptMethodReference(
        messageSend.selector, 
        args == null ? 0 : args.length, 
        (int)(messageSend.nameSourcePosition >>> 32));
  }

  protected void consumeMethodInvocationNameWithTypeArguments()
  {
    super.consumeMethodInvocationNameWithTypeArguments();

    MessageSend messageSend = (MessageSend)this.expressionStack[this.expressionPtr];
    Expression[] args = messageSend.arguments;
    if (this.reportReferenceInfo)
      this.requestor.acceptMethodReference(
        messageSend.selector, 
        args == null ? 0 : args.length, 
        (int)(messageSend.nameSourcePosition >>> 32));
  }

  protected void consumeMethodInvocationPrimary()
  {
    super.consumeMethodInvocationPrimary();
    MessageSend messageSend = (MessageSend)this.expressionStack[this.expressionPtr];
    Expression[] args = messageSend.arguments;
    if (this.reportReferenceInfo)
      this.requestor.acceptMethodReference(
        messageSend.selector, 
        args == null ? 0 : args.length, 
        (int)(messageSend.nameSourcePosition >>> 32));
  }

  protected void consumeMethodInvocationPrimaryWithTypeArguments()
  {
    super.consumeMethodInvocationPrimaryWithTypeArguments();
    MessageSend messageSend = (MessageSend)this.expressionStack[this.expressionPtr];
    Expression[] args = messageSend.arguments;
    if (this.reportReferenceInfo)
      this.requestor.acceptMethodReference(
        messageSend.selector, 
        args == null ? 0 : args.length, 
        (int)(messageSend.nameSourcePosition >>> 32));
  }

  protected void consumeMethodInvocationSuper()
  {
    super.consumeMethodInvocationSuper();
    MessageSend messageSend = (MessageSend)this.expressionStack[this.expressionPtr];
    Expression[] args = messageSend.arguments;
    if (this.reportReferenceInfo)
      this.requestor.acceptMethodReference(
        messageSend.selector, 
        args == null ? 0 : args.length, 
        (int)(messageSend.nameSourcePosition >>> 32));
  }

  protected void consumeMethodInvocationSuperWithTypeArguments()
  {
    super.consumeMethodInvocationSuperWithTypeArguments();
    MessageSend messageSend = (MessageSend)this.expressionStack[this.expressionPtr];
    Expression[] args = messageSend.arguments;
    if (this.reportReferenceInfo)
      this.requestor.acceptMethodReference(
        messageSend.selector, 
        args == null ? 0 : args.length, 
        (int)(messageSend.nameSourcePosition >>> 32));
  }

  protected void consumeNormalAnnotation() {
    super.consumeNormalAnnotation();
    Annotation annotation = (Annotation)this.expressionStack[this.expressionPtr];
    if (this.reportReferenceInfo)
      this.requestor.acceptTypeReference(annotation.type.getTypeName(), annotation.sourceStart, annotation.sourceEnd);
  }

  protected void consumeSingleMemberAnnotation() {
    super.consumeSingleMemberAnnotation();
    SingleMemberAnnotation member = (SingleMemberAnnotation)this.expressionStack[this.expressionPtr];
    if (this.reportReferenceInfo)
      this.requestor.acceptMethodReference(TypeConstants.VALUE, 0, member.sourceStart);
  }

  protected void consumeSingleStaticImportDeclarationName()
  {
    int length;
    char[][] tokens = new char[length = this.identifierLengthStack[(this.identifierLengthPtr--)]][];
    this.identifierPtr -= length;
    long[] positions = new long[length];
    System.arraycopy(this.identifierStack, this.identifierPtr + 1, tokens, 0, length);
    System.arraycopy(this.identifierPositionStack, this.identifierPtr + 1, positions, 0, length);
    ImportReference impt;
    pushOnAstStack(impt = newImportReference(tokens, positions, false, 8));

    this.modifiers = 0;
    this.modifiersSourceStart = -1;

    if (this.currentToken == 27)
      impt.declarationSourceEnd = (this.scanner.currentPosition - 1);
    else {
      impt.declarationSourceEnd = impt.sourceEnd;
    }
    impt.declarationEnd = impt.declarationSourceEnd;

    impt.declarationSourceStart = this.intStack[(this.intPtr--)];

    if ((!this.statementRecoveryActivated) && 
      (this.options.sourceLevel < 3211264L) && 
      (this.lastErrorEndPositionBeforeRecovery < this.scanner.currentPosition)) {
      impt.modifiers = 0;
      problemReporter().invalidUsageOfStaticImports(impt);
    }

    if (this.currentElement != null) {
      this.lastCheckPoint = (impt.declarationSourceEnd + 1);
      this.currentElement = this.currentElement.add(impt, 0);
      this.lastIgnoredToken = -1;
      this.restartRecovery = true;
    }
    if (this.reportReferenceInfo)
    {
      int tokensLength = impt.tokens.length - 1;
      int start = (int)(impt.sourcePositions[tokensLength] >>> 32);
      char[] last = impt.tokens[tokensLength];

      this.requestor.acceptFieldReference(last, start);
      this.requestor.acceptMethodReference(last, 0, start);
      this.requestor.acceptTypeReference(last, start);

      if (tokensLength > 0) {
        char[][] compoundName = new char[tokensLength][];
        System.arraycopy(impt.tokens, 0, compoundName, 0, tokensLength);
        int end = (int)impt.sourcePositions[(tokensLength - 1)];
        this.requestor.acceptTypeReference(compoundName, impt.sourceStart, end);
      }
    }
  }

  protected void consumeSingleTypeImportDeclarationName()
  {
    int length;
    char[][] tokens = new char[length = this.identifierLengthStack[(this.identifierLengthPtr--)]][];
    this.identifierPtr -= length;
    long[] positions = new long[length];
    System.arraycopy(this.identifierStack, this.identifierPtr + 1, tokens, 0, length);
    System.arraycopy(this.identifierPositionStack, this.identifierPtr + 1, positions, 0, length);
    ImportReference impt;
    pushOnAstStack(impt = newImportReference(tokens, positions, false, 0));

    if (this.currentToken == 27)
      impt.declarationSourceEnd = (this.scanner.currentPosition - 1);
    else {
      impt.declarationSourceEnd = impt.sourceEnd;
    }
    impt.declarationEnd = impt.declarationSourceEnd;

    impt.declarationSourceStart = this.intStack[(this.intPtr--)];

    if (this.currentElement != null) {
      this.lastCheckPoint = (impt.declarationSourceEnd + 1);
      this.currentElement = this.currentElement.add(impt, 0);
      this.lastIgnoredToken = -1;
      this.restartRecovery = true;
    }
    if (this.reportReferenceInfo)
      this.requestor.acceptTypeReference(impt.tokens, impt.sourceStart, impt.sourceEnd);
  }

  protected void consumeStaticImportOnDemandDeclarationName()
  {
    int length;
    char[][] tokens = new char[length = this.identifierLengthStack[(this.identifierLengthPtr--)]][];
    this.identifierPtr -= length;
    long[] positions = new long[length];
    System.arraycopy(this.identifierStack, this.identifierPtr + 1, tokens, 0, length);
    System.arraycopy(this.identifierPositionStack, this.identifierPtr + 1, positions, 0, length);
    ImportReference impt;
    pushOnAstStack(impt = new ImportReference(tokens, positions, true, 8));

    this.modifiers = 0;
    this.modifiersSourceStart = -1;

    if (this.currentToken == 27)
      impt.declarationSourceEnd = (this.scanner.currentPosition - 1);
    else {
      impt.declarationSourceEnd = impt.sourceEnd;
    }
    impt.declarationEnd = impt.declarationSourceEnd;

    impt.declarationSourceStart = this.intStack[(this.intPtr--)];

    if ((!this.statementRecoveryActivated) && 
      (this.options.sourceLevel < 3211264L) && 
      (this.lastErrorEndPositionBeforeRecovery < this.scanner.currentPosition)) {
      impt.modifiers = 0;
      problemReporter().invalidUsageOfStaticImports(impt);
    }

    if (this.currentElement != null) {
      this.lastCheckPoint = (impt.declarationSourceEnd + 1);
      this.currentElement = this.currentElement.add(impt, 0);
      this.lastIgnoredToken = -1;
      this.restartRecovery = true;
    }
    if (this.reportReferenceInfo)
      this.requestor.acceptTypeReference(impt.tokens, impt.sourceStart, impt.sourceEnd);
  }

  protected void consumeTypeImportOnDemandDeclarationName()
  {
    int length;
    char[][] tokens = new char[length = this.identifierLengthStack[(this.identifierLengthPtr--)]][];
    this.identifierPtr -= length;
    long[] positions = new long[length];
    System.arraycopy(this.identifierStack, this.identifierPtr + 1, tokens, 0, length);
    System.arraycopy(this.identifierPositionStack, this.identifierPtr + 1, positions, 0, length);
    ImportReference impt;
    pushOnAstStack(impt = new ImportReference(tokens, positions, true, 0));

    if (this.currentToken == 27)
      impt.declarationSourceEnd = (this.scanner.currentPosition - 1);
    else {
      impt.declarationSourceEnd = impt.sourceEnd;
    }
    impt.declarationEnd = impt.declarationSourceEnd;

    impt.declarationSourceStart = this.intStack[(this.intPtr--)];

    if (this.currentElement != null) {
      this.lastCheckPoint = (impt.declarationSourceEnd + 1);
      this.currentElement = this.currentElement.add(impt, 0);
      this.lastIgnoredToken = -1;
      this.restartRecovery = true;
    }
    if (this.reportReferenceInfo)
      this.requestor.acceptUnknownReference(impt.tokens, impt.sourceStart, impt.sourceEnd);
  }

  public MethodDeclaration convertToMethodDeclaration(ConstructorDeclaration c, CompilationResult compilationResult) {
    MethodDeclaration methodDeclaration = super.convertToMethodDeclaration(c, compilationResult);
    int selectorSourceEnd = this.sourceEnds.removeKey(c);
    if (selectorSourceEnd != -1)
      this.sourceEnds.put(methodDeclaration, selectorSourceEnd);
    char[][] categories = (char[][])this.nodesToCategories.remove(c);
    if (categories != null) {
      this.nodesToCategories.put(methodDeclaration, categories);
    }
    return methodDeclaration;
  }
  protected CompilationUnitDeclaration endParse(int act) {
    if (this.sourceType != null) {
      switch (TypeDeclaration.kind(this.sourceType.getModifiers())) {
      case 1:
        consumeClassDeclaration();
        break;
      case 2:
        consumeInterfaceDeclaration();
        break;
      case 3:
        consumeEnumDeclaration();
        break;
      case 4:
        consumeAnnotationTypeDeclaration();
      }
    }

    if (this.compilationUnit != null) {
      CompilationUnitDeclaration result = super.endParse(act);
      return result;
    }
    return null;
  }

  private ISourceElementRequestor.TypeParameterInfo[] getTypeParameterInfos(TypeParameter[] typeParameters) {
    if (typeParameters == null) return null;
    int typeParametersLength = typeParameters.length;
    ISourceElementRequestor.TypeParameterInfo[] result = new ISourceElementRequestor.TypeParameterInfo[typeParametersLength];
    for (int i = 0; i < typeParametersLength; i++) {
      TypeParameter typeParameter = typeParameters[i];
      TypeReference firstBound = typeParameter.type;
      TypeReference[] otherBounds = typeParameter.bounds;
      char[][] typeParameterBounds = (char[][])null;
      if (firstBound != null) {
        if (otherBounds != null) {
          int otherBoundsLength = otherBounds.length;
          char[][] boundNames = new char[otherBoundsLength + 1][];
          boundNames[0] = CharOperation.concatWith(firstBound.getParameterizedTypeName(), '.');
          for (int j = 0; j < otherBoundsLength; j++) {
            boundNames[(j + 1)] = 
              CharOperation.concatWith(otherBounds[j].getParameterizedTypeName(), '.');
          }
          typeParameterBounds = boundNames;
        } else {
          typeParameterBounds = new char[][] { CharOperation.concatWith(firstBound.getParameterizedTypeName(), '.') };
        }
      }
      else typeParameterBounds = CharOperation.NO_CHAR_CHAR;

      ISourceElementRequestor.TypeParameterInfo typeParameterInfo = new ISourceElementRequestor.TypeParameterInfo();
      typeParameterInfo.declarationStart = typeParameter.declarationSourceStart;
      typeParameterInfo.declarationEnd = typeParameter.declarationSourceEnd;
      typeParameterInfo.name = typeParameter.name;
      typeParameterInfo.nameSourceStart = typeParameter.sourceStart;
      typeParameterInfo.nameSourceEnd = typeParameter.sourceEnd;
      typeParameterInfo.bounds = typeParameterBounds;
      result[i] = typeParameterInfo;
    }
    return result;
  }

  public TypeReference getTypeReference(int dim)
  {
    int length = this.identifierLengthStack[(this.identifierLengthPtr--)];
    if (length < 0) {
      TypeReference ref = TypeReference.baseTypeReference(-length, dim);
      ref.sourceStart = this.intStack[(this.intPtr--)];
      if (dim == 0) {
        ref.sourceEnd = this.intStack[(this.intPtr--)];
      } else {
        this.intPtr -= 1;
        ref.sourceEnd = this.endPosition;
      }
      if (this.reportReferenceInfo) {
        this.requestor.acceptTypeReference(ref.getParameterizedTypeName(), ref.sourceStart, ref.sourceEnd);
      }
      return ref;
    }
    int numberOfIdentifiers = this.genericsIdentifiersLengthStack[(this.genericsIdentifiersLengthPtr--)];
    if ((length != numberOfIdentifiers) || (this.genericsLengthStack[this.genericsLengthPtr] != 0))
    {
      TypeReference ref = getTypeReferenceForGenericType(dim, length, numberOfIdentifiers);
      if (this.reportReferenceInfo) {
        if ((length == 1) && (numberOfIdentifiers == 1)) {
          ParameterizedSingleTypeReference parameterizedSingleTypeReference = (ParameterizedSingleTypeReference)ref;
          this.requestor.acceptTypeReference(parameterizedSingleTypeReference.token, parameterizedSingleTypeReference.sourceStart);
        } else {
          ParameterizedQualifiedTypeReference parameterizedQualifiedTypeReference = (ParameterizedQualifiedTypeReference)ref;
          this.requestor.acceptTypeReference(parameterizedQualifiedTypeReference.tokens, parameterizedQualifiedTypeReference.sourceStart, parameterizedQualifiedTypeReference.sourceEnd);
        }
      }
      return ref;
    }if (length == 1)
    {
      this.genericsLengthPtr -= 1;
      if (dim == 0) {
        SingleTypeReference ref = 
          new SingleTypeReference(
          this.identifierStack[this.identifierPtr], 
          this.identifierPositionStack[(this.identifierPtr--)]);
        if (this.reportReferenceInfo) {
          this.requestor.acceptTypeReference(ref.token, ref.sourceStart);
        }
        return ref;
      }
      ArrayTypeReference ref = 
        new ArrayTypeReference(
        this.identifierStack[this.identifierPtr], 
        dim, 
        this.identifierPositionStack[(this.identifierPtr--)]);
      ref.sourceEnd = this.endPosition;
      if (this.reportReferenceInfo) {
        this.requestor.acceptTypeReference(ref.token, ref.sourceStart);
      }
      return ref;
    }

    this.genericsLengthPtr -= 1;
    char[][] tokens = new char[length][];
    this.identifierPtr -= length;
    long[] positions = new long[length];
    System.arraycopy(this.identifierStack, this.identifierPtr + 1, tokens, 0, length);
    System.arraycopy(
      this.identifierPositionStack, 
      this.identifierPtr + 1, 
      positions, 
      0, 
      length);
    if (dim == 0) {
      QualifiedTypeReference ref = new QualifiedTypeReference(tokens, positions);
      if (this.reportReferenceInfo) {
        this.requestor.acceptTypeReference(ref.tokens, ref.sourceStart, ref.sourceEnd);
      }
      return ref;
    }
    ArrayQualifiedTypeReference ref = 
      new ArrayQualifiedTypeReference(tokens, dim, positions);
    ref.sourceEnd = this.endPosition;
    if (this.reportReferenceInfo) {
      this.requestor.acceptTypeReference(ref.tokens, ref.sourceStart, ref.sourceEnd);
    }
    return ref;
  }

  public NameReference getUnspecifiedReference()
  {
    int length;
    if ((length = this.identifierLengthStack[(this.identifierLengthPtr--)]) == 1)
    {
      SingleNameReference ref = 
        newSingleNameReference(
        this.identifierStack[this.identifierPtr], 
        this.identifierPositionStack[(this.identifierPtr--)]);
      if (this.reportReferenceInfo) {
        addUnknownRef(ref);
      }
      return ref;
    }

    char[][] tokens = new char[length][];
    this.identifierPtr -= length;
    System.arraycopy(this.identifierStack, this.identifierPtr + 1, tokens, 0, length);
    long[] positions = new long[length];
    System.arraycopy(this.identifierPositionStack, this.identifierPtr + 1, positions, 0, length);
    QualifiedNameReference ref = 
      newQualifiedNameReference(
      tokens, 
      positions, 
      (int)(this.identifierPositionStack[(this.identifierPtr + 1)] >> 32), 
      (int)this.identifierPositionStack[(this.identifierPtr + length)]);
    if (this.reportReferenceInfo) {
      addUnknownRef(ref);
    }
    return ref;
  }

  public NameReference getUnspecifiedReferenceOptimized()
  {
    int length;
    if ((length = this.identifierLengthStack[(this.identifierLengthPtr--)]) == 1)
    {
      SingleNameReference ref = 
        newSingleNameReference(
        this.identifierStack[this.identifierPtr], 
        this.identifierPositionStack[(this.identifierPtr--)]);
      ref.bits &= -8;
      ref.bits |= 3;
      if (this.reportReferenceInfo) {
        addUnknownRef(ref);
      }
      return ref;
    }

    char[][] tokens = new char[length][];
    this.identifierPtr -= length;
    System.arraycopy(this.identifierStack, this.identifierPtr + 1, tokens, 0, length);
    long[] positions = new long[length];
    System.arraycopy(this.identifierPositionStack, this.identifierPtr + 1, positions, 0, length);
    QualifiedNameReference ref = 
      newQualifiedNameReference(
      tokens, 
      positions, 
      (int)(this.identifierPositionStack[(this.identifierPtr + 1)] >> 32), 
      (int)this.identifierPositionStack[(this.identifierPtr + length)]);
    ref.bits &= -8;
    ref.bits |= 3;
    if (this.reportReferenceInfo) {
      addUnknownRef(ref);
    }
    return ref;
  }

  private boolean hasDeprecatedAnnotation(Annotation[] annotations)
  {
    if (annotations != null) {
      int i = 0; for (int length = annotations.length; i < length; i++) {
        Annotation annotation = annotations[i];
        if (CharOperation.equals(annotation.type.getLastToken(), TypeConstants.JAVA_LANG_DEPRECATED[2])) {
          return true;
        }
      }
    }
    return false;
  }

  protected ImportReference newImportReference(char[][] tokens, long[] positions, boolean onDemand, int mod) {
    return new ImportReference(tokens, positions, onDemand, mod);
  }
  protected QualifiedNameReference newQualifiedNameReference(char[][] tokens, long[] positions, int sourceStart, int sourceEnd) {
    return new QualifiedNameReference(tokens, positions, sourceStart, sourceEnd);
  }
  protected SingleNameReference newSingleNameReference(char[] source, long positions) {
    return new SingleNameReference(source, positions);
  }

  public void notifySourceElementRequestor(CompilationUnitDeclaration parsedUnit)
  {
    if (parsedUnit == null)
    {
      if ((this.astStack[0] instanceof AbstractMethodDeclaration)) {
        notifySourceElementRequestor((AbstractMethodDeclaration)this.astStack[0]);
        return;
      }
      return;
    }

    boolean isInRange = 
      (this.scanner.initialPosition <= parsedUnit.sourceStart) && 
      (this.scanner.eofPosition >= parsedUnit.sourceEnd);

    int length = 0;
    ASTNode[] nodes = (ASTNode[])null;
    if (this.sourceType == null) {
      if (isInRange) {
        this.requestor.enterCompilationUnit();
      }
      ImportReference currentPackage = parsedUnit.currentPackage;
      ImportReference[] imports = parsedUnit.imports;
      TypeDeclaration[] types = parsedUnit.types;
      length = 
        (currentPackage == null ? 0 : 1) + (
        imports == null ? 0 : imports.length) + (
        types == null ? 0 : types.length);
      nodes = new ASTNode[length];
      int index = 0;
      if (currentPackage != null) {
        nodes[(index++)] = currentPackage;
      }
      if (imports != null) {
        int i = 0; for (int max = imports.length; i < max; i++) {
          nodes[(index++)] = imports[i];
        }
      }
      if (types != null) {
        int i = 0; for (int max = types.length; i < max; i++)
          nodes[(index++)] = types[i];
      }
    }
    else {
      TypeDeclaration[] types = parsedUnit.types;
      if (types != null) {
        length = types.length;
        nodes = new ASTNode[length];
        int i = 0; for (int max = types.length; i < max; i++) {
          nodes[i] = types[i];
        }
      }

    }

    if ((nodes != null) && (length > 0)) {
      quickSort(nodes, 0, length - 1);
      for (int i = 0; i < length; i++) {
        ASTNode node = nodes[i];
        if ((node instanceof ImportReference)) {
          ImportReference importRef = (ImportReference)node;
          if (node == parsedUnit.currentPackage)
            notifySourceElementRequestor(importRef, true);
          else
            notifySourceElementRequestor(importRef, false);
        }
        else {
          notifySourceElementRequestor((TypeDeclaration)node, this.sourceType == null, null);
        }
      }
    }

    if ((this.sourceType == null) && 
      (isInRange))
      this.requestor.exitCompilationUnit(parsedUnit.sourceEnd);
  }

  public void notifySourceElementRequestor(AbstractMethodDeclaration methodDeclaration)
  {
    boolean isInRange = 
      (this.scanner.initialPosition <= methodDeclaration.declarationSourceStart) && 
      (this.scanner.eofPosition >= methodDeclaration.declarationSourceEnd);

    if (methodDeclaration.isClinit()) {
      visitIfNeeded(methodDeclaration);
      return;
    }

    if (methodDeclaration.isDefaultConstructor()) {
      if (this.reportReferenceInfo) {
        ConstructorDeclaration constructorDeclaration = (ConstructorDeclaration)methodDeclaration;
        ExplicitConstructorCall constructorCall = constructorDeclaration.constructorCall;
        if (constructorCall != null) {
          switch (constructorCall.accessMode) {
          case 3:
            this.requestor.acceptConstructorReference(
              this.typeNames[(this.nestedTypeIndex - 1)], 
              constructorCall.arguments == null ? 0 : constructorCall.arguments.length, 
              constructorCall.sourceStart);
            break;
          case 1:
          case 2:
            this.requestor.acceptConstructorReference(
              this.superTypeNames[(this.nestedTypeIndex - 1)], 
              constructorCall.arguments == null ? 0 : constructorCall.arguments.length, 
              constructorCall.sourceStart);
          }
        }
      }

      return;
    }
    char[][] argumentTypes = (char[][])null;
    char[][] argumentNames = (char[][])null;
    boolean isVarArgs = false;
    Argument[] arguments = methodDeclaration.arguments;
    if (arguments != null) {
      int argumentLength = arguments.length;
      argumentTypes = new char[argumentLength][];
      argumentNames = new char[argumentLength][];
      for (int i = 0; i < argumentLength; i++) {
        argumentTypes[i] = CharOperation.concatWith(arguments[i].type.getParameterizedTypeName(), '.');
        argumentNames[i] = arguments[i].name;
      }
      isVarArgs = arguments[(argumentLength - 1)].isVarArgs();
    }
    char[][] thrownExceptionTypes = (char[][])null;
    TypeReference[] thrownExceptions = methodDeclaration.thrownExceptions;
    if (thrownExceptions != null) {
      int thrownExceptionLength = thrownExceptions.length;
      thrownExceptionTypes = new char[thrownExceptionLength][];
      for (int i = 0; i < thrownExceptionLength; i++) {
        thrownExceptionTypes[i] = 
          CharOperation.concatWith(thrownExceptions[i].getParameterizedTypeName(), '.');
      }
    }

    int selectorSourceEnd = -1;
    if (methodDeclaration.isConstructor()) {
      selectorSourceEnd = this.sourceEnds.get(methodDeclaration);
      if (isInRange) {
        int currentModifiers = methodDeclaration.modifiers;
        if (isVarArgs) {
          currentModifiers |= 128;
        }

        boolean deprecated = ((currentModifiers & 0x100000) != 0) || (hasDeprecatedAnnotation(methodDeclaration.annotations));

        ISourceElementRequestor.MethodInfo methodInfo = new ISourceElementRequestor.MethodInfo();
        methodInfo.isConstructor = true;
        methodInfo.declarationStart = methodDeclaration.declarationSourceStart;
        methodInfo.modifiers = (deprecated ? currentModifiers & 0xFFFF | 0x100000 : currentModifiers & 0xFFFF);
        methodInfo.name = methodDeclaration.selector;
        methodInfo.nameSourceStart = methodDeclaration.sourceStart;
        methodInfo.nameSourceEnd = selectorSourceEnd;
        methodInfo.parameterTypes = argumentTypes;
        methodInfo.parameterNames = argumentNames;
        methodInfo.exceptionTypes = thrownExceptionTypes;
        methodInfo.typeParameters = getTypeParameterInfos(methodDeclaration.typeParameters());
        methodInfo.annotationPositions = collectAnnotationPositions(methodDeclaration.annotations);
        methodInfo.categories = ((char[][])this.nodesToCategories.get(methodDeclaration));
        this.requestor.enterConstructor(methodInfo);
      }
      if (this.reportReferenceInfo) {
        ConstructorDeclaration constructorDeclaration = (ConstructorDeclaration)methodDeclaration;
        ExplicitConstructorCall constructorCall = constructorDeclaration.constructorCall;
        if (constructorCall != null) {
          switch (constructorCall.accessMode) {
          case 3:
            this.requestor.acceptConstructorReference(
              this.typeNames[(this.nestedTypeIndex - 1)], 
              constructorCall.arguments == null ? 0 : constructorCall.arguments.length, 
              constructorCall.sourceStart);
            break;
          case 1:
          case 2:
            this.requestor.acceptConstructorReference(
              this.superTypeNames[(this.nestedTypeIndex - 1)], 
              constructorCall.arguments == null ? 0 : constructorCall.arguments.length, 
              constructorCall.sourceStart);
          }
        }
      }

      visitIfNeeded(methodDeclaration);
      if (isInRange) {
        this.requestor.exitConstructor(methodDeclaration.declarationSourceEnd);
      }
      return;
    }
    selectorSourceEnd = this.sourceEnds.get(methodDeclaration);
    if (isInRange) {
      int currentModifiers = methodDeclaration.modifiers;
      if (isVarArgs) {
        currentModifiers |= 128;
      }

      boolean deprecated = ((currentModifiers & 0x100000) != 0) || (hasDeprecatedAnnotation(methodDeclaration.annotations));

      TypeReference returnType = (methodDeclaration instanceof MethodDeclaration) ? 
        ((MethodDeclaration)methodDeclaration).returnType : 
        null;
      ISourceElementRequestor.MethodInfo methodInfo = new ISourceElementRequestor.MethodInfo();
      methodInfo.isAnnotation = (methodDeclaration instanceof AnnotationMethodDeclaration);
      methodInfo.declarationStart = methodDeclaration.declarationSourceStart;
      methodInfo.modifiers = (deprecated ? currentModifiers & 0xFFFF | 0x100000 : currentModifiers & 0xFFFF);
      methodInfo.returnType = (returnType == null ? null : CharOperation.concatWith(returnType.getParameterizedTypeName(), '.'));
      methodInfo.name = methodDeclaration.selector;
      methodInfo.nameSourceStart = methodDeclaration.sourceStart;
      methodInfo.nameSourceEnd = selectorSourceEnd;
      methodInfo.parameterTypes = argumentTypes;
      methodInfo.parameterNames = argumentNames;
      methodInfo.exceptionTypes = thrownExceptionTypes;
      methodInfo.typeParameters = getTypeParameterInfos(methodDeclaration.typeParameters());
      methodInfo.annotationPositions = collectAnnotationPositions(methodDeclaration.annotations);
      methodInfo.categories = ((char[][])this.nodesToCategories.get(methodDeclaration));
      this.requestor.enterMethod(methodInfo);
    }

    visitIfNeeded(methodDeclaration);

    if (isInRange) {
      if ((methodDeclaration instanceof AnnotationMethodDeclaration)) {
        AnnotationMethodDeclaration annotationMethodDeclaration = (AnnotationMethodDeclaration)methodDeclaration;
        Expression expression = annotationMethodDeclaration.defaultValue;
        if (expression != null) {
          this.requestor.exitMethod(methodDeclaration.declarationSourceEnd, expression.sourceStart, expression.sourceEnd);
          return;
        }
      }
      this.requestor.exitMethod(methodDeclaration.declarationSourceEnd, -1, -1);
    }
  }

  public void notifySourceElementRequestor(FieldDeclaration fieldDeclaration, TypeDeclaration declaringType)
  {
    boolean isInRange = 
      (this.scanner.initialPosition <= fieldDeclaration.declarationSourceStart) && 
      (this.scanner.eofPosition >= fieldDeclaration.declarationSourceEnd);

    switch (fieldDeclaration.getKind())
    {
    case 3:
      if ((fieldDeclaration.initialization instanceof AllocationExpression)) {
        AllocationExpression alloc = (AllocationExpression)fieldDeclaration.initialization;
        this.requestor.acceptConstructorReference(
          declaringType.name, 
          alloc.arguments == null ? 0 : alloc.arguments.length, 
          alloc.sourceStart);
      }

    case 1:
      int fieldEndPosition = this.sourceEnds.get(fieldDeclaration);
      if (fieldEndPosition == -1)
      {
        fieldEndPosition = fieldDeclaration.declarationSourceEnd;
      }
      if (isInRange) {
        int currentModifiers = fieldDeclaration.modifiers;

        boolean deprecated = ((currentModifiers & 0x100000) != 0) || (hasDeprecatedAnnotation(fieldDeclaration.annotations));

        char[] typeName = (char[])null;
        if (fieldDeclaration.type == null)
        {
          typeName = declaringType.name;
          currentModifiers |= 16384;
        }
        else {
          typeName = CharOperation.concatWith(fieldDeclaration.type.getParameterizedTypeName(), '.');
        }
        ISourceElementRequestor.FieldInfo fieldInfo = new ISourceElementRequestor.FieldInfo();
        fieldInfo.declarationStart = fieldDeclaration.declarationSourceStart;
        fieldInfo.name = fieldDeclaration.name;
        fieldInfo.modifiers = (deprecated ? currentModifiers & 0xFFFF | 0x100000 : currentModifiers & 0xFFFF);
        fieldInfo.type = typeName;
        fieldInfo.nameSourceStart = fieldDeclaration.sourceStart;
        fieldInfo.nameSourceEnd = fieldDeclaration.sourceEnd;
        fieldInfo.annotationPositions = collectAnnotationPositions(fieldDeclaration.annotations);
        fieldInfo.categories = ((char[][])this.nodesToCategories.get(fieldDeclaration));
        this.requestor.enterField(fieldInfo);
      }
      visitIfNeeded(fieldDeclaration, declaringType);
      if (isInRange) {
        this.requestor.exitField(
          (fieldDeclaration.initialization == null) || 
          ((fieldDeclaration.initialization instanceof ArrayInitializer)) || 
          ((fieldDeclaration.initialization instanceof AllocationExpression)) || 
          ((fieldDeclaration.initialization instanceof ArrayAllocationExpression)) || 
          ((fieldDeclaration.initialization instanceof Assignment)) || 
          ((fieldDeclaration.initialization instanceof ClassLiteralAccess)) || 
          ((fieldDeclaration.initialization instanceof MessageSend)) || 
          ((fieldDeclaration.initialization instanceof ArrayReference)) || 
          ((fieldDeclaration.initialization instanceof ThisReference)) ? 
          -1 : 
          fieldDeclaration.initialization.sourceStart, 
          fieldEndPosition, 
          fieldDeclaration.declarationSourceEnd);
      }
      break;
    case 2:
      if (isInRange) {
        this.requestor.enterInitializer(
          fieldDeclaration.declarationSourceStart, 
          fieldDeclaration.modifiers);
      }
      visitIfNeeded((Initializer)fieldDeclaration);
      if (isInRange)
        this.requestor.exitInitializer(fieldDeclaration.declarationSourceEnd);
      break;
    }
  }

  public void notifySourceElementRequestor(ImportReference importReference, boolean isPackage)
  {
    if (isPackage)
      this.requestor.acceptPackage(
        importReference.declarationSourceStart, 
        importReference.declarationSourceEnd, 
        CharOperation.concatWith(importReference.getImportName(), '.'));
    else
      this.requestor.acceptImport(
        importReference.declarationSourceStart, 
        importReference.declarationSourceEnd, 
        importReference.tokens, 
        (importReference.bits & 0x20000) != 0, 
        importReference.modifiers);
  }

  public void notifySourceElementRequestor(TypeDeclaration typeDeclaration, boolean notifyTypePresence, TypeDeclaration declaringType)
  {
    if (CharOperation.equals(TypeConstants.PACKAGE_INFO_NAME, typeDeclaration.name)) return;

    boolean isInRange = 
      (this.scanner.initialPosition <= typeDeclaration.declarationSourceStart) && 
      (this.scanner.eofPosition >= typeDeclaration.declarationSourceEnd);

    FieldDeclaration[] fields = typeDeclaration.fields;
    AbstractMethodDeclaration[] methods = typeDeclaration.methods;
    TypeDeclaration[] memberTypes = typeDeclaration.memberTypes;
    int fieldCounter = fields == null ? 0 : fields.length;
    int methodCounter = methods == null ? 0 : methods.length;
    int memberTypeCounter = memberTypes == null ? 0 : memberTypes.length;
    int fieldIndex = 0;
    int methodIndex = 0;
    int memberTypeIndex = 0;

    if (notifyTypePresence) {
      char[][] interfaceNames = (char[][])null;
      int superInterfacesLength = 0;
      TypeReference[] superInterfaces = typeDeclaration.superInterfaces;
      if (superInterfaces != null) {
        superInterfacesLength = superInterfaces.length;
        interfaceNames = new char[superInterfacesLength][];
      }
      else if ((typeDeclaration.bits & 0x200) != 0)
      {
        QualifiedAllocationExpression alloc = typeDeclaration.allocation;
        if ((alloc != null) && (alloc.type != null)) {
          superInterfaces = new TypeReference[] { alloc.type };
          superInterfacesLength = 1;
          interfaceNames = new char[1][];
        }
      }

      if (superInterfaces != null) {
        for (int i = 0; i < superInterfacesLength; i++) {
          interfaceNames[i] = 
            CharOperation.concatWith(superInterfaces[i].getParameterizedTypeName(), '.');
        }
      }
      int kind = TypeDeclaration.kind(typeDeclaration.modifiers);
      char[] implicitSuperclassName = TypeConstants.CharArray_JAVA_LANG_OBJECT;
      if (isInRange) {
        int currentModifiers = typeDeclaration.modifiers;

        boolean deprecated = ((currentModifiers & 0x100000) != 0) || (hasDeprecatedAnnotation(typeDeclaration.annotations));

        boolean isEnumInit = (typeDeclaration.allocation != null) && (typeDeclaration.allocation.enumConstant != null);
        char[] superclassName;
        char[] superclassName;
        if (isEnumInit) {
          currentModifiers |= 16384;
          superclassName = declaringType.name;
        } else {
          TypeReference superclass = typeDeclaration.superclass;
          superclassName = superclass != null ? CharOperation.concatWith(superclass.getParameterizedTypeName(), '.') : null;
        }
        ISourceElementRequestor.TypeInfo typeInfo = new ISourceElementRequestor.TypeInfo();
        typeInfo.declarationStart = typeDeclaration.declarationSourceStart;
        typeInfo.modifiers = (deprecated ? currentModifiers & 0xFFFF | 0x100000 : currentModifiers & 0xFFFF);
        typeInfo.name = typeDeclaration.name;
        typeInfo.nameSourceStart = typeDeclaration.sourceStart;
        typeInfo.nameSourceEnd = sourceEnd(typeDeclaration);
        typeInfo.superclass = superclassName;
        typeInfo.superinterfaces = interfaceNames;
        typeInfo.typeParameters = getTypeParameterInfos(typeDeclaration.typeParameters);
        typeInfo.annotationPositions = collectAnnotationPositions(typeDeclaration.annotations);
        typeInfo.categories = ((char[][])this.nodesToCategories.get(typeDeclaration));
        typeInfo.secondary = typeDeclaration.isSecondary();
        typeInfo.anonymousMember = ((typeDeclaration.allocation != null) && (typeDeclaration.allocation.enclosingInstance != null));
        this.requestor.enterType(typeInfo);
        switch (kind) {
        case 1:
          if (superclassName != null)
            implicitSuperclassName = superclassName;
          break;
        case 2:
          implicitSuperclassName = TypeConstants.CharArray_JAVA_LANG_OBJECT;
          break;
        case 3:
          implicitSuperclassName = TypeConstants.CharArray_JAVA_LANG_ENUM;
          break;
        case 4:
          implicitSuperclassName = TypeConstants.CharArray_JAVA_LANG_ANNOTATION_ANNOTATION;
        }
      }

      if (this.nestedTypeIndex == this.typeNames.length)
      {
        System.arraycopy(this.typeNames, 0, this.typeNames = new char[this.nestedTypeIndex * 2][], 0, this.nestedTypeIndex);
        System.arraycopy(this.superTypeNames, 0, this.superTypeNames = new char[this.nestedTypeIndex * 2][], 0, this.nestedTypeIndex);
      }
      this.typeNames[this.nestedTypeIndex] = typeDeclaration.name;
      this.superTypeNames[(this.nestedTypeIndex++)] = implicitSuperclassName;
    }
    while ((fieldIndex < fieldCounter) || 
      (memberTypeIndex < memberTypeCounter) || 
      (methodIndex < methodCounter)) {
      FieldDeclaration nextFieldDeclaration = null;
      AbstractMethodDeclaration nextMethodDeclaration = null;
      TypeDeclaration nextMemberDeclaration = null;

      int position = 2147483647;
      int nextDeclarationType = -1;
      if (fieldIndex < fieldCounter) {
        nextFieldDeclaration = fields[fieldIndex];
        if (nextFieldDeclaration.declarationSourceStart < position) {
          position = nextFieldDeclaration.declarationSourceStart;
          nextDeclarationType = 0;
        }
      }
      if (methodIndex < methodCounter) {
        nextMethodDeclaration = methods[methodIndex];
        if (nextMethodDeclaration.declarationSourceStart < position) {
          position = nextMethodDeclaration.declarationSourceStart;
          nextDeclarationType = 1;
        }
      }
      if (memberTypeIndex < memberTypeCounter) {
        nextMemberDeclaration = memberTypes[memberTypeIndex];
        if (nextMemberDeclaration.declarationSourceStart < position) {
          position = nextMemberDeclaration.declarationSourceStart;
          nextDeclarationType = 2;
        }
      }
      switch (nextDeclarationType) {
      case 0:
        fieldIndex++;
        notifySourceElementRequestor(nextFieldDeclaration, typeDeclaration);
        break;
      case 1:
        methodIndex++;
        notifySourceElementRequestor(nextMethodDeclaration);
        break;
      case 2:
        memberTypeIndex++;
        notifySourceElementRequestor(nextMemberDeclaration, true, null);
      }
    }
    if (notifyTypePresence) {
      if (isInRange) {
        this.requestor.exitType(typeDeclaration.declarationSourceEnd);
      }
      this.nestedTypeIndex -= 1;
    }
  }

  public void parseCompilationUnit(ICompilationUnit unit, int start, int end, boolean fullParse)
  {
    this.reportReferenceInfo = fullParse;
    boolean old = this.diet;
    try
    {
      this.diet = true;
      CompilationResult compilationUnitResult = new CompilationResult(unit, 0, 0, this.options.maxProblemsPerUnit);
      CompilationUnitDeclaration parsedUnit = parse(unit, compilationUnitResult, start, end);
      if (this.scanner.recordLineSeparator) {
        this.requestor.acceptLineSeparatorPositions(compilationUnitResult.getLineSeparatorPositions());
      }
      if ((this.localDeclarationVisitor != null) || (fullParse)) {
        this.diet = false;
        getMethodBodies(parsedUnit);
      }
      this.scanner.resetTo(start, end);
      notifySourceElementRequestor(parsedUnit);
    } catch (AbortCompilation localAbortCompilation) {
    }
    finally {
      this.diet = old;
      reset();
    }
  }

  public CompilationUnitDeclaration parseCompilationUnit(ICompilationUnit unit, boolean fullParse)
  {
    boolean old = this.diet;
    try
    {
      this.diet = true;
      this.reportReferenceInfo = fullParse;
      CompilationResult compilationUnitResult = new CompilationResult(unit, 0, 0, this.options.maxProblemsPerUnit);
      CompilationUnitDeclaration parsedUnit = parse(unit, compilationUnitResult);
      if (this.scanner.recordLineSeparator) {
        this.requestor.acceptLineSeparatorPositions(compilationUnitResult.getLineSeparatorPositions());
      }
      int initialStart = this.scanner.initialPosition;
      int initialEnd = this.scanner.eofPosition;
      if ((this.localDeclarationVisitor != null) || (fullParse)) {
        this.diet = false;
        getMethodBodies(parsedUnit);
      }
      this.scanner.resetTo(initialStart, initialEnd);
      notifySourceElementRequestor(parsedUnit);
      return parsedUnit;
    } catch (AbortCompilation localAbortCompilation) {
    }
    finally {
      this.diet = old;
      reset();
    }
    return null;
  }

  public void parseTypeMemberDeclarations(ISourceType type, ICompilationUnit sourceUnit, int start, int end, boolean needReferenceInfo)
  {
    boolean old = this.diet;

    CompilationResult compilationUnitResult = 
      new CompilationResult(sourceUnit, 0, 0, this.options.maxProblemsPerUnit);
    try {
      this.diet = (!needReferenceInfo);
      this.reportReferenceInfo = needReferenceInfo;
      CompilationUnitDeclaration unit = 
        SourceTypeConverter.buildCompilationUnit(
        new ISourceType[] { type }, 
        0, 
        problemReporter(), 
        compilationUnitResult);
      if ((unit == null) || (unit.types == null) || (unit.types.length != 1))
        return;
      CompilationUnitDeclaration unit;
      this.sourceType = type;
      try
      {
        initialize();
        goForClassBodyDeclarations();

        this.scanner.setSource(sourceUnit.getContents());
        this.scanner.resetTo(start, end);

        this.referenceContext = (this.compilationUnit = unit);

        pushOnAstStack(unit.types[0]);

        parse();
        notifySourceElementRequestor(unit);
      } finally {
        unit = this.compilationUnit;
        this.compilationUnit = null;
      }
    } catch (AbortCompilation localAbortCompilation) {
    }
    finally {
      if (this.scanner.recordLineSeparator) {
        this.requestor.acceptLineSeparatorPositions(compilationUnitResult.getLineSeparatorPositions());
      }
      this.diet = old;
      reset();
    }
  }

  public void parseTypeMemberDeclarations(char[] contents, int start, int end)
  {
    boolean old = this.diet;
    try
    {
      this.diet = true;

      initialize();
      goForClassBodyDeclarations();

      this.scanner.setSource(contents);
      this.scanner.recordLineSeparator = false;
      this.scanner.taskTags = null;
      this.scanner.taskPriorities = null;
      this.scanner.resetTo(start, end);

      this.referenceContext = null;

      parse();
      notifySourceElementRequestor(null);
    } catch (AbortCompilation localAbortCompilation) {
    }
    finally {
      this.diet = old;
      reset();
    }
  }

  private static void quickSort(ASTNode[] sortedCollection, int left, int right)
  {
    int original_left = left;
    int original_right = right;
    ASTNode mid = sortedCollection[(left + (right - left) / 2)];
    do {
      while (sortedCollection[left].sourceStart < mid.sourceStart) {
        left++;
      }
      while (mid.sourceStart < sortedCollection[right].sourceStart) {
        right--;
      }
      if (left <= right) {
        ASTNode tmp = sortedCollection[left];
        sortedCollection[left] = sortedCollection[right];
        sortedCollection[right] = tmp;
        left++;
        right--;
      }
    }
    while (left <= right);
    if (original_left < right) {
      quickSort(sortedCollection, original_left, right);
    }
    if (left < original_right)
      quickSort(sortedCollection, left, original_right);
  }

  private void rememberCategories() {
    if (this.useSourceJavadocParser) {
      SourceJavadocParser sourceJavadocParser = (SourceJavadocParser)this.javadocParser;
      char[][] categories = sourceJavadocParser.categories;
      if (categories.length > 0) {
        this.nodesToCategories.put(this.astStack[this.astPtr], categories);
        sourceJavadocParser.categories = CharOperation.NO_CHAR_CHAR;
      }
    }
  }

  private void reset() { this.sourceEnds = new HashtableOfObjectToInt();
    this.nodesToCategories = new HashMap();
    this.typeNames = new char[4][];
    this.superTypeNames = new char[4][];
    this.nestedTypeIndex = 0; }

  private int sourceEnd(TypeDeclaration typeDeclaration) {
    if ((typeDeclaration.bits & 0x200) != 0) {
      QualifiedAllocationExpression allocation = typeDeclaration.allocation;
      if (allocation.type == null)
        return typeDeclaration.sourceEnd;
      return allocation.type.sourceEnd;
    }
    return typeDeclaration.sourceEnd;
  }

  private void visitIfNeeded(AbstractMethodDeclaration method) {
    if ((this.localDeclarationVisitor != null) && 
      ((method.bits & 0x2) != 0)) {
      if ((method instanceof ConstructorDeclaration)) {
        ConstructorDeclaration constructorDeclaration = (ConstructorDeclaration)method;
        if (constructorDeclaration.constructorCall != null) {
          constructorDeclaration.constructorCall.traverse(this.localDeclarationVisitor, method.scope);
        }
      }
      if (method.statements != null) {
        int statementsLength = method.statements.length;
        for (int i = 0; i < statementsLength; i++)
          method.statements[i].traverse(this.localDeclarationVisitor, method.scope);
      }
    }
  }

  private void visitIfNeeded(FieldDeclaration field, TypeDeclaration declaringType) {
    if ((this.localDeclarationVisitor != null) && 
      ((field.bits & 0x2) != 0) && 
      (field.initialization != null))
      try {
        this.localDeclarationVisitor.pushDeclaringType(declaringType);
        field.initialization.traverse(this.localDeclarationVisitor, null);
      } finally {
        this.localDeclarationVisitor.popDeclaringType();
      }
  }

  private void visitIfNeeded(Initializer initializer)
  {
    if ((this.localDeclarationVisitor != null) && 
      ((initializer.bits & 0x2) != 0) && 
      (initializer.block != null))
      initializer.block.traverse(this.localDeclarationVisitor, null);
  }

  public class LocalDeclarationVisitor extends ASTVisitor
  {
    ArrayList declaringTypes;

    public LocalDeclarationVisitor()
    {
    }

    public void pushDeclaringType(TypeDeclaration declaringType)
    {
      if (this.declaringTypes == null) {
        this.declaringTypes = new ArrayList();
      }
      this.declaringTypes.add(declaringType);
    }
    public void popDeclaringType() {
      this.declaringTypes.remove(this.declaringTypes.size() - 1);
    }
    public TypeDeclaration peekDeclaringType() {
      if (this.declaringTypes == null) return null;
      int size = this.declaringTypes.size();
      if (size == 0) return null;
      return (TypeDeclaration)this.declaringTypes.get(size - 1);
    }
    public boolean visit(TypeDeclaration typeDeclaration, BlockScope scope) {
      SourceElementParser.this.notifySourceElementRequestor(typeDeclaration, SourceElementParser.this.sourceType == null, peekDeclaringType());
      return false;
    }
    public boolean visit(TypeDeclaration typeDeclaration, ClassScope scope) {
      SourceElementParser.this.notifySourceElementRequestor(typeDeclaration, SourceElementParser.this.sourceType == null, peekDeclaringType());
      return false;
    }
  }
}