package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class CastExpression extends Expression
{
  public Expression expression;
  public Expression type;
  public TypeBinding expectedType;

  public CastExpression(Expression expression, Expression type)
  {
    this.expression = expression;
    this.type = type;
    type.bits |= 1073741824;
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    return this.expression
      .analyseCode(currentScope, flowContext, flowInfo)
      .unconditionalInits();
  }

  public static void checkNeedForAssignedCast(BlockScope scope, TypeBinding expectedType, CastExpression rhs)
  {
    if (scope.compilerOptions().getSeverity(67108864L) == -1) return;

    TypeBinding castedExpressionType = rhs.expression.resolvedType;

    if ((castedExpressionType == null) || (rhs.resolvedType.isBaseType())) return;

    if (castedExpressionType.isCompatibleWith(expectedType))
      scope.problemReporter().unnecessaryCast(rhs);
  }

  public static void checkNeedForEnclosingInstanceCast(BlockScope scope, Expression enclosingInstance, TypeBinding enclosingInstanceType, TypeBinding memberType)
  {
    if (scope.compilerOptions().getSeverity(67108864L) == -1) return;

    TypeBinding castedExpressionType = ((CastExpression)enclosingInstance).expression.resolvedType;
    if (castedExpressionType == null) return;

    if (castedExpressionType == enclosingInstanceType) {
      scope.problemReporter().unnecessaryCast((CastExpression)enclosingInstance); } else {
      if (castedExpressionType == TypeBinding.NULL) {
        return;
      }
      TypeBinding alternateEnclosingInstanceType = castedExpressionType;
      if ((castedExpressionType.isBaseType()) || (castedExpressionType.isArrayType())) return;
      if (memberType == scope.getMemberType(memberType.sourceName(), (ReferenceBinding)alternateEnclosingInstanceType))
        scope.problemReporter().unnecessaryCast((CastExpression)enclosingInstance);
    }
  }

  public static void checkNeedForArgumentCast(BlockScope scope, int operator, int operatorSignature, Expression expression, int expressionTypeId)
  {
    if (scope.compilerOptions().getSeverity(67108864L) == -1) return;

    int alternateLeftTypeId = expressionTypeId;
    if (((expression.bits & 0x4000) == 0) && (expression.resolvedType.isBaseType()))
    {
      return;
    }
    TypeBinding alternateLeftType = ((CastExpression)expression).expression.resolvedType;
    if (alternateLeftType == null) return;
    if ((alternateLeftTypeId = alternateLeftType.id) == expressionTypeId) {
      scope.problemReporter().unnecessaryCast((CastExpression)expression);
      return;
    }if (alternateLeftTypeId == 12) {
      alternateLeftTypeId = expressionTypeId;
      return;
    }
  }

  public static void checkNeedForArgumentCasts(BlockScope scope, Expression receiver, TypeBinding receiverType, MethodBinding binding, Expression[] arguments, TypeBinding[] argumentTypes, InvocationSite invocationSite)
  {
    if (scope.compilerOptions().getSeverity(67108864L) == -1) return;

    int length = argumentTypes.length;

    TypeBinding[] rawArgumentTypes = argumentTypes;
    for (int i = 0; i < length; i++) {
      Expression argument = arguments[i];
      if ((argument instanceof CastExpression))
      {
        if (((argument.bits & 0x4000) != 0) || (!argument.resolvedType.isBaseType()))
        {
          TypeBinding castedExpressionType = ((CastExpression)argument).expression.resolvedType;
          if (castedExpressionType == null) return;

          if (castedExpressionType == argumentTypes[i])
            scope.problemReporter().unnecessaryCast((CastExpression)argument);
          else if (castedExpressionType != TypeBinding.NULL)
          {
            if ((argument.implicitConversion & 0x200) == 0)
            {
              if (rawArgumentTypes == argumentTypes) {
                System.arraycopy(rawArgumentTypes, 0, rawArgumentTypes = new TypeBinding[length], 0, length);
              }

              rawArgumentTypes[i] = castedExpressionType;
            }
          }
        }
      }
    }
    if (rawArgumentTypes != argumentTypes)
      checkAlternateBinding(scope, receiver, receiverType, binding, arguments, argumentTypes, rawArgumentTypes, invocationSite);
  }

  public static void checkNeedForArgumentCasts(BlockScope scope, int operator, int operatorSignature, Expression left, int leftTypeId, boolean leftIsCast, Expression right, int rightTypeId, boolean rightIsCast)
  {
    if (scope.compilerOptions().getSeverity(67108864L) == -1) return;

    int alternateLeftTypeId = leftTypeId;
    if (leftIsCast) {
      if (((left.bits & 0x4000) == 0) && (left.resolvedType.isBaseType()))
      {
        leftIsCast = false;
      } else {
        TypeBinding alternateLeftType = ((CastExpression)left).expression.resolvedType;
        if (alternateLeftType == null) return;
        if ((alternateLeftTypeId = alternateLeftType.id) == leftTypeId) {
          scope.problemReporter().unnecessaryCast((CastExpression)left);
          leftIsCast = false;
        } else if (alternateLeftTypeId == 12) {
          alternateLeftTypeId = leftTypeId;
          leftIsCast = false;
        }
      }
    }

    int alternateRightTypeId = rightTypeId;
    if (rightIsCast) {
      if (((right.bits & 0x4000) == 0) && (right.resolvedType.isBaseType()))
      {
        rightIsCast = false;
      } else {
        TypeBinding alternateRightType = ((CastExpression)right).expression.resolvedType;
        if (alternateRightType == null) return;
        if ((alternateRightTypeId = alternateRightType.id) == rightTypeId) {
          scope.problemReporter().unnecessaryCast((CastExpression)right);
          rightIsCast = false;
        } else if (alternateRightTypeId == 12) {
          alternateRightTypeId = rightTypeId;
          rightIsCast = false;
        }
      }
    }
    if ((leftIsCast) || (rightIsCast)) {
      if ((alternateLeftTypeId > 15) || (alternateRightTypeId > 15)) {
        if (alternateLeftTypeId == 11)
          alternateRightTypeId = 1;
        else if (alternateRightTypeId == 11)
          alternateLeftTypeId = 1;
        else {
          return;
        }
      }
      int alternateOperatorSignature = OperatorExpression.OperatorSignatures[operator][((alternateLeftTypeId << 4) + alternateRightTypeId)];

      if ((operatorSignature & 0xF0F0F) == (alternateOperatorSignature & 0xF0F0F)) {
        if (leftIsCast) scope.problemReporter().unnecessaryCast((CastExpression)left);
        if (rightIsCast) scope.problemReporter().unnecessaryCast((CastExpression)right);
      }
    }
  }

  private static void checkAlternateBinding(BlockScope scope, Expression receiver, TypeBinding receiverType, MethodBinding binding, Expression[] arguments, TypeBinding[] originalArgumentTypes, TypeBinding[] alternateArgumentTypes, InvocationSite invocationSite)
  {
    InvocationSite fakeInvocationSite = new InvocationSite() {
      public TypeBinding[] genericTypeArguments() { return null; } 
      public boolean isSuperAccess() { return CastExpression.this.isSuperAccess(); } 
      public boolean isTypeAccess() { return CastExpression.this.isTypeAccess(); } 
      public void setActualReceiverType(ReferenceBinding actualReceiverType) {  } 
      public void setDepth(int depth) {  } 
      public void setFieldIndex(int depth) {  } 
      public int sourceStart() { return 0; } 
      public int sourceEnd() { return 0; }

    };
    MethodBinding bindingIfNoCast;
    MethodBinding bindingIfNoCast;
    if (binding.isConstructor())
      bindingIfNoCast = scope.getConstructor((ReferenceBinding)receiverType, alternateArgumentTypes, fakeInvocationSite);
    else {
      bindingIfNoCast = receiver.isImplicitThis() ? 
        scope.getImplicitMethod(binding.selector, alternateArgumentTypes, fakeInvocationSite) : 
        scope.getMethod(receiverType, binding.selector, alternateArgumentTypes, fakeInvocationSite);
    }
    if (bindingIfNoCast == binding) {
      int argumentLength = originalArgumentTypes.length;
      if (binding.isVarargs()) {
        int paramLength = binding.parameters.length;
        if (paramLength == argumentLength) {
          int varargsIndex = paramLength - 1;
          ArrayBinding varargsType = (ArrayBinding)binding.parameters[varargsIndex];
          TypeBinding lastArgType = alternateArgumentTypes[varargsIndex];

          if (varargsType.dimensions != lastArgType.dimensions()) {
            return;
          }
          if ((lastArgType.isCompatibleWith(varargsType.elementsType())) && 
            (lastArgType.isCompatibleWith(varargsType))) {
            return;
          }
        }
      }
      for (int i = 0; i < argumentLength; i++)
        if (originalArgumentTypes[i] != alternateArgumentTypes[i])
          scope.problemReporter().unnecessaryCast((CastExpression)arguments[i]);
    }
  }

  public boolean checkUnsafeCast(Scope scope, TypeBinding castType, TypeBinding expressionType, TypeBinding match, boolean isNarrowing)
  {
    if (match == castType) {
      if ((!isNarrowing) && (match == this.resolvedType.leafComponentType())) {
        tagAsUnnecessaryCast(scope, castType);
      }
      return true;
    }
    if (match != null) if (match.isProvablyDistinctFrom(isNarrowing ? expressionType : castType, 0)) {
        return false;
      }
    switch (castType.kind()) {
    case 260:
      if (castType.isBoundParameterizedType()) {
        if (match == null) {
          this.bits |= 128;
          return true;
        }
        switch (match.kind()) {
        case 260:
          if (isNarrowing)
          {
            if ((expressionType.isRawType()) || (!expressionType.isEquivalentTo(match))) {
              this.bits |= 128;
              return true;
            }

            TypeBinding genericCastType = castType.erasure();
            TypeBinding genericMatch = genericCastType.findSuperTypeWithSameErasure(expressionType);
            if (genericMatch == match) {
              this.bits |= 128;
            }
            return true;
          }

          if (!match.isEquivalentTo(castType)) {
            this.bits |= 128;
            return true;
          }

          break;
        case 1028:
          this.bits |= 128;
          return true;
        default:
          if (isNarrowing)
          {
            this.bits |= 128;
            return true;
          }
          break;
        }
      }
      break;
    case 68:
      TypeBinding leafType = castType.leafComponentType();
      if ((isNarrowing) && ((leafType.isBoundParameterizedType()) || (leafType.isTypeVariable()))) {
        this.bits |= 128;
        return true;
      }
      break;
    case 4100:
      this.bits |= 128;
      return true;
    }
    if ((!isNarrowing) && (match == this.resolvedType.leafComponentType())) {
      tagAsUnnecessaryCast(scope, castType);
    }
    return true;
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired)
  {
    int pc = codeStream.position;
    boolean needRuntimeCheckcast = (this.bits & 0x40) != 0;
    if (this.constant != Constant.NotAConstant) {
      if ((valueRequired) || (needRuntimeCheckcast)) {
        codeStream.generateConstant(this.constant, this.implicitConversion);
        if (needRuntimeCheckcast) {
          codeStream.checkcast(this.resolvedType);
        }
        if (!valueRequired)
        {
          codeStream.pop();
        }
      }
      codeStream.recordPositionsFrom(pc, this.sourceStart);
      return;
    }
    this.expression.generateCode(currentScope, codeStream, (valueRequired) || (needRuntimeCheckcast));
    if ((needRuntimeCheckcast) && (this.expression.postConversionType(currentScope) != this.resolvedType.erasure())) {
      codeStream.checkcast(this.resolvedType);
    }
    if (valueRequired)
      codeStream.generateImplicitConversion(this.implicitConversion);
    else if (needRuntimeCheckcast) {
      codeStream.pop();
    }
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }

  public Expression innermostCastedExpression() {
    Expression current = this.expression;
    while ((current instanceof CastExpression)) {
      current = ((CastExpression)current).expression;
    }
    return current;
  }

  public LocalVariableBinding localVariableBinding()
  {
    return this.expression.localVariableBinding();
  }

  public int nullStatus(FlowInfo flowInfo) {
    return this.expression.nullStatus(flowInfo);
  }

  public Constant optimizedBooleanConstant()
  {
    switch (this.resolvedType.id) {
    case 5:
    case 33:
      return this.expression.optimizedBooleanConstant();
    }
    return Constant.NotAConstant;
  }

  public StringBuffer printExpression(int indent, StringBuffer output)
  {
    output.append('(');
    this.type.print(0, output).append(") ");
    return this.expression.printExpression(0, output);
  }

  public TypeBinding resolveType(BlockScope scope)
  {
    this.constant = Constant.NotAConstant;
    this.implicitConversion = 0;

    if (((this.type instanceof TypeReference)) || (((this.type instanceof NameReference)) && 
      ((this.type.bits & 0x1FE00000) >> 21 == 0)))
    {
      TypeBinding castType = this.resolvedType = this.type.resolveType(scope);

      TypeBinding expressionType = this.expression.resolveType(scope);
      if (castType != null) {
        if (expressionType != null) {
          boolean isLegal = checkCastTypesCompatibility(scope, castType, expressionType, this.expression);
          if (isLegal) {
            this.expression.computeConversion(scope, castType, expressionType);
            if ((this.bits & 0x80) != 0) {
              scope.problemReporter().unsafeCast(this, scope);
            } else {
              if ((castType.isRawType()) && (scope.compilerOptions().getSeverity(35184372088832L) != -1)) {
                scope.problemReporter().rawTypeReference(this.type, castType);
              }
              if (((this.bits & 0x4020) == 16384) && 
                (!isIndirectlyUsed()))
                scope.problemReporter().unnecessaryCast(this);
            }
          }
          else {
            scope.problemReporter().typeCastError(this, castType, expressionType);
            this.bits |= 32;
          }
        }
        this.resolvedType = castType.capture(scope, this.sourceEnd);
      }
      return this.resolvedType;
    }
    TypeBinding expressionType = this.expression.resolveType(scope);
    if (expressionType == null) return null;
    scope.problemReporter().invalidTypeReference(this.type);
    return null;
  }

  public void setExpectedType(TypeBinding expectedType)
  {
    this.expectedType = expectedType;
  }

  private boolean isIndirectlyUsed()
  {
    if ((this.expression instanceof MessageSend)) {
      MethodBinding method = ((MessageSend)this.expression).binding;
      if (((method instanceof ParameterizedGenericMethodBinding)) && 
        (((ParameterizedGenericMethodBinding)method).inferredReturnType)) {
        if (this.expectedType == null)
          return true;
        if (this.resolvedType != this.expectedType)
          return true;
      }
    }
    if ((this.expectedType != null) && (this.resolvedType.isBaseType()) && (!this.resolvedType.isCompatibleWith(this.expectedType)))
    {
      return true;
    }
    return false;
  }

  public void tagAsNeedCheckCast()
  {
    this.bits |= 64;
  }

  public void tagAsUnnecessaryCast(Scope scope, TypeBinding castType)
  {
    if (this.expression.resolvedType == null) return;
    this.bits |= 16384;
  }

  public void traverse(ASTVisitor visitor, BlockScope blockScope)
  {
    if (visitor.visit(this, blockScope)) {
      this.type.traverse(visitor, blockScope);
      this.expression.traverse(visitor, blockScope);
    }
    visitor.endVisit(this, blockScope);
  }
}