package org.eclipse.jdt.internal.compiler.ast;

import java.util.ArrayList;
import org.eclipse.jdt.core.compiler.CategorizedProblem;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.ExceptionHandlingFlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.InitializationFlowContext;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.NestedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticArgumentBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.parser.Parser;
import org.eclipse.jdt.internal.compiler.problem.AbortMethod;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class ConstructorDeclaration extends AbstractMethodDeclaration
{
  public ExplicitConstructorCall constructorCall;
  public TypeParameter[] typeParameters;

  public ConstructorDeclaration(CompilationResult compilationResult)
  {
    super(compilationResult);
  }

  /** @deprecated */
  public void analyseCode(ClassScope classScope, InitializationFlowContext initializerFlowContext, FlowInfo flowInfo)
  {
    analyseCode(classScope, initializerFlowContext, flowInfo, 0);
  }

  public void analyseCode(ClassScope classScope, InitializationFlowContext initializerFlowContext, FlowInfo flowInfo, int initialReachMode)
  {
    if (this.ignoreFurtherInvestigation) {
      return;
    }
    int nonStaticFieldInfoReachMode = flowInfo.reachMode();
    flowInfo.setReachMode(initialReachMode);
    MethodBinding constructorBinding;
    if (((constructorBinding = this.binding) != null) && 
      ((this.bits & 0x80) == 0) && 
      (!constructorBinding.isUsed()) && 
      (constructorBinding.isPrivate() ? 
      (this.binding.declaringClass.tagBits & 0x0) != 0L : 
      (this.binding.declaringClass.tagBits & 0x30) == 16L))
    {
      this.scope.problemReporter().unusedPrivateConstructor(this);
    }

    if (isRecursive(null)) {
      this.scope.problemReporter().recursiveConstructorInvocation(this.constructorCall);
    }
    try
    {
      ExceptionHandlingFlowContext constructorContext = 
        new ExceptionHandlingFlowContext(
        initializerFlowContext.parent, 
        this, 
        this.binding.thrownExceptions, 
        this.scope, 
        FlowInfo.DEAD_END);
      initializerFlowContext.checkInitializerExceptions(
        this.scope, 
        constructorContext, 
        flowInfo);

      if (this.binding.declaringClass.isAnonymousType()) {
        ArrayList computedExceptions = constructorContext.extendedExceptions;
        if (computedExceptions != null)
        {
          int size;
          if ((size = computedExceptions.size()) > 0)
          {
            ReferenceBinding[] actuallyThrownExceptions;
            computedExceptions.toArray(actuallyThrownExceptions = new ReferenceBinding[size]);
            this.binding.thrownExceptions = actuallyThrownExceptions;
          }
        }

      }

      if (this.arguments != null) {
        int i = 0; for (int count = this.arguments.length; i < count; i++) {
          flowInfo.markAsDefinitelyAssigned(this.arguments[i].binding);
        }

      }

      if (this.constructorCall != null)
      {
        if (this.constructorCall.accessMode == 3) {
          FieldBinding[] fields = this.binding.declaringClass.fields();
          int i = 0; for (int count = fields.length; i < count; i++)
          {
            FieldBinding field;
            if (!(field = fields[i]).isStatic()) {
              flowInfo.markAsDefinitelyAssigned(field);
            }
          }
        }
        flowInfo = this.constructorCall.analyseCode(this.scope, constructorContext, flowInfo);
      }

      flowInfo.setReachMode(nonStaticFieldInfoReachMode);

      if (this.statements != null) {
        boolean didAlreadyComplain = false;
        int i = 0; for (int count = this.statements.length; i < count; i++) {
          Statement stat = this.statements[i];
          if (!stat.complainIfUnreachable(flowInfo, this.scope, didAlreadyComplain))
            flowInfo = stat.analyseCode(this.scope, constructorContext, flowInfo);
          else {
            didAlreadyComplain = true;
          }
        }
      }

      if ((flowInfo.tagBits & 0x1) == 0) {
        this.bits |= 64;
      }

      flowInfo.setReachMode(initialReachMode);

      if ((this.constructorCall != null) && 
        (this.constructorCall.accessMode != 3)) {
        flowInfo = flowInfo.mergedWith(constructorContext.initsOnReturn);
        FieldBinding[] fields = this.binding.declaringClass.fields();
        int i = 0; for (int count = fields.length; i < count; i++)
        {
          FieldBinding field;
          if ((!(field = fields[i]).isStatic()) && 
            (field.isFinal()) && 
            (!flowInfo.isDefinitelyAssigned(fields[i]))) {
            this.scope.problemReporter().uninitializedBlankFinalField(
              field, 
              (this.bits & 0x80) != 0 ? this.scope.referenceType() : this);
          }
        }
      }

      constructorContext.complainIfUnusedExceptionHandlers(this);
    } catch (AbortMethod localAbortMethod) {
      this.ignoreFurtherInvestigation = true;
    }
  }

  public void generateCode(ClassScope classScope, ClassFile classFile)
  {
    int problemResetPC = 0;
    if (this.ignoreFurtherInvestigation) {
      if (this.binding == null) {
        return;
      }
      CategorizedProblem[] problems = 
        this.scope.referenceCompilationUnit().compilationResult.getProblems();
      int problemsLength;
      CategorizedProblem[] problemsCopy = new CategorizedProblem[problemsLength = problems.length];
      System.arraycopy(problems, 0, problemsCopy, 0, problemsLength);
      classFile.addProblemConstructor(this, this.binding, problemsCopy);
      return;
    }
    try {
      problemResetPC = classFile.contentsOffset;
      internalGenerateCode(classScope, classFile);
    } catch (AbortMethod e) {
      if (e.compilationResult == CodeStream.RESTART_IN_WIDE_MODE)
      {
        try {
          classFile.contentsOffset = problemResetPC;
          classFile.methodCount -= 1;
          classFile.codeStream.wideMode = true;
          internalGenerateCode(classScope, classFile);
        }
        catch (AbortMethod localAbortMethod1) {
          CategorizedProblem[] problems = 
            this.scope.referenceCompilationUnit().compilationResult.getAllProblems();
          int problemsLength;
          CategorizedProblem[] problemsCopy = new CategorizedProblem[problemsLength = problems.length];
          System.arraycopy(problems, 0, problemsCopy, 0, problemsLength);
          classFile.addProblemConstructor(this, this.binding, problemsCopy, problemResetPC);
        }
      }
      else {
        CategorizedProblem[] problems = 
          this.scope.referenceCompilationUnit().compilationResult.getAllProblems();
        int problemsLength;
        CategorizedProblem[] problemsCopy = new CategorizedProblem[problemsLength = problems.length];
        System.arraycopy(problems, 0, problemsCopy, 0, problemsLength);
        classFile.addProblemConstructor(this, this.binding, problemsCopy, problemResetPC);
      }
    }
  }

  public void generateSyntheticFieldInitializationsIfNecessary(MethodScope methodScope, CodeStream codeStream, ReferenceBinding declaringClass) {
    if (!declaringClass.isNestedType()) return;

    NestedTypeBinding nestedType = (NestedTypeBinding)declaringClass;

    SyntheticArgumentBinding[] syntheticArgs = nestedType.syntheticEnclosingInstances();
    int i = 0; for (int max = syntheticArgs == null ? 0 : syntheticArgs.length; i < max; i++)
    {
      SyntheticArgumentBinding syntheticArg;
      if ((syntheticArg = syntheticArgs[i]).matchingField != null) {
        codeStream.aload_0();
        codeStream.load(syntheticArg);
        codeStream.putfield(syntheticArg.matchingField);
      }
    }
    syntheticArgs = nestedType.syntheticOuterLocalVariables();
    int i = 0; for (int max = syntheticArgs == null ? 0 : syntheticArgs.length; i < max; i++)
    {
      SyntheticArgumentBinding syntheticArg;
      if ((syntheticArg = syntheticArgs[i]).matchingField != null) {
        codeStream.aload_0();
        codeStream.load(syntheticArg);
        codeStream.putfield(syntheticArg.matchingField);
      }
    }
  }

  private void internalGenerateCode(ClassScope classScope, ClassFile classFile) {
    classFile.generateMethodInfoHeader(this.binding);
    int methodAttributeOffset = classFile.contentsOffset;
    int attributeNumber = classFile.generateMethodInfoAttribute(this.binding);
    if ((!this.binding.isNative()) && (!this.binding.isAbstract()))
    {
      TypeDeclaration declaringType = classScope.referenceContext;
      int codeAttributeOffset = classFile.contentsOffset;
      classFile.generateCodeAttributeHeader();
      CodeStream codeStream = classFile.codeStream;
      codeStream.reset(this, classFile);

      ReferenceBinding declaringClass = this.binding.declaringClass;

      int enumOffset = declaringClass.isEnum() ? 2 : 0;
      int argSlotSize = 1 + enumOffset;

      if (declaringClass.isNestedType()) {
        NestedTypeBinding nestedType = (NestedTypeBinding)declaringClass;
        this.scope.extraSyntheticArguments = nestedType.syntheticOuterLocalVariables();
        this.scope.computeLocalVariablePositions(
          nestedType.enclosingInstancesSlotSize + 1 + enumOffset, 
          codeStream);
        argSlotSize += nestedType.enclosingInstancesSlotSize;
        argSlotSize += nestedType.outerLocalVariablesSlotSize;
      } else {
        this.scope.computeLocalVariablePositions(1 + enumOffset, codeStream);
      }

      if (this.arguments != null) {
        int i = 0; for (int max = this.arguments.length; i < max; i++)
        {
          LocalVariableBinding argBinding;
          codeStream.addVisibleLocalVariable(argBinding = this.arguments[i].binding);
          argBinding.recordInitializationStartPC(0);
          TypeBinding argType;
          if (((argType = argBinding.type) == TypeBinding.LONG) || (argType == TypeBinding.DOUBLE))
            argSlotSize += 2;
          else {
            argSlotSize++;
          }
        }
      }

      MethodScope initializerScope = declaringType.initializerScope;
      initializerScope.computeLocalVariablePositions(argSlotSize, codeStream);

      boolean needFieldInitializations = (this.constructorCall == null) || (this.constructorCall.accessMode != 3);

      boolean preInitSyntheticFields = this.scope.compilerOptions().targetJDK >= 3145728L;

      if ((needFieldInitializations) && (preInitSyntheticFields)) {
        generateSyntheticFieldInitializationsIfNecessary(this.scope, codeStream, declaringClass);
      }

      if (this.constructorCall != null) {
        this.constructorCall.generateCode(this.scope, codeStream);
      }

      if (needFieldInitializations) {
        if (!preInitSyntheticFields) {
          generateSyntheticFieldInitializationsIfNecessary(this.scope, codeStream, declaringClass);
        }

        if (declaringType.fields != null) {
          int i = 0; for (int max = declaringType.fields.length; i < max; i++)
          {
            FieldDeclaration fieldDecl;
            if (!(fieldDecl = declaringType.fields[i]).isStatic()) {
              fieldDecl.generateCode(initializerScope, codeStream);
            }
          }
        }
      }

      if (this.statements != null) {
        int i = 0; for (int max = this.statements.length; i < max; i++) {
          this.statements[i].generateCode(this.scope, codeStream);
        }
      }
      if ((this.bits & 0x40) != 0) {
        codeStream.return_();
      }

      codeStream.exitUserScope(this.scope);
      codeStream.recordPositionsFrom(0, this.bodyEnd);
      classFile.completeCodeAttribute(codeAttributeOffset);
      attributeNumber++;
    }
    classFile.completeMethodInfo(methodAttributeOffset, attributeNumber);

    if (this.ignoreFurtherInvestigation)
      throw new AbortMethod(this.scope.referenceCompilationUnit().compilationResult, null);
  }

  public boolean isConstructor()
  {
    return true;
  }

  public boolean isDefaultConstructor() {
    return (this.bits & 0x80) != 0;
  }

  public boolean isInitializationMethod() {
    return true;
  }

  public boolean isRecursive(ArrayList visited)
  {
    if ((this.binding == null) || 
      (this.constructorCall == null) || 
      (this.constructorCall.binding == null) || 
      (this.constructorCall.isSuperAccess()) || 
      (!this.constructorCall.binding.isValidBinding())) {
      return false;
    }

    ConstructorDeclaration targetConstructor = 
      (ConstructorDeclaration)this.scope.referenceType().declarationOf(this.constructorCall.binding.original());
    if (this == targetConstructor) return true;

    if (visited == null) {
      visited = new ArrayList(1);
    } else {
      int index = visited.indexOf(this);
      if (index >= 0) return index == 0;
    }
    visited.add(this);

    return targetConstructor.isRecursive(visited);
  }

  public void parseStatements(Parser parser, CompilationUnitDeclaration unit)
  {
    if (this.ignoreFurtherInvestigation)
      return;
    if (((this.bits & 0x80) != 0) && (this.constructorCall == null)) {
      this.constructorCall = SuperReference.implicitSuperConstructorCall();
      this.constructorCall.sourceStart = this.sourceStart;
      this.constructorCall.sourceEnd = this.sourceEnd;
      return;
    }
    parser.parse(this, unit);
  }

  public StringBuffer printBody(int indent, StringBuffer output)
  {
    output.append(" {");
    if (this.constructorCall != null) {
      output.append('\n');
      this.constructorCall.printStatement(indent, output);
    }
    if (this.statements != null) {
      for (int i = 0; i < this.statements.length; i++) {
        output.append('\n');
        this.statements[i].printStatement(indent, output);
      }
    }
    output.append('\n');
    printIndent(indent == 0 ? 0 : indent - 1, output).append('}');
    return output;
  }

  public void resolveJavadoc() {
    if ((this.binding == null) || (this.javadoc != null))
      super.resolveJavadoc();
    else if ((this.bits & 0x80) == 0)
      this.scope.problemReporter().javadocMissing(this.sourceStart, this.sourceEnd, this.binding.modifiers);
  }

  public void resolveStatements()
  {
    SourceTypeBinding sourceType = this.scope.enclosingSourceType();
    if (!CharOperation.equals(sourceType.sourceName, this.selector)) {
      this.scope.problemReporter().missingReturnType(this);
    }
    if (this.typeParameters != null) {
      int i = 0; for (int length = this.typeParameters.length; i < length; i++) {
        this.typeParameters[i].resolve(this.scope);
      }
    }
    if ((this.binding != null) && (!this.binding.isPrivate())) {
      sourceType.tagBits |= 4503599627370496L;
    }

    if (this.constructorCall != null) {
      if ((sourceType.id == 1) && 
        (this.constructorCall.accessMode != 3))
      {
        if (this.constructorCall.accessMode == 2) {
          this.scope.problemReporter().cannotUseSuperInJavaLangObject(this.constructorCall);
        }
        this.constructorCall = null;
      } else {
        this.constructorCall.resolve(this.scope);
      }
    }
    if ((this.modifiers & 0x1000000) != 0) {
      this.scope.problemReporter().methodNeedBody(this);
    }
    super.resolveStatements();
  }

  public void traverse(ASTVisitor visitor, ClassScope classScope) {
    if (visitor.visit(this, classScope)) {
      if (this.javadoc != null) {
        this.javadoc.traverse(visitor, this.scope);
      }
      if (this.annotations != null) {
        int annotationsLength = this.annotations.length;
        for (int i = 0; i < annotationsLength; i++)
          this.annotations[i].traverse(visitor, this.scope);
      }
      if (this.typeParameters != null) {
        int typeParametersLength = this.typeParameters.length;
        for (int i = 0; i < typeParametersLength; i++) {
          this.typeParameters[i].traverse(visitor, this.scope);
        }
      }
      if (this.arguments != null) {
        int argumentLength = this.arguments.length;
        for (int i = 0; i < argumentLength; i++)
          this.arguments[i].traverse(visitor, this.scope);
      }
      if (this.thrownExceptions != null) {
        int thrownExceptionsLength = this.thrownExceptions.length;
        for (int i = 0; i < thrownExceptionsLength; i++)
          this.thrownExceptions[i].traverse(visitor, this.scope);
      }
      if (this.constructorCall != null)
        this.constructorCall.traverse(visitor, this.scope);
      if (this.statements != null) {
        int statementsLength = this.statements.length;
        for (int i = 0; i < statementsLength; i++)
          this.statements[i].traverse(visitor, this.scope);
      }
    }
    visitor.endVisit(this, classScope);
  }
  public TypeParameter[] typeParameters() {
    return this.typeParameters;
  }
}