package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.parser.Parser;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class Initializer extends FieldDeclaration
{
  public Block block;
  public int lastVisibleFieldID;
  public int bodyStart;
  public int bodyEnd;

  public Initializer(Block block, int modifiers)
  {
    this.block = block;
    this.modifiers = modifiers;

    this.declarationSourceStart = (this.sourceStart = block.sourceStart);
  }

  public FlowInfo analyseCode(MethodScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    return this.block.analyseCode(currentScope, flowContext, flowInfo);
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream)
  {
    if ((this.bits & 0x80000000) == 0) {
      return;
    }
    int pc = codeStream.position;
    this.block.generateCode(currentScope, codeStream);
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }

  public int getKind()
  {
    return 2;
  }

  public boolean isStatic()
  {
    return (this.modifiers & 0x8) != 0;
  }

  public void parseStatements(Parser parser, TypeDeclaration typeDeclaration, CompilationUnitDeclaration unit)
  {
    parser.parse(this, typeDeclaration, unit);
  }

  public StringBuffer printStatement(int indent, StringBuffer output)
  {
    if (this.modifiers != 0) {
      printIndent(indent, output);
      printModifiers(this.modifiers, output);
      if (this.annotations != null) printAnnotations(this.annotations, output);
      output.append("{\n");
      this.block.printBody(indent, output);
      printIndent(indent, output).append('}');
      return output;
    }
    return this.block.printStatement(indent, output);
  }

  public void resolve(MethodScope scope)
  {
    FieldBinding previousField = scope.initializedField;
    int previousFieldID = scope.lastVisibleFieldID;
    try {
      scope.initializedField = null;
      scope.lastVisibleFieldID = this.lastVisibleFieldID;
      if (isStatic()) {
        ReferenceBinding declaringType = scope.enclosingSourceType();
        if ((declaringType.isNestedType()) && (!declaringType.isStatic()))
          scope.problemReporter().innerTypesCannotDeclareStaticInitializers(
            declaringType, 
            this);
      }
      this.block.resolve(scope);
    } finally {
      scope.initializedField = previousField;
      scope.lastVisibleFieldID = previousFieldID;
    }
  }

  public void traverse(ASTVisitor visitor, MethodScope scope)
  {
    if (visitor.visit(this, scope)) {
      this.block.traverse(visitor, scope);
    }
    visitor.endVisit(this, scope);
  }
}