package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.LabelFlowContext;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class LabeledStatement extends Statement
{
  public Statement statement;
  public char[] label;
  public BranchLabel targetLabel;
  public int labelEnd;
  int mergedInitStateIndex = -1;

  public LabeledStatement(char[] label, Statement statement, long labelPosition, int sourceEnd)
  {
    this.statement = statement;

    if ((statement instanceof EmptyStatement)) statement.bits |= 1;
    this.label = label;
    this.sourceStart = ((int)(labelPosition >>> 32));
    this.labelEnd = ((int)labelPosition);
    this.sourceEnd = sourceEnd;
  }

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    if (this.statement == null)
      return flowInfo;
    LabelFlowContext labelContext;
    FlowInfo statementInfo;
    FlowInfo mergedInfo;
    FlowInfo mergedInfo;
    if (((statementInfo = this.statement
      .analyseCode(
      currentScope, 
      labelContext = 
      new LabelFlowContext(
      flowContext, 
      this, 
      this.label, 
      this.targetLabel = new BranchLabel(), 
      currentScope), 
      flowInfo)).tagBits & 0x1) != 0)
    {
      FlowInfo mergedInfo;
      if ((labelContext.initsOnBreak.tagBits & 0x1) == 0)
      {
        mergedInfo = flowInfo.unconditionalCopy()
          .addInitializationsFrom(labelContext.initsOnBreak);
      }
      else mergedInfo = labelContext.initsOnBreak; 
    }
    else
    {
      mergedInfo = statementInfo.mergedWith(labelContext.initsOnBreak);
    }
    this.mergedInitStateIndex = 
      currentScope.methodScope().recordInitializationStates(mergedInfo);
    if ((this.bits & 0x40) == 0) {
      currentScope.problemReporter().unusedLabel(this);
    }
    return mergedInfo;
  }

  public ASTNode concreteStatement()
  {
    return this.statement;
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream)
  {
    if ((this.bits & 0x80000000) == 0) {
      return;
    }
    int pc = codeStream.position;
    if (this.targetLabel != null) {
      this.targetLabel.initialize(codeStream);
      if (this.statement != null) {
        this.statement.generateCode(currentScope, codeStream);
      }
      this.targetLabel.place();
    }

    if (this.mergedInitStateIndex != -1) {
      codeStream.removeNotDefinitelyAssignedVariables(currentScope, this.mergedInitStateIndex);
      codeStream.addDefinitelyAssignedVariables(currentScope, this.mergedInitStateIndex);
    }
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }

  public StringBuffer printStatement(int tab, StringBuffer output)
  {
    printIndent(tab, output).append(this.label).append(": ");
    if (this.statement == null)
      output.append(';');
    else
      this.statement.printStatement(0, output);
    return output;
  }

  public void resolve(BlockScope scope)
  {
    if (this.statement != null)
      this.statement.resolve(scope);
  }

  public void traverse(ASTVisitor visitor, BlockScope blockScope)
  {
    if ((visitor.visit(this, blockScope)) && 
      (this.statement != null)) this.statement.traverse(visitor, blockScope);

    visitor.endVisit(this, blockScope);
  }
}