package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.InvocationSite;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ProblemMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class MessageSend extends Expression
  implements InvocationSite
{
  public Expression receiver;
  public char[] selector;
  public Expression[] arguments;
  public MethodBinding binding;
  protected MethodBinding codegenBinding;
  MethodBinding syntheticAccessor;
  public TypeBinding expectedType;
  public long nameSourcePosition;
  public TypeBinding actualReceiverType;
  public TypeBinding receiverGenericCast;
  public TypeBinding valueCast;
  public TypeReference[] typeArguments;
  public TypeBinding[] genericTypeArguments;

  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo)
  {
    boolean nonStatic = !this.binding.isStatic();
    flowInfo = this.receiver.analyseCode(currentScope, flowContext, flowInfo, nonStatic).unconditionalInits();
    if (nonStatic) {
      this.receiver.checkNPE(currentScope, flowContext, flowInfo);
    }

    if (this.arguments != null) {
      int length = this.arguments.length;
      for (int i = 0; i < length; i++)
        flowInfo = this.arguments[i].analyseCode(currentScope, flowContext, flowInfo).unconditionalInits();
    }
    ReferenceBinding[] thrownExceptions;
    if ((thrownExceptions = this.binding.thrownExceptions) != Binding.NO_EXCEPTIONS)
    {
      flowContext.checkExceptionHandlers(thrownExceptions, this, flowInfo.copy(), currentScope);
    }

    manageSyntheticAccessIfNecessary(currentScope, flowInfo);
    return flowInfo;
  }

  public void computeConversion(Scope scope, TypeBinding runtimeTimeType, TypeBinding compileTimeType)
  {
    if ((runtimeTimeType == null) || (compileTimeType == null)) {
      return;
    }
    if ((this.binding != null) && (this.binding.isValidBinding())) {
      MethodBinding originalBinding = this.binding.original();
      TypeBinding originalType = originalBinding.returnType;

      if ((originalBinding != this.binding) && 
        (originalType != this.binding.returnType) && 
        (runtimeTimeType.id != 1) && 
        ((originalType.tagBits & 0x20000000) != 0L)) {
        TypeBinding targetType = (!compileTimeType.isBaseType()) && (runtimeTimeType.isBaseType()) ? 
          compileTimeType : 
          runtimeTimeType;
        this.valueCast = originalType.genericCast(targetType);
      } else if ((this.actualReceiverType.isArrayType()) && 
        (runtimeTimeType.id != 1) && 
        (this.binding.parameters == Binding.NO_PARAMETERS) && 
        (scope.compilerOptions().complianceLevel >= 3211264L) && 
        (CharOperation.equals(this.binding.selector, TypeConstants.CLONE)))
      {
        this.valueCast = runtimeTimeType;
      }
    }
    super.computeConversion(scope, runtimeTimeType, compileTimeType);
  }

  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired)
  {
    int pc = codeStream.position;

    boolean isStatic = this.codegenBinding.isStatic();
    if (isStatic) {
      this.receiver.generateCode(currentScope, codeStream, false);
      codeStream.recordPositionsFrom(pc, this.sourceStart);
    } else if (((this.bits & 0x1FE0) != 0) && (this.receiver.isImplicitThis()))
    {
      ReferenceBinding targetType = currentScope.enclosingSourceType().enclosingTypeAt((this.bits & 0x1FE0) >> 5);
      Object[] path = currentScope.getEmulationPath(targetType, true, false);
      codeStream.generateOuterAccess(path, this, targetType, currentScope);
    } else {
      this.receiver.generateCode(currentScope, codeStream, true);
      if (this.receiverGenericCast != null)
        codeStream.checkcast(this.receiverGenericCast);
      codeStream.recordPositionsFrom(pc, this.sourceStart);
    }

    generateArguments(this.binding, this.arguments, currentScope, codeStream);

    if (this.syntheticAccessor == null) {
      if (isStatic) {
        codeStream.invokestatic(this.codegenBinding);
      }
      else if ((this.receiver.isSuper()) || (this.codegenBinding.isPrivate())) {
        codeStream.invokespecial(this.codegenBinding);
      }
      else if (this.codegenBinding.declaringClass.isInterface())
        codeStream.invokeinterface(this.codegenBinding);
      else {
        codeStream.invokevirtual(this.codegenBinding);
      }
    }
    else
    {
      codeStream.invokestatic(this.syntheticAccessor);
    }

    if (this.valueCast != null) codeStream.checkcast(this.valueCast);
    if (valueRequired)
    {
      codeStream.generateImplicitConversion(this.implicitConversion);
    } else {
      boolean isUnboxing = (this.implicitConversion & 0x400) != 0;

      if (isUnboxing) codeStream.generateImplicitConversion(this.implicitConversion);
      switch (isUnboxing ? postConversionType(currentScope).id : this.codegenBinding.returnType.id) {
      case 7:
      case 8:
        codeStream.pop2();
        break;
      case 6:
        break;
      default:
        codeStream.pop();
      }
    }
    codeStream.recordPositionsFrom(pc, (int)(this.nameSourcePosition >>> 32));
  }

  public TypeBinding[] genericTypeArguments()
  {
    return this.genericTypeArguments;
  }

  public boolean isSuperAccess() {
    return this.receiver.isSuper();
  }
  public boolean isTypeAccess() {
    return (this.receiver != null) && (this.receiver.isTypeReference());
  }

  public void manageSyntheticAccessIfNecessary(BlockScope currentScope, FlowInfo flowInfo) {
    if ((flowInfo.tagBits & 0x1) != 0) return;

    this.codegenBinding = this.binding.original();
    if (this.binding.isPrivate())
    {
      if (currentScope.enclosingSourceType() != this.codegenBinding.declaringClass)
      {
        this.syntheticAccessor = ((SourceTypeBinding)this.codegenBinding.declaringClass).addSyntheticMethod(this.codegenBinding, isSuperAccess());
        currentScope.problemReporter().needToEmulateMethodAccess(this.codegenBinding, this);
      }
    }
    else {
      if ((this.receiver instanceof QualifiedSuperReference))
      {
        SourceTypeBinding destinationType = (SourceTypeBinding)((QualifiedSuperReference)this.receiver).currentCompatibleType;
        this.syntheticAccessor = destinationType.addSyntheticMethod(this.codegenBinding, isSuperAccess());
        currentScope.problemReporter().needToEmulateMethodAccess(this.codegenBinding, this);
        return;
      }
      if (this.binding.isProtected())
      {
        SourceTypeBinding enclosingSourceType;
        if (((this.bits & 0x1FE0) != 0) && 
          (this.codegenBinding.declaringClass.getPackage() != 
          (enclosingSourceType = currentScope.enclosingSourceType()).getPackage()))
        {
          SourceTypeBinding currentCompatibleType = (SourceTypeBinding)enclosingSourceType.enclosingTypeAt((this.bits & 0x1FE0) >> 5);
          this.syntheticAccessor = currentCompatibleType.addSyntheticMethod(this.codegenBinding, isSuperAccess());
          currentScope.problemReporter().needToEmulateMethodAccess(this.codegenBinding, this);
          return;
        }

      }

    }

    if ((this.binding.declaringClass != this.actualReceiverType) && 
      (this.receiverGenericCast == null) && 
      (!this.actualReceiverType.isArrayType())) {
      CompilerOptions options = currentScope.compilerOptions();
      if (((options.targetJDK >= 3014656L) && 
        ((options.complianceLevel >= 3145728L) || (!this.receiver.isImplicitThis()) || (!this.codegenBinding.isStatic())) && 
        (this.binding.declaringClass.id != 1)) || 
        (!this.binding.declaringClass.canBeSeenBy(currentScope)))
      {
        this.codegenBinding = currentScope.enclosingSourceType()
          .getUpdatedMethodBinding(this.codegenBinding, (ReferenceBinding)this.actualReceiverType.erasure());
      }
    }
  }

  public int nullStatus(FlowInfo flowInfo)
  {
    return 0;
  }

  public TypeBinding postConversionType(Scope scope)
  {
    TypeBinding convertedType = this.resolvedType;
    if (this.valueCast != null)
      convertedType = this.valueCast;
    int runtimeType = (this.implicitConversion & 0xFF) >> 4;
    switch (runtimeType) {
    case 5:
      convertedType = TypeBinding.BOOLEAN;
      break;
    case 3:
      convertedType = TypeBinding.BYTE;
      break;
    case 4:
      convertedType = TypeBinding.SHORT;
      break;
    case 2:
      convertedType = TypeBinding.CHAR;
      break;
    case 10:
      convertedType = TypeBinding.INT;
      break;
    case 9:
      convertedType = TypeBinding.FLOAT;
      break;
    case 7:
      convertedType = TypeBinding.LONG;
      break;
    case 8:
      convertedType = TypeBinding.DOUBLE;
    case 6:
    }

    if ((this.implicitConversion & 0x200) != 0) {
      convertedType = scope.environment().computeBoxingType(convertedType);
    }
    return convertedType;
  }

  public StringBuffer printExpression(int indent, StringBuffer output)
  {
    if (!this.receiver.isImplicitThis()) this.receiver.printExpression(0, output).append('.');
    if (this.typeArguments != null) {
      output.append('<');
      int max = this.typeArguments.length - 1;
      for (int j = 0; j < max; j++) {
        this.typeArguments[j].print(0, output);
        output.append(", ");
      }
      this.typeArguments[max].print(0, output);
      output.append('>');
    }
    output.append(this.selector).append('(');
    if (this.arguments != null) {
      for (int i = 0; i < this.arguments.length; i++) {
        if (i > 0) output.append(", ");
        this.arguments[i].printExpression(0, output);
      }
    }
    return output.append(')');
  }

  public TypeBinding resolveType(BlockScope scope)
  {
    this.constant = Constant.NotAConstant;
    boolean receiverCast = false; boolean argsContainCast = false;
    if ((this.receiver instanceof CastExpression)) {
      this.receiver.bits |= 32;
      receiverCast = true;
    }
    this.actualReceiverType = this.receiver.resolveType(scope);
    boolean receiverIsType = ((this.receiver instanceof NameReference)) && ((((NameReference)this.receiver).bits & 0x4) != 0);
    if ((receiverCast) && (this.actualReceiverType != null))
    {
      if (((CastExpression)this.receiver).expression.resolvedType == this.actualReceiverType) {
        scope.problemReporter().unnecessaryCast((CastExpression)this.receiver);
      }
    }

    if (this.typeArguments != null) {
      int length = this.typeArguments.length;
      boolean argHasError = false;
      this.genericTypeArguments = new TypeBinding[length];
      for (int i = 0; i < length; i++) {
        TypeReference typeReference = this.typeArguments[i];
        if ((this.genericTypeArguments[i] =  = typeReference.resolveType(scope, true)) == null) {
          argHasError = true;
        }
        if ((argHasError) && ((typeReference instanceof Wildcard))) {
          scope.problemReporter().illegalUsageOfWildcard(typeReference);
        }
      }
      if (argHasError) {
        return null;
      }
    }

    TypeBinding[] argumentTypes = Binding.NO_PARAMETERS;
    if (this.arguments != null) {
      boolean argHasError = false;
      int length = this.arguments.length;
      argumentTypes = new TypeBinding[length];
      for (int i = 0; i < length; i++) {
        Expression argument = this.arguments[i];
        if ((argument instanceof CastExpression)) {
          argument.bits |= 32;
          argsContainCast = true;
        }
        if ((argumentTypes[i] =  = argument.resolveType(scope)) == null) {
          argHasError = true;
        }
      }
      if (argHasError) {
        if ((this.actualReceiverType instanceof ReferenceBinding))
        {
          TypeBinding[] pseudoArgs = new TypeBinding[length];
          int i = length;
          do { pseudoArgs[i] = (argumentTypes[i] == null ? TypeBinding.NULL : argumentTypes[i]);

            i--; } while (i >= 0);

          this.binding = 
            (this.receiver.isImplicitThis() ? 
            scope.getImplicitMethod(this.selector, pseudoArgs, this) : 
            scope.findMethod((ReferenceBinding)this.actualReceiverType, this.selector, pseudoArgs, this));
          if ((this.binding != null) && (!this.binding.isValidBinding())) {
            MethodBinding closestMatch = ((ProblemMethodBinding)this.binding).closestMatch;

            if (closestMatch != null) {
              if (closestMatch.original().typeVariables != Binding.NO_TYPE_VARIABLES)
              {
                closestMatch = scope.environment().createParameterizedGenericMethod(closestMatch.original(), null);
              }
              this.binding = closestMatch;
              MethodBinding closestMatchOriginal = closestMatch.original();
              if (((closestMatchOriginal.isPrivate()) || (closestMatchOriginal.declaringClass.isLocalType())) && (!scope.isDefinedInMethod(closestMatchOriginal)))
              {
                closestMatchOriginal.modifiers |= 134217728;
              }
            }
          }
        }
        return null;
      }
    }
    if (this.actualReceiverType == null) {
      return null;
    }

    if (this.actualReceiverType.isBaseType()) {
      scope.problemReporter().errorNoMethodFor(this, this.actualReceiverType, argumentTypes);
      return null;
    }
    this.binding = 
      (this.receiver.isImplicitThis() ? 
      scope.getImplicitMethod(this.selector, argumentTypes, this) : 
      scope.getMethod(this.actualReceiverType, this.selector, argumentTypes, this));
    if (!this.binding.isValidBinding()) {
      if (this.binding.declaringClass == null) {
        if ((this.actualReceiverType instanceof ReferenceBinding)) {
          this.binding.declaringClass = ((ReferenceBinding)this.actualReceiverType);
        } else {
          scope.problemReporter().errorNoMethodFor(this, this.actualReceiverType, argumentTypes);
          return null;
        }
      }
      scope.problemReporter().invalidMethod(this, this.binding);
      MethodBinding closestMatch = ((ProblemMethodBinding)this.binding).closestMatch;
      switch (this.binding.problemId()) {
      case 3:
        break;
      case 2:
      case 6:
      case 7:
      case 8:
      case 10:
        if (closestMatch != null) this.resolvedType = closestMatch.returnType; break;
      case 4:
      case 5:
      case 9:
      }if (closestMatch != null) {
        this.binding = closestMatch;
        MethodBinding closestMatchOriginal = closestMatch.original();
        if (((closestMatchOriginal.isPrivate()) || (closestMatchOriginal.declaringClass.isLocalType())) && (!scope.isDefinedInMethod(closestMatchOriginal)))
        {
          closestMatchOriginal.modifiers |= 134217728;
        }
      }
      return this.resolvedType;
    }
    CompilerOptions compilerOptions = scope.compilerOptions();
    if (!this.binding.isStatic())
    {
      if (receiverIsType) {
        scope.problemReporter().mustUseAStaticMethod(this, this.binding);
        if ((this.actualReceiverType.isRawType()) && 
          ((this.receiver.bits & 0x40000000) == 0) && 
          (compilerOptions.getSeverity(35184372088832L) != -1))
          scope.problemReporter().rawTypeReference(this.receiver, this.actualReceiverType);
      }
      else {
        this.receiver.computeConversion(scope, this.actualReceiverType, this.actualReceiverType);

        TypeBinding receiverErasure = this.actualReceiverType.erasure();
        if (((receiverErasure instanceof ReferenceBinding)) && 
          (receiverErasure.findSuperTypeWithSameErasure(this.binding.declaringClass) == null)) {
          this.receiverGenericCast = this.binding.declaringClass;
        }
      }
    }
    else
    {
      if ((!this.receiver.isImplicitThis()) && (!this.receiver.isSuper()) && (!receiverIsType)) {
        scope.problemReporter().nonStaticAccessToStaticMethod(this, this.binding);
      }
      if ((!this.receiver.isImplicitThis()) && (this.binding.declaringClass != this.actualReceiverType)) {
        scope.problemReporter().indirectAccessToStaticMethod(this, this.binding);
      }
    }
    checkInvocationArguments(scope, this.receiver, this.actualReceiverType, this.binding, this.arguments, argumentTypes, argsContainCast, this);

    if ((this.binding.isAbstract()) && 
      (this.receiver.isSuper())) {
      scope.problemReporter().cannotDireclyInvokeAbstractMethod(this, this.binding);
    }

    if (isMethodUseDeprecated(this.binding, scope, true)) {
      scope.problemReporter().deprecatedMethod(this.binding, this);
    }

    if ((this.actualReceiverType.isArrayType()) && 
      (this.binding.parameters == Binding.NO_PARAMETERS) && 
      (compilerOptions.complianceLevel >= 3211264L) && 
      (CharOperation.equals(this.binding.selector, TypeConstants.CLONE))) {
      this.resolvedType = this.actualReceiverType;
    } else {
      TypeBinding returnType = this.binding.returnType;
      if (returnType != null) returnType = returnType.capture(scope, this.sourceEnd);
      this.resolvedType = returnType;
    }
    if ((this.receiver.isSuper()) && (compilerOptions.getSeverity(562949953421312L) != -1)) {
      ReferenceContext referenceContext = scope.methodScope().referenceContext;
      if ((referenceContext instanceof AbstractMethodDeclaration)) {
        AbstractMethodDeclaration abstractMethodDeclaration = (AbstractMethodDeclaration)referenceContext;
        MethodBinding enclosingMethodBinding = abstractMethodDeclaration.binding;
        if ((enclosingMethodBinding.isOverriding()) && 
          (CharOperation.equals(this.binding.selector, enclosingMethodBinding.selector)) && 
          (this.binding.areParametersEqual(enclosingMethodBinding))) {
          abstractMethodDeclaration.bits |= 16;
        }
      }
    }
    return this.resolvedType;
  }

  public void setActualReceiverType(ReferenceBinding receiverType) {
    if (receiverType == null) return;
    this.actualReceiverType = receiverType;
  }
  public void setDepth(int depth) {
    this.bits &= -8161;
    if (depth > 0)
      this.bits |= (depth & 0xFF) << 5;
  }

  public void setExpectedType(TypeBinding expectedType)
  {
    this.expectedType = expectedType;
  }

  public void setFieldIndex(int depth) {
  }

  public void traverse(ASTVisitor visitor, BlockScope blockScope) {
    if (visitor.visit(this, blockScope)) {
      this.receiver.traverse(visitor, blockScope);
      if (this.typeArguments != null) {
        int i = 0; for (int typeArgumentsLength = this.typeArguments.length; i < typeArgumentsLength; i++) {
          this.typeArguments[i].traverse(visitor, blockScope);
        }
      }
      if (this.arguments != null) {
        int argumentsLength = this.arguments.length;
        for (int i = 0; i < argumentsLength; i++)
          this.arguments[i].traverse(visitor, blockScope);
      }
    }
    visitor.endVisit(this, blockScope);
  }
}