package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.PackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class QualifiedTypeReference extends TypeReference
{
  public char[][] tokens;
  public long[] sourcePositions;

  public QualifiedTypeReference(char[][] sources, long[] poss)
  {
    this.tokens = sources;
    this.sourcePositions = poss;
    this.sourceStart = ((int)(this.sourcePositions[0] >>> 32));
    this.sourceEnd = ((int)(this.sourcePositions[(this.sourcePositions.length - 1)] & 0xFFFFFFFF));
  }

  public TypeReference copyDims(int dim)
  {
    return new ArrayQualifiedTypeReference(this.tokens, dim, this.sourcePositions);
  }

  protected TypeBinding findNextTypeBinding(int tokenIndex, Scope scope, PackageBinding packageBinding) {
    LookupEnvironment env = scope.environment();
    try {
      env.missingClassFileLocation = this;
      if (this.resolvedType == null) {
        this.resolvedType = scope.getType(this.tokens[tokenIndex], packageBinding);
      } else {
        this.resolvedType = scope.getMemberType(this.tokens[tokenIndex], (ReferenceBinding)this.resolvedType);
        if ((this.resolvedType instanceof ProblemReferenceBinding)) {
          ProblemReferenceBinding problemBinding = (ProblemReferenceBinding)this.resolvedType;
          this.resolvedType = new ProblemReferenceBinding(
            CharOperation.subarray(this.tokens, 0, tokenIndex + 1), 
            problemBinding.closestMatch(), 
            this.resolvedType.problemId());
        }
      }
      return this.resolvedType;
    } catch (AbortCompilation e) {
      e.updateContext(this, scope.referenceCompilationUnit().compilationResult);
      throw e;
    } finally {
      env.missingClassFileLocation = null;
    }
  }

  public char[] getLastToken() {
    return this.tokens[(this.tokens.length - 1)];
  }

  protected TypeBinding getTypeBinding(Scope scope) {
    if (this.resolvedType != null) {
      return this.resolvedType;
    }
    Binding binding = scope.getPackage(this.tokens);
    if ((binding != null) && (!binding.isValidBinding())) {
      return (ReferenceBinding)binding;
    }
    PackageBinding packageBinding = binding == null ? null : (PackageBinding)binding;
    boolean isClassScope = scope.kind == 3;
    ReferenceBinding qualifiedType = null;
    int i = packageBinding == null ? 0 : packageBinding.compoundName.length; int max = this.tokens.length; for (int last = max - 1; i < max; i++) {
      findNextTypeBinding(i, scope, packageBinding);
      if (!this.resolvedType.isValidBinding())
        return this.resolvedType;
      if ((i == 0) && (this.resolvedType.isTypeVariable()) && (((TypeVariableBinding)this.resolvedType).firstBound == null)) {
        scope.problemReporter().illegalAccessFromTypeVariable((TypeVariableBinding)this.resolvedType, this);
        return this.resolvedType = null;
      }
      if ((i < last) && (isTypeUseDeprecated(this.resolvedType, scope))) {
        reportDeprecatedType(this.resolvedType, scope);
      }
      if ((isClassScope) && 
        (((ClassScope)scope).detectHierarchyCycle(this.resolvedType, this)))
        return null;
      ReferenceBinding currentType = (ReferenceBinding)this.resolvedType;
      if (qualifiedType != null)
      {
        if (currentType.isGenericType()) {
          qualifiedType = scope.environment().createRawType(currentType, qualifiedType);
        }
        else
        {
          boolean rawQualified;
          if (((rawQualified = qualifiedType.isRawType())) && (!currentType.isStatic()))
            qualifiedType = scope.environment().createRawType((ReferenceBinding)currentType.erasure(), qualifiedType);
          else if (((rawQualified) || (qualifiedType.isParameterizedType())) && (qualifiedType.erasure() == currentType.enclosingType().erasure()))
            qualifiedType = scope.environment().createParameterizedType((ReferenceBinding)currentType.erasure(), null, qualifiedType);
          else
            qualifiedType = currentType;
        }
      }
      else qualifiedType = currentType.isGenericType() ? (ReferenceBinding)scope.environment().convertToRawType(currentType) : currentType;
    }

    this.resolvedType = qualifiedType;
    return this.resolvedType;
  }

  public char[][] getTypeName()
  {
    return this.tokens;
  }

  public StringBuffer printExpression(int indent, StringBuffer output)
  {
    for (int i = 0; i < this.tokens.length; i++) {
      if (i > 0) output.append('.');
      output.append(this.tokens[i]);
    }
    return output;
  }

  public void traverse(ASTVisitor visitor, BlockScope scope)
  {
    visitor.visit(this, scope);
    visitor.endVisit(this, scope);
  }

  public void traverse(ASTVisitor visitor, ClassScope scope)
  {
    visitor.visit(this, scope);
    visitor.endVisit(this, scope);
  }
}