package org.eclipse.jdt.internal.compiler.batch;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.classfmt.ClassFileReader;
import org.eclipse.jdt.internal.compiler.classfmt.ClassFormatException;
import org.eclipse.jdt.internal.compiler.env.AccessRuleSet;
import org.eclipse.jdt.internal.compiler.env.NameEnvironmentAnswer;
import org.eclipse.jdt.internal.compiler.util.Util;

public class ClasspathJar extends ClasspathLocation
{
  protected File file;
  protected ZipFile zipFile;
  protected boolean closeZipFileAtEnd;
  protected Hashtable packageCache;
  protected char[] normalizedPath;

  public ClasspathJar(File file, boolean closeZipFileAtEnd, AccessRuleSet accessRuleSet, String destinationPath)
  {
    super(accessRuleSet, destinationPath);
    this.file = file;
    this.closeZipFileAtEnd = closeZipFileAtEnd;
  }
  public NameEnvironmentAnswer findClass(char[] typeName, String qualifiedPackageName, String qualifiedBinaryFileName) {
    return findClass(typeName, qualifiedPackageName, qualifiedBinaryFileName, false);
  }
  public NameEnvironmentAnswer findClass(char[] typeName, String qualifiedPackageName, String qualifiedBinaryFileName, boolean asBinaryOnly) {
    if (!isPackage(qualifiedPackageName))
      return null;
    try
    {
      ClassFileReader reader = ClassFileReader.read(this.zipFile, qualifiedBinaryFileName);
      if (reader != null)
        return new NameEnvironmentAnswer(reader, fetchAccessRestriction(qualifiedBinaryFileName));
    }
    catch (ClassFormatException localClassFormatException) {
    }
    catch (IOException localIOException) {
    }
    return null;
  }
  public char[][][] findTypeNames(String qualifiedPackageName) {
    if (!isPackage(qualifiedPackageName)) {
      return null;
    }
    ArrayList answers = new ArrayList();
    for (Enumeration e = this.zipFile.entries(); e.hasMoreElements(); ) {
      String fileName = ((ZipEntry)e.nextElement()).getName();

      int last = fileName.lastIndexOf('/');
      while (last > 0)
      {
        String packageName = fileName.substring(0, last);
        if (!qualifiedPackageName.equals(packageName))
          break;
        int indexOfDot = fileName.lastIndexOf('.');
        if (indexOfDot != -1) {
          String typeName = fileName.substring(last + 1, indexOfDot);
          char[] packageArray = packageName.toCharArray();
          answers.add(
            CharOperation.arrayConcat(
            CharOperation.splitOn('/', packageArray), 
            typeName.toCharArray()));
        }
      }
    }
    int size = answers.size();
    if (size != 0) {
      char[][][] result = new char[size][][];
      answers.toArray(result);
      return null;
    }
    return null;
  }
  public void initialize() throws IOException {
    this.zipFile = new ZipFile(this.file);
  }
  public boolean isPackage(String qualifiedPackageName) {
    if (this.packageCache != null) {
      return this.packageCache.containsKey(qualifiedPackageName);
    }
    this.packageCache = new Hashtable(41);
    this.packageCache.put(Util.EMPTY_STRING, Util.EMPTY_STRING);

    for (Enumeration e = this.zipFile.entries(); e.hasMoreElements(); ) {
      String fileName = ((ZipEntry)e.nextElement()).getName();

      int last = fileName.lastIndexOf('/');
      while (last > 0)
      {
        String packageName = fileName.substring(0, last);
        if (this.packageCache.containsKey(packageName))
          break;
        this.packageCache.put(packageName, packageName);
        last = packageName.lastIndexOf('/');
      }
    }
    return this.packageCache.containsKey(qualifiedPackageName);
  }
  public void reset() {
    if ((this.zipFile != null) && (this.closeZipFileAtEnd)) {
      try {
        this.zipFile.close();
      }
      catch (IOException localIOException) {
      }
      this.zipFile = null;
    }
    this.packageCache = null;
  }
  public String toString() {
    return "Classpath for jar file " + this.file.getPath();
  }
  public char[] normalizedPath() {
    if (this.normalizedPath == null) {
      char[] rawName = this.file.getPath().toCharArray();
      if (File.separatorChar == '\\') {
        CharOperation.replace(rawName, '\\', '/');
      }
      this.normalizedPath = CharOperation.subarray(rawName, 0, CharOperation.lastIndexOf('.', rawName));
    }
    return this.normalizedPath;
  }
  public String getPath() {
    return this.file.getPath();
  }
}