package org.eclipse.jdt.internal.compiler.flow;

import java.util.ArrayList;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.ast.SubRoutineStatement;
import org.eclipse.jdt.internal.compiler.ast.TryStatement;
import org.eclipse.jdt.internal.compiler.codegen.ObjectCache;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class ExceptionHandlingFlowContext extends FlowContext
{
  public static final int BitCacheSize = 32;
  public ReferenceBinding[] handledExceptions;
  int[] isReached;
  int[] isNeeded;
  UnconditionalFlowInfo[] initsOnExceptions;
  ObjectCache indexes = new ObjectCache();
  boolean isMethodContext;
  public UnconditionalFlowInfo initsOnReturn;
  public ArrayList extendedExceptions;

  public ExceptionHandlingFlowContext(FlowContext parent, ASTNode associatedNode, ReferenceBinding[] handledExceptions, BlockScope scope, UnconditionalFlowInfo flowInfo)
  {
    super(parent, associatedNode);
    this.isMethodContext = (scope == scope.methodScope());
    this.handledExceptions = handledExceptions;
    int count = handledExceptions.length; int cacheSize = count / 32 + 1;
    this.isReached = new int[cacheSize];
    this.isNeeded = new int[cacheSize];
    this.initsOnExceptions = new UnconditionalFlowInfo[count];
    for (int i = 0; i < count; i++) {
      this.indexes.put(handledExceptions[i], i);
      int cacheIndex = i / 32; int bitMask = 1 << i % 32;
      if (handledExceptions[i].isUncheckedException(true)) {
        this.isReached[cacheIndex] |= bitMask;
        this.initsOnExceptions[i] = flowInfo.unconditionalCopy();
      } else {
        this.initsOnExceptions[i] = FlowInfo.DEAD_END;
      }
    }
    System.arraycopy(this.isReached, 0, this.isNeeded, 0, cacheSize);
    this.initsOnReturn = FlowInfo.DEAD_END;
  }

  public void complainIfUnusedExceptionHandlers(AbstractMethodDeclaration method) {
    MethodScope scope = method.scope;

    if (((method.binding.modifiers & 0x30000000) != 0) && 
      (!scope.compilerOptions().reportUnusedDeclaredThrownExceptionWhenOverriding)) {
      return;
    }

    int i = 0; for (int count = this.handledExceptions.length; i < count; i++) {
      int index = this.indexes.get(this.handledExceptions[i]);
      int cacheIndex = index / 32;
      int bitMask = 1 << index % 32;
      if ((this.isReached[cacheIndex] & bitMask) == 0)
        scope.problemReporter().unusedDeclaredThrownException(
          this.handledExceptions[index], 
          method, 
          method.thrownExceptions[index]);
    }
  }

  public void complainIfUnusedExceptionHandlers(BlockScope scope, TryStatement tryStatement)
  {
    int i = 0; for (int count = this.handledExceptions.length; i < count; i++) {
      int index = this.indexes.get(this.handledExceptions[i]);
      int cacheIndex = index / 32;
      int bitMask = 1 << index % 32;
      if ((this.isReached[cacheIndex] & bitMask) == 0) {
        scope.problemReporter().unreachableCatchBlock(
          this.handledExceptions[index], 
          tryStatement.catchArguments[index].type);
      }
      else if ((this.isNeeded[cacheIndex] & bitMask) == 0)
        scope.problemReporter().hiddenCatchBlock(
          this.handledExceptions[index], 
          tryStatement.catchArguments[index].type);
    }
  }

  public String individualToString()
  {
    StringBuffer buffer = new StringBuffer("Exception flow context");
    int length = this.handledExceptions.length;
    for (int i = 0; i < length; i++) {
      int cacheIndex = i / 32;
      int bitMask = 1 << i % 32;
      buffer.append('[').append(this.handledExceptions[i].readableName());
      if ((this.isReached[cacheIndex] & bitMask) != 0) {
        if ((this.isNeeded[cacheIndex] & bitMask) == 0)
          buffer.append("-masked");
        else
          buffer.append("-reached");
      }
      else {
        buffer.append("-not reached");
      }
      buffer.append('-').append(this.initsOnExceptions[i].toString()).append(']');
    }
    buffer.append("[initsOnReturn -").append(this.initsOnReturn.toString()).append(']');
    return buffer.toString();
  }

  public UnconditionalFlowInfo initsOnException(ReferenceBinding exceptionType)
  {
    int index;
    if ((index = this.indexes.get(exceptionType)) < 0) {
      return FlowInfo.DEAD_END;
    }
    return this.initsOnExceptions[index];
  }

  public UnconditionalFlowInfo initsOnReturn() {
    return this.initsOnReturn;
  }

  public void mergeUnhandledException(TypeBinding newException)
  {
    if (this.extendedExceptions == null) {
      this.extendedExceptions = new ArrayList(5);
      for (int i = 0; i < this.handledExceptions.length; i++) {
        this.extendedExceptions.add(this.handledExceptions[i]);
      }
    }
    boolean isRedundant = false;

    for (int i = this.extendedExceptions.size() - 1; i >= 0; i--) {
      switch (Scope.compareTypes(newException, (TypeBinding)this.extendedExceptions.get(i))) {
      case 1:
        this.extendedExceptions.remove(i);
        break;
      case -1:
        isRedundant = true;
      case 0:
      }

    }

    if (!isRedundant)
      this.extendedExceptions.add(newException);
  }

  public void recordHandlingException(ReferenceBinding exceptionType, UnconditionalFlowInfo flowInfo, TypeBinding raisedException, ASTNode invocationSite, boolean wasAlreadyDefinitelyCaught)
  {
    int index = this.indexes.get(exceptionType);

    int cacheIndex = index / 32;
    int bitMask = 1 << index % 32;
    if (!wasAlreadyDefinitelyCaught) {
      this.isNeeded[cacheIndex] |= bitMask;
    }
    this.isReached[cacheIndex] |= bitMask;

    this.initsOnExceptions[index] = 
      ((this.initsOnExceptions[index].tagBits & 0x1) == 0 ? 
      this.initsOnExceptions[index].mergedWith(flowInfo) : 
      flowInfo.unconditionalCopy());
  }

  public void recordReturnFrom(UnconditionalFlowInfo flowInfo) {
    if ((flowInfo.tagBits & 0x1) == 0)
      if ((this.initsOnReturn.tagBits & 0x1) == 0) {
        this.initsOnReturn = this.initsOnReturn.mergedWith(flowInfo);
      }
      else
        this.initsOnReturn = ((UnconditionalFlowInfo)flowInfo.copy());
  }

  public SubRoutineStatement subroutine()
  {
    if ((this.associatedNode instanceof SubRoutineStatement))
    {
      if (this.parent.subroutine() == this.associatedNode)
        return null;
      return (SubRoutineStatement)this.associatedNode;
    }
    return null;
  }
}