package org.eclipse.jdt.internal.compiler.flow;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.LabeledStatement;
import org.eclipse.jdt.internal.compiler.ast.Reference;
import org.eclipse.jdt.internal.compiler.ast.SubRoutineStatement;
import org.eclipse.jdt.internal.compiler.ast.TryStatement;
import org.eclipse.jdt.internal.compiler.codegen.BranchLabel;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class FlowContext
  implements TypeConstants
{
  public static final FlowContext NotContinuableContext = new FlowContext(null, null);
  public ASTNode associatedNode;
  public FlowContext parent;
  public NullInfoRegistry initsOnFinally;
  boolean deferNullDiagnostic;
  boolean preemptNullDiagnostic;
  public static final int CAN_ONLY_NULL_NON_NULL = 0;
  public static final int CAN_ONLY_NULL = 1;
  public static final int CAN_ONLY_NON_NULL = 2;
  public static final int MAY_NULL = 3;
  public static final int CHECK_MASK = 255;
  public static final int IN_COMPARISON_NULL = 256;
  public static final int IN_COMPARISON_NON_NULL = 512;
  public static final int IN_ASSIGNMENT = 768;
  public static final int IN_INSTANCEOF = 1024;
  public static final int CONTEXT_MASK = -256;

  public FlowContext(FlowContext parent, ASTNode associatedNode)
  {
    this.parent = parent;
    this.associatedNode = associatedNode;
    if (parent != null) {
      this.deferNullDiagnostic = 
        ((parent.deferNullDiagnostic) || (parent.preemptNullDiagnostic));
      this.initsOnFinally = parent.initsOnFinally;
    }
  }

  public BranchLabel breakLabel() {
    return null;
  }

  public void checkExceptionHandlers(TypeBinding raisedException, ASTNode location, FlowInfo flowInfo, BlockScope scope)
  {
    FlowContext traversedContext = this;
    while (traversedContext != null)
    {
      SubRoutineStatement sub;
      if (((sub = traversedContext.subroutine()) != null) && (sub.isSubRoutineEscaping()))
      {
        return;
      }

      if ((traversedContext instanceof ExceptionHandlingFlowContext)) {
        ExceptionHandlingFlowContext exceptionContext = 
          (ExceptionHandlingFlowContext)traversedContext;
        ReferenceBinding[] caughtExceptions;
        if ((caughtExceptions = exceptionContext.handledExceptions) != Binding.NO_EXCEPTIONS) {
          boolean definitelyCaught = false;
          int caughtIndex = 0; int caughtCount = caughtExceptions.length;
          for (; caughtIndex < caughtCount; 
            caughtIndex++) {
            ReferenceBinding caughtException = caughtExceptions[caughtIndex];
            int state = caughtException == null ? 
              -1 : 
              Scope.compareTypes(raisedException, caughtException);
            switch (state) {
            case -1:
              exceptionContext.recordHandlingException(
                caughtException, 
                flowInfo.unconditionalInits(), 
                raisedException, 
                location, 
                definitelyCaught);

              definitelyCaught = true;
              break;
            case 1:
              exceptionContext.recordHandlingException(
                caughtException, 
                flowInfo.unconditionalInits(), 
                raisedException, 
                location, 
                false);
            case 0:
            }
          }
          if (definitelyCaught) {
            return;
          }
        }
        if (exceptionContext.isMethodContext) {
          if (raisedException.isUncheckedException(false)) {
            return;
          }

          if (!(exceptionContext.associatedNode instanceof AbstractMethodDeclaration)) break;
          AbstractMethodDeclaration method = (AbstractMethodDeclaration)exceptionContext.associatedNode;
          if ((!method.isConstructor()) || (!method.binding.declaringClass.isAnonymousType()))
            break;
          exceptionContext.mergeUnhandledException(raisedException);
          return;
        }

      }

      traversedContext.recordReturnFrom(flowInfo.unconditionalInits());

      if ((traversedContext instanceof InsideSubRoutineFlowContext)) {
        ASTNode node = traversedContext.associatedNode;
        if ((node instanceof TryStatement)) {
          TryStatement tryStatement = (TryStatement)node;
          flowInfo.addInitializationsFrom(tryStatement.subRoutineInits);
        }
      }
      traversedContext = traversedContext.parent;
    }

    scope.problemReporter().unhandledException(raisedException, location);
  }

  public void checkExceptionHandlers(TypeBinding[] raisedExceptions, ASTNode location, FlowInfo flowInfo, BlockScope scope)
  {
    int raisedCount;
    if ((raisedExceptions == null) || 
      ((raisedCount = raisedExceptions.length) == 0))
      return;
    int raisedCount;
    int remainingCount = raisedCount;

    System.arraycopy(
      raisedExceptions, 
      0, 
      raisedExceptions = new TypeBinding[raisedCount], 
      0, 
      raisedCount);
    FlowContext traversedContext = this;

    while (traversedContext != null)
    {
      SubRoutineStatement sub;
      if (((sub = traversedContext.subroutine()) != null) && (sub.isSubRoutineEscaping()))
      {
        return;
      }

      if ((traversedContext instanceof ExceptionHandlingFlowContext)) {
        ExceptionHandlingFlowContext exceptionContext = 
          (ExceptionHandlingFlowContext)traversedContext;
        ReferenceBinding[] caughtExceptions;
        if ((caughtExceptions = exceptionContext.handledExceptions) != Binding.NO_EXCEPTIONS) {
          int caughtCount = caughtExceptions.length;
          boolean[] locallyCaught = new boolean[raisedCount];

          for (int caughtIndex = 0; caughtIndex < caughtCount; caughtIndex++) {
            ReferenceBinding caughtException = caughtExceptions[caughtIndex];
            for (int raisedIndex = 0; raisedIndex < raisedCount; raisedIndex++)
            {
              TypeBinding raisedException;
              if ((raisedException = raisedExceptions[raisedIndex]) != null) {
                int state = caughtException == null ? 
                  -1 : 
                  Scope.compareTypes(raisedException, caughtException);
                switch (state) {
                case -1:
                  exceptionContext.recordHandlingException(
                    caughtException, 
                    flowInfo.unconditionalInits(), 
                    raisedException, 
                    location, 
                    locallyCaught[raisedIndex]);

                  if (locallyCaught[raisedIndex] == 0) {
                    locallyCaught[raisedIndex] = true;

                    remainingCount--;
                  }
                  break;
                case 1:
                  exceptionContext.recordHandlingException(
                    caughtException, 
                    flowInfo.unconditionalInits(), 
                    raisedException, 
                    location, 
                    false);
                case 0:
                }
              }
            }
          }

          for (int i = 0; i < raisedCount; i++) {
            if (locallyCaught[i] != 0) {
              raisedExceptions[i] = null;
            }
          }
        }

        if (exceptionContext.isMethodContext) {
          for (int i = 0; i < raisedCount; i++)
          {
            TypeBinding raisedException;
            if (((raisedException = raisedExceptions[i]) != null) && 
              (raisedException.isUncheckedException(false))) {
              remainingCount--;
              raisedExceptions[i] = null;
            }

          }

          if (!(exceptionContext.associatedNode instanceof AbstractMethodDeclaration)) break;
          AbstractMethodDeclaration method = (AbstractMethodDeclaration)exceptionContext.associatedNode;
          if ((!method.isConstructor()) || (!method.binding.declaringClass.isAnonymousType()))
            break;
          for (int i = 0; i < raisedCount; i++)
          {
            TypeBinding raisedException;
            if ((raisedException = raisedExceptions[i]) != null) {
              exceptionContext.mergeUnhandledException(raisedException);
            }
          }
          return;
        }

      }

      if (remainingCount == 0) {
        return;
      }
      traversedContext.recordReturnFrom(flowInfo.unconditionalInits());

      if ((traversedContext instanceof InsideSubRoutineFlowContext)) {
        ASTNode node = traversedContext.associatedNode;
        if ((node instanceof TryStatement)) {
          TryStatement tryStatement = (TryStatement)node;
          flowInfo.addInitializationsFrom(tryStatement.subRoutineInits);
        }
      }
      traversedContext = traversedContext.parent;
    }

    for (int i = 0; i < raisedCount; i++)
    {
      TypeBinding exception;
      if ((exception = raisedExceptions[i]) != null)
      {
        int j = 0;
        while (raisedExceptions[j] != exception)
        {
          j++; if (j >= i)
          {
            scope.problemReporter().unhandledException(exception, location); } 
        }
      }
    }
  }

  public BranchLabel continueLabel() { return null; }


  public FlowContext getTargetContextForBreakLabel(char[] labelName)
  {
    FlowContext current = this; FlowContext lastNonReturningSubRoutine = null;
    while (current != null) {
      if (current.isNonReturningContext())
        lastNonReturningSubRoutine = current;
      char[] currentLabelName;
      if (((currentLabelName = current.labelName()) != null) && 
        (CharOperation.equals(currentLabelName, labelName))) {
        ((LabeledStatement)current.associatedNode).bits |= 64;
        if (lastNonReturningSubRoutine == null)
          return current;
        return lastNonReturningSubRoutine;
      }
      current = current.parent;
    }

    return null;
  }

  public FlowContext getTargetContextForContinueLabel(char[] labelName)
  {
    FlowContext current = this;
    FlowContext lastContinuable = null;
    FlowContext lastNonReturningSubRoutine = null;

    while (current != null) {
      if (current.isNonReturningContext()) {
        lastNonReturningSubRoutine = current;
      }
      else if (current.isContinuable())
        lastContinuable = current;
      char[] currentLabelName;
      if (((currentLabelName = current.labelName()) != null) && (CharOperation.equals(currentLabelName, labelName))) {
        ((LabeledStatement)current.associatedNode).bits |= 64;

        if ((lastContinuable != null) && 
          (current.associatedNode.concreteStatement() == lastContinuable.associatedNode))
        {
          if (lastNonReturningSubRoutine == null) return lastContinuable;
          return lastNonReturningSubRoutine;
        }

        return NotContinuableContext;
      }
      current = current.parent;
    }

    return null;
  }

  public FlowContext getTargetContextForDefaultBreak()
  {
    FlowContext current = this; FlowContext lastNonReturningSubRoutine = null;
    while (current != null) {
      if (current.isNonReturningContext()) {
        lastNonReturningSubRoutine = current;
      }
      if ((current.isBreakable()) && (current.labelName() == null)) {
        if (lastNonReturningSubRoutine == null) return current;
        return lastNonReturningSubRoutine;
      }
      current = current.parent;
    }

    return null;
  }

  public FlowContext getTargetContextForDefaultContinue()
  {
    FlowContext current = this; FlowContext lastNonReturningSubRoutine = null;
    while (current != null) {
      if (current.isNonReturningContext()) {
        lastNonReturningSubRoutine = current;
      }
      if (current.isContinuable()) {
        if (lastNonReturningSubRoutine == null)
          return current;
        return lastNonReturningSubRoutine;
      }
      current = current.parent;
    }

    return null;
  }

  public String individualToString() {
    return "Flow context";
  }

  public FlowInfo initsOnBreak() {
    return FlowInfo.DEAD_END;
  }

  public UnconditionalFlowInfo initsOnReturn() {
    return FlowInfo.DEAD_END;
  }

  public boolean isBreakable() {
    return false;
  }

  public boolean isContinuable() {
    return false;
  }

  public boolean isNonReturningContext() {
    return false;
  }

  public boolean isSubRoutine() {
    return false;
  }

  public char[] labelName() {
    return null;
  }

  public void recordBreakFrom(FlowInfo flowInfo)
  {
  }

  public void recordBreakTo(FlowContext targetContext)
  {
  }

  public void recordContinueFrom(FlowContext innerFlowContext, FlowInfo flowInfo)
  {
  }

  protected boolean recordFinalAssignment(VariableBinding variable, Reference finalReference) {
    return true;
  }

  protected void recordNullReference(LocalVariableBinding local, Expression expression, int status)
  {
  }

  public void recordReturnFrom(UnconditionalFlowInfo flowInfo)
  {
  }

  public void recordSettingFinal(VariableBinding variable, Reference finalReference, FlowInfo flowInfo)
  {
    if ((flowInfo.tagBits & 0x1) == 0)
    {
      FlowContext context = this;
      while (context != null) {
        if (!context.recordFinalAssignment(variable, finalReference)) {
          break;
        }
        context = context.parent;
      }
    }
  }

  public void recordUsingNullReference(Scope scope, LocalVariableBinding local, Expression reference, int checkType, FlowInfo flowInfo)
  {
    if (((flowInfo.tagBits & 0x1) != 0) || 
      (flowInfo.isDefinitelyUnknown(local))) {
      return;
    }
    switch (checkType) {
    case 256:
    case 512:
      if (flowInfo.isDefinitelyNonNull(local)) {
        if (checkType == 512)
          scope.problemReporter().localVariableRedundantCheckOnNonNull(local, reference);
        else {
          scope.problemReporter().localVariableNonNullComparedToNull(local, reference);
        }
        return;
      }
      if (flowInfo.cannotBeDefinitelyNullOrNonNull(local)) {
        return;
      }
    case 257:
    case 513:
    case 769:
    case 1025:
      if (flowInfo.isDefinitelyNull(local))
        switch (checkType & 0xFFFFFF00) {
        case 256:
          scope.problemReporter().localVariableRedundantCheckOnNull(local, reference);
          return;
        case 512:
          scope.problemReporter().localVariableNullComparedToNonNull(local, reference);
          return;
        case 768:
          scope.problemReporter().localVariableRedundantNullAssignment(local, reference);
          return;
        case 1024:
          scope.problemReporter().localVariableNullInstanceof(local, reference);
          return;
        }
      else if (flowInfo.cannotBeDefinitelyNullOrNonNull(local)) {
        return;
      }
      break;
    case 3:
      if (flowInfo.isDefinitelyNull(local)) {
        scope.problemReporter().localVariableNullReference(local, reference);
        return;
      }
      if (flowInfo.isPotentiallyNull(local))
      {
        scope.problemReporter().localVariablePotentialNullReference(local, reference);
        return;
      }
      break;
    }
    if (this.parent != null)
      this.parent.recordUsingNullReference(scope, local, reference, checkType, 
        flowInfo);
  }

  void removeFinalAssignmentIfAny(Reference reference)
  {
  }

  public SubRoutineStatement subroutine()
  {
    return null;
  }

  public String toString() {
    StringBuffer buffer = new StringBuffer();
    FlowContext current = this;
    int parentsCount = 0;
    while ((current = current.parent) != null) {
      parentsCount++;
    }
    FlowContext[] parents = new FlowContext[parentsCount + 1];
    current = this;
    int index = parentsCount;
    while (index >= 0) {
      parents[(index--)] = current;
      current = current.parent;
    }
    for (int i = 0; i < parentsCount; i++) {
      for (int j = 0; j < i; j++)
        buffer.append('\t');
      buffer.append(parents[i].individualToString()).append('\n');
    }
    buffer.append('*');
    for (int j = 0; j < parentsCount + 1; j++)
      buffer.append('\t');
    buffer.append(individualToString()).append('\n');
    return buffer.toString();
  }
}