package org.eclipse.jdt.internal.compiler.lookup;

import java.util.ArrayList;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.env.ClassSignature;
import org.eclipse.jdt.internal.compiler.env.EnumConstantSignature;
import org.eclipse.jdt.internal.compiler.env.IBinaryAnnotation;
import org.eclipse.jdt.internal.compiler.env.IBinaryElementValuePair;
import org.eclipse.jdt.internal.compiler.env.IBinaryField;
import org.eclipse.jdt.internal.compiler.env.IBinaryMethod;
import org.eclipse.jdt.internal.compiler.env.IBinaryNestedType;
import org.eclipse.jdt.internal.compiler.env.IBinaryType;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.util.SimpleLookupTable;

public class BinaryTypeBinding extends ReferenceBinding
{
  protected ReferenceBinding superclass;
  protected ReferenceBinding enclosingType;
  protected ReferenceBinding[] superInterfaces;
  protected FieldBinding[] fields;
  protected MethodBinding[] methods;
  protected ReferenceBinding[] memberTypes;
  protected TypeVariableBinding[] typeVariables;
  protected LookupEnvironment environment;
  protected SimpleLookupTable storedAnnotations = null;

  static Object convertMemberValue(Object binaryValue, LookupEnvironment env) {
    if (binaryValue == null) return null;
    if ((binaryValue instanceof Constant))
      return binaryValue;
    if ((binaryValue instanceof ClassSignature))
      return env.getTypeFromSignature(((ClassSignature)binaryValue).getTypeName(), 0, -1, false, null);
    if ((binaryValue instanceof IBinaryAnnotation))
      return createAnnotation((IBinaryAnnotation)binaryValue, env);
    if ((binaryValue instanceof EnumConstantSignature)) {
      EnumConstantSignature ref = (EnumConstantSignature)binaryValue;
      ReferenceBinding enumType = 
        (ReferenceBinding)env.getTypeFromSignature(ref.getTypeName(), 0, -1, false, null);
      enumType = resolveType(enumType, env, false);
      return enumType.getField(ref.getEnumConstantName(), false);
    }
    if ((binaryValue instanceof Object[])) {
      Object[] objects = (Object[])binaryValue;
      int length = objects.length;
      if (length == 0) return objects;
      Object[] values = new Object[length];
      for (int i = 0; i < length; i++)
        values[i] = convertMemberValue(objects[i], env);
      return values;
    }

    throw new IllegalStateException();
  }
  static AnnotationBinding createAnnotation(IBinaryAnnotation annotationInfo, LookupEnvironment env) {
    IBinaryElementValuePair[] binaryPairs = annotationInfo.getElementValuePairs();
    int length = binaryPairs == null ? 0 : binaryPairs.length;
    ElementValuePair[] pairs = length == 0 ? Binding.NO_ELEMENT_VALUE_PAIRS : new ElementValuePair[length];
    for (int i = 0; i < length; i++) {
      pairs[i] = new ElementValuePair(binaryPairs[i].getName(), convertMemberValue(binaryPairs[i].getValue(), env), null);
    }
    char[] typeName = annotationInfo.getTypeName();
    ReferenceBinding annotationType = env.getTypeFromConstantPoolName(typeName, 1, typeName.length - 1, false);
    return new UnresolvedAnnotationBinding(annotationType, pairs, env);
  }
  public static AnnotationBinding[] createAnnotations(IBinaryAnnotation[] annotationInfos, LookupEnvironment env) {
    int length = annotationInfos == null ? 0 : annotationInfos.length;
    AnnotationBinding[] result = length == 0 ? Binding.NO_ANNOTATIONS : new AnnotationBinding[length];
    for (int i = 0; i < length; i++)
      result[i] = createAnnotation(annotationInfos[i], env);
    return result;
  }
  public static ReferenceBinding resolveType(ReferenceBinding type, LookupEnvironment environment, boolean convertGenericToRawType) {
    if ((type instanceof UnresolvedReferenceBinding))
      return ((UnresolvedReferenceBinding)type).resolve(environment, convertGenericToRawType);
    if (type.isParameterizedType())
      return ((ParameterizedTypeBinding)type).resolve();
    if (type.isWildcard()) {
      return ((WildcardBinding)type).resolve();
    }
    if (convertGenericToRawType)
      return (ReferenceBinding)environment.convertUnresolvedBinaryToRawType(type);
    return type;
  }
  public static TypeBinding resolveType(TypeBinding type, LookupEnvironment environment, ParameterizedTypeBinding parameterizedType, int rank) {
    switch (type.kind())
    {
    case 260:
      return ((ParameterizedTypeBinding)type).resolve();
    case 516:
      return ((WildcardBinding)type).resolve();
    case 68:
      resolveType(((ArrayBinding)type).leafComponentType, environment, parameterizedType, rank);
      break;
    case 4100:
      ((TypeVariableBinding)type).resolve(environment);
      break;
    case 2052:
      if (parameterizedType == null) {
        return environment.convertUnresolvedBinaryToRawType(type);
      }
      break;
    default:
      if ((type instanceof UnresolvedReferenceBinding))
        return ((UnresolvedReferenceBinding)type).resolve(environment, parameterizedType == null); break;
    }
    return type;
  }

  protected BinaryTypeBinding()
  {
  }

  public BinaryTypeBinding(PackageBinding packageBinding, IBinaryType binaryType, LookupEnvironment environment)
  {
    this.compoundName = CharOperation.splitOn('/', binaryType.getName());
    computeId();

    this.tagBits |= 64L;
    this.environment = environment;
    this.fPackage = packageBinding;
    this.fileName = binaryType.getFileName();

    char[] typeSignature = environment.globalOptions.sourceLevel >= 3211264L ? binaryType.getGenericSignature() : null;
    this.typeVariables = ((typeSignature != null) && (typeSignature.length > 0) && (typeSignature[0] == '<') ? 
      null : 
      Binding.NO_TYPE_VARIABLES);

    this.sourceName = binaryType.getSourceName();
    this.modifiers = binaryType.getModifiers();

    if ((binaryType.getTagBits() & 0x80) != 0L) {
      this.tagBits |= 131072L;
    }
    if (binaryType.isAnonymous())
      this.tagBits |= 52L;
    else if (binaryType.isLocal())
      this.tagBits |= 20L;
    else if (binaryType.isMember()) {
      this.tagBits |= 12L;
    }

    char[] enclosingTypeName = binaryType.getEnclosingTypeName();
    if (enclosingTypeName != null)
    {
      this.enclosingType = environment.getTypeFromConstantPoolName(enclosingTypeName, 0, -1, true);
      this.tagBits |= 12L;
      this.tagBits |= 134217728L;
      if (enclosingType().isStrictfp())
        this.modifiers |= 2048;
      if (enclosingType().isDeprecated())
        this.modifiers |= 2097152;
    }
  }

  public FieldBinding[] availableFields()
  {
    if ((this.tagBits & 0x2000) != 0L) {
      return this.fields;
    }

    if ((this.tagBits & 0x1000) == 0L) {
      int length = this.fields.length;
      if (length > 1)
        ReferenceBinding.sortFields(this.fields, 0, length);
      this.tagBits |= 4096L;
    }
    FieldBinding[] availableFields = new FieldBinding[this.fields.length];
    int count = 0;
    for (int i = 0; i < this.fields.length; i++)
      try {
        availableFields[count] = resolveTypeFor(this.fields[i]);
        count++;
      }
      catch (AbortCompilation localAbortCompilation)
      {
      }
    if (count < availableFields.length)
      System.arraycopy(availableFields, 0, availableFields = new FieldBinding[count], 0, count);
    return availableFields;
  }

  public MethodBinding[] availableMethods()
  {
    if ((this.tagBits & 0x8000) != 0L) {
      return this.methods;
    }

    if ((this.tagBits & 0x4000) == 0L) {
      int length = this.methods.length;
      if (length > 1)
        ReferenceBinding.sortMethods(this.methods, 0, length);
      this.tagBits |= 16384L;
    }
    MethodBinding[] availableMethods = new MethodBinding[this.methods.length];
    int count = 0;
    for (int i = 0; i < this.methods.length; i++)
      try {
        availableMethods[count] = resolveTypesFor(this.methods[i]);
        count++;
      }
      catch (AbortCompilation localAbortCompilation)
      {
      }
    if (count < availableMethods.length)
      System.arraycopy(availableMethods, 0, availableMethods = new MethodBinding[count], 0, count);
    return availableMethods;
  }

  void cachePartsFrom(IBinaryType binaryType, boolean needFieldsAndMethods)
  {
    this.typeVariables = Binding.NO_TYPE_VARIABLES;
    this.superInterfaces = Binding.NO_SUPERINTERFACES;

    this.memberTypes = Binding.NO_MEMBER_TYPES;
    IBinaryNestedType[] memberTypeStructures = binaryType.getMemberTypes();
    if (memberTypeStructures != null) {
      int size = memberTypeStructures.length;
      if (size > 0) {
        this.memberTypes = new ReferenceBinding[size];
        for (int i = 0; i < size; i++)
        {
          this.memberTypes[i] = this.environment.getTypeFromConstantPoolName(memberTypeStructures[i].getName(), 0, -1, false);
        }this.tagBits |= 268435456L;
      }
    }

    long sourceLevel = this.environment.globalOptions.sourceLevel;
    char[] typeSignature = (char[])null;
    if (sourceLevel >= 3211264L) {
      typeSignature = binaryType.getGenericSignature();
      this.tagBits |= binaryType.getTagBits();
    }
    if (typeSignature == null) {
      char[] superclassName = binaryType.getSuperclassName();
      if (superclassName != null)
      {
        this.superclass = this.environment.getTypeFromConstantPoolName(superclassName, 0, -1, false);
        this.tagBits |= 33554432L;
      }

      this.superInterfaces = Binding.NO_SUPERINTERFACES;
      char[][] interfaceNames = binaryType.getInterfaceNames();
      if (interfaceNames != null) {
        int size = interfaceNames.length;
        if (size > 0) {
          this.superInterfaces = new ReferenceBinding[size];
          for (int i = 0; i < size; i++)
          {
            this.superInterfaces[i] = this.environment.getTypeFromConstantPoolName(interfaceNames[i], 0, -1, false);
          }this.tagBits |= 67108864L;
        }
      }
    }
    else {
      SignatureWrapper wrapper = new SignatureWrapper(typeSignature);
      if (wrapper.signature[wrapper.start] == '<')
      {
        wrapper.start += 1;
        this.typeVariables = createTypeVariables(wrapper, true);
        wrapper.start += 1;
        this.tagBits |= 16777216L;
        this.modifiers |= 1073741824;
      }

      this.superclass = ((ReferenceBinding)this.environment.getTypeFromTypeSignature(wrapper, Binding.NO_TYPE_VARIABLES, this));
      this.tagBits |= 33554432L;

      this.superInterfaces = Binding.NO_SUPERINTERFACES;
      if (!wrapper.atEnd())
      {
        ArrayList types = new ArrayList(2);
        do
          types.add(this.environment.getTypeFromTypeSignature(wrapper, Binding.NO_TYPE_VARIABLES, this));
        while (!
          wrapper.atEnd());
        this.superInterfaces = new ReferenceBinding[types.size()];
        types.toArray(this.superInterfaces);
        this.tagBits |= 67108864L;
      }
    }

    if (needFieldsAndMethods) {
      createFields(binaryType.getFields(), sourceLevel);
      createMethods(binaryType.getMethods(), sourceLevel);
    } else {
      this.fields = Binding.NO_FIELDS;
      this.methods = Binding.NO_METHODS;
    }
    if (this.environment.globalOptions.storeAnnotations)
      setAnnotations(createAnnotations(binaryType.getAnnotations(), this.environment)); 
  }

  private void createFields(IBinaryField[] iFields, long sourceLevel) { this.fields = Binding.NO_FIELDS;
    if (iFields != null) {
      int size = iFields.length;
      if (size > 0) {
        this.fields = new FieldBinding[size];
        boolean use15specifics = sourceLevel >= 3211264L;
        boolean isViewedAsDeprecated = isViewedAsDeprecated();
        boolean hasRestrictedAccess = hasRestrictedAccess();
        int firstAnnotatedFieldIndex = -1;
        for (int i = 0; i < size; i++) {
          IBinaryField binaryField = iFields[i];
          char[] fieldSignature = use15specifics ? binaryField.getGenericSignature() : null;
          TypeBinding type = fieldSignature == null ? 
            this.environment.getTypeFromSignature(binaryField.getTypeName(), 0, -1, false, this) : 
            this.environment.getTypeFromTypeSignature(new SignatureWrapper(fieldSignature), Binding.NO_TYPE_VARIABLES, this);
          FieldBinding field = 
            new FieldBinding(
            binaryField.getName(), 
            type, 
            binaryField.getModifiers() | 0x2000000, 
            this, 
            binaryField.getConstant());
          if ((firstAnnotatedFieldIndex < 0) && 
            (this.environment.globalOptions.storeAnnotations) && 
            (binaryField.getAnnotations() != null)) {
            firstAnnotatedFieldIndex = i;
          }
          field.id = i;
          if (use15specifics)
            field.tagBits |= binaryField.getTagBits();
          if ((isViewedAsDeprecated) && (!field.isDeprecated()))
            field.modifiers |= 2097152;
          if (hasRestrictedAccess)
            field.modifiers |= 262144;
          if (fieldSignature != null)
            field.modifiers |= 1073741824;
          this.fields[i] = field;
        }

        if (firstAnnotatedFieldIndex >= 0)
          for (int i = firstAnnotatedFieldIndex; i < size; i++)
            this.fields[i].setAnnotations(createAnnotations(iFields[i].getAnnotations(), this.environment));
      }
    }
  }

  private MethodBinding createMethod(IBinaryMethod method, long sourceLevel)
  {
    int methodModifiers = method.getModifiers() | 0x2000000;
    if (sourceLevel < 3211264L)
      methodModifiers &= -129;
    ReferenceBinding[] exceptions = Binding.NO_EXCEPTIONS;
    TypeBinding[] parameters = Binding.NO_PARAMETERS;
    TypeVariableBinding[] typeVars = Binding.NO_TYPE_VARIABLES;
    AnnotationBinding[][] paramAnnotations = (AnnotationBinding[][])null;
    TypeBinding returnType = null;

    boolean use15specifics = sourceLevel >= 3211264L;
    char[] methodSignature = use15specifics ? method.getGenericSignature() : null;
    if (methodSignature == null) {
      char[] methodDescriptor = method.getMethodDescriptor();
      int numOfParams = 0;

      int index = 0;
      char nextChar;
      while ((nextChar = methodDescriptor[(++index)]) != ')')
      {
        char nextChar;
        if (nextChar != '[')
        {
          numOfParams++;
          while ((nextChar == 'L') && 
            ((nextChar = methodDescriptor[(++index)]) != ';'));
        }
      }
      int startIndex = (method.isConstructor()) && (isMemberType()) && (!isStatic()) ? 1 : 0;
      int size = numOfParams - startIndex;
      if (size > 0) {
        parameters = new TypeBinding[size];
        if (this.environment.globalOptions.storeAnnotations)
          paramAnnotations = new AnnotationBinding[size][];
        index = 1;
        int end = 0;
        int i = 0; while (i < numOfParams) {
          while ((nextChar = methodDescriptor[(++end)]) == '[');
          while ((nextChar == 'L') && 
            ((nextChar = methodDescriptor[(++end)]) != ';'));
          if (i >= startIndex) {
            parameters[(i - startIndex)] = this.environment.getTypeFromSignature(methodDescriptor, index, end, false, this);

            if (paramAnnotations != null)
              paramAnnotations[(i - startIndex)] = createAnnotations(method.getParameterAnnotations(i), this.environment);
          }
          index = end + 1;

          i++;
        }

      }

      char[][] exceptionTypes = method.getExceptionTypeNames();
      if (exceptionTypes != null) {
        size = exceptionTypes.length;
        if (size > 0) {
          exceptions = new ReferenceBinding[size];
          for (int i = 0; i < size; i++) {
            exceptions[i] = this.environment.getTypeFromConstantPoolName(exceptionTypes[i], 0, -1, false);
          }
        }
      }
      if (!method.isConstructor())
        returnType = this.environment.getTypeFromSignature(methodDescriptor, index + 1, -1, false, this);
    } else {
      methodModifiers |= 1073741824;

      SignatureWrapper wrapper = new SignatureWrapper(methodSignature);
      if (wrapper.signature[wrapper.start] == '<')
      {
        wrapper.start += 1;
        typeVars = createTypeVariables(wrapper, false);
        wrapper.start += 1;
      }

      if (wrapper.signature[wrapper.start] == '(') {
        wrapper.start += 1;
        if (wrapper.signature[wrapper.start] == ')') {
          wrapper.start += 1;
        } else {
          ArrayList types = new ArrayList(2);
          while (wrapper.signature[wrapper.start] != ')')
            types.add(this.environment.getTypeFromTypeSignature(wrapper, typeVars, this));
          wrapper.start += 1;
          int numParam = types.size();
          parameters = new TypeBinding[numParam];
          types.toArray(parameters);
          if (this.environment.globalOptions.storeAnnotations) {
            paramAnnotations = new AnnotationBinding[numParam][];
            for (int i = 0; i < numParam; i++) {
              paramAnnotations[i] = createAnnotations(method.getParameterAnnotations(i), this.environment);
            }
          }
        }
      }

      returnType = this.environment.getTypeFromTypeSignature(wrapper, typeVars, this);

      if ((!wrapper.atEnd()) && (wrapper.signature[wrapper.start] == '^'))
      {
        ArrayList types = new ArrayList(2);
        do {
          wrapper.start += 1;
          types.add(this.environment.getTypeFromTypeSignature(wrapper, typeVars, this));
        }while ((!wrapper.atEnd()) && (wrapper.signature[wrapper.start] == '^'));
        exceptions = new ReferenceBinding[types.size()];
        types.toArray(exceptions);
      } else {
        char[][] exceptionTypes = method.getExceptionTypeNames();
        if (exceptionTypes != null) {
          int size = exceptionTypes.length;
          if (size > 0) {
            exceptions = new ReferenceBinding[size];
            for (int i = 0; i < size; i++) {
              exceptions[i] = this.environment.getTypeFromConstantPoolName(exceptionTypes[i], 0, -1, false);
            }
          }
        }
      }
    }
    MethodBinding result = method.isConstructor() ? 
      new MethodBinding(methodModifiers, parameters, exceptions, this) : 
      new MethodBinding(methodModifiers, method.getSelector(), returnType, parameters, exceptions, this);
    if (this.environment.globalOptions.storeAnnotations) {
      result.setAnnotations(
        createAnnotations(method.getAnnotations(), this.environment), 
        paramAnnotations, 
        isAnnotationType() ? convertMemberValue(method.getDefaultValue(), this.environment) : null);
    }
    if (use15specifics)
      result.tagBits |= method.getTagBits();
    result.typeVariables = typeVars;

    int i = 0; for (int length = typeVars.length; i < length; i++)
      typeVars[i].declaringElement = result;
    return result;
  }

  private void createMethods(IBinaryMethod[] iMethods, long sourceLevel)
  {
    int total = 0; int initialTotal = 0; int iClinit = -1;
    int[] toSkip = (int[])null;
    if (iMethods != null) {
      total = initialTotal = iMethods.length;
      boolean keepBridgeMethods = (sourceLevel < 3211264L) && 
        (this.environment.globalOptions.complianceLevel >= 3211264L);
      int i = total;
      do { IBinaryMethod method = iMethods[i];
        if ((method.getModifiers() & 0x1000) != 0) {
          if ((!keepBridgeMethods) || ((method.getModifiers() & 0x40) == 0))
          {
            if (toSkip == null) toSkip = new int[iMethods.length];
            toSkip[i] = -1;
            total--;
          } } else if (iClinit == -1) {
          char[] methodName = method.getSelector();
          if ((methodName.length == 8) && (methodName[0] == '<'))
          {
            iClinit = i;
            total--;
          }
        }
        i--; } while (i >= 0);
    }

    if (total == 0) {
      this.methods = Binding.NO_METHODS;
      return;
    }

    boolean isViewedAsDeprecated = isViewedAsDeprecated();
    boolean hasRestrictedAccess = hasRestrictedAccess();
    this.methods = new MethodBinding[total];
    if (total == initialTotal) {
      for (int i = 0; i < initialTotal; i++) {
        MethodBinding method = createMethod(iMethods[i], sourceLevel);
        if ((isViewedAsDeprecated) && (!method.isDeprecated()))
          method.modifiers |= 2097152;
        if (hasRestrictedAccess)
          method.modifiers |= 262144;
        this.methods[i] = method;
      }
    } else {
      int i = 0; for (int index = 0; i < initialTotal; i++)
        if ((iClinit != i) && ((toSkip == null) || (toSkip[i] != -1))) {
          MethodBinding method = createMethod(iMethods[i], sourceLevel);
          if ((isViewedAsDeprecated) && (!method.isDeprecated()))
            method.modifiers |= 2097152;
          if (hasRestrictedAccess)
            method.modifiers |= 262144;
          this.methods[(index++)] = method;
        }
    }
  }

  private TypeVariableBinding[] createTypeVariables(SignatureWrapper wrapper, boolean assignVariables)
  {
    char[] typeSignature = wrapper.signature;
    int depth = 0; int length = typeSignature.length;
    int rank = 0;
    ArrayList variables = new ArrayList(1);
    depth = 0;
    boolean pendingVariable = true;

    for (int i = 1; i < length; i++)
      switch (typeSignature[i]) {
      case '<':
        depth++;
        break;
      case '>':
        depth--; if (depth >= 0) break;
        break;
      case ';':
        if ((depth == 0) && (i + 1 < length) && (typeSignature[(i + 1)] != ':'))
          pendingVariable = true;
        break;
      case '=':
      default:
        if (pendingVariable) {
          pendingVariable = false;
          int colon = CharOperation.indexOf(':', typeSignature, i);
          char[] variableName = CharOperation.subarray(typeSignature, i, colon);
          variables.add(new TypeVariableBinding(variableName, this, rank++));
        }
        break;
      }
    TypeVariableBinding[] result;
    variables.toArray(result = new TypeVariableBinding[rank]);

    if (assignVariables)
      this.typeVariables = result;
    for (int i = 0; i < rank; i++) {
      initializeTypeVariable(result[i], result, wrapper);
    }
    return result;
  }

  public ReferenceBinding enclosingType()
  {
    if ((this.tagBits & 0x8000000) == 0L) {
      return this.enclosingType;
    }

    this.enclosingType = resolveType(this.enclosingType, this.environment, false);
    this.tagBits &= -134217729L;
    return this.enclosingType;
  }

  public FieldBinding[] fields() {
    if ((this.tagBits & 0x2000) != 0L) {
      return this.fields;
    }

    if ((this.tagBits & 0x1000) == 0L) {
      int length = this.fields.length;
      if (length > 1)
        ReferenceBinding.sortFields(this.fields, 0, length);
      this.tagBits |= 4096L;
    }
    int i = this.fields.length;
    do { resolveTypeFor(this.fields[i]);

      i--; } while (i >= 0);

    this.tagBits |= 8192L;
    return this.fields;
  }

  public char[] genericTypeSignature()
  {
    return computeGenericTypeSignature(this.typeVariables);
  }

  public MethodBinding getExactConstructor(TypeBinding[] argumentTypes)
  {
    if ((this.tagBits & 0x4000) == 0L) {
      int length = this.methods.length;
      if (length > 1)
        ReferenceBinding.sortMethods(this.methods, 0, length);
      this.tagBits |= 16384L;
    }
    int argCount = argumentTypes.length;
    long range;
    if ((range = ReferenceBinding.binarySearch(TypeConstants.INIT, this.methods)) >= 0L) {
      int imethod = (int)range; for (int end = (int)(range >> 32); imethod <= end; imethod++) {
        MethodBinding method = this.methods[imethod];
        if (method.parameters.length == argCount) {
          resolveTypesFor(method);
          TypeBinding[] toMatch = method.parameters;
          int iarg = 0;
          while (toMatch[iarg] == argumentTypes[iarg])
          {
            iarg++; if (iarg >= argCount)
            {
              return method;
            }
          }
        }
      }
    }
    return null;
  }

  public MethodBinding getExactMethod(char[] selector, TypeBinding[] argumentTypes, CompilationUnitScope refScope)
  {
    if ((this.tagBits & 0x4000) == 0L) {
      int length = this.methods.length;
      if (length > 1)
        ReferenceBinding.sortMethods(this.methods, 0, length);
      this.tagBits |= 16384L;
    }

    int argCount = argumentTypes.length;
    boolean foundNothing = true;
    long range;
    if ((range = ReferenceBinding.binarySearch(selector, this.methods)) >= 0L) {
      int imethod = (int)range; for (int end = (int)(range >> 32); imethod <= end; imethod++) {
        MethodBinding method = this.methods[imethod];
        foundNothing = false;
        if (method.parameters.length == argCount) {
          resolveTypesFor(method);
          TypeBinding[] toMatch = method.parameters;
          int iarg = 0;
          while (toMatch[iarg] == argumentTypes[iarg])
          {
            iarg++; if (iarg >= argCount)
            {
              return method;
            }
          }
        }
      }
    }
    if (foundNothing) {
      if (isInterface()) {
        if (superInterfaces().length == 1) {
          if (refScope != null)
            refScope.recordTypeReference(this.superInterfaces[0]);
          return this.superInterfaces[0].getExactMethod(selector, argumentTypes, refScope);
        }
      } else if (superclass() != null) {
        if (refScope != null)
          refScope.recordTypeReference(this.superclass);
        return this.superclass.getExactMethod(selector, argumentTypes, refScope);
      }
    }
    return null;
  }

  public FieldBinding getField(char[] fieldName, boolean needResolve)
  {
    if ((this.tagBits & 0x1000) == 0L) {
      int length = this.fields.length;
      if (length > 1)
        ReferenceBinding.sortFields(this.fields, 0, length);
      this.tagBits |= 4096L;
    }
    FieldBinding field = ReferenceBinding.binarySearch(fieldName, this.fields);
    return (needResolve) && (field != null) ? resolveTypeFor(field) : field;
  }

  public ReferenceBinding getMemberType(char[] typeName)
  {
    int i = this.memberTypes.length;
    do { ReferenceBinding memberType = this.memberTypes[i];
      if ((memberType instanceof UnresolvedReferenceBinding)) {
        char[] name = memberType.sourceName;
        int prefixLength = this.compoundName[(this.compoundName.length - 1)].length + 1;
        if ((name.length == prefixLength + typeName.length) && 
          (CharOperation.fragmentEquals(typeName, name, prefixLength, true)))
          return this.memberTypes[i] =  = resolveType(memberType, this.environment, false);
      } else if (CharOperation.equals(typeName, memberType.sourceName)) {
        return memberType;
      }
      i--; } while (i >= 0);

    return null;
  }

  public MethodBinding[] getMethods(char[] selector) {
    if ((this.tagBits & 0x8000) != 0L)
    {
      long range;
      if ((range = ReferenceBinding.binarySearch(selector, this.methods)) >= 0L) {
        int start = (int)range; int end = (int)(range >> 32);
        int length = end - start + 1;
        if ((this.tagBits & 0x8000) != 0L)
        {
          MethodBinding[] result;
          System.arraycopy(this.methods, start, result = new MethodBinding[length], 0, length);
          return result;
        }
      }
      return Binding.NO_METHODS;
    }

    if ((this.tagBits & 0x4000) == 0L) {
      int length = this.methods.length;
      if (length > 1)
        ReferenceBinding.sortMethods(this.methods, 0, length);
      this.tagBits |= 16384L;
    }
    long range;
    if ((range = ReferenceBinding.binarySearch(selector, this.methods)) >= 0L) {
      int start = (int)range; int end = (int)(range >> 32);
      int length = end - start + 1;
      MethodBinding[] result = new MethodBinding[length];

      int i = start; for (int index = 0; i <= end; index++) {
        result[index] = resolveTypesFor(this.methods[i]);

        i++;
      }
      return result;
    }
    return Binding.NO_METHODS;
  }
  public boolean hasMemberTypes() {
    return this.memberTypes.length > 0;
  }

  public TypeVariableBinding getTypeVariable(char[] variableName) {
    TypeVariableBinding variable = super.getTypeVariable(variableName);
    variable.resolve(this.environment);
    return variable;
  }

  private void initializeTypeVariable(TypeVariableBinding variable, TypeVariableBinding[] existingVariables, SignatureWrapper wrapper)
  {
    int colon = CharOperation.indexOf(':', wrapper.signature, wrapper.start);
    wrapper.start = (colon + 1);
    ReferenceBinding firstBound = null;
    ReferenceBinding type;
    ReferenceBinding type;
    if (wrapper.signature[wrapper.start] == ':') {
      type = this.environment.getResolvedType(TypeConstants.JAVA_LANG_OBJECT, null);
    } else {
      type = (ReferenceBinding)this.environment.getTypeFromTypeSignature(wrapper, existingVariables, this);
      firstBound = type;
    }

    variable.modifiers |= 33554432;
    variable.superclass = type;

    ReferenceBinding[] bounds = (ReferenceBinding[])null;
    if (wrapper.signature[wrapper.start] == ':') {
      ArrayList types = new ArrayList(2);
      do {
        wrapper.start += 1;
        types.add(this.environment.getTypeFromTypeSignature(wrapper, existingVariables, this));
      }while (wrapper.signature[wrapper.start] == ':');
      bounds = new ReferenceBinding[types.size()];
      types.toArray(bounds);
    }

    variable.superInterfaces = (bounds == null ? Binding.NO_SUPERINTERFACES : bounds);
    if (firstBound == null) {
      firstBound = variable.superInterfaces.length == 0 ? null : variable.superInterfaces[0];
    }
    variable.firstBound = firstBound;
  }

  public boolean isEquivalentTo(TypeBinding otherType)
  {
    if (this == otherType) return true;
    if (otherType == null) return false;
    switch (otherType.kind()) {
    case 516:
      return ((WildcardBinding)otherType).boundCheck(this);
    case 1028:
      return otherType.erasure() == this;
    }
    return false;
  }
  public boolean isGenericType() {
    return this.typeVariables != Binding.NO_TYPE_VARIABLES;
  }
  public int kind() {
    if (this.typeVariables != Binding.NO_TYPE_VARIABLES)
      return 2052;
    return 4;
  }

  public ReferenceBinding[] memberTypes() {
    if ((this.tagBits & 0x10000000) == 0L) {
      return this.memberTypes;
    }
    int i = this.memberTypes.length;
    do { this.memberTypes[i] = resolveType(this.memberTypes[i], this.environment, false);

      i--; } while (i >= 0);

    this.tagBits &= -268435457L;
    return this.memberTypes;
  }

  public MethodBinding[] methods() {
    if ((this.tagBits & 0x8000) != 0L) {
      return this.methods;
    }

    if ((this.tagBits & 0x4000) == 0L) {
      int length = this.methods.length;
      if (length > 1)
        ReferenceBinding.sortMethods(this.methods, 0, length);
      this.tagBits |= 16384L;
    }
    int i = this.methods.length;
    do { resolveTypesFor(this.methods[i]);

      i--; } while (i >= 0);

    this.tagBits |= 32768L;
    return this.methods;
  }
  private FieldBinding resolveTypeFor(FieldBinding field) {
    if ((field.modifiers & 0x2000000) == 0) {
      return field;
    }
    field.type = resolveType(field.type, this.environment, null, 0);
    field.modifiers &= -33554433;
    return field;
  }
  MethodBinding resolveTypesFor(MethodBinding method) {
    if ((method.modifiers & 0x2000000) == 0) {
      return method;
    }
    if (!method.isConstructor())
      method.returnType = resolveType(method.returnType, this.environment, null, 0);
    int i = method.parameters.length;
    do { method.parameters[i] = resolveType(method.parameters[i], this.environment, null, 0);

      i--; } while (i >= 0);

    int i = method.thrownExceptions.length;
    do { method.thrownExceptions[i] = resolveType(method.thrownExceptions[i], this.environment, true);

      i--; } while (i >= 0);

    int i = method.typeVariables.length;
    do { method.typeVariables[i].resolve(this.environment);

      i--; } while (i >= 0);

    method.modifiers &= -33554433;
    return method;
  }
  AnnotationBinding[] retrieveAnnotations(Binding binding) {
    return AnnotationBinding.addStandardAnnotations(super.retrieveAnnotations(binding), binding.getAnnotationTagBits(), this.environment);
  }
  SimpleLookupTable storedAnnotations(boolean forceInitialize) {
    if ((forceInitialize) && (this.storedAnnotations == null)) {
      if (!this.environment.globalOptions.storeAnnotations)
        return null;
      this.storedAnnotations = new SimpleLookupTable(3);
    }
    return this.storedAnnotations;
  }

  public ReferenceBinding superclass()
  {
    if ((this.tagBits & 0x2000000) == 0L) {
      return this.superclass;
    }

    this.superclass = resolveType(this.superclass, this.environment, true);
    this.tagBits &= -33554433L;
    if (this.superclass.problemId() == 1)
      this.tagBits |= 131072L;
    return this.superclass;
  }

  public ReferenceBinding[] superInterfaces() {
    if ((this.tagBits & 0x4000000) == 0L) {
      return this.superInterfaces;
    }
    int i = this.superInterfaces.length;
    do { this.superInterfaces[i] = resolveType(this.superInterfaces[i], this.environment, true);
      if (this.superInterfaces[i].problemId() == 1)
        this.tagBits |= 131072L;
      i--; } while (i >= 0);

    this.tagBits &= -67108865L;
    return this.superInterfaces;
  }
  public TypeVariableBinding[] typeVariables() {
    if ((this.tagBits & 0x1000000) == 0L) {
      return this.typeVariables;
    }
    int i = this.typeVariables.length;
    do { this.typeVariables[i].resolve(this.environment);

      i--; } while (i >= 0);

    this.tagBits &= -16777217L;
    return this.typeVariables;
  }
  public String toString() {
    StringBuffer buffer = new StringBuffer();

    if (isDeprecated()) buffer.append("deprecated ");
    if (isPublic()) buffer.append("public ");
    if (isProtected()) buffer.append("protected ");
    if (isPrivate()) buffer.append("private ");
    if ((isAbstract()) && (isClass())) buffer.append("abstract ");
    if ((isStatic()) && (isNestedType())) buffer.append("static ");
    if (isFinal()) buffer.append("final ");

    if (isEnum()) buffer.append("enum ");
    else if (isAnnotationType()) buffer.append("@interface ");
    else if (isClass()) buffer.append("class "); else
      buffer.append("interface ");
    buffer.append(this.compoundName != null ? CharOperation.toString(this.compoundName) : "UNNAMED TYPE");

    buffer.append("\n\textends ");
    buffer.append(this.superclass != null ? this.superclass.debugName() : "NULL TYPE");

    if (this.superInterfaces != null) {
      if (this.superInterfaces != Binding.NO_SUPERINTERFACES) {
        buffer.append("\n\timplements : ");
        int i = 0; for (int length = this.superInterfaces.length; i < length; i++) {
          if (i > 0)
            buffer.append(", ");
          buffer.append(this.superInterfaces[i] != null ? this.superInterfaces[i].debugName() : "NULL TYPE");
        }
      }
    }
    else buffer.append("NULL SUPERINTERFACES");

    if (this.enclosingType != null) {
      buffer.append("\n\tenclosing type : ");
      buffer.append(this.enclosingType.debugName());
    }

    if (this.fields != null) {
      if (this.fields != Binding.NO_FIELDS) {
        buffer.append("\n/*   fields   */");
        int i = 0; for (int length = this.fields.length; i < length; i++)
          buffer.append(this.fields[i] != null ? "\n" + this.fields[i].toString() : "\nNULL FIELD");
      }
    }
    else buffer.append("NULL FIELDS");

    if (this.methods != null) {
      if (this.methods != Binding.NO_METHODS) {
        buffer.append("\n/*   methods   */");
        int i = 0; for (int length = this.methods.length; i < length; i++)
          buffer.append(this.methods[i] != null ? "\n" + this.methods[i].toString() : "\nNULL METHOD");
      }
    }
    else buffer.append("NULL METHODS");

    if (this.memberTypes != null) {
      if (this.memberTypes != Binding.NO_MEMBER_TYPES) {
        buffer.append("\n/*   members   */");
        int i = 0; for (int length = this.memberTypes.length; i < length; i++)
          buffer.append(this.memberTypes[i] != null ? "\n" + this.memberTypes[i].toString() : "\nNULL TYPE");
      }
    }
    else buffer.append("NULL MEMBER TYPES");

    buffer.append("\n\n\n");
    return buffer.toString();
  }
  MethodBinding[] unResolvedMethods() {
    return this.methods;
  }
}