package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.ast.CaseStatement;
import org.eclipse.jdt.internal.compiler.ast.LocalDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;

public class BlockScope extends Scope
{
  public LocalVariableBinding[] locals;
  public int localIndex;
  public int startIndex;
  public int offset;
  public int maxOffset;
  public BlockScope[] shiftScopes;
  public Scope[] subscopes = new Scope[1];
  public int subscopeCount = 0;
  public CaseStatement enclosingCase;
  public static final VariableBinding[] EmulationPathToImplicitThis = new VariableBinding[0];
  public static final VariableBinding[] NoEnclosingInstanceInConstructorCall = new VariableBinding[0];

  public static final VariableBinding[] NoEnclosingInstanceInStaticContext = new VariableBinding[0];

  public BlockScope(BlockScope parent) {
    this(parent, true);
  }

  public BlockScope(BlockScope parent, boolean addToParentScope) {
    this(1, parent);
    this.locals = new LocalVariableBinding[5];
    if (addToParentScope) parent.addSubscope(this);
    this.startIndex = parent.localIndex;
  }

  public BlockScope(BlockScope parent, int variableCount) {
    this(1, parent);
    this.locals = new LocalVariableBinding[variableCount];
    parent.addSubscope(this);
    this.startIndex = parent.localIndex;
  }

  protected BlockScope(int kind, Scope parent) {
    super(kind, parent);
  }

  public final void addAnonymousType(TypeDeclaration anonymousType, ReferenceBinding superBinding)
  {
    ClassScope anonymousClassScope = new ClassScope(this, anonymousType);
    anonymousClassScope.buildAnonymousTypeBinding(
      enclosingSourceType(), 
      superBinding);
  }

  public final void addLocalType(TypeDeclaration localType)
  {
    ClassScope localTypeScope = new ClassScope(this, localType);
    addSubscope(localTypeScope);
    localTypeScope.buildLocalTypeBinding(enclosingSourceType());
  }

  public final void addLocalVariable(LocalVariableBinding binding)
  {
    checkAndSetModifiersForVariable(binding);

    if (this.localIndex == this.locals.length)
      System.arraycopy(
        this.locals, 
        0, 
        this.locals = new LocalVariableBinding[this.localIndex * 2], 
        0, 
        this.localIndex);
    this.locals[(this.localIndex++)] = binding;

    binding.declaringScope = this;
    binding.id = (outerMostMethodScope().analysisIndex++);
  }

  public void addSubscope(Scope childScope)
  {
    if (this.subscopeCount == this.subscopes.length)
      System.arraycopy(
        this.subscopes, 
        0, 
        this.subscopes = new Scope[this.subscopeCount * 2], 
        0, 
        this.subscopeCount);
    this.subscopes[(this.subscopeCount++)] = childScope;
  }

  public final boolean allowBlankFinalFieldAssignment(FieldBinding binding)
  {
    if (enclosingReceiverType() != binding.declaringClass) {
      return false;
    }
    MethodScope methodScope = methodScope();
    if (methodScope.isStatic != binding.isStatic()) {
      return false;
    }
    return (methodScope.isInsideInitializer()) || 
      (((AbstractMethodDeclaration)methodScope.referenceContext).isInitializationMethod());
  }
  String basicToString(int tab) {
    String newLine = "\n";
    int i = tab;
    do { newLine = newLine + "\t";

      i--; } while (i >= 0);

    String s = newLine + "--- Block Scope ---";
    newLine = newLine + "\t";
    s = s + newLine + "locals:";
    for (int i = 0; i < this.localIndex; i++)
      s = s + newLine + "\t" + this.locals[i].toString();
    s = s + newLine + "startIndex = " + this.startIndex;
    return s;
  }

  private void checkAndSetModifiersForVariable(LocalVariableBinding varBinding) {
    int modifiers = varBinding.modifiers;
    if (((modifiers & 0x400000) != 0) && (varBinding.declaration != null)) {
      problemReporter().duplicateModifierForVariable(varBinding.declaration, this instanceof MethodScope);
    }
    int realModifiers = modifiers & 0xFFFF;

    int unexpectedModifiers = -17;
    if (((realModifiers & unexpectedModifiers) != 0) && (varBinding.declaration != null)) {
      problemReporter().illegalModifierForVariable(varBinding.declaration, this instanceof MethodScope);
    }
    varBinding.modifiers = modifiers;
  }

  void computeLocalVariablePositions(int ilocal, int initOffset, CodeStream codeStream)
  {
    this.offset = initOffset;
    this.maxOffset = initOffset;

    int maxLocals = this.localIndex;
    boolean hasMoreVariables = ilocal < maxLocals;

    int iscope = 0; int maxScopes = this.subscopeCount;
    boolean hasMoreScopes = maxScopes > 0;

    while ((hasMoreVariables) || (hasMoreScopes)) {
      if ((hasMoreScopes) && (
        (!hasMoreVariables) || (this.subscopes[iscope].startIndex() <= ilocal)))
      {
        if ((this.subscopes[iscope] instanceof BlockScope)) {
          BlockScope subscope = (BlockScope)this.subscopes[iscope];
          int subOffset = subscope.shiftScopes == null ? this.offset : subscope.maxShiftedOffset();
          subscope.computeLocalVariablePositions(0, subOffset, codeStream);
          if (subscope.maxOffset > this.maxOffset)
            this.maxOffset = subscope.maxOffset;
        }
        iscope++; hasMoreScopes = iscope < maxScopes;
      }
      else
      {
        LocalVariableBinding local = this.locals[ilocal];

        boolean generateCurrentLocalVar = (local.useFlag == 1) && (local.constant() == Constant.NotAConstant);

        if ((local.useFlag == 0) && 
          (local.declaration != null) && 
          ((local.declaration.bits & 0x40000000) != 0))
        {
          if (!(local.declaration instanceof Argument)) {
            problemReporter().unusedLocalVariable(local.declaration);
          }
        }

        if ((!generateCurrentLocalVar) && 
          (local.declaration != null) && (compilerOptions().preserveAllLocalVariables)) {
          generateCurrentLocalVar = true;
          local.useFlag = 1;
        }

        if (generateCurrentLocalVar)
        {
          if (local.declaration != null) {
            codeStream.record(local);
          }

          local.resolvedPosition = this.offset;

          if ((local.type == TypeBinding.LONG) || (local.type == TypeBinding.DOUBLE))
            this.offset += 2;
          else {
            this.offset += 1;
          }
          if (this.offset > 65535)
            problemReporter().noMoreAvailableSpaceForLocal(
              local, 
              local.declaration == null ? (ASTNode)methodScope().referenceContext : local.declaration);
        }
        else {
          local.resolvedPosition = -1;
        }
        ilocal++; hasMoreVariables = ilocal < maxLocals;
      }
    }
    if (this.offset > this.maxOffset)
      this.maxOffset = this.offset;
  }

  public void emulateOuterAccess(LocalVariableBinding outerLocalVariable)
  {
    BlockScope outerVariableScope = outerLocalVariable.declaringScope;
    if (outerVariableScope == null)
      return;
    MethodScope currentMethodScope = methodScope();
    if (outerVariableScope.methodScope() != currentMethodScope) {
      NestedTypeBinding currentType = (NestedTypeBinding)enclosingSourceType();

      if (!currentType.isLocalType()) {
        return;
      }

      if (!currentMethodScope.isInsideInitializerOrConstructor())
        currentType.addSyntheticArgumentAndField(outerLocalVariable);
      else
        currentType.addSyntheticArgument(outerLocalVariable);
    }
  }

  public final ReferenceBinding findLocalType(char[] name)
  {
    long compliance = compilerOptions().complianceLevel;
    for (int i = this.subscopeCount - 1; i >= 0; i--)
      if ((this.subscopes[i] instanceof ClassScope)) {
        LocalTypeBinding sourceType = (LocalTypeBinding)((ClassScope)this.subscopes[i]).referenceContext.binding;

        if ((compliance < 3145728L) || (sourceType.enclosingCase == null) || 
          (isInsideCase(sourceType.enclosingCase)))
        {
          if (CharOperation.equals(sourceType.sourceName(), name))
            return sourceType;
        }
      }
    return null;
  }

  public LocalDeclaration[] findLocalVariableDeclarations(int position)
  {
    int ilocal = 0; int maxLocals = this.localIndex;
    boolean hasMoreVariables = maxLocals > 0;
    LocalDeclaration[] localDeclarations = (LocalDeclaration[])null;
    int declPtr = 0;

    int iscope = 0; int maxScopes = this.subscopeCount;
    boolean hasMoreScopes = maxScopes > 0;

    while ((hasMoreVariables) || (hasMoreScopes)) {
      if ((hasMoreScopes) && (
        (!hasMoreVariables) || (this.subscopes[iscope].startIndex() <= ilocal)))
      {
        Scope subscope = this.subscopes[iscope];
        if (subscope.kind == 1) {
          localDeclarations = ((BlockScope)subscope).findLocalVariableDeclarations(position);
          if (localDeclarations != null) {
            return localDeclarations;
          }
        }
        iscope++; hasMoreScopes = iscope < maxScopes;
      }
      else {
        LocalVariableBinding local = this.locals[ilocal];
        if (local != null) {
          LocalDeclaration localDecl = local.declaration;
          if (localDecl != null) {
            if (localDecl.declarationSourceStart <= position) {
              if (position <= localDecl.declarationSourceEnd) {
                if (localDeclarations == null) {
                  localDeclarations = new LocalDeclaration[maxLocals];
                }
                localDeclarations[(declPtr++)] = localDecl;
              }
            }
            else return localDeclarations;
          }
        }

        ilocal++; hasMoreVariables = ilocal < maxLocals;
        if ((!hasMoreVariables) && (localDeclarations != null)) {
          return localDeclarations;
        }
      }
    }
    return null;
  }

  public LocalVariableBinding findVariable(char[] variableName) {
    int varLength = variableName.length;
    for (int i = this.localIndex - 1; i >= 0; i--)
    {
      LocalVariableBinding local;
      char[] localName;
      if (((localName = (local = this.locals[i]).name).length == varLength) && (CharOperation.equals(localName, variableName)))
        return local;
    }
    return null;
  }

  public Binding getBinding(char[][] compoundName, int mask, InvocationSite invocationSite, boolean needResolve)
  {
    Binding binding = getBinding(compoundName[0], mask | 0x4 | 0x10, invocationSite, needResolve);
    invocationSite.setFieldIndex(1);
    if ((binding instanceof VariableBinding)) return binding;
    CompilationUnitScope unitScope = compilationUnitScope();

    unitScope.recordQualifiedReference(compoundName);
    if (!binding.isValidBinding()) return binding;

    int length = compoundName.length;
    int currentIndex = 1;
    if ((binding instanceof PackageBinding)) {
      PackageBinding packageBinding = (PackageBinding)binding;
      while (currentIndex < length) {
        unitScope.recordReference(packageBinding.compoundName, compoundName[currentIndex]);
        binding = packageBinding.getTypeOrPackage(compoundName[(currentIndex++)]);
        invocationSite.setFieldIndex(currentIndex);
        if (binding == null) {
          if (currentIndex == length)
          {
            return new ProblemReferenceBinding(
              CharOperation.subarray(compoundName, 0, currentIndex), 
              null, 
              1);
          }
          return new ProblemBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            1);
        }
        if ((binding instanceof ReferenceBinding)) {
          if (!binding.isValidBinding())
            return new ProblemReferenceBinding(
              CharOperation.subarray(compoundName, 0, currentIndex), 
              ((ReferenceBinding)binding).closestMatch(), 
              binding.problemId());
          if (((ReferenceBinding)binding).canBeSeenBy(this)) break;
          return new ProblemReferenceBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            (ReferenceBinding)binding, 
            2);
        }

        packageBinding = (PackageBinding)binding;
      }

      return new ProblemReferenceBinding(
        CharOperation.subarray(compoundName, 0, currentIndex), 
        null, 
        1);
    }

    ReferenceBinding referenceBinding = (ReferenceBinding)binding;
    binding = environment().convertToRawType(referenceBinding);
    if ((invocationSite instanceof ASTNode)) {
      ASTNode invocationNode = (ASTNode)invocationSite;
      if (invocationNode.isTypeUseDeprecated(referenceBinding, this)) {
        problemReporter().deprecatedType(referenceBinding, invocationNode);
      }
    }
    while (currentIndex < length) {
      referenceBinding = (ReferenceBinding)binding;
      char[] nextName = compoundName[(currentIndex++)];
      invocationSite.setFieldIndex(currentIndex);
      invocationSite.setActualReceiverType(referenceBinding);
      if (((mask & 0x1) != 0) && ((binding = findField(referenceBinding, nextName, invocationSite, true)) != null)) {
        if (binding.isValidBinding()) break;
        return new ProblemFieldBinding(
          ((ProblemFieldBinding)binding).closestMatch, 
          ((ProblemFieldBinding)binding).declaringClass, 
          CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), 
          binding.problemId());
      }

      if ((binding = findMemberType(nextName, referenceBinding)) == null) {
        if ((mask & 0x1) != 0) {
          return new ProblemBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            referenceBinding, 
            1);
        }
        return new ProblemReferenceBinding(
          CharOperation.subarray(compoundName, 0, currentIndex), 
          referenceBinding, 
          1);
      }

      if (!binding.isValidBinding())
        return new ProblemReferenceBinding(
          CharOperation.subarray(compoundName, 0, currentIndex), 
          ((ReferenceBinding)binding).closestMatch(), 
          binding.problemId());
      if ((invocationSite instanceof ASTNode)) {
        referenceBinding = (ReferenceBinding)binding;
        ASTNode invocationNode = (ASTNode)invocationSite;
        if (invocationNode.isTypeUseDeprecated(referenceBinding, this)) {
          problemReporter().deprecatedType(referenceBinding, invocationNode);
        }
      }
    }
    if (((mask & 0x1) != 0) && ((binding instanceof FieldBinding)))
    {
      FieldBinding field = (FieldBinding)binding;
      if (!field.isStatic())
        return new ProblemFieldBinding(
          field, 
          field.declaringClass, 
          CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), 
          7);
      return binding;
    }
    if (((mask & 0x4) != 0) && ((binding instanceof ReferenceBinding)))
    {
      return binding;
    }

    return new ProblemBinding(
      CharOperation.subarray(compoundName, 0, currentIndex), 
      1);
  }

  public final Binding getBinding(char[][] compoundName, InvocationSite invocationSite)
  {
    int currentIndex = 0;
    int length = compoundName.length;
    Binding binding = 
      getBinding(
      compoundName[(currentIndex++)], 
      23, 
      invocationSite, 
      true);
    if (!binding.isValidBinding()) {
      return binding;
    }
    if ((binding instanceof PackageBinding)) {
      while (currentIndex < length) {
        PackageBinding packageBinding = (PackageBinding)binding;
        binding = packageBinding.getTypeOrPackage(compoundName[(currentIndex++)]);
        if (binding == null) {
          if (currentIndex == length)
          {
            return new ProblemReferenceBinding(
              CharOperation.subarray(compoundName, 0, currentIndex), 
              null, 
              1);
          }
          return new ProblemBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            1);
        }
        if ((binding instanceof ReferenceBinding)) {
          if (!binding.isValidBinding())
            return new ProblemReferenceBinding(
              CharOperation.subarray(compoundName, 0, currentIndex), 
              ((ReferenceBinding)binding).closestMatch(), 
              binding.problemId());
          if (((ReferenceBinding)binding).canBeSeenBy(this)) break;
          return new ProblemReferenceBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            (ReferenceBinding)binding, 
            2);
        }
      }

      return binding;
    }

    if ((binding instanceof ReferenceBinding)) {
      while (currentIndex < length) {
        ReferenceBinding typeBinding = (ReferenceBinding)binding;
        char[] nextName = compoundName[(currentIndex++)];
        if ((binding = findField(typeBinding, nextName, invocationSite, true)) != null) {
          if (!binding.isValidBinding()) {
            return new ProblemFieldBinding(
              (FieldBinding)binding, 
              ((FieldBinding)binding).declaringClass, 
              CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), 
              binding.problemId());
          }
          if (((FieldBinding)binding).isStatic()) break;
          return new ProblemFieldBinding(
            (FieldBinding)binding, 
            ((FieldBinding)binding).declaringClass, 
            CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), 
            7);
        }

        if ((binding = findMemberType(nextName, typeBinding)) == null) {
          return new ProblemBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            typeBinding, 
            1);
        }
        if (!binding.isValidBinding()) {
          return new ProblemReferenceBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            ((ReferenceBinding)binding).closestMatch(), 
            binding.problemId());
        }
      }
      return binding;
    }

    VariableBinding variableBinding = (VariableBinding)binding;
    while (currentIndex < length) {
      TypeBinding typeBinding = variableBinding.type;
      if (typeBinding == null) {
        return new ProblemFieldBinding(
          null, 
          null, 
          CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), 
          1);
      }
      variableBinding = findField(typeBinding, compoundName[(currentIndex++)], invocationSite, true);
      if (variableBinding == null) {
        return new ProblemFieldBinding(
          null, 
          null, 
          CharOperation.concatWith(CharOperation.subarray(compoundName, 0, currentIndex), '.'), 
          1);
      }
      if (!variableBinding.isValidBinding())
        return variableBinding;
    }
    return variableBinding;
  }

  public VariableBinding[] getEmulationPath(LocalVariableBinding outerLocalVariable)
  {
    MethodScope currentMethodScope = methodScope();
    SourceTypeBinding sourceType = currentMethodScope.enclosingSourceType();

    BlockScope variableScope = outerLocalVariable.declaringScope;
    if ((variableScope == null) || (currentMethodScope == variableScope.methodScope())) {
      return new VariableBinding[] { outerLocalVariable };
    }

    if ((currentMethodScope.isInsideInitializerOrConstructor()) && 
      (sourceType.isNestedType()))
    {
      SyntheticArgumentBinding syntheticArg;
      if ((syntheticArg = ((NestedTypeBinding)sourceType).getSyntheticArgument(outerLocalVariable)) != null) {
        return new VariableBinding[] { syntheticArg };
      }
    }

    if (!currentMethodScope.isStatic)
    {
      FieldBinding syntheticField;
      if ((syntheticField = sourceType.getSyntheticField(outerLocalVariable)) != null) {
        return new VariableBinding[] { syntheticField };
      }
    }
    return null;
  }

  public Object[] getEmulationPath(ReferenceBinding targetEnclosingType, boolean onlyExactMatch, boolean denyEnclosingArgInConstructorCall)
  {
    MethodScope currentMethodScope = methodScope();
    SourceTypeBinding sourceType = currentMethodScope.enclosingSourceType();

    if ((!currentMethodScope.isStatic) && (!currentMethodScope.isConstructorCall) && (
      (sourceType == targetEnclosingType) || ((!onlyExactMatch) && (sourceType.findSuperTypeWithSameErasure(targetEnclosingType) != null)))) {
      return EmulationPathToImplicitThis;
    }

    if ((!sourceType.isNestedType()) || (sourceType.isStatic())) {
      if (currentMethodScope.isConstructorCall)
        return NoEnclosingInstanceInConstructorCall;
      if (currentMethodScope.isStatic) {
        return NoEnclosingInstanceInStaticContext;
      }
      return null;
    }
    boolean insideConstructor = currentMethodScope.isInsideInitializerOrConstructor();

    if (insideConstructor)
    {
      SyntheticArgumentBinding syntheticArg;
      if ((syntheticArg = ((NestedTypeBinding)sourceType).getSyntheticArgument(targetEnclosingType, onlyExactMatch)) != null)
      {
        if ((denyEnclosingArgInConstructorCall) && 
          (currentMethodScope.isConstructorCall) && (
          (sourceType == targetEnclosingType) || ((!onlyExactMatch) && (sourceType.findSuperTypeWithSameErasure(targetEnclosingType) != null)))) {
          return NoEnclosingInstanceInConstructorCall;
        }
        return new Object[] { syntheticArg };
      }

    }

    if (currentMethodScope.isStatic) {
      return NoEnclosingInstanceInStaticContext;
    }
    if (sourceType.isAnonymousType()) {
      ReferenceBinding enclosingType = sourceType.enclosingType();
      if (enclosingType.isNestedType()) {
        NestedTypeBinding nestedEnclosingType = (NestedTypeBinding)enclosingType;
        SyntheticArgumentBinding enclosingArgument = nestedEnclosingType.getSyntheticArgument(nestedEnclosingType.enclosingType(), onlyExactMatch);
        if (enclosingArgument != null) {
          FieldBinding syntheticField = sourceType.getSyntheticField(enclosingArgument);
          if ((syntheticField != null) && (
            (syntheticField.type == targetEnclosingType) || ((!onlyExactMatch) && (((ReferenceBinding)syntheticField.type).findSuperTypeWithSameErasure(targetEnclosingType) != null)))) {
            return new Object[] { syntheticField };
          }
        }
      }
    }
    FieldBinding syntheticField = sourceType.getSyntheticField(targetEnclosingType, onlyExactMatch);
    if (syntheticField != null) {
      if (currentMethodScope.isConstructorCall) {
        return NoEnclosingInstanceInConstructorCall;
      }
      return new Object[] { syntheticField };
    }

    Object[] path = new Object[2];
    ReferenceBinding currentType = sourceType.enclosingType();
    if (insideConstructor) {
      path[0] = ((NestedTypeBinding)sourceType).getSyntheticArgument(currentType, onlyExactMatch);
    } else {
      if (currentMethodScope.isConstructorCall) {
        return NoEnclosingInstanceInConstructorCall;
      }
      path[0] = sourceType.getSyntheticField(currentType, onlyExactMatch);
    }
    if (path[0] != null)
    {
      int count = 1;
      ReferenceBinding currentEnclosingType;
      while ((currentEnclosingType = currentType.enclosingType()) != null)
      {
        ReferenceBinding currentEnclosingType;
        if ((currentType == targetEnclosingType) || (
          (!onlyExactMatch) && (currentType.findSuperTypeWithSameErasure(targetEnclosingType) != null)))
          break;
        if (currentMethodScope != null) {
          currentMethodScope = currentMethodScope.enclosingMethodScope();
          if ((currentMethodScope != null) && (currentMethodScope.isConstructorCall)) {
            return NoEnclosingInstanceInConstructorCall;
          }
          if ((currentMethodScope != null) && (currentMethodScope.isStatic)) {
            return NoEnclosingInstanceInStaticContext;
          }
        }

        syntheticField = ((NestedTypeBinding)currentType).getSyntheticField(currentEnclosingType, onlyExactMatch);
        if (syntheticField == null) {
          break;
        }
        if (count == path.length) {
          System.arraycopy(path, 0, path = new Object[count + 1], 0, count);
        }

        path[(count++)] = ((SourceTypeBinding)syntheticField.declaringClass).addSyntheticMethod(syntheticField, true);
        currentType = currentEnclosingType;
      }
      if ((currentType == targetEnclosingType) || (
        (!onlyExactMatch) && (currentType.findSuperTypeWithSameErasure(targetEnclosingType) != null))) {
        return path;
      }
    }
    return null;
  }

  public final boolean isDuplicateLocalVariable(char[] name)
  {
    BlockScope current = this;
    while (true) {
      for (int i = 0; i < this.localIndex; i++) {
        if (CharOperation.equals(name, current.locals[i].name))
          return true;
      }
      if (current.kind != 1) return false;
      current = (BlockScope)current.parent;
    }
  }

  public int maxShiftedOffset() {
    int max = -1;
    if (this.shiftScopes != null) {
      int i = 0; for (int length = this.shiftScopes.length; i < length; i++) {
        int subMaxOffset = this.shiftScopes[i].maxOffset;
        if (subMaxOffset > max) max = subMaxOffset;
      }
    }
    return max;
  }

  public ProblemReporter problemReporter()
  {
    return outerMostMethodScope().problemReporter();
  }

  public void propagateInnerEmulation(ReferenceBinding targetType, boolean isEnclosingInstanceSupplied)
  {
    SyntheticArgumentBinding[] syntheticArguments;
    if ((syntheticArguments = targetType.syntheticOuterLocalVariables()) != null) {
      int i = 0; for (int max = syntheticArguments.length; i < max; i++) {
        SyntheticArgumentBinding syntheticArg = syntheticArguments[i];

        if ((!isEnclosingInstanceSupplied) || 
          (syntheticArg.type != targetType.enclosingType()))
          emulateOuterAccess(syntheticArg.actualOuterLocalVariable);
      }
    }
  }

  public TypeDeclaration referenceType()
  {
    return methodScope().referenceType();
  }

  public int scopeIndex()
  {
    if ((this instanceof MethodScope)) return -1;
    BlockScope parentScope = (BlockScope)this.parent;
    Scope[] parentSubscopes = parentScope.subscopes;
    int i = 0; for (int max = parentScope.subscopeCount; i < max; i++) {
      if (parentSubscopes[i] == this) return i;
    }
    return -1;
  }

  int startIndex()
  {
    return this.startIndex;
  }

  public String toString() {
    return toString(0);
  }

  public String toString(int tab) {
    String s = basicToString(tab);
    for (int i = 0; i < this.subscopeCount; i++)
      if ((this.subscopes[i] instanceof BlockScope))
        s = s + ((BlockScope)this.subscopes[i]).toString(tab + 1) + "\n";
    return s;
  }
}