package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.codegen.ConstantPool;

public class MethodBinding extends Binding
{
  public int modifiers;
  public char[] selector;
  public TypeBinding returnType;
  public TypeBinding[] parameters;
  public ReferenceBinding[] thrownExceptions;
  public ReferenceBinding declaringClass;
  public TypeVariableBinding[] typeVariables = Binding.NO_TYPE_VARIABLES;
  char[] signature;
  public long tagBits;

  protected MethodBinding()
  {
  }

  public MethodBinding(int modifiers, char[] selector, TypeBinding returnType, TypeBinding[] parameters, ReferenceBinding[] thrownExceptions, ReferenceBinding declaringClass)
  {
    this.modifiers = modifiers;
    this.selector = selector;
    this.returnType = returnType;
    this.parameters = ((parameters == null) || (parameters.length == 0) ? Binding.NO_PARAMETERS : parameters);
    this.thrownExceptions = ((thrownExceptions == null) || (thrownExceptions.length == 0) ? Binding.NO_EXCEPTIONS : thrownExceptions);
    this.declaringClass = declaringClass;

    if ((this.declaringClass != null) && 
      (this.declaringClass.isStrictfp()) && 
      (!isNative()) && (!isAbstract()))
      this.modifiers |= 2048;
  }

  public MethodBinding(int modifiers, TypeBinding[] parameters, ReferenceBinding[] thrownExceptions, ReferenceBinding declaringClass) {
    this(modifiers, TypeConstants.INIT, TypeBinding.VOID, parameters, thrownExceptions, declaringClass);
  }

  public MethodBinding(MethodBinding initialMethodBinding, ReferenceBinding declaringClass) {
    this.modifiers = initialMethodBinding.modifiers;
    this.selector = initialMethodBinding.selector;
    this.returnType = initialMethodBinding.returnType;
    this.parameters = initialMethodBinding.parameters;
    this.thrownExceptions = initialMethodBinding.thrownExceptions;
    this.declaringClass = declaringClass;
    declaringClass.storeAnnotationHolder(this, initialMethodBinding.declaringClass.retrieveAnnotationHolder(initialMethodBinding, true));
  }

  public final boolean areParameterErasuresEqual(MethodBinding method)
  {
    TypeBinding[] args = method.parameters;
    if (this.parameters == args) {
      return true;
    }
    int length = this.parameters.length;
    if (length != args.length) {
      return false;
    }
    for (int i = 0; i < length; i++)
      if ((this.parameters[i] != args[i]) && (this.parameters[i].erasure() != args[i].erasure()))
        return false;
    return true;
  }

  public final boolean areParametersEqual(MethodBinding method)
  {
    TypeBinding[] args = method.parameters;
    if (this.parameters == args) {
      return true;
    }
    int length = this.parameters.length;
    if (length != args.length) {
      return false;
    }
    for (int i = 0; i < length; i++)
      if (this.parameters[i] != args[i])
        return false;
    return true;
  }

  public final boolean areParametersCompatibleWith(TypeBinding[] arguments)
  {
    int paramLength = this.parameters.length;
    int argLength = arguments.length;
    int lastIndex = argLength;
    if (isVarargs()) {
      lastIndex = paramLength - 1;
      if (paramLength == argLength) {
        TypeBinding varArgType = this.parameters[lastIndex];
        TypeBinding lastArgument = arguments[lastIndex];
        if ((varArgType != lastArgument) && (!lastArgument.isCompatibleWith(varArgType)))
          return false;
      } else if (paramLength < argLength) {
        TypeBinding varArgType = ((ArrayBinding)this.parameters[lastIndex]).elementsType();
        for (int i = lastIndex; i < argLength; i++)
          if ((varArgType != arguments[i]) && (!arguments[i].isCompatibleWith(varArgType)))
            return false;
      } else if (lastIndex != argLength) {
        return false;
      }
    }

    for (int i = 0; i < lastIndex; i++)
      if ((this.parameters[i] != arguments[i]) && (!arguments[i].isCompatibleWith(this.parameters[i])))
        return false;
    return true;
  }

  public final int kind()
  {
    return 8;
  }

  public final boolean canBeSeenBy(PackageBinding invocationPackage)
  {
    if (isPublic()) return true;
    if (isPrivate()) return false;

    return invocationPackage == this.declaringClass.getPackage();
  }

  public final boolean areTypeVariableErasuresEqual(MethodBinding method)
  {
    TypeVariableBinding[] vars = method.typeVariables;
    if (this.typeVariables == vars) {
      return true;
    }
    int length = this.typeVariables.length;
    if (length != vars.length) {
      return false;
    }
    for (int i = 0; i < length; i++)
      if ((this.typeVariables[i] != vars[i]) && (this.typeVariables[i].erasure() != vars[i].erasure()))
        return false;
    return true;
  }

  public final boolean canBeSeenBy(InvocationSite invocationSite, Scope scope)
  {
    if (isPublic()) return true;

    SourceTypeBinding invocationType = scope.enclosingSourceType();
    if (invocationType == this.declaringClass) return true;

    if (isProtected())
    {
      if (invocationType.fPackage == this.declaringClass.fPackage) return true;
      return invocationSite.isSuperAccess();
    }

    if (isPrivate())
    {
      ReferenceBinding outerInvocationType = invocationType;
      ReferenceBinding temp = outerInvocationType.enclosingType();
      while (temp != null) {
        outerInvocationType = temp;
        temp = temp.enclosingType();
      }

      ReferenceBinding outerDeclaringClass = (ReferenceBinding)this.declaringClass.erasure();
      temp = outerDeclaringClass.enclosingType();
      while (temp != null) {
        outerDeclaringClass = temp;
        temp = temp.enclosingType();
      }
      return outerInvocationType == outerDeclaringClass;
    }

    return invocationType.fPackage == this.declaringClass.fPackage;
  }

  public final boolean canBeSeenBy(TypeBinding receiverType, InvocationSite invocationSite, Scope scope)
  {
    if (isPublic()) return true;

    SourceTypeBinding invocationType = scope.enclosingSourceType();
    if ((invocationType == this.declaringClass) && (invocationType == receiverType)) return true;

    if (invocationType == null) {
      return (!isPrivate()) && (scope.getCurrentPackage() == this.declaringClass.fPackage);
    }
    if (isProtected())
    {
      if (invocationType == this.declaringClass) return true;
      if (invocationType.fPackage == this.declaringClass.fPackage) return true;

      ReferenceBinding currentType = invocationType;
      TypeBinding receiverErasure = receiverType.erasure();
      ReferenceBinding declaringErasure = (ReferenceBinding)this.declaringClass.erasure();
      int depth = 0;
      do {
        if (currentType.findSuperTypeWithSameErasure(declaringErasure) != null) {
          if (invocationSite.isSuperAccess()) {
            return true;
          }
          if ((receiverType instanceof ArrayBinding))
            return false;
          if (isStatic()) {
            if (depth > 0) invocationSite.setDepth(depth);
            return true;
          }
          if ((currentType == receiverErasure) || (receiverErasure.findSuperTypeWithSameErasure(currentType) != null)) {
            if (depth > 0) invocationSite.setDepth(depth);
            return true;
          }
        }
        depth++;
        currentType = currentType.enclosingType();
      }while (currentType != null);
      return false;
    }

    if (isPrivate())
    {
      if (receiverType != this.declaringClass)
      {
        if ((!receiverType.isTypeVariable()) || (!((TypeVariableBinding)receiverType).isErasureBoundTo(this.declaringClass.erasure())))
        {
          return false;
        }
      }

      if (invocationType != this.declaringClass) {
        ReferenceBinding outerInvocationType = invocationType;
        ReferenceBinding temp = outerInvocationType.enclosingType();
        while (temp != null) {
          outerInvocationType = temp;
          temp = temp.enclosingType();
        }

        ReferenceBinding outerDeclaringClass = (ReferenceBinding)this.declaringClass.erasure();
        temp = outerDeclaringClass.enclosingType();
        while (temp != null) {
          outerDeclaringClass = temp;
          temp = temp.enclosingType();
        }
        if (outerInvocationType != outerDeclaringClass) return false;
      }
      return true;
    }

    PackageBinding declaringPackage = this.declaringClass.fPackage;
    if (invocationType.fPackage != declaringPackage) return false;

    if ((receiverType instanceof ArrayBinding))
      return false;
    ReferenceBinding currentType = (ReferenceBinding)receiverType;
    do {
      if (this.declaringClass == currentType) return true;
      PackageBinding currentPackage = currentType.fPackage;

      if ((currentPackage != null) && (currentPackage != declaringPackage)) return false; 
    }
    while ((currentType = currentType.superclass()) != null);
    return false;
  }
  MethodBinding computeSubstitutedMethod(MethodBinding method, LookupEnvironment env) {
    int length = this.typeVariables.length;
    TypeVariableBinding[] vars = method.typeVariables;
    if (length != vars.length) {
      return null;
    }

    ParameterizedGenericMethodBinding substitute = 
      env.createParameterizedGenericMethod(method, this.typeVariables);
    for (int i = 0; i < length; i++)
      if (!this.typeVariables[i].isInterchangeableWith(vars[i], substitute))
        return null;
    return substitute;
  }

  public char[] computeUniqueKey(boolean isLeaf)
  {
    char[] declaringKey = this.declaringClass.computeUniqueKey(false);
    int declaringLength = declaringKey.length;

    int selectorLength = this.selector == TypeConstants.INIT ? 0 : this.selector.length;

    char[] sig = genericSignature();
    boolean isGeneric = sig != null;
    if (!isGeneric) sig = signature();
    int signatureLength = sig.length;

    int thrownExceptionsLength = this.thrownExceptions.length;
    int thrownExceptionsSignatureLength = 0;
    char[][] thrownExceptionsSignatures = (char[][])null;
    boolean addThrownExceptions = (thrownExceptionsLength > 0) && ((!isGeneric) || (CharOperation.lastIndexOf('^', sig) < 0));
    if (addThrownExceptions) {
      thrownExceptionsSignatures = new char[thrownExceptionsLength][];
      for (int i = 0; i < thrownExceptionsLength; i++) {
        if (this.thrownExceptions[i] != null) {
          thrownExceptionsSignatures[i] = this.thrownExceptions[i].signature();
          thrownExceptionsSignatureLength += thrownExceptionsSignatures[i].length + 1;
        }
      }
    }

    char[] uniqueKey = new char[declaringLength + 1 + selectorLength + signatureLength + thrownExceptionsSignatureLength];
    int index = 0;
    System.arraycopy(declaringKey, 0, uniqueKey, index, declaringLength);
    index = declaringLength;
    uniqueKey[(index++)] = '.';
    System.arraycopy(this.selector, 0, uniqueKey, index, selectorLength);
    index += selectorLength;
    System.arraycopy(sig, 0, uniqueKey, index, signatureLength);
    if (thrownExceptionsSignatureLength > 0) {
      index += signatureLength;
      for (int i = 0; i < thrownExceptionsLength; i++) {
        char[] thrownExceptionSignature = thrownExceptionsSignatures[i];
        if (thrownExceptionSignature != null) {
          uniqueKey[(index++)] = '|';
          int length = thrownExceptionSignature.length;
          System.arraycopy(thrownExceptionSignature, 0, uniqueKey, index, length);
          index += length;
        }
      }
    }

    return uniqueKey;
  }

  public TypeBinding constantPoolDeclaringClass()
  {
    return this.declaringClass;
  }

  public final char[] constantPoolName()
  {
    return this.selector;
  }

  public char[] genericSignature()
  {
    if ((this.modifiers & 0x40000000) == 0) return null;
    StringBuffer sig = new StringBuffer(10);
    if (this.typeVariables != Binding.NO_TYPE_VARIABLES) {
      sig.append('<');
      int i = 0; for (int length = this.typeVariables.length; i < length; i++) {
        sig.append(this.typeVariables[i].genericSignature());
      }
      sig.append('>');
    }
    sig.append('(');
    int i = 0; for (int length = this.parameters.length; i < length; i++) {
      sig.append(this.parameters[i].genericTypeSignature());
    }
    sig.append(')');
    if (this.returnType != null) {
      sig.append(this.returnType.genericTypeSignature());
    }

    boolean needExceptionSignatures = false;
    int length = this.thrownExceptions.length;
    for (int i = 0; i < length; i++) {
      if ((this.thrownExceptions[i].modifiers & 0x40000000) != 0) {
        needExceptionSignatures = true;
        break;
      }
    }
    if (needExceptionSignatures) {
      for (int i = 0; i < length; i++) {
        sig.append('^');
        sig.append(this.thrownExceptions[i].genericTypeSignature());
      }
    }
    int sigLength = sig.length();
    char[] genericSignature = new char[sigLength];
    sig.getChars(0, sigLength, genericSignature, 0);
    return genericSignature;
  }
  public AnnotationBinding[] getAnnotations() {
    MethodBinding originalMethod = original();
    return originalMethod.declaringClass.retrieveAnnotations(originalMethod);
  }

  public AnnotationBinding[] getParameterAnnotations(int index)
  {
    MethodBinding originalMethod = original();
    AnnotationHolder holder = originalMethod.declaringClass.retrieveAnnotationHolder(originalMethod, true);
    return holder == null ? Binding.NO_ANNOTATIONS : holder.getParameterAnnotations(index);
  }
  public final int getAccessFlags() {
    return this.modifiers & 0xFFFF;
  }

  public long getAnnotationTagBits()
  {
    MethodBinding originalMethod = original();
    if (((originalMethod.tagBits & 0x0) == 0L) && ((originalMethod.declaringClass instanceof SourceTypeBinding))) {
      ClassScope scope = ((SourceTypeBinding)originalMethod.declaringClass).scope;
      if (scope != null) {
        TypeDeclaration typeDecl = scope.referenceContext;
        AbstractMethodDeclaration methodDecl = typeDecl.declarationOf(originalMethod);
        if (methodDecl != null)
          ASTNode.resolveAnnotations(methodDecl.scope, methodDecl.annotations, originalMethod);
      }
    }
    return originalMethod.tagBits;
  }

  public Object getDefaultValue()
  {
    MethodBinding originalMethod = original();
    if ((originalMethod.tagBits & 0x0) == 0L)
    {
      if ((originalMethod.declaringClass instanceof SourceTypeBinding)) {
        SourceTypeBinding sourceType = (SourceTypeBinding)originalMethod.declaringClass;
        if (sourceType.scope != null) {
          AbstractMethodDeclaration methodDeclaration = originalMethod.sourceMethod();
          if ((methodDeclaration != null) && (methodDeclaration.isAnnotationMethod())) {
            methodDeclaration.resolve(sourceType.scope);
          }
        }
      }
      originalMethod.tagBits |= 2251799813685248L;
    }
    AnnotationHolder holder = originalMethod.declaringClass.retrieveAnnotationHolder(originalMethod, true);
    return holder == null ? null : holder.getDefaultValue();
  }
  public TypeVariableBinding getTypeVariable(char[] variableName) {
    int i = this.typeVariables.length;
    do { if (CharOperation.equals(this.typeVariables[i].sourceName, variableName))
        return this.typeVariables[i];
      i--; } while (i >= 0);

    return null;
  }

  public boolean hasSubstitutedParameters()
  {
    return false;
  }

  public boolean hasSubstitutedReturnType()
  {
    return false;
  }

  public final boolean isAbstract()
  {
    return (this.modifiers & 0x400) != 0;
  }

  public final boolean isBridge()
  {
    return (this.modifiers & 0x40) != 0;
  }

  public final boolean isConstructor()
  {
    return this.selector == TypeConstants.INIT;
  }

  public final boolean isDefault()
  {
    return (!isPublic()) && (!isProtected()) && (!isPrivate());
  }

  public final boolean isDefaultAbstract()
  {
    return (this.modifiers & 0x80000) != 0;
  }

  public final boolean isDeprecated()
  {
    return (this.modifiers & 0x100000) != 0;
  }

  public final boolean isFinal()
  {
    return (this.modifiers & 0x10) != 0;
  }

  public final boolean isImplementing()
  {
    return (this.modifiers & 0x20000000) != 0;
  }

  public final boolean isNative()
  {
    return (this.modifiers & 0x100) != 0;
  }

  public final boolean isOverriding()
  {
    return (this.modifiers & 0x10000000) != 0;
  }

  public final boolean isMain()
  {
    if ((this.selector.length == 4) && (CharOperation.equals(this.selector, TypeConstants.MAIN)) && 
      ((this.modifiers & 0x9) != 0) && 
      (TypeBinding.VOID == this.returnType) && 
      (this.parameters.length == 1)) {
      TypeBinding paramType = this.parameters[0];
      if ((paramType.dimensions() == 1) && (paramType.leafComponentType().id == 11)) {
        return true;
      }
    }
    return false;
  }

  public final boolean isPrivate()
  {
    return (this.modifiers & 0x2) != 0;
  }

  public final boolean isUsed()
  {
    return (this.modifiers & 0x8000000) != 0;
  }

  public final boolean isProtected()
  {
    return (this.modifiers & 0x4) != 0;
  }

  public final boolean isPublic()
  {
    return (this.modifiers & 0x1) != 0;
  }

  public final boolean isStatic()
  {
    return (this.modifiers & 0x8) != 0;
  }

  public final boolean isStrictfp()
  {
    return (this.modifiers & 0x800) != 0;
  }

  public final boolean isSynchronized()
  {
    return (this.modifiers & 0x20) != 0;
  }

  public final boolean isSynthetic()
  {
    return (this.modifiers & 0x1000) != 0;
  }

  public final boolean isVarargs()
  {
    return (this.modifiers & 0x80) != 0;
  }

  public final boolean isViewedAsDeprecated()
  {
    return (this.modifiers & 0x300000) != 0;
  }

  public MethodBinding original()
  {
    return this;
  }

  public char[] readableName() {
    StringBuffer buffer = new StringBuffer(this.parameters.length + 20);
    if (isConstructor())
      buffer.append(this.declaringClass.sourceName());
    else
      buffer.append(this.selector);
    buffer.append('(');
    if (this.parameters != Binding.NO_PARAMETERS) {
      int i = 0; for (int length = this.parameters.length; i < length; i++) {
        if (i > 0)
          buffer.append(", ");
        buffer.append(this.parameters[i].sourceName());
      }
    }
    buffer.append(')');
    return buffer.toString().toCharArray();
  }
  public void setAnnotations(AnnotationBinding[] annotations) {
    this.declaringClass.storeAnnotations(this, annotations);
  }
  public void setAnnotations(AnnotationBinding[] annotations, AnnotationBinding[][] parameterAnnotations, Object defaultValue) {
    this.declaringClass.storeAnnotationHolder(this, AnnotationHolder.storeAnnotations(annotations, parameterAnnotations, defaultValue));
  }
  public void setDefaultValue(Object defaultValue) {
    MethodBinding originalMethod = original();
    originalMethod.tagBits |= 2251799813685248L;

    AnnotationHolder holder = this.declaringClass.retrieveAnnotationHolder(this, false);
    if (holder == null)
      setAnnotations(null, null, defaultValue);
    else
      setAnnotations(holder.getAnnotations(), holder.getParameterAnnotations(), defaultValue); 
  }

  public void setParameterAnnotations(AnnotationBinding[][] parameterAnnotations) { AnnotationHolder holder = this.declaringClass.retrieveAnnotationHolder(this, false);
    if (holder == null)
      setAnnotations(null, parameterAnnotations, null);
    else
      setAnnotations(holder.getAnnotations(), parameterAnnotations, holder.getDefaultValue());
  }

  public char[] shortReadableName()
  {
    StringBuffer buffer = new StringBuffer(this.parameters.length + 20);
    if (isConstructor())
      buffer.append(this.declaringClass.shortReadableName());
    else
      buffer.append(this.selector);
    buffer.append('(');
    if (this.parameters != Binding.NO_PARAMETERS) {
      int i = 0; for (int length = this.parameters.length; i < length; i++) {
        if (i > 0)
          buffer.append(", ");
        buffer.append(this.parameters[i].shortReadableName());
      }
    }
    buffer.append(')');
    int nameLength = buffer.length();
    char[] shortReadableName = new char[nameLength];
    buffer.getChars(0, nameLength, shortReadableName, 0);
    return shortReadableName;
  }

  protected final void setSelector(char[] selector) {
    this.selector = selector;
    this.signature = null;
  }

  public final char[] signature()
  {
    if (this.signature != null) {
      return this.signature;
    }
    StringBuffer buffer = new StringBuffer(this.parameters.length + 20);
    buffer.append('(');

    TypeBinding[] targetParameters = this.parameters;
    boolean isConstructor = isConstructor();
    if ((isConstructor) && (this.declaringClass.isEnum())) {
      buffer.append(ConstantPool.JavaLangStringSignature);
      buffer.append(TypeBinding.INT.signature());
    }
    boolean needSynthetics = (isConstructor) && (this.declaringClass.isNestedType());
    if (needSynthetics)
    {
      ReferenceBinding[] syntheticArgumentTypes = this.declaringClass.syntheticEnclosingInstanceTypes();
      if (syntheticArgumentTypes != null) {
        int i = 0; for (int count = syntheticArgumentTypes.length; i < count; i++) {
          buffer.append(syntheticArgumentTypes[i].signature());
        }
      }

      if ((this instanceof SyntheticMethodBinding)) {
        targetParameters = ((SyntheticMethodBinding)this).targetMethod.parameters;
      }
    }

    if (targetParameters != Binding.NO_PARAMETERS) {
      for (int i = 0; i < targetParameters.length; i++) {
        buffer.append(targetParameters[i].signature());
      }
    }
    if (needSynthetics) {
      SyntheticArgumentBinding[] syntheticOuterArguments = this.declaringClass.syntheticOuterLocalVariables();
      int count = syntheticOuterArguments == null ? 0 : syntheticOuterArguments.length;
      for (int i = 0; i < count; i++) {
        buffer.append(syntheticOuterArguments[i].type.signature());
      }

      int i = targetParameters.length; for (int extraLength = this.parameters.length; i < extraLength; i++) {
        buffer.append(this.parameters[i].signature());
      }
    }
    buffer.append(')');
    if (this.returnType != null)
      buffer.append(this.returnType.signature());
    int nameLength = buffer.length();
    this.signature = new char[nameLength];
    buffer.getChars(0, nameLength, this.signature, 0);

    return this.signature;
  }

  public final char[] signature(ClassFile classFile)
  {
    if (this.signature != null) {
      if ((this.tagBits & 0x200) != 0L)
      {
        boolean isConstructor = isConstructor();
        TypeBinding[] targetParameters = this.parameters;
        boolean needSynthetics = (isConstructor) && (this.declaringClass.isNestedType());
        if (needSynthetics)
        {
          ReferenceBinding[] syntheticArgumentTypes = this.declaringClass.syntheticEnclosingInstanceTypes();
          if (syntheticArgumentTypes != null) {
            int i = 0; for (int count = syntheticArgumentTypes.length; i < count; i++) {
              ReferenceBinding syntheticArgumentType = syntheticArgumentTypes[i];
              if (syntheticArgumentType.isNestedType()) {
                classFile.recordInnerClasses(syntheticArgumentType);
              }
            }
          }
          if ((this instanceof SyntheticMethodBinding)) {
            targetParameters = ((SyntheticMethodBinding)this).targetMethod.parameters;
          }
        }

        if (targetParameters != Binding.NO_PARAMETERS) {
          for (int i = 0; i < targetParameters.length; i++) {
            TypeBinding targetParameter = targetParameters[i];
            TypeBinding leafTargetParameterType = targetParameter.leafComponentType();
            if (leafTargetParameterType.isNestedType()) {
              classFile.recordInnerClasses(leafTargetParameterType);
            }
          }
        }
        if (needSynthetics)
        {
          int i = targetParameters.length; for (int extraLength = this.parameters.length; i < extraLength; i++) {
            TypeBinding parameter = this.parameters[i];
            TypeBinding leafParameterType = parameter.leafComponentType();
            if (leafParameterType.isNestedType()) {
              classFile.recordInnerClasses(leafParameterType);
            }
          }
        }
        if (this.returnType != null) {
          TypeBinding ret = this.returnType.leafComponentType();
          if (ret.isNestedType()) {
            classFile.recordInnerClasses(ret);
          }
        }
      }
      return this.signature;
    }

    StringBuffer buffer = new StringBuffer(this.parameters.length + 20);
    buffer.append('(');

    TypeBinding[] targetParameters = this.parameters;
    boolean isConstructor = isConstructor();
    if ((isConstructor) && (this.declaringClass.isEnum())) {
      buffer.append(ConstantPool.JavaLangStringSignature);
      buffer.append(TypeBinding.INT.signature());
    }
    boolean needSynthetics = (isConstructor) && (this.declaringClass.isNestedType());
    if (needSynthetics)
    {
      ReferenceBinding[] syntheticArgumentTypes = this.declaringClass.syntheticEnclosingInstanceTypes();
      if (syntheticArgumentTypes != null) {
        int i = 0; for (int count = syntheticArgumentTypes.length; i < count; i++) {
          ReferenceBinding syntheticArgumentType = syntheticArgumentTypes[i];
          if (syntheticArgumentType.isNestedType()) {
            this.tagBits |= 512L;
            classFile.recordInnerClasses(syntheticArgumentType);
          }
          buffer.append(syntheticArgumentType.signature());
        }
      }

      if ((this instanceof SyntheticMethodBinding)) {
        targetParameters = ((SyntheticMethodBinding)this).targetMethod.parameters;
      }
    }

    if (targetParameters != Binding.NO_PARAMETERS) {
      for (int i = 0; i < targetParameters.length; i++) {
        TypeBinding targetParameter = targetParameters[i];
        TypeBinding leafTargetParameterType = targetParameter.leafComponentType();
        if (leafTargetParameterType.isNestedType()) {
          this.tagBits |= 512L;
          classFile.recordInnerClasses(leafTargetParameterType);
        }
        buffer.append(targetParameter.signature());
      }
    }
    if (needSynthetics) {
      SyntheticArgumentBinding[] syntheticOuterArguments = this.declaringClass.syntheticOuterLocalVariables();
      int count = syntheticOuterArguments == null ? 0 : syntheticOuterArguments.length;
      for (int i = 0; i < count; i++) {
        buffer.append(syntheticOuterArguments[i].type.signature());
      }

      int i = targetParameters.length; for (int extraLength = this.parameters.length; i < extraLength; i++) {
        TypeBinding parameter = this.parameters[i];
        TypeBinding leafParameterType = parameter.leafComponentType();
        if (leafParameterType.isNestedType()) {
          this.tagBits |= 512L;
          classFile.recordInnerClasses(leafParameterType);
        }
        buffer.append(parameter.signature());
      }
    }
    buffer.append(')');
    if (this.returnType != null) {
      TypeBinding ret = this.returnType.leafComponentType();
      if (ret.isNestedType()) {
        this.tagBits |= 512L;
        classFile.recordInnerClasses(ret);
      }
      buffer.append(this.returnType.signature());
    }
    int nameLength = buffer.length();
    this.signature = new char[nameLength];
    buffer.getChars(0, nameLength, this.signature, 0);

    return this.signature;
  }
  public final int sourceEnd() {
    AbstractMethodDeclaration method = sourceMethod();
    if (method == null) {
      if ((this.declaringClass instanceof SourceTypeBinding))
        return ((SourceTypeBinding)this.declaringClass).sourceEnd();
      return 0;
    }
    return method.sourceEnd;
  }

  public AbstractMethodDeclaration sourceMethod() {
    try {
      sourceType = (SourceTypeBinding)this.declaringClass;
    }
    catch (ClassCastException localClassCastException)
    {
      SourceTypeBinding sourceType;
      return null;
    }
    SourceTypeBinding sourceType;
    AbstractMethodDeclaration[] methods = sourceType.scope.referenceContext.methods;
    int i = methods.length;
    do { if (this == methods[i].binding)
        return methods[i];
      i--; } while (i >= 0);

    return null;
  }
  public final int sourceStart() {
    AbstractMethodDeclaration method = sourceMethod();
    if (method == null) {
      if ((this.declaringClass instanceof SourceTypeBinding))
        return ((SourceTypeBinding)this.declaringClass).sourceStart();
      return 0;
    }
    return method.sourceStart;
  }

  public String toString() {
    String s = this.returnType != null ? this.returnType.debugName() : "NULL TYPE";
    s = s + " ";
    s = s + (this.selector != null ? new String(this.selector) : "UNNAMED METHOD");

    s = s + "(";
    if (this.parameters != null) {
      if (this.parameters != Binding.NO_PARAMETERS) {
        int i = 0; for (int length = this.parameters.length; i < length; i++) {
          if (i > 0)
            s = s + ", ";
          s = s + (this.parameters[i] != null ? this.parameters[i].debugName() : "NULL TYPE");
        }
      }
    }
    else s = s + "NULL PARAMETERS";

    s = s + ") ";

    if (this.thrownExceptions != null) {
      if (this.thrownExceptions != Binding.NO_EXCEPTIONS) {
        s = s + "throws ";
        int i = 0; for (int length = this.thrownExceptions.length; i < length; i++) {
          if (i > 0)
            s = s + ", ";
          s = s + (this.thrownExceptions[i] != null ? this.thrownExceptions[i].debugName() : "NULL TYPE");
        }
      }
    }
    else s = s + "NULL THROWN EXCEPTIONS";

    return s;
  }

  public MethodBinding tiebreakMethod()
  {
    return this;
  }
  public TypeVariableBinding[] typeVariables() {
    return this.typeVariables;
  }
}