package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.internal.compiler.ast.MessageSend;

public class ParameterizedGenericMethodBinding extends ParameterizedMethodBinding
  implements Substitution
{
  public TypeBinding[] typeArguments;
  private LookupEnvironment environment;
  public boolean inferredReturnType;
  public boolean wasInferred;
  public boolean isRaw;
  private MethodBinding tiebreakMethod;
  public boolean isUnchecked;

  public static MethodBinding computeCompatibleMethod(MethodBinding originalMethod, TypeBinding[] arguments, Scope scope, InvocationSite invocationSite)
  {
    TypeVariableBinding[] typeVariables = originalMethod.typeVariables;
    TypeBinding[] substitutes = invocationSite.genericTypeArguments();
    ParameterizedGenericMethodBinding methodSubstitute;
    ParameterizedGenericMethodBinding methodSubstitute;
    if (substitutes != null)
    {
      if (substitutes.length != typeVariables.length)
      {
        return new ProblemMethodBinding(originalMethod, originalMethod.selector, substitutes, 11);
      }
      methodSubstitute = scope.environment().createParameterizedGenericMethod(originalMethod, substitutes);
    }
    else
    {
      TypeBinding[] parameters = originalMethod.parameters;
      InferenceContext inferenceContext = new InferenceContext(originalMethod);
      methodSubstitute = inferFromArgumentTypes(scope, originalMethod, arguments, parameters, inferenceContext);
      if (methodSubstitute == null) {
        return null;
      }

      if (inferenceContext.hasUnresolvedTypeArgument()) {
        if (methodSubstitute.returnType != TypeBinding.VOID) {
          TypeBinding expectedType = null;

          if ((invocationSite instanceof MessageSend)) {
            MessageSend message = (MessageSend)invocationSite;
            expectedType = message.expectedType;
          }
          if (expectedType != null)
          {
            inferenceContext.hasExplicitExpectedType = true;
          }
          else expectedType = scope.getJavaLangObject();

          inferenceContext.expectedType = expectedType;
        }
        methodSubstitute = methodSubstitute.inferFromExpectedType(scope, inferenceContext);
        if (methodSubstitute == null) {
          return null;
        }
      }
    }

    if (!methodSubstitute.isRaw) {
      int i = 0; for (int length = typeVariables.length; i < length; i++) {
        TypeVariableBinding typeVariable = typeVariables[i];
        TypeBinding substitute = methodSubstitute.typeArguments[i];
        switch (typeVariable.boundCheck(methodSubstitute, substitute))
        {
        case 2:
          int argLength = arguments.length;
          TypeBinding[] augmentedArguments = new TypeBinding[argLength + 2];
          System.arraycopy(arguments, 0, augmentedArguments, 0, argLength);
          augmentedArguments[argLength] = substitute;
          augmentedArguments[(argLength + 1)] = typeVariable;
          return new ProblemMethodBinding(methodSubstitute, originalMethod.selector, augmentedArguments, 10);
        case 1:
          methodSubstitute.isUnchecked = true;
        }
      }
    }

    return methodSubstitute;
  }

  private static ParameterizedGenericMethodBinding inferFromArgumentTypes(Scope scope, MethodBinding originalMethod, TypeBinding[] arguments, TypeBinding[] parameters, InferenceContext inferenceContext)
  {
    if (originalMethod.isVarargs()) {
      int paramLength = parameters.length;
      int minArgLength = paramLength - 1;
      int argLength = arguments.length;

      for (int i = 0; i < minArgLength; i++) {
        parameters[i].collectSubstitutes(scope, arguments[i], inferenceContext, 1);
        if (inferenceContext.status == 1) return null;
      }

      if (minArgLength < argLength) {
        TypeBinding varargType = parameters[minArgLength];
        TypeBinding lastArgument = arguments[minArgLength];

        if (paramLength == argLength) {
          if (lastArgument == TypeBinding.NULL) break label158;
          switch (lastArgument.dimensions()) {
          case 0:
            break;
          case 1:
            if (lastArgument.leafComponentType().isBaseType()) break; break;
          default:
            break;
          }
        }

        varargType = ((ArrayBinding)varargType).elementsType();

        label158: for (int i = minArgLength; i < argLength; i++) {
          varargType.collectSubstitutes(scope, arguments[i], inferenceContext, 1);
          if (inferenceContext.status == 1) return null; 
        }
      }
    }
    else { int paramLength = parameters.length;
      for (int i = 0; i < paramLength; i++) {
        parameters[i].collectSubstitutes(scope, arguments[i], inferenceContext, 1);
        if (inferenceContext.status == 1) return null;
      }
    }
    if (inferenceContext.status == 2)
    {
      return scope.environment().createParameterizedGenericMethod(originalMethod, null);
    }
    TypeVariableBinding[] originalVariables = originalMethod.typeVariables;
    if (!resolveSubstituteConstraints(scope, originalVariables, inferenceContext, false)) {
      return null;
    }

    TypeBinding[] inferredSustitutes = inferenceContext.substitutes;
    TypeBinding[] actualSubstitutes = inferredSustitutes;
    int i = 0; for (int varLength = originalVariables.length; i < varLength; i++) {
      if (inferredSustitutes[i] == null) {
        if (actualSubstitutes == inferredSustitutes) {
          System.arraycopy(inferredSustitutes, 0, actualSubstitutes = new TypeBinding[varLength], 0, i);
        }
        actualSubstitutes[i] = originalVariables[i];
      } else if (actualSubstitutes != inferredSustitutes) {
        actualSubstitutes[i] = inferredSustitutes[i];
      }
    }
    ParameterizedGenericMethodBinding paramMethod = scope.environment().createParameterizedGenericMethod(originalMethod, actualSubstitutes);
    return paramMethod;
  }

  private static boolean resolveSubstituteConstraints(Scope scope, TypeVariableBinding[] typeVariables, InferenceContext inferenceContext, boolean considerEXTENDSConstraints) {
    TypeBinding[] substitutes = inferenceContext.substitutes;
    int varLength = typeVariables.length;

    for (int i = 0; i < varLength; i++) {
      TypeVariableBinding current = typeVariables[i];
      TypeBinding substitute = substitutes[i];
      if (substitute == null) {
        TypeBinding[] equalSubstitutes = inferenceContext.getSubstitutes(current, 0);
        if (equalSubstitutes != null)
        {
          int j = 0; for (int equalLength = equalSubstitutes.length; j < equalLength; j++) {
            TypeBinding equalSubstitute = equalSubstitutes[j];
            if (equalSubstitute != null) {
              if (equalSubstitute == current)
              {
                for (int k = j + 1; k < equalLength; k++) {
                  equalSubstitute = equalSubstitutes[k];
                  if ((equalSubstitute != current) && (equalSubstitute != null)) {
                    substitutes[i] = equalSubstitute;
                    break;
                  }
                }
                substitutes[i] = current;
                break;
              }

              substitutes[i] = equalSubstitute;
              break;
            }
          }
        }
      }
    }
    if (inferenceContext.hasUnresolvedTypeArgument())
    {
      for (int i = 0; i < varLength; i++) {
        TypeVariableBinding current = typeVariables[i];
        TypeBinding substitute = substitutes[i];
        if (substitute == null) {
          TypeBinding[] bounds = inferenceContext.getSubstitutes(current, 2);
          if (bounds != null) {
            TypeBinding mostSpecificSubstitute = scope.lowerUpperBound(bounds);
            if (mostSpecificSubstitute == null) {
              return false;
            }
            if (mostSpecificSubstitute != TypeBinding.VOID)
              substitutes[i] = mostSpecificSubstitute; 
          }
        }
      }
    }
    if ((considerEXTENDSConstraints) && (inferenceContext.hasUnresolvedTypeArgument()))
    {
      for (int i = 0; i < varLength; i++) {
        TypeVariableBinding current = typeVariables[i];
        TypeBinding substitute = substitutes[i];
        if (substitute == null) {
          TypeBinding[] bounds = inferenceContext.getSubstitutes(current, 1);
          if (bounds != null) {
            TypeBinding[] glb = Scope.greaterLowerBound(bounds);
            TypeBinding mostSpecificSubstitute = null;
            if (glb != null) mostSpecificSubstitute = glb[0];

            if (mostSpecificSubstitute != null)
              substitutes[i] = mostSpecificSubstitute; 
          }
        }
      }
    }
    return true;
  }

  public ParameterizedGenericMethodBinding(MethodBinding originalMethod, RawTypeBinding rawType, LookupEnvironment environment)
  {
    TypeVariableBinding[] originalVariables = originalMethod.typeVariables;
    int length = originalVariables.length;
    TypeBinding[] rawArguments = new TypeBinding[length];
    for (int i = 0; i < length; i++) {
      rawArguments[i] = environment.convertToRawType(originalVariables[i].erasure());
    }
    this.isRaw = true;
    this.tagBits = originalMethod.tagBits;
    this.environment = environment;
    this.modifiers = originalMethod.modifiers;
    this.selector = originalMethod.selector;
    this.declaringClass = (rawType == null ? originalMethod.declaringClass : rawType);
    this.typeVariables = Binding.NO_TYPE_VARIABLES;
    this.typeArguments = rawArguments;
    this.originalMethod = originalMethod;
    boolean ignoreRawTypeSubstitution = (rawType == null) || (originalMethod.isStatic());
    this.parameters = 
      Scope.substitute(this, ignoreRawTypeSubstitution ? 
      originalMethod.parameters : 
      Scope.substitute(rawType, originalMethod.parameters));
    this.thrownExceptions = 
      Scope.substitute(this, ignoreRawTypeSubstitution ? 
      originalMethod.thrownExceptions : 
      Scope.substitute(rawType, originalMethod.thrownExceptions));
    this.returnType = 
      Scope.substitute(this, ignoreRawTypeSubstitution ? 
      originalMethod.returnType : 
      Scope.substitute(rawType, originalMethod.returnType));
    this.wasInferred = false;
  }

  public ParameterizedGenericMethodBinding(MethodBinding originalMethod, TypeBinding[] typeArguments, LookupEnvironment environment)
  {
    this.environment = environment;
    this.modifiers = originalMethod.modifiers;
    this.selector = originalMethod.selector;
    this.declaringClass = originalMethod.declaringClass;
    this.typeVariables = Binding.NO_TYPE_VARIABLES;
    this.typeArguments = typeArguments;
    this.isRaw = false;
    this.tagBits = originalMethod.tagBits;
    this.originalMethod = originalMethod;
    this.parameters = Scope.substitute(this, originalMethod.parameters);
    this.thrownExceptions = Scope.substitute(this, originalMethod.thrownExceptions);
    this.returnType = Scope.substitute(this, originalMethod.returnType);
    this.wasInferred = true;
  }

  public char[] computeUniqueKey(boolean isLeaf)
  {
    StringBuffer buffer = new StringBuffer();
    buffer.append(this.originalMethod.computeUniqueKey(false));
    buffer.append('%');
    buffer.append('<');
    if (!this.isRaw) {
      int length = this.typeArguments.length;
      for (int i = 0; i < length; i++) {
        TypeBinding typeArgument = this.typeArguments[i];
        buffer.append(typeArgument.computeUniqueKey(false));
      }
    }
    buffer.append('>');
    int resultLength = buffer.length();
    char[] result = new char[resultLength];
    buffer.getChars(0, resultLength, result, 0);
    return result;
  }

  public LookupEnvironment environment()
  {
    return this.environment;
  }

  public boolean hasSubstitutedParameters()
  {
    if (this.wasInferred)
      return this.originalMethod.hasSubstitutedParameters();
    return super.hasSubstitutedParameters();
  }

  public boolean hasSubstitutedReturnType()
  {
    if (this.inferredReturnType)
      return this.originalMethod.hasSubstitutedReturnType();
    return super.hasSubstitutedReturnType();
  }

  private ParameterizedGenericMethodBinding inferFromExpectedType(Scope scope, InferenceContext inferenceContext)
  {
    TypeVariableBinding[] originalVariables = this.originalMethod.typeVariables;
    int varLength = originalVariables.length;

    if (inferenceContext.expectedType != null) {
      this.returnType.collectSubstitutes(scope, inferenceContext.expectedType, inferenceContext, 2);
      if (inferenceContext.status == 1) return null;
    }

    for (int i = 0; i < varLength; i++) {
      TypeVariableBinding originalVariable = originalVariables[i];
      TypeBinding argument = this.typeArguments[i];
      boolean argAlreadyInferred = argument != originalVariable;
      if (originalVariable.firstBound == originalVariable.superclass) {
        TypeBinding substitutedBound = Scope.substitute(this, originalVariable.superclass);
        argument.collectSubstitutes(scope, substitutedBound, inferenceContext, 2);
        if (inferenceContext.status == 1) return null;

        if (argAlreadyInferred) {
          substitutedBound.collectSubstitutes(scope, argument, inferenceContext, 1);
          if (inferenceContext.status == 1) return null;
        }
      }
      int j = 0; for (int max = originalVariable.superInterfaces.length; j < max; j++) {
        TypeBinding substitutedBound = Scope.substitute(this, originalVariable.superInterfaces[j]);
        argument.collectSubstitutes(scope, substitutedBound, inferenceContext, 2);
        if (inferenceContext.status == 1) return null;

        if (argAlreadyInferred) {
          substitutedBound.collectSubstitutes(scope, argument, inferenceContext, 1);
          if (inferenceContext.status == 1) return null;
        }
      }
    }
    if (inferenceContext.status == 2)
    {
      this.isRaw = true;
      this.isUnchecked = false;
      for (int i = 0; i < varLength; i++)
        this.typeArguments[i] = originalVariables[i].upperBound();
    }
    else
    {
      if (!resolveSubstituteConstraints(scope, originalVariables, inferenceContext, true)) {
        return null;
      }
      for (int i = 0; i < varLength; i++) {
        TypeBinding substitute = inferenceContext.substitutes[i];
        if (substitute != null) {
          this.typeArguments[i] = inferenceContext.substitutes[i];
        }
        else {
          this.typeArguments[i] = originalVariables[i].upperBound();
        }
      }

    }

    this.typeArguments = Scope.substitute(this, this.typeArguments);

    TypeBinding oldReturnType = this.returnType;
    this.returnType = Scope.substitute(this, this.returnType);
    this.inferredReturnType = ((inferenceContext.hasExplicitExpectedType) && (this.returnType != oldReturnType));
    this.parameters = Scope.substitute(this, this.parameters);
    this.thrownExceptions = Scope.substitute(this, this.thrownExceptions);
    return this;
  }

  public boolean isRawSubstitution()
  {
    return this.isRaw;
  }

  public TypeBinding substitute(TypeVariableBinding originalVariable)
  {
    TypeVariableBinding[] variables = this.originalMethod.typeVariables;
    int length = variables.length;

    if ((originalVariable.rank < length) && (variables[originalVariable.rank] == originalVariable)) {
      return this.typeArguments[originalVariable.rank];
    }
    if ((!isStatic()) && ((this.declaringClass instanceof Substitution))) {
      return ((Substitution)this.declaringClass).substitute(originalVariable);
    }
    return originalVariable;
  }

  public MethodBinding tiebreakMethod()
  {
    if (this.tiebreakMethod == null) {
      TypeVariableBinding[] originalVariables = this.originalMethod.typeVariables;
      int length = originalVariables.length;
      TypeBinding[] newArguments = new TypeBinding[length];
      for (int i = 0; i < length; i++)
        newArguments[i] = this.environment.convertToRawType(originalVariables[i].upperBound());
      this.tiebreakMethod = this.environment.createParameterizedGenericMethod(this.originalMethod, newArguments);
    }
    return this.tiebreakMethod;
  }
}