package org.eclipse.jdt.internal.compiler.lookup;

import java.util.Arrays;
import java.util.Comparator;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.env.IDependent;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.SimpleLookupTable;

public abstract class ReferenceBinding extends TypeBinding
  implements IDependent
{
  public char[][] compoundName;
  public char[] sourceName;
  public int modifiers;
  public PackageBinding fPackage;
  char[] fileName;
  char[] constantPoolName;
  char[] signature;
  private SimpleLookupTable compatibleCache;
  public static ReferenceBinding LUB_GENERIC = new ReferenceBinding() { } ;

  private static final Comparator FIELD_COMPARATOR = new Comparator() {
    public int compare(Object o1, Object o2) {
      char[] n1 = ((FieldBinding)o1).name;
      char[] n2 = ((FieldBinding)o2).name;
      return ReferenceBinding.compare(n1, n2, n1.length, n2.length);
    }
  };

  private static final Comparator METHOD_COMPARATOR = new Comparator() {
    public int compare(Object o1, Object o2) {
      MethodBinding m1 = (MethodBinding)o1;
      MethodBinding m2 = (MethodBinding)o2;
      char[] s1 = m1.selector;
      char[] s2 = m2.selector;
      int c = ReferenceBinding.compare(s1, s2, s1.length, s2.length);
      return c == 0 ? m1.parameters.length - m2.parameters.length : c;
    }
  };

  public static FieldBinding binarySearch(char[] name, FieldBinding[] sortedFields)
  {
    if (sortedFields == null)
      return null;
    int max = sortedFields.length;
    if (max == 0)
      return null;
    int left = 0; int right = max - 1; int nameLength = name.length;
    int mid = 0;

    while (left <= right) {
      mid = left + (right - left) / 2;
      char[] midName;
      int compare = compare(name, midName = sortedFields[mid].name, nameLength, midName.length);
      if (compare < 0)
        right = mid - 1;
      else if (compare > 0)
        left = mid + 1;
      else {
        return sortedFields[mid];
      }
    }
    return null;
  }

  public static long binarySearch(char[] selector, MethodBinding[] sortedMethods)
  {
    if (sortedMethods == null)
      return -1L;
    int max = sortedMethods.length;
    if (max == 0)
      return -1L;
    int left = 0; int right = max - 1; int selectorLength = selector.length;
    int mid = 0;

    while (left <= right) {
      mid = left + (right - left) / 2;
      char[] midSelector;
      int compare = compare(selector, midSelector = sortedMethods[mid].selector, selectorLength, midSelector.length);
      if (compare < 0) {
        right = mid - 1;
      } else if (compare > 0) {
        left = mid + 1;
      } else {
        int start = mid; int end = mid;
        do {
          start--; if (start <= left) break;  } while (CharOperation.equals(sortedMethods[(start - 1)].selector, selector));

        while ((end < right) && (CharOperation.equals(sortedMethods[(end + 1)].selector, selector))) end++;
        return start + (end << 32);
      }
    }
    return -1L;
  }

  static int compare(char[] str1, char[] str2, int len1, int len2)
  {
    int n = Math.min(len1, len2);
    int i = 0;
    while (n-- != 0) {
      char c1 = str1[i];
      char c2 = str2[(i++)];
      if (c1 != c2) {
        return c1 - c2;
      }
    }
    return len1 - len2;
  }

  public static void sortFields(FieldBinding[] sortedFields, int left, int right)
  {
    Arrays.sort(sortedFields, left, right, FIELD_COMPARATOR);
  }

  public static void sortMethods(MethodBinding[] sortedMethods, int left, int right)
  {
    Arrays.sort(sortedMethods, left, right, METHOD_COMPARATOR);
  }

  public FieldBinding[] availableFields()
  {
    return fields();
  }

  public MethodBinding[] availableMethods()
  {
    return methods();
  }

  public boolean canBeInstantiated()
  {
    return (this.modifiers & 0x6600) == 0;
  }

  public final boolean canBeSeenBy(PackageBinding invocationPackage)
  {
    if (isPublic()) return true;
    if (isPrivate()) return false;

    return invocationPackage == this.fPackage;
  }

  public final boolean canBeSeenBy(ReferenceBinding receiverType, ReferenceBinding invocationType)
  {
    if (isPublic()) return true;

    if ((invocationType == this) && (invocationType == receiverType)) return true;

    if (isProtected())
    {
      if (invocationType == this) return true;
      if (invocationType.fPackage == this.fPackage) return true;

      ReferenceBinding currentType = invocationType;
      ReferenceBinding declaringClass = enclosingType();
      if (declaringClass == invocationType) return true;
      if (declaringClass == null) return false;
      do
      {
        if (currentType.findSuperTypeWithSameErasure(declaringClass) != null) return true;

        currentType = currentType.enclosingType();
      }while (currentType != null);
      return false;
    }

    if (isPrivate())
    {
      if ((receiverType != this) && (receiverType != enclosingType()))
      {
        if (receiverType.isTypeVariable()) { TypeVariableBinding typeVariable = (TypeVariableBinding)receiverType;
          if ((typeVariable.isErasureBoundTo(erasure())) || (typeVariable.isErasureBoundTo(enclosingType().erasure())));
        }
        else {
          return false;
        }
      }

      if (invocationType != this) {
        ReferenceBinding outerInvocationType = invocationType;
        ReferenceBinding temp = outerInvocationType.enclosingType();
        while (temp != null) {
          outerInvocationType = temp;
          temp = temp.enclosingType();
        }

        ReferenceBinding outerDeclaringClass = (ReferenceBinding)erasure();
        temp = outerDeclaringClass.enclosingType();
        while (temp != null) {
          outerDeclaringClass = temp;
          temp = temp.enclosingType();
        }
        if (outerInvocationType != outerDeclaringClass) return false;
      }
      return true;
    }

    if (invocationType.fPackage != this.fPackage) return false;

    ReferenceBinding currentType = receiverType;
    ReferenceBinding declaringClass = enclosingType() == null ? this : enclosingType();
    do {
      if (declaringClass == currentType) return true;
      PackageBinding currentPackage = currentType.fPackage;

      if ((currentPackage != null) && (currentPackage != this.fPackage)) return false; 
    }
    while ((currentType = currentType.superclass()) != null);
    return false;
  }

  public final boolean canBeSeenBy(Scope scope)
  {
    if (isPublic()) return true;

    SourceTypeBinding invocationType = scope.enclosingSourceType();
    if (invocationType == this) return true;

    if (invocationType == null) {
      return (!isPrivate()) && (scope.getCurrentPackage() == this.fPackage);
    }
    if (isProtected())
    {
      if (invocationType.fPackage == this.fPackage) return true;

      ReferenceBinding currentType = invocationType;
      ReferenceBinding declaringClass = enclosingType();
      if (declaringClass == null) return false;
      do
      {
        if (declaringClass == invocationType) return true;
        if (declaringClass.isSuperclassOf(currentType)) return true;

        currentType = currentType.enclosingType();
      }while (currentType != null);
      return false;
    }
    if (isPrivate())
    {
      ReferenceBinding outerInvocationType = invocationType;
      ReferenceBinding temp = outerInvocationType.enclosingType();
      while (temp != null) {
        outerInvocationType = temp;
        temp = temp.enclosingType();
      }

      ReferenceBinding outerDeclaringClass = (ReferenceBinding)erasure();
      temp = outerDeclaringClass.enclosingType();
      while (temp != null) {
        outerDeclaringClass = temp;
        temp = temp.enclosingType();
      }
      return outerInvocationType == outerDeclaringClass;
    }

    return invocationType.fPackage == this.fPackage;
  }

  public ReferenceBinding closestMatch()
  {
    return this;
  }

  public char[] computeGenericTypeSignature(TypeVariableBinding[] typeVariables)
  {
    boolean isMemberOfGeneric = (isMemberType()) && ((enclosingType().modifiers & 0x40000000) != 0);
    if ((typeVariables == Binding.NO_TYPE_VARIABLES) && (!isMemberOfGeneric)) {
      return signature();
    }
    StringBuffer sig = new StringBuffer(10);
    if (isMemberOfGeneric) {
      char[] typeSig = enclosingType().genericTypeSignature();
      for (int i = 0; i < typeSig.length - 1; i++) {
        sig.append(typeSig[i]);
      }
      sig.append('.');
      sig.append(this.sourceName);
    } else {
      char[] typeSig = signature();
      for (int i = 0; i < typeSig.length - 1; i++) {
        sig.append(typeSig[i]);
      }
    }
    if (typeVariables == Binding.NO_TYPE_VARIABLES) {
      sig.append(';');
    } else {
      sig.append('<');
      int i = 0; for (int length = typeVariables.length; i < length; i++) {
        sig.append(typeVariables[i].genericTypeSignature());
      }
      sig.append(">;");
    }
    int sigLength = sig.length();
    char[] result = new char[sigLength];
    sig.getChars(0, sigLength, result, 0);
    return result;
  }

  public void computeId() {
    switch (this.compoundName.length)
    {
    case 3:
      if (!CharOperation.equals(TypeConstants.JAVA, this.compoundName[0])) {
        return;
      }

      if (!CharOperation.equals(TypeConstants.LANG, this.compoundName[1])) {
        if (CharOperation.equals(TypeConstants.IO, this.compoundName[1])) {
          if (CharOperation.equals(TypeConstants.JAVA_IO_PRINTSTREAM[2], this.compoundName[2]))
            this.id = 53;
          else if (CharOperation.equals(TypeConstants.JAVA_IO_SERIALIZABLE[2], this.compoundName[2]))
            this.id = 37;
          else if (CharOperation.equals(TypeConstants.JAVA_IO_EXTERNALIZABLE[2], this.compoundName[2]))
            this.id = 56;
          else if (CharOperation.equals(TypeConstants.JAVA_IO_OBJECTSTREAMEXCEPTION[2], this.compoundName[2]))
            this.id = 57;
          else if (CharOperation.equals(TypeConstants.JAVA_IO_IOEXCEPTION[2], this.compoundName[2]))
            this.id = 58;
        } else if ((CharOperation.equals(TypeConstants.UTIL, this.compoundName[1])) && 
          (CharOperation.equals(TypeConstants.JAVA_UTIL_ITERATOR[2], this.compoundName[2]))) {
          this.id = 39;
        }
        return;
      }

      char[] typeName = this.compoundName[2];
      if (typeName.length == 0) return;
      switch (typeName[0]) {
      case 'A':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ASSERTIONERROR[2]))
          this.id = 35;
        return;
      case 'B':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_BOOLEAN[2]))
          this.id = 33;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_BYTE[2]))
          this.id = 26;
        return;
      case 'C':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_CHARACTER[2]))
          this.id = 28;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_CLASS[2]))
          this.id = 16;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_CLASSNOTFOUNDEXCEPTION[2]))
          this.id = 23;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_CLONEABLE[2]))
          this.id = 36;
        return;
      case 'D':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_DOUBLE[2]))
          this.id = 32;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_DEPRECATED[2]))
          this.id = 44;
        return;
      case 'E':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ERROR[2]))
          this.id = 19;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_EXCEPTION[2]))
          this.id = 25;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ENUM[2]))
          this.id = 41;
        return;
      case 'F':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_FLOAT[2]))
          this.id = 31;
        return;
      case 'I':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_INTEGER[2]))
          this.id = 29;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ITERABLE[2]))
          this.id = 38;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ILLEGALARGUMENTEXCEPTION[2]))
          this.id = 42;
        return;
      case 'L':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_LONG[2]))
          this.id = 30;
        return;
      case 'N':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_NOCLASSDEFERROR[2]))
          this.id = 22;
        return;
      case 'O':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_OBJECT[2]))
          this.id = 1;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_OVERRIDE[2]))
          this.id = 47;
        return;
      case 'R':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_RUNTIMEEXCEPTION[2]))
          this.id = 24;
        break;
      case 'S':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_STRING[2]))
          this.id = 11;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_STRINGBUFFER[2]))
          this.id = 17;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_STRINGBUILDER[2]))
          this.id = 40;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_SYSTEM[2]))
          this.id = 18;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_SHORT[2]))
          this.id = 27;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_SUPPRESSWARNINGS[2]))
          this.id = 49;
        return;
      case 'T':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_THROWABLE[2]))
          this.id = 21;
        return;
      case 'V':
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_VOID[2]))
          this.id = 34;
        return;
      case 'G':
      case 'H':
      case 'J':
      case 'K':
      case 'M':
      case 'P':
      case 'Q':
      case 'U': } break;
    case 4:
      if (!CharOperation.equals(TypeConstants.JAVA, this.compoundName[0]))
        return;
      if (!CharOperation.equals(TypeConstants.LANG, this.compoundName[1]))
        return;
      char[] packageName = this.compoundName[2];
      if (packageName.length == 0) return;
      char[] typeName = this.compoundName[3];
      if (typeName.length == 0) return;
      if (CharOperation.equals(packageName, TypeConstants.REFLECT)) {
        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_REFLECT_CONSTRUCTOR[3]))
          this.id = 20;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_REFLECT_FIELD[3]))
          this.id = 54;
        else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_REFLECT_METHOD[3])) {
          this.id = 55;
        }
        return;
      }if (CharOperation.equals(packageName, TypeConstants.ANNOTATION))
        switch (typeName[0]) {
        case 'A':
          if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_ANNOTATION[3]))
            this.id = 43;
          return;
        case 'D':
          if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_DOCUMENTED[3]))
            this.id = 45;
          return;
        case 'E':
          if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_ELEMENTTYPE[3]))
            this.id = 52;
          return;
        case 'I':
          if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_INHERITED[3]))
            this.id = 46;
          return;
        case 'R':
          if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_RETENTION[3]))
            this.id = 48;
          else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_RETENTIONPOLICY[3]))
            this.id = 51;
          return;
        case 'T':
          if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_TARGET[3]))
            this.id = 50;
          return;
        }
      break;
    }
  }

  public char[] computeUniqueKey(boolean isLeaf)
  {
    if (!isLeaf) return signature();
    return genericTypeSignature();
  }

  public char[] constantPoolName()
  {
    if (this.constantPoolName != null) return this.constantPoolName;
    return this.constantPoolName = CharOperation.concatWith(this.compoundName, '/');
  }
  public String debugName() {
    return this.compoundName != null ? new String(readableName()) : "UNNAMED TYPE";
  }
  public final int depth() {
    int depth = 0;
    ReferenceBinding current = this;
    while ((current = current.enclosingType()) != null)
      depth++;
    return depth;
  }
  public boolean detectAnnotationCycle() {
    if ((this.tagBits & 0x0) != 0L) return false;
    if ((this.tagBits & 0x80000000) != 0L) return true;

    this.tagBits |= 2147483648L;
    MethodBinding[] currentMethods = methods();
    boolean inCycle = false;
    int i = 0; for (int l = currentMethods.length; i < l; i++) {
      TypeBinding returnType = currentMethods[i].returnType.leafComponentType();
      if (this == returnType) {
        if ((this instanceof SourceTypeBinding)) {
          MethodDeclaration decl = (MethodDeclaration)currentMethods[i].sourceMethod();
          ((SourceTypeBinding)this).scope.problemReporter().annotationCircularity(this, this, decl != null ? decl.returnType : null);
        }
      } else if ((returnType.isAnnotationType()) && (((ReferenceBinding)returnType).detectAnnotationCycle())) {
        if ((this instanceof SourceTypeBinding)) {
          MethodDeclaration decl = (MethodDeclaration)currentMethods[i].sourceMethod();
          ((SourceTypeBinding)this).scope.problemReporter().annotationCircularity(this, returnType, decl != null ? decl.returnType : null);
        }
        inCycle = true;
      }
    }
    if (inCycle)
      return true;
    this.tagBits |= 4294967296L;
    return false;
  }

  public final ReferenceBinding enclosingTypeAt(int relativeDepth) {
    ReferenceBinding current = this;
    while ((relativeDepth-- > 0) && (current != null))
      current = current.enclosingType();
    return current;
  }

  public int enumConstantCount() {
    int count = 0;
    FieldBinding[] fields = fields();
    int i = 0; for (int length = fields.length; i < length; i++) {
      if ((fields[i].modifiers & 0x4000) != 0) count++;
    }
    return count;
  }

  public int fieldCount() {
    return fields().length;
  }

  public FieldBinding[] fields() {
    return Binding.NO_FIELDS;
  }

  public final int getAccessFlags() {
    return this.modifiers & 0xFFFF;
  }

  public AnnotationBinding[] getAnnotations()
  {
    return retrieveAnnotations(this);
  }

  public long getAnnotationTagBits()
  {
    return this.tagBits;
  }
  public MethodBinding getExactConstructor(TypeBinding[] argumentTypes) {
    return null;
  }
  public MethodBinding getExactMethod(char[] selector, TypeBinding[] argumentTypes, CompilationUnitScope refScope) {
    return null;
  }
  public FieldBinding getField(char[] fieldName, boolean needResolve) {
    return null;
  }

  public char[] getFileName()
  {
    return this.fileName;
  }
  public ReferenceBinding getMemberType(char[] typeName) {
    ReferenceBinding[] memberTypes = memberTypes();
    int i = memberTypes.length;
    do { if (CharOperation.equals(memberTypes[i].sourceName, typeName))
        return memberTypes[i];
      i--; } while (i >= 0);

    return null;
  }

  public MethodBinding[] getMethods(char[] selector) {
    return Binding.NO_METHODS;
  }

  public PackageBinding getPackage() {
    return this.fPackage;
  }

  public TypeVariableBinding getTypeVariable(char[] variableName) {
    TypeVariableBinding[] typeVariables = typeVariables();
    int i = typeVariables.length;
    do { if (CharOperation.equals(typeVariables[i].sourceName, variableName))
        return typeVariables[i];
      i--; } while (i >= 0);

    return null;
  }

  public int hashCode()
  {
    return (this.compoundName == null) || (this.compoundName.length == 0) ? 
      super.hashCode() : 
      CharOperation.hashCode(this.compoundName[(this.compoundName.length - 1)]);
  }

  public boolean hasIncompatibleSuperType(ReferenceBinding otherType)
  {
    if (this == otherType) return false; 
ReferenceBinding[] interfacesToVisit = (ReferenceBinding[])null;
    int nextPosition = 0;
    ReferenceBinding currentType = this;
    TypeBinding match;
    do { match = otherType.findSuperTypeWithSameErasure(currentType);
      if ((match != null) && (!match.isIntersectingWith(currentType))) {
        return true;
      }
      ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
      if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES))
        if (interfacesToVisit == null) {
          interfacesToVisit = itsInterfaces;
          nextPosition = interfacesToVisit.length;
        } else {
          int itsLength = itsInterfaces.length;
          if (nextPosition + itsLength >= interfacesToVisit.length)
            System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
          for (int a = 0; a < itsLength; a++) {
            ReferenceBinding next = itsInterfaces[a];
            int b = 0;
            while (next != interfacesToVisit[b])
            {
              b++; if (b >= nextPosition)
              {
                interfacesToVisit[(nextPosition++)] = next;
              }
            }
          }
        } } while ((currentType = currentType.superclass()) != null);

    for (int i = 0; i < nextPosition; i++) {
      currentType = interfacesToVisit[i];
      if (currentType == otherType) return false;
      match = otherType.findSuperTypeWithSameErasure(currentType);
      if ((match != null) && (!match.isIntersectingWith(currentType))) {
        return true;
      }
      ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
      if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES)) {
        int itsLength = itsInterfaces.length;
        if (nextPosition + itsLength >= interfacesToVisit.length)
          System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
        for (int a = 0; a < itsLength; a++) {
          ReferenceBinding next = itsInterfaces[a];
          int b = 0;
          while (next != interfacesToVisit[b])
          {
            b++; if (b >= nextPosition)
            {
              interfacesToVisit[(nextPosition++)] = next;
            }
          }
        }
      }
    }
    return false;
  }
  public boolean hasMemberTypes() {
    return false;
  }
  public final boolean hasRestrictedAccess() {
    return (this.modifiers & 0x40000) != 0;
  }

  public boolean implementsInterface(ReferenceBinding anInterface, boolean searchHierarchy)
  {
    if (this == anInterface) {
      return true;
    }
    ReferenceBinding[] interfacesToVisit = (ReferenceBinding[])null;
    int nextPosition = 0;
    ReferenceBinding currentType = this;
    do {
      ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
      if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES))
        if (interfacesToVisit == null) {
          interfacesToVisit = itsInterfaces;
          nextPosition = interfacesToVisit.length;
        } else {
          int itsLength = itsInterfaces.length;
          if (nextPosition + itsLength >= interfacesToVisit.length)
            System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
          for (int a = 0; a < itsLength; a++) {
            ReferenceBinding next = itsInterfaces[a];
            int b = 0;
            while (next != interfacesToVisit[b])
            {
              b++; if (b >= nextPosition)
              {
                interfacesToVisit[(nextPosition++)] = next;
              }
            }
          }
        }
    }
    while ((searchHierarchy) && ((currentType = currentType.superclass()) != null));

    for (int i = 0; i < nextPosition; i++) {
      currentType = interfacesToVisit[i];
      if (currentType.isEquivalentTo(anInterface)) {
        return true;
      }
      ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
      if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES)) {
        int itsLength = itsInterfaces.length;
        if (nextPosition + itsLength >= interfacesToVisit.length)
          System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
        for (int a = 0; a < itsLength; a++) {
          ReferenceBinding next = itsInterfaces[a];
          int b = 0;
          while (next != interfacesToVisit[b])
          {
            b++; if (b >= nextPosition)
            {
              interfacesToVisit[(nextPosition++)] = next;
            }
          }
        }
      }
    }
    return false;
  }

  boolean implementsMethod(MethodBinding method)
  {
    char[] selector = method.selector;
    ReferenceBinding type = this;
    while (type != null) {
      MethodBinding[] methods = type.methods();
      long range;
      if ((range = binarySearch(selector, methods)) >= 0L) {
        int start = (int)range; int end = (int)(range >> 32);
        for (int i = start; i <= end; i++) {
          if (methods[i].areParametersEqual(method))
            return true;
        }
      }
      type = type.superclass();
    }
    return false;
  }

  public final boolean isAbstract()
  {
    return (this.modifiers & 0x400) != 0;
  }

  public boolean isAnnotationType() {
    return (this.modifiers & 0x2000) != 0;
  }

  public final boolean isBinaryBinding() {
    return (this.tagBits & 0x40) != 0L;
  }

  public boolean isClass() {
    return (this.modifiers & 0x6200) == 0;
  }

  public boolean isCompatibleWith(TypeBinding otherType)
  {
    if (otherType == this)
      return true;
    if (otherType.id == 1)
      return true;
    Object result;
    if (this.compatibleCache == null) {
      this.compatibleCache = new SimpleLookupTable(3);
      result = null;
    } else {
      Object result = this.compatibleCache.get(otherType);
      if (result != null) {
        return result == Boolean.TRUE;
      }
    }
    this.compatibleCache.put(otherType, Boolean.FALSE);
    if (isCompatibleWith0(otherType)) {
      this.compatibleCache.put(otherType, Boolean.TRUE);
      return true;
    }
    return false;
  }

  private boolean isCompatibleWith0(TypeBinding otherType)
  {
    if (otherType == this)
      return true;
    if (otherType.id == 1) {
      return true;
    }

    if (isEquivalentTo(otherType))
      return true;
    switch (otherType.kind()) {
    case 516:
      return false;
    case 4100:
      if (otherType.isCapture()) {
        CaptureBinding otherCapture = (CaptureBinding)otherType;
        TypeBinding otherLowerBound;
        if ((otherLowerBound = otherCapture.lowerBound) != null) {
          if (otherLowerBound.isArrayType()) return false;
          return isCompatibleWith(otherLowerBound);
        }
      }
    case 4:
    case 260:
    case 1028:
    case 2052:
      switch (kind()) {
      case 260:
      case 1028:
      case 2052:
        if (erasure() == otherType.erasure())
          return false;
        break;
      }
      ReferenceBinding otherReferenceType = (ReferenceBinding)otherType;
      if (otherReferenceType.isInterface())
        return implementsInterface(otherReferenceType, true);
      if (isInterface())
      {
        return false;
      }return otherReferenceType.isSuperclassOf(this);
    }
    return false;
  }

  public final boolean isDefault()
  {
    return (this.modifiers & 0x7) == 0;
  }

  public final boolean isDeprecated()
  {
    return (this.modifiers & 0x100000) != 0;
  }

  public boolean isEnum() {
    return (this.modifiers & 0x4000) != 0;
  }

  public final boolean isFinal()
  {
    return (this.modifiers & 0x10) != 0;
  }

  public boolean isHierarchyBeingConnected()
  {
    return ((this.tagBits & 0x200) == 0L) && ((this.tagBits & 0x100) != 0L);
  }

  public boolean isInterface()
  {
    return (this.modifiers & 0x200) != 0;
  }

  public final boolean isPrivate()
  {
    return (this.modifiers & 0x2) != 0;
  }

  public final boolean isProtected()
  {
    return (this.modifiers & 0x4) != 0;
  }

  public final boolean isPublic()
  {
    return (this.modifiers & 0x1) != 0;
  }

  public final boolean isStatic()
  {
    return ((this.modifiers & 0x208) != 0) || ((this.tagBits & 0x4) == 0L);
  }

  public final boolean isStrictfp()
  {
    return (this.modifiers & 0x800) != 0;
  }

  public boolean isSuperclassOf(ReferenceBinding otherType)
  {
    while ((otherType = otherType.superclass()) != null) {
      if (otherType.isEquivalentTo(this)) return true;
    }
    return false;
  }

  public boolean isThrowable()
  {
    ReferenceBinding current = this;
    do
      switch (current.id) {
      case 19:
      case 21:
      case 24:
      case 25:
        return true;
      case 20:
      case 22:
      case 23: }  while ((current = current.superclass()) != null);
    return false;
  }

  public boolean isUncheckedException(boolean includeSupertype)
  {
    switch (this.id) {
    case 19:
    case 24:
      return true;
    case 21:
    case 25:
      return includeSupertype;
    case 20:
    case 22:
    case 23: } ReferenceBinding current = this;
    while ((current = current.superclass()) != null)
      switch (current.id) {
      case 19:
      case 24:
        return true;
      case 21:
      case 25:
        return false;
      case 20:
      case 22:
      case 23: }  return false;
  }

  public final boolean isUsed()
  {
    return (this.modifiers & 0x8000000) != 0;
  }

  public final boolean isViewedAsDeprecated()
  {
    return ((this.modifiers & 0x300000) != 0) || 
      ((getPackage().tagBits & 0x0) != 0L);
  }
  public ReferenceBinding[] memberTypes() {
    return Binding.NO_MEMBER_TYPES;
  }
  public MethodBinding[] methods() {
    return Binding.NO_METHODS;
  }
  public final ReferenceBinding outermostEnclosingType() { ReferenceBinding current = this;
    ReferenceBinding last;
    do last = current;
    while ((current = current.enclosingType()) != null);
    return last;
  }

  public char[] qualifiedSourceName()
  {
    if (isMemberType())
      return CharOperation.concat(enclosingType().qualifiedSourceName(), sourceName(), '.');
    return sourceName();
  }

  public char[] readableName()
  {
    char[] readableName;
    char[] readableName;
    if (isMemberType())
      readableName = CharOperation.concat(enclosingType().readableName(), this.sourceName, '.');
    else
      readableName = CharOperation.concatWith(this.compoundName, '.');
    TypeVariableBinding[] typeVars;
    if ((typeVars = typeVariables()) != Binding.NO_TYPE_VARIABLES) {
      StringBuffer nameBuffer = new StringBuffer(10);
      nameBuffer.append(readableName).append('<');
      int i = 0; for (int length = typeVars.length; i < length; i++) {
        if (i > 0) nameBuffer.append(',');
        nameBuffer.append(typeVars[i].readableName());
      }
      nameBuffer.append('>');
      int nameLength = nameBuffer.length();
      readableName = new char[nameLength];
      nameBuffer.getChars(0, nameLength, readableName, 0);
    }
    return readableName;
  }

  public AnnotationHolder retrieveAnnotationHolder(Binding binding, boolean forceInitialization) {
    SimpleLookupTable store = storedAnnotations(false);
    return store == null ? null : (AnnotationHolder)store.get(binding);
  }
  AnnotationBinding[] retrieveAnnotations(Binding binding) {
    AnnotationHolder holder = retrieveAnnotationHolder(binding, true);
    return holder == null ? Binding.NO_ANNOTATIONS : holder.getAnnotations();
  }
  public void setAnnotations(AnnotationBinding[] annotations) {
    storeAnnotations(this, annotations);
  }

  public char[] shortReadableName()
  {
    char[] shortReadableName;
    char[] shortReadableName;
    if (isMemberType())
      shortReadableName = CharOperation.concat(enclosingType().shortReadableName(), this.sourceName, '.');
    else
      shortReadableName = this.sourceName;
    TypeVariableBinding[] typeVars;
    if ((typeVars = typeVariables()) != Binding.NO_TYPE_VARIABLES) {
      StringBuffer nameBuffer = new StringBuffer(10);
      nameBuffer.append(shortReadableName).append('<');
      int i = 0; for (int length = typeVars.length; i < length; i++) {
        if (i > 0) nameBuffer.append(',');
        nameBuffer.append(typeVars[i].shortReadableName());
      }
      nameBuffer.append('>');
      int nameLength = nameBuffer.length();
      shortReadableName = new char[nameLength];
      nameBuffer.getChars(0, nameLength, shortReadableName, 0);
    }
    return shortReadableName;
  }
  public char[] signature() {
    if (this.signature != null) {
      return this.signature;
    }
    return this.signature = CharOperation.concat('L', constantPoolName(), ';');
  }
  public char[] sourceName() {
    return this.sourceName;
  }
  void storeAnnotationHolder(Binding binding, AnnotationHolder holder) {
    if (holder == null) {
      SimpleLookupTable store = storedAnnotations(false);
      if (store != null)
        store.removeKey(binding);
    } else {
      SimpleLookupTable store = storedAnnotations(true);
      if (store != null)
        store.put(binding, holder); 
    }
  }

  void storeAnnotations(Binding binding, AnnotationBinding[] annotations) { AnnotationHolder holder = null;
    if ((annotations == null) || (annotations.length == 0)) {
      SimpleLookupTable store = storedAnnotations(false);
      if (store != null)
        holder = (AnnotationHolder)store.get(binding);
      if (holder != null);
    } else {
      SimpleLookupTable store = storedAnnotations(true);
      if (store == null) return;
      holder = (AnnotationHolder)store.get(binding);
      if (holder == null)
        holder = new AnnotationHolder();
    }
    storeAnnotationHolder(binding, holder.setAnnotations(annotations)); }

  SimpleLookupTable storedAnnotations(boolean forceInitialize) {
    return null;
  }
  public ReferenceBinding superclass() {
    return null;
  }
  public ReferenceBinding[] superInterfaces() {
    return Binding.NO_SUPERINTERFACES;
  }
  public ReferenceBinding[] syntheticEnclosingInstanceTypes() {
    if (isStatic()) return null;

    ReferenceBinding enclosingType = enclosingType();
    if (enclosingType == null)
      return null;
    return new ReferenceBinding[] { enclosingType };
  }
  public SyntheticArgumentBinding[] syntheticOuterLocalVariables() {
    return null;
  }

  MethodBinding[] unResolvedMethods() {
    return methods();
  }
}