package org.eclipse.jdt.internal.compiler.lookup;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.CaseStatement;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ImportReference;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeParameter;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.HashtableOfObject;
import org.eclipse.jdt.internal.compiler.util.ObjectVector;
import org.eclipse.jdt.internal.compiler.util.SimpleSet;

public abstract class Scope
  implements TypeConstants, TypeIds
{
  public static final int BLOCK_SCOPE = 1;
  public static final int CLASS_SCOPE = 3;
  public static final int COMPILATION_UNIT_SCOPE = 4;
  public static final int METHOD_SCOPE = 2;
  public static final int NOT_COMPATIBLE = -1;
  public static final int COMPATIBLE = 0;
  public static final int AUTOBOX_COMPATIBLE = 1;
  public static final int VARARGS_COMPATIBLE = 2;
  public static final int EQUAL_OR_MORE_SPECIFIC = -1;
  public static final int NOT_RELATED = 0;
  public static final int MORE_GENERIC = 1;
  public int kind;
  public Scope parent;

  public static int compareTypes(TypeBinding left, TypeBinding right)
  {
    if (left.isCompatibleWith(right))
      return -1;
    if (right.isCompatibleWith(left))
      return 1;
    return 0;
  }

  public static TypeBinding getBaseType(char[] name) {
    int length = name.length;
    if ((length > 2) && (length < 8))
      switch (name[0]) {
      case 'i':
        if ((length == 3) && (name[1] == 'n') && (name[2] == 't'))
          return TypeBinding.INT;
        break;
      case 'v':
        if ((length == 4) && (name[1] == 'o') && (name[2] == 'i') && (name[3] == 'd'))
          return TypeBinding.VOID;
        break;
      case 'b':
        if ((length == 7) && 
          (name[1] == 'o') && 
          (name[2] == 'o') && 
          (name[3] == 'l') && 
          (name[4] == 'e') && 
          (name[5] == 'a') && 
          (name[6] == 'n'))
          return TypeBinding.BOOLEAN;
        if ((length == 4) && (name[1] == 'y') && (name[2] == 't') && (name[3] == 'e'))
          return TypeBinding.BYTE;
        break;
      case 'c':
        if ((length == 4) && (name[1] == 'h') && (name[2] == 'a') && (name[3] == 'r'))
          return TypeBinding.CHAR;
        break;
      case 'd':
        if ((length == 6) && 
          (name[1] == 'o') && 
          (name[2] == 'u') && 
          (name[3] == 'b') && 
          (name[4] == 'l') && 
          (name[5] == 'e'))
          return TypeBinding.DOUBLE;
        break;
      case 'f':
        if ((length == 5) && 
          (name[1] == 'l') && 
          (name[2] == 'o') && 
          (name[3] == 'a') && 
          (name[4] == 't'))
          return TypeBinding.FLOAT;
        break;
      case 'l':
        if ((length == 4) && (name[1] == 'o') && (name[2] == 'n') && (name[3] == 'g'))
          return TypeBinding.LONG;
        break;
      case 's':
        if ((length == 5) && 
          (name[1] == 'h') && 
          (name[2] == 'o') && 
          (name[3] == 'r') && 
          (name[4] == 't'))
          return TypeBinding.SHORT;
        break;
      }
    return null;
  }

  public static ReferenceBinding[] greaterLowerBound(ReferenceBinding[] types)
  {
    if (types == null) return null;
    int length = types.length;
    if (length == 0) return null;
    ReferenceBinding[] result = types;
    int removed = 0;
    for (int i = 0; i < length; i++) {
      ReferenceBinding iType = result[i];
      if (iType != null)
        for (int j = 0; j < length; j++)
          if (i != j) {
            ReferenceBinding jType = result[j];
            if ((jType != null) && 
              (iType.isCompatibleWith(jType))) {
              if (result == types) {
                System.arraycopy(result, 0, result = new ReferenceBinding[length], 0, length);
              }
              result[j] = null;
              removed++;
            }
          }
    }
    if (removed == 0) return result;
    if (length == removed) return null;
    ReferenceBinding[] trimmedResult = new ReferenceBinding[length - removed];
    int i = 0; for (int index = 0; i < length; i++) {
      ReferenceBinding iType = result[i];
      if (iType != null) {
        trimmedResult[(index++)] = iType;
      }
    }
    return trimmedResult;
  }

  public static TypeBinding[] greaterLowerBound(TypeBinding[] types)
  {
    if (types == null) return null;
    int length = types.length;
    if (length == 0) return null;
    TypeBinding[] result = types;
    int removed = 0;
    for (int i = 0; i < length; i++) {
      TypeBinding iType = result[i];
      if (iType != null)
        for (int j = 0; j < length; j++)
          if (i != j) {
            TypeBinding jType = result[j];
            if ((jType != null) && 
              (iType.isCompatibleWith(jType))) {
              if (result == types) {
                System.arraycopy(result, 0, result = new TypeBinding[length], 0, length);
              }
              result[j] = null;
              removed++;
            }
          }
    }
    if (removed == 0) return result;
    if (length == removed) return null;
    TypeBinding[] trimmedResult = new TypeBinding[length - removed];
    int i = 0; for (int index = 0; i < length; i++) {
      TypeBinding iType = result[i];
      if (iType != null) {
        trimmedResult[(index++)] = iType;
      }
    }
    return trimmedResult;
  }

  public static ReferenceBinding[] substitute(Substitution substitution, ReferenceBinding[] originalTypes)
  {
    if (originalTypes == null) return null;
    ReferenceBinding[] substitutedTypes = originalTypes;
    int i = 0; for (int length = originalTypes.length; i < length; i++) {
      ReferenceBinding originalType = originalTypes[i];
      ReferenceBinding substitutedParameter = (ReferenceBinding)substitute(substitution, originalType);
      if (substitutedParameter != originalType) {
        if (substitutedTypes == originalTypes) {
          System.arraycopy(originalTypes, 0, substitutedTypes = new ReferenceBinding[length], 0, i);
        }
        substitutedTypes[i] = substitutedParameter;
      } else if (substitutedTypes != originalTypes) {
        substitutedTypes[i] = originalType;
      }
    }
    return substitutedTypes;
  }

  public static TypeBinding substitute(Substitution substitution, TypeBinding originalType)
  {
    if (originalType == null) return null;
    switch (originalType.kind())
    {
    case 4100:
      return substitution.substitute((TypeVariableBinding)originalType);
    case 260:
      ParameterizedTypeBinding originalParameterizedType = (ParameterizedTypeBinding)originalType;
      ReferenceBinding originalEnclosing = originalType.enclosingType();
      ReferenceBinding substitutedEnclosing = originalEnclosing;
      if (originalEnclosing != null) {
        substitutedEnclosing = (ReferenceBinding)substitute(substitution, originalEnclosing);
      }
      TypeBinding[] originalArguments = originalParameterizedType.arguments;
      TypeBinding[] substitutedArguments = originalArguments;
      if (originalArguments != null) {
        if (substitution.isRawSubstitution()) {
          return originalParameterizedType.environment.createRawType(originalParameterizedType.genericType(), substitutedEnclosing);
        }
        substitutedArguments = substitute(substitution, originalArguments);
      }
      if ((substitutedArguments != originalArguments) || (substitutedEnclosing != originalEnclosing))
      {
        return originalParameterizedType.environment.createParameterizedType(
          originalParameterizedType.genericType(), substitutedArguments, substitutedEnclosing);
      }

      break;
    case 68:
      ArrayBinding originalArrayType = (ArrayBinding)originalType;
      TypeBinding originalLeafComponentType = originalArrayType.leafComponentType;
      TypeBinding substitute = substitute(substitution, originalLeafComponentType);
      if (substitute != originalLeafComponentType) {
        return originalArrayType.environment.createArrayType(substitute.leafComponentType(), substitute.dimensions() + originalType.dimensions());
      }

      break;
    case 516:
      WildcardBinding wildcard = (WildcardBinding)originalType;
      if (wildcard.boundKind != 0) {
        TypeBinding originalBound = wildcard.bound;
        TypeBinding substitutedBound = substitute(substitution, originalBound);
        TypeBinding[] originalOtherBounds = wildcard.otherBounds;
        TypeBinding[] substitutedOtherBounds = substitute(substitution, originalOtherBounds);
        if ((substitutedBound != originalBound) || (originalOtherBounds != substitutedOtherBounds)) {
          return wildcard.environment.createWildcard(wildcard.genericType, wildcard.rank, substitutedBound, substitutedOtherBounds, wildcard.boundKind);
        }
      }

      break;
    case 4:
      if (originalType.isMemberType()) {
        ReferenceBinding originalReferenceType = (ReferenceBinding)originalType;
        ReferenceBinding originalEnclosing = originalType.enclosingType();
        ReferenceBinding substitutedEnclosing = originalEnclosing;
        if (originalEnclosing != null) {
          substitutedEnclosing = (ReferenceBinding)substitute(substitution, originalEnclosing);
        }

        if (substitutedEnclosing != originalEnclosing)
          return substitution.isRawSubstitution() ? 
            substitution.environment().createRawType(originalReferenceType, substitutedEnclosing) : 
            substitution.environment().createParameterizedType(originalReferenceType, null, substitutedEnclosing);
      }
      break;
    case 2052:
      ReferenceBinding originalReferenceType = (ReferenceBinding)originalType;
      ReferenceBinding originalEnclosing = originalType.enclosingType();
      ReferenceBinding substitutedEnclosing = originalEnclosing;
      if (originalEnclosing != null) {
        substitutedEnclosing = (ReferenceBinding)substitute(substitution, originalEnclosing);
      }

      if (substitution.isRawSubstitution()) {
        return substitution.environment().createRawType(originalReferenceType, substitutedEnclosing);
      }

      TypeBinding[] originalArguments = originalReferenceType.typeVariables();
      TypeBinding[] substitutedArguments = substitute(substitution, originalArguments);

      return substitution.environment().createParameterizedType(originalReferenceType, substitutedArguments, substitutedEnclosing);
    }

    return originalType;
  }

  public static TypeBinding[] substitute(Substitution substitution, TypeBinding[] originalTypes)
  {
    if (originalTypes == null) return null;
    TypeBinding[] substitutedTypes = originalTypes;
    int i = 0; for (int length = originalTypes.length; i < length; i++) {
      TypeBinding originalType = originalTypes[i];
      TypeBinding substitutedParameter = substitute(substitution, originalType);
      if (substitutedParameter != originalType) {
        if (substitutedTypes == originalTypes) {
          System.arraycopy(originalTypes, 0, substitutedTypes = new TypeBinding[length], 0, i);
        }
        substitutedTypes[i] = substitutedParameter;
      } else if (substitutedTypes != originalTypes) {
        substitutedTypes[i] = originalType;
      }
    }
    return substitutedTypes;
  }

  protected Scope(int kind, Scope parent) {
    this.kind = kind;
    this.parent = parent;
  }

  public TypeBinding boxing(TypeBinding type)
  {
    if (type.isBaseType())
      return environment().computeBoxingType(type);
    return type;
  }

  public final ClassScope classScope() {
    Scope scope = this;
    do {
      if ((scope instanceof ClassScope))
        return (ClassScope)scope;
      scope = scope.parent;
    }while (scope != null);
    return null;
  }

  public final CompilationUnitScope compilationUnitScope() {
    Scope lastScope = null;
    Scope scope = this;
    do {
      lastScope = scope;
      scope = scope.parent;
    }while (scope != null);
    return (CompilationUnitScope)lastScope;
  }

  public final CompilerOptions compilerOptions()
  {
    return compilationUnitScope().environment.globalOptions;
  }

  protected final MethodBinding computeCompatibleMethod(MethodBinding method, TypeBinding[] arguments, InvocationSite invocationSite)
  {
    TypeBinding[] genericTypeArguments = invocationSite.genericTypeArguments();
    TypeBinding[] parameters = method.parameters;
    TypeVariableBinding[] typeVariables = method.typeVariables;
    if ((parameters == arguments) && 
      ((method.returnType.tagBits & 0x20000000) == 0L) && 
      (genericTypeArguments == null) && 
      (typeVariables == Binding.NO_TYPE_VARIABLES)) {
      return method;
    }
    int argLength = arguments.length;
    int paramLength = parameters.length;
    boolean isVarArgs = method.isVarargs();
    if ((argLength != paramLength) && (
      (!isVarArgs) || (argLength < paramLength - 1))) {
      return null;
    }
    if (typeVariables != Binding.NO_TYPE_VARIABLES) {
      TypeBinding[] newArgs = (TypeBinding[])null;
      for (int i = 0; i < argLength; i++) {
        TypeBinding param = i < paramLength ? parameters[i] : parameters[(paramLength - 1)];
        if (arguments[i].isBaseType() != param.isBaseType()) {
          if (newArgs == null) {
            newArgs = new TypeBinding[argLength];
            System.arraycopy(arguments, 0, newArgs, 0, argLength);
          }
          newArgs[i] = environment().computeBoxingType(arguments[i]);
        }
      }
      if (newArgs != null)
        arguments = newArgs;
      method = ParameterizedGenericMethodBinding.computeCompatibleMethod(method, arguments, this, invocationSite);
      if (method == null) return null;
      if (!method.isValidBinding()) return method;
      parameters = method.parameters;
    } else if (genericTypeArguments != null) {
      if ((method instanceof ParameterizedGenericMethodBinding)) {
        if (!((ParameterizedGenericMethodBinding)method).wasInferred)
        {
          return new ProblemMethodBinding(method, method.selector, genericTypeArguments, 13);
        }
      }
      else return new ProblemMethodBinding(method, method.selector, genericTypeArguments, 11);

    }

    if (parameterCompatibilityLevel(method, arguments) > -1)
      return method;
    if (genericTypeArguments != null)
      return new ProblemMethodBinding(method, method.selector, arguments, 12);
    return null;
  }

  protected boolean connectTypeVariables(TypeParameter[] typeParameters, boolean checkForErasedCandidateCollisions) {
    if ((typeParameters == null) || (compilerOptions().sourceLevel < 3211264L)) return true;
    boolean noProblems = true;
    Map invocations = new HashMap(2);
    int i = 0; for (int paramLength = typeParameters.length; i < paramLength; i++) {
      TypeParameter typeParameter = typeParameters[i];
      TypeVariableBinding typeVariable = typeParameter.binding;
      if (typeVariable == null) return false;

      typeVariable.superclass = getJavaLangObject();
      typeVariable.superInterfaces = Binding.NO_SUPERINTERFACES;

      typeVariable.firstBound = null;

      TypeReference typeRef = typeParameter.type;
      if (typeRef != null)
      {
        TypeBinding superType = this.kind == 2 ? 
          typeRef.resolveType((BlockScope)this, false) : 
          typeRef.resolveType((ClassScope)this);
        if (superType == null) {
          typeVariable.tagBits |= 131072L;
          noProblems = false;
        }
        else {
          typeRef.resolvedType = superType;
          if (superType.isArrayType()) {
            problemReporter().boundCannotBeArray(typeRef, superType);
          }
          else {
            boolean isTypeVariableFirstBound = superType.isTypeVariable();
            if (isTypeVariableFirstBound) {
              TypeVariableBinding varSuperType = (TypeVariableBinding)superType;
              if ((varSuperType.rank >= typeVariable.rank) && (varSuperType.declaringElement == typeVariable.declaringElement)) {
                problemReporter().forwardTypeVariableReference(typeParameter, varSuperType);
                typeVariable.tagBits |= 131072L;
                noProblems = false;
                continue;
              }
            }
            ReferenceBinding superRefType = (ReferenceBinding)superType;
            if (superRefType.isFinal())
              problemReporter().finalVariableBound(typeVariable, typeRef);
            if (!superType.isInterface())
              typeVariable.superclass = superRefType;
            else {
              typeVariable.superInterfaces = new ReferenceBinding[] { superRefType };
            }
            typeVariable.firstBound = superRefType;
            TypeReference[] boundRefs = typeParameter.bounds;
            if (boundRefs != null) {
              int j = 0; for (int boundLength = boundRefs.length; j < boundLength; j++) {
                typeRef = boundRefs[j];
                superType = this.kind == 2 ? 
                  typeRef.resolveType((BlockScope)this, false) : 
                  typeRef.resolveType((ClassScope)this);
                if (superType == null) {
                  typeVariable.tagBits |= 131072L;
                  noProblems = false;
                  break;
                }
                typeRef.resolvedType = superType;
                if ((isTypeVariableFirstBound) && (j == 0)) {
                  problemReporter().noAdditionalBoundAfterTypeVariable(typeRef);
                }
                if (superType.isArrayType()) {
                  problemReporter().boundCannotBeArray(typeRef, superType);
                  break;
                }
                superRefType = (ReferenceBinding)superType;
                if (!superType.isInterface()) {
                  problemReporter().boundMustBeAnInterface(typeRef, superType);
                  typeVariable.tagBits |= 131072L;
                  noProblems = false;
                  break;
                }

                if ((checkForErasedCandidateCollisions) && (typeVariable.firstBound == typeVariable.superclass) && 
                  (hasErasedCandidatesCollisions(superType, typeVariable.superclass, invocations, typeVariable, typeRef))) {
                  noProblems = false;
                  break;
                }

                int index = typeVariable.superInterfaces.length;
                do { ReferenceBinding previousInterface = typeVariable.superInterfaces[index];
                  if (previousInterface == superRefType) {
                    problemReporter().duplicateBounds(typeRef, superType);
                    typeVariable.tagBits |= 131072L;
                    noProblems = false;
                    break;
                  }
                  if ((checkForErasedCandidateCollisions) && 
                    (hasErasedCandidatesCollisions(superType, previousInterface, invocations, typeVariable, typeRef))) {
                    noProblems = false;
                    break;
                  }
                  index--; } while (index >= 0);

                int size = typeVariable.superInterfaces.length;
                System.arraycopy(typeVariable.superInterfaces, 0, typeVariable.superInterfaces = new ReferenceBinding[size + 1], 0, size);
                typeVariable.superInterfaces[size] = superRefType;
              }
            }
          }
        }
      }
    }
    return noProblems;
  }

  public ArrayBinding createArrayType(TypeBinding type, int dimension) {
    if (type.isValidBinding()) {
      return environment().createArrayType(type, dimension);
    }
    return new ArrayBinding(type, dimension, environment());
  }

  public TypeVariableBinding[] createTypeVariables(TypeParameter[] typeParameters, Binding declaringElement)
  {
    if ((typeParameters == null) || (compilerOptions().sourceLevel < 3211264L)) {
      return Binding.NO_TYPE_VARIABLES;
    }
    PackageBinding unitPackage = compilationUnitScope().fPackage;
    int length = typeParameters.length;
    TypeVariableBinding[] typeVariableBindings = new TypeVariableBinding[length];
    int count = 0;
    for (int i = 0; i < length; i++) {
      TypeParameter typeParameter = typeParameters[i];
      TypeVariableBinding parameterBinding = new TypeVariableBinding(typeParameter.name, declaringElement, i);
      parameterBinding.fPackage = unitPackage;
      typeParameter.binding = parameterBinding;

      for (int j = 0; j < count; j++) {
        TypeVariableBinding knownVar = typeVariableBindings[j];
        if (CharOperation.equals(knownVar.sourceName, typeParameter.name))
          problemReporter().duplicateTypeParameterInType(typeParameter);
      }
      typeVariableBindings[(count++)] = parameterBinding;
    }

    if (count != length)
      System.arraycopy(typeVariableBindings, 0, typeVariableBindings = new TypeVariableBinding[count], 0, count);
    return typeVariableBindings;
  }

  public final ClassScope enclosingClassScope() {
    Scope scope = this;
    while ((scope = scope.parent) != null) {
      if ((scope instanceof ClassScope)) return (ClassScope)scope;
    }
    return null;
  }

  public final MethodScope enclosingMethodScope() {
    Scope scope = this;
    while ((scope = scope.parent) != null) {
      if ((scope instanceof MethodScope)) return (MethodScope)scope;
    }
    return null;
  }

  public final ReferenceBinding enclosingReceiverType()
  {
    Scope scope = this;
    do {
      if ((scope instanceof ClassScope)) {
        return environment().convertToParameterizedType(((ClassScope)scope).referenceContext.binding);
      }
      scope = scope.parent;
    }while (scope != null);
    return null;
  }

  public ReferenceContext enclosingReferenceContext()
  {
    Scope current = this;
    while ((current = 
      current.parent) != null) {
      switch (current.kind) {
      case 2:
        return ((MethodScope)current).referenceContext;
      case 3:
        return ((ClassScope)current).referenceContext;
      case 4:
        return ((CompilationUnitScope)current).referenceContext;
      }
    }
    return null;
  }

  public final SourceTypeBinding enclosingSourceType()
  {
    Scope scope = this;
    do {
      if ((scope instanceof ClassScope))
        return ((ClassScope)scope).referenceContext.binding;
      scope = scope.parent;
    }while (scope != null);
    return null;
  }

  public final LookupEnvironment environment() {
    Scope unitScope = this;
    Scope scope;
    while ((scope = unitScope.parent) != null)
    {
      Scope scope;
      unitScope = scope;
    }return ((CompilationUnitScope)unitScope).environment;
  }

  protected MethodBinding findDefaultAbstractMethod(ReferenceBinding receiverType, char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite, ReferenceBinding classHierarchyStart, ObjectVector found, MethodBinding concreteMatch)
  {
    int startFoundSize = found.size;
    ReferenceBinding currentType = classHierarchyStart;
    while (currentType != null) {
      findMethodInSuperInterfaces(currentType, selector, found);
      currentType = currentType.superclass();
    }
    MethodBinding[] candidates = (MethodBinding[])null;
    int candidatesCount = 0;
    MethodBinding problemMethod = null;
    int foundSize = found.size;
    if (foundSize > startFoundSize)
    {
      for (int i = startFoundSize; i < foundSize; i++) {
        MethodBinding methodBinding = (MethodBinding)found.elementAt(i);
        MethodBinding compatibleMethod = computeCompatibleMethod(methodBinding, argumentTypes, invocationSite);
        if (compatibleMethod != null) {
          if (compatibleMethod.isValidBinding()) {
            if ((concreteMatch == null) || (concreteMatch.declaringClass.findSuperTypeWithSameErasure(compatibleMethod.declaringClass) == null) || 
              (!environment().methodVerifier().doesMethodOverride(concreteMatch, compatibleMethod)))
            {
              if (candidatesCount == 0) {
                candidates = new MethodBinding[foundSize - startFoundSize + 1];
                if (concreteMatch != null)
                  candidates[(candidatesCount++)] = concreteMatch;
              }
              candidates[(candidatesCount++)] = compatibleMethod;
            } } else if (problemMethod == null) {
            problemMethod = compatibleMethod;
          }
        }
      }
    }

    if (candidatesCount < 2) {
      if (concreteMatch == null) {
        if (candidatesCount == 0)
          return problemMethod;
        concreteMatch = candidates[0];
      }
      compilationUnitScope().recordTypeReferences(concreteMatch.thrownExceptions);
      return concreteMatch;
    }

    if (compilerOptions().complianceLevel >= 3145728L)
      return mostSpecificMethodBinding(candidates, candidatesCount, argumentTypes, invocationSite, receiverType);
    return mostSpecificInterfaceMethodBinding(candidates, candidatesCount, invocationSite);
  }

  public ReferenceBinding findDirectMemberType(char[] typeName, ReferenceBinding enclosingType)
  {
    if ((enclosingType.tagBits & 0x10000) != 0L) {
      return null;
    }
    ReferenceBinding enclosingReceiverType = enclosingReceiverType();
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordReference(enclosingType, typeName);
    ReferenceBinding memberType = enclosingType.getMemberType(typeName);
    if (memberType != null) {
      unitScope.recordTypeReference(memberType);
      if (enclosingReceiverType == null ? 
        memberType.canBeSeenBy(getCurrentPackage()) : 
        memberType.canBeSeenBy(enclosingType, enclosingReceiverType))
        return memberType;
      return new ProblemReferenceBinding(typeName, memberType, 2);
    }
    return null;
  }

  public MethodBinding findExactMethod(ReferenceBinding receiverType, char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite)
  {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordTypeReferences(argumentTypes);
    MethodBinding exactMethod = receiverType.getExactMethod(selector, argumentTypes, unitScope);
    if ((exactMethod != null) && (exactMethod.typeVariables == Binding.NO_TYPE_VARIABLES) && (!exactMethod.isBridge()))
    {
      unitScope.recordTypeReferences(exactMethod.thrownExceptions);

      if ((receiverType.isInterface()) || (exactMethod.canBeSeenBy(receiverType, invocationSite, this))) {
        if ((receiverType.id != 1) && 
          (argumentTypes == Binding.NO_PARAMETERS) && 
          (CharOperation.equals(selector, GETCLASS)) && 
          (exactMethod.returnType.isParameterizedType())) {
          return ParameterizedMethodBinding.instantiateGetClass(receiverType, exactMethod, this);
        }

        if (invocationSite.genericTypeArguments() != null) {
          exactMethod = computeCompatibleMethod(exactMethod, argumentTypes, invocationSite);
        }
        return exactMethod;
      }
    }
    return null;
  }

  public FieldBinding findField(TypeBinding receiverType, char[] fieldName, InvocationSite invocationSite, boolean needResolve)
  {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordTypeReference(receiverType);
    TypeBinding leafType;
    TypeBinding leafType;
    switch (receiverType.kind()) {
    case 132:
      return null;
    case 516:
    case 4100:
      TypeBinding receiverErasure = receiverType.erasure();
      if (!receiverErasure.isArrayType())
        break label153;
      leafType = receiverErasure.leafComponentType();
      break;
    case 68:
      leafType = receiverType.leafComponentType();
      break;
    default:
      break;
    }
    TypeBinding leafType;
    if (((leafType instanceof ReferenceBinding)) && 
      (!((ReferenceBinding)leafType).canBeSeenBy(this)))
      return new ProblemFieldBinding((ReferenceBinding)leafType, fieldName, 8);
    if (CharOperation.equals(fieldName, LENGTH))
      return ArrayBinding.ArrayLength;
    return null;

    label153: ReferenceBinding currentType = (ReferenceBinding)receiverType;
    if (!currentType.canBeSeenBy(this)) {
      return new ProblemFieldBinding(currentType, fieldName, 8);
    }
    FieldBinding field = currentType.getField(fieldName, needResolve);
    if (field != null) {
      if (invocationSite == null ? 
        field.canBeSeenBy(getCurrentPackage()) : 
        field.canBeSeenBy(currentType, invocationSite, this))
        return field;
      return new ProblemFieldBinding(field, field.declaringClass, fieldName, 2);
    }

    ReferenceBinding[] interfacesToVisit = (ReferenceBinding[])null;
    int nextPosition = 0;
    FieldBinding visibleField = null;
    boolean keepLooking = true;
    FieldBinding notVisibleField = null;

    while (keepLooking) {
      ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
      if (itsInterfaces == null) {
        ((SourceTypeBinding)currentType).scope.connectTypeHierarchy();
        itsInterfaces = currentType.superInterfaces();
      }
      if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES))
        if (interfacesToVisit == null) {
          interfacesToVisit = itsInterfaces;
          nextPosition = interfacesToVisit.length;
        } else {
          int itsLength = itsInterfaces.length;
          if (nextPosition + itsLength >= interfacesToVisit.length)
            System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
          for (int a = 0; a < itsLength; a++) {
            ReferenceBinding next = itsInterfaces[a];
            int b = 0;
            while (next != interfacesToVisit[b])
            {
              b++; if (b >= nextPosition)
              {
                interfacesToVisit[(nextPosition++)] = next;
              }
            }
          }
        }
      if ((currentType = currentType.superclass()) == null) {
        break;
      }
      unitScope.recordTypeReference(currentType);
      if ((field = currentType.getField(fieldName, needResolve)) != null) {
        keepLooking = false;
        if (field.canBeSeenBy(receiverType, invocationSite, this)) {
          if (visibleField == null)
            visibleField = field;
          else
            return new ProblemFieldBinding(visibleField, visibleField.declaringClass, fieldName, 3);
        }
        else if (notVisibleField == null) {
          notVisibleField = field;
        }
      }

    }

    if (interfacesToVisit != null) {
      ProblemFieldBinding ambiguous = null;
      for (int i = 0; i < nextPosition; i++) {
        ReferenceBinding anInterface = interfacesToVisit[i];
        unitScope.recordTypeReference(anInterface);
        if ((field = anInterface.getField(fieldName, true)) != null) {
          if (visibleField == null) {
            visibleField = field;
          } else {
            ambiguous = new ProblemFieldBinding(visibleField, visibleField.declaringClass, fieldName, 3);
            break;
          }
        } else {
          ReferenceBinding[] itsInterfaces = anInterface.superInterfaces();
          if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES)) {
            int itsLength = itsInterfaces.length;
            if (nextPosition + itsLength >= interfacesToVisit.length)
              System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
            for (int a = 0; a < itsLength; a++) {
              ReferenceBinding next = itsInterfaces[a];
              int b = 0;
              while (next != interfacesToVisit[b])
              {
                b++; if (b >= nextPosition)
                {
                  interfacesToVisit[(nextPosition++)] = next;
                }
              }
            }
          }
        }
      }
      if (ambiguous != null) {
        return ambiguous;
      }
    }
    if (visibleField != null)
      return visibleField;
    if (notVisibleField != null) {
      return new ProblemFieldBinding(notVisibleField, currentType, fieldName, 2);
    }
    return null;
  }

  public ReferenceBinding findMemberType(char[] typeName, ReferenceBinding enclosingType)
  {
    if ((enclosingType.tagBits & 0x10000) != 0L) {
      return null;
    }
    ReferenceBinding enclosingSourceType = enclosingSourceType();
    PackageBinding currentPackage = getCurrentPackage();
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordReference(enclosingType, typeName);
    ReferenceBinding memberType = enclosingType.getMemberType(typeName);
    if (memberType != null) {
      unitScope.recordTypeReference(memberType);
      if (enclosingSourceType == null ? 
        memberType.canBeSeenBy(currentPackage) : 
        memberType.canBeSeenBy(enclosingType, enclosingSourceType))
        return memberType;
      return new ProblemReferenceBinding(typeName, memberType, 2);
    }

    ReferenceBinding currentType = enclosingType;
    ReferenceBinding[] interfacesToVisit = (ReferenceBinding[])null;
    int nextPosition = 0;
    ReferenceBinding visibleMemberType = null;
    boolean keepLooking = true;
    ReferenceBinding notVisible = null;

    while (keepLooking) {
      ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
      if (itsInterfaces == null) {
        ReferenceBinding sourceType = currentType.isParameterizedType() ? 
          ((ParameterizedTypeBinding)currentType).genericType() : 
          currentType;
        if (sourceType.isHierarchyBeingConnected())
          return null;
        ((SourceTypeBinding)sourceType).scope.connectTypeHierarchy();
        itsInterfaces = currentType.superInterfaces();
      }
      if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES))
        if (interfacesToVisit == null) {
          interfacesToVisit = itsInterfaces;
          nextPosition = interfacesToVisit.length;
        } else {
          int itsLength = itsInterfaces.length;
          if (nextPosition + itsLength >= interfacesToVisit.length)
            System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
          for (int a = 0; a < itsLength; a++) {
            ReferenceBinding next = itsInterfaces[a];
            int b = 0;
            while (next != interfacesToVisit[b])
            {
              b++; if (b >= nextPosition)
              {
                interfacesToVisit[(nextPosition++)] = next;
              }
            }
          }
        }
      if ((currentType = currentType.superclass()) == null) {
        break;
      }
      unitScope.recordReference(currentType, typeName);
      if ((memberType = currentType.getMemberType(typeName)) != null) {
        unitScope.recordTypeReference(memberType);
        keepLooking = false;
        if (enclosingSourceType == null ? 
          memberType.canBeSeenBy(currentPackage) : 
          memberType.canBeSeenBy(enclosingType, enclosingSourceType)) {
          if (visibleMemberType == null)
            visibleMemberType = memberType;
          else
            return new ProblemReferenceBinding(typeName, visibleMemberType, 3);
        }
        else notVisible = memberType;
      }

    }

    if (interfacesToVisit != null) {
      ProblemReferenceBinding ambiguous = null;
      for (int i = 0; i < nextPosition; i++) {
        ReferenceBinding anInterface = interfacesToVisit[i];
        unitScope.recordReference(anInterface, typeName);
        if ((memberType = anInterface.getMemberType(typeName)) != null) {
          unitScope.recordTypeReference(memberType);
          if (visibleMemberType == null) {
            visibleMemberType = memberType;
          } else {
            ambiguous = new ProblemReferenceBinding(typeName, visibleMemberType, 3);
            break;
          }
        } else {
          ReferenceBinding[] itsInterfaces = anInterface.superInterfaces();
          if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES)) {
            int itsLength = itsInterfaces.length;
            if (nextPosition + itsLength >= interfacesToVisit.length)
              System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
            for (int a = 0; a < itsLength; a++) {
              ReferenceBinding next = itsInterfaces[a];
              int b = 0;
              while (next != interfacesToVisit[b])
              {
                b++; if (b >= nextPosition)
                {
                  interfacesToVisit[(nextPosition++)] = next;
                }
              }
            }
          }
        }
      }
      if (ambiguous != null)
        return ambiguous;
    }
    if (visibleMemberType != null)
      return visibleMemberType;
    if (notVisible != null)
      return new ProblemReferenceBinding(typeName, notVisible, 2);
    return null;
  }

  public MethodBinding findMethod(ReferenceBinding receiverType, char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite)
  {
    ReferenceBinding currentType = receiverType;
    boolean receiverTypeIsInterface = receiverType.isInterface();
    ObjectVector found = new ObjectVector(3);
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordTypeReferences(argumentTypes);

    if (receiverTypeIsInterface) {
      unitScope.recordTypeReference(receiverType);
      MethodBinding[] receiverMethods = receiverType.getMethods(selector);
      if (receiverMethods.length > 0)
        found.addAll(receiverMethods);
      findMethodInSuperInterfaces(receiverType, selector, found);
      currentType = getJavaLangObject();
    }

    long complianceLevel = compilerOptions().complianceLevel;
    boolean isCompliant14 = complianceLevel >= 3145728L;
    boolean isCompliant15 = complianceLevel >= 3211264L;
    ReferenceBinding classHierarchyStart = currentType;
    MethodVerifier verifier = environment().methodVerifier();
    while (currentType != null) {
      unitScope.recordTypeReference(currentType);
      MethodBinding[] currentMethods = currentType.getMethods(selector);
      int currentLength = currentMethods.length;
      if (currentLength > 0) {
        if ((isCompliant14) && ((receiverTypeIsInterface) || (found.size > 0))) {
          int i = 0; for (int l = currentLength; i < l; i++) {
            MethodBinding currentMethod = currentMethods[i];
            if (currentMethod != null)
              if ((receiverTypeIsInterface) && (!currentMethod.isPublic())) {
                currentLength--;
                currentMethods[i] = null;
              }
              else
              {
                int j = 0; for (int max = found.size; j < max; j++) {
                  MethodBinding matchingMethod = (MethodBinding)found.elementAt(j);
                  if (verifier.doesMethodOverride(matchingMethod.original(), currentMethod.original())) {
                    if ((isCompliant15) && 
                      (matchingMethod.isBridge()) && (!currentMethod.isBridge())) {
                      break;
                    }
                    currentLength--;
                    currentMethods[i] = null;
                    break;
                  }
                }
              }
          }
        }
        if (currentLength > 0)
        {
          if (currentMethods.length == currentLength) {
            found.addAll(currentMethods);
          } else {
            int i = 0; for (int max = currentMethods.length; i < max; i++) {
              MethodBinding currentMethod = currentMethods[i];
              if (currentMethod != null)
                found.add(currentMethod);
            }
          }
        }
      }
      currentType = currentType.superclass();
    }

    int foundSize = found.size;
    MethodBinding[] candidates = (MethodBinding[])null;
    int candidatesCount = 0;
    MethodBinding problemMethod = null;
    boolean searchForDefaultAbstractMethod = (isCompliant14) && (!receiverTypeIsInterface) && ((receiverType.isAbstract()) || (receiverType.isTypeVariable()));
    if (foundSize > 0)
    {
      for (int i = 0; i < foundSize; i++) {
        MethodBinding methodBinding = (MethodBinding)found.elementAt(i);
        MethodBinding compatibleMethod = computeCompatibleMethod(methodBinding, argumentTypes, invocationSite);
        if (compatibleMethod != null) {
          if (compatibleMethod.isValidBinding()) {
            if ((foundSize == 1) && (compatibleMethod.canBeSeenBy(receiverType, invocationSite, this)))
            {
              if (searchForDefaultAbstractMethod)
                return findDefaultAbstractMethod(receiverType, selector, argumentTypes, invocationSite, classHierarchyStart, found, compatibleMethod);
              unitScope.recordTypeReferences(compatibleMethod.thrownExceptions);
              return compatibleMethod;
            }
            if (candidatesCount == 0)
              candidates = new MethodBinding[foundSize];
            candidates[(candidatesCount++)] = compatibleMethod;
          } else if (problemMethod == null) {
            problemMethod = compatibleMethod;
          }
        }
      }

    }

    if (candidatesCount == 0)
    {
      MethodBinding interfaceMethod = 
        findDefaultAbstractMethod(receiverType, selector, argumentTypes, invocationSite, classHierarchyStart, found, null);
      if (interfaceMethod != null) return interfaceMethod;
      if (found.size == 0) return null;
      if (problemMethod != null) return problemMethod;

      int bestArgMatches = -1;
      MethodBinding bestGuess = (MethodBinding)found.elementAt(0);
      int argLength = argumentTypes.length;
      foundSize = found.size;
      for (int i = 0; i < foundSize; i++) {
        MethodBinding methodBinding = (MethodBinding)found.elementAt(i);
        TypeBinding[] params = methodBinding.parameters;
        int paramLength = params.length;
        int argMatches = 0;
        for (int a = 0; a < argLength; a++) {
          TypeBinding arg = argumentTypes[a];
          for (int p = a == 0 ? 0 : a - 1; (p < paramLength) && (p < a + 1); p++) {
            if (params[p] == arg) {
              argMatches++;
              break;
            }
          }
        }
        if (argMatches >= bestArgMatches)
        {
          if (argMatches == bestArgMatches) { int diff1 = paramLength < argLength ? 2 * (argLength - paramLength) : paramLength - argLength;
            int bestLength = bestGuess.parameters.length;
            int diff2 = bestLength < argLength ? 2 * (argLength - bestLength) : bestLength - argLength;
            if (diff1 >= diff2); } else {
            bestArgMatches = argMatches;
            bestGuess = methodBinding;
          }
        }
      }
      return new ProblemMethodBinding(bestGuess, bestGuess.selector, argumentTypes, 1);
    }

    int visiblesCount = 0;
    if (receiverTypeIsInterface) {
      if (candidatesCount == 1) {
        unitScope.recordTypeReferences(candidates[0].thrownExceptions);
        return candidates[0];
      }
      visiblesCount = candidatesCount;
    } else {
      for (int i = 0; i < candidatesCount; i++) {
        MethodBinding methodBinding = candidates[i];
        if (methodBinding.canBeSeenBy(receiverType, invocationSite, this)) {
          if (visiblesCount != i) {
            candidates[i] = null;
            candidates[visiblesCount] = methodBinding;
          }
          visiblesCount++;
        }
      }
      if (visiblesCount == 1) {
        if (searchForDefaultAbstractMethod)
          return findDefaultAbstractMethod(receiverType, selector, argumentTypes, invocationSite, classHierarchyStart, found, candidates[0]);
        unitScope.recordTypeReferences(candidates[0].thrownExceptions);
        return candidates[0];
      }
      if (visiblesCount == 0) {
        MethodBinding interfaceMethod = 
          findDefaultAbstractMethod(receiverType, selector, argumentTypes, invocationSite, classHierarchyStart, found, null);
        if (interfaceMethod != null) return interfaceMethod;
        return new ProblemMethodBinding(candidates[0], candidates[0].selector, candidates[0].parameters, 2);
      }
    }

    if (complianceLevel <= 3080192L) {
      ReferenceBinding declaringClass = candidates[0].declaringClass;
      return !declaringClass.isInterface() ? 
        mostSpecificClassMethodBinding(candidates, visiblesCount, invocationSite) : 
        mostSpecificInterfaceMethodBinding(candidates, visiblesCount, invocationSite);
    }

    if (compilerOptions().sourceLevel >= 3211264L) {
      for (int i = 0; i < visiblesCount; i++) {
        MethodBinding current = candidates[i];
        if ((current instanceof ParameterizedGenericMethodBinding))
          current = ((ParameterizedGenericMethodBinding)current).originalMethod;
        if ((current instanceof ParameterizedMethodBinding)) {
          for (int j = i + 1; j < visiblesCount; j++)
            if ((current.declaringClass == candidates[j].declaringClass) && (current.areParametersEqual(candidates[j])))
              return new ProblemMethodBinding(candidates[i], candidates[i].selector, candidates[i].parameters, 3);
        }
      }
    }
    MethodBinding mostSpecificMethod = mostSpecificMethodBinding(candidates, visiblesCount, argumentTypes, invocationSite, receiverType);
    if (searchForDefaultAbstractMethod) {
      if (mostSpecificMethod.isValidBinding())
      {
        return findDefaultAbstractMethod(receiverType, selector, argumentTypes, invocationSite, classHierarchyStart, found, mostSpecificMethod);
      }
      MethodBinding interfaceMethod = findDefaultAbstractMethod(receiverType, selector, argumentTypes, invocationSite, classHierarchyStart, found, null);
      if ((interfaceMethod != null) && (interfaceMethod.isValidBinding()))
        return interfaceMethod;
    }
    return mostSpecificMethod;
  }

  public MethodBinding findMethodForArray(ArrayBinding receiverType, char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite)
  {
    TypeBinding leafType = receiverType.leafComponentType();
    if (((leafType instanceof ReferenceBinding)) && 
      (!((ReferenceBinding)leafType).canBeSeenBy(this))) {
      return new ProblemMethodBinding(selector, Binding.NO_PARAMETERS, (ReferenceBinding)leafType, 8);
    }

    ReferenceBinding object = getJavaLangObject();
    MethodBinding methodBinding = object.getExactMethod(selector, argumentTypes, null);
    if (methodBinding != null)
    {
      if (argumentTypes == Binding.NO_PARAMETERS)
        switch (selector[0]) {
        case 'c':
          if (CharOperation.equals(selector, CLONE)) {
            return new UpdatedMethodBinding(
              compilerOptions().targetJDK >= 3145728L ? receiverType : object, 
              methodBinding.modifiers & 0xFFFFFFFB | 0x1, 
              CLONE, 
              methodBinding.returnType, 
              argumentTypes, 
              null, 
              object);
          }
          break;
        case 'g':
          if ((CharOperation.equals(selector, GETCLASS)) && (methodBinding.returnType.isParameterizedType()))
            return ParameterizedMethodBinding.instantiateGetClass(receiverType, methodBinding, this); break;
        case 'd':
        case 'e':
        case 'f':
        }
      if (methodBinding.canBeSeenBy(receiverType, invocationSite, this))
        return methodBinding;
    }
    methodBinding = findMethod(object, selector, argumentTypes, invocationSite);
    if (methodBinding == null)
      return new ProblemMethodBinding(selector, argumentTypes, 1);
    return methodBinding;
  }

  protected void findMethodInSuperInterfaces(ReferenceBinding currentType, char[] selector, ObjectVector found) {
    ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
    if ((itsInterfaces != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES)) {
      ReferenceBinding[] interfacesToVisit = itsInterfaces;
      int nextPosition = interfacesToVisit.length;
      for (int i = 0; i < nextPosition; i++) {
        currentType = interfacesToVisit[i];
        compilationUnitScope().recordTypeReference(currentType);
        MethodBinding[] currentMethods = currentType.getMethods(selector);
        if (currentMethods.length > 0) {
          int foundSize = found.size;
          if (foundSize > 0)
          {
            int c = 0; for (int l = currentMethods.length; c < l; c++) {
              MethodBinding current = currentMethods[c];
              int f = 0;
              while (current != found.elementAt(f))
              {
                f++; if (f >= foundSize)
                {
                  found.add(current);
                }
              }
            } } else { found.addAll(currentMethods); }

        }
        if (((itsInterfaces = currentType.superInterfaces()) != null) && (itsInterfaces != Binding.NO_SUPERINTERFACES)) {
          int itsLength = itsInterfaces.length;
          if (nextPosition + itsLength >= interfacesToVisit.length)
            System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
          for (int a = 0; a < itsLength; a++) {
            ReferenceBinding next = itsInterfaces[a];
            int b = 0;
            while (next != interfacesToVisit[b])
            {
              b++; if (b >= nextPosition)
              {
                interfacesToVisit[(nextPosition++)] = next;
              }
            }
          }
        }
      }
    }
  }

  public ReferenceBinding findType(char[] typeName, PackageBinding declarationPackage, PackageBinding invocationPackage)
  {
    compilationUnitScope().recordReference(declarationPackage.compoundName, typeName);
    ReferenceBinding typeBinding = declarationPackage.getType(typeName);
    if (typeBinding == null) {
      return null;
    }
    if ((typeBinding.isValidBinding()) && 
      (declarationPackage != invocationPackage) && (!typeBinding.canBeSeenBy(invocationPackage))) {
      return new ProblemReferenceBinding(typeName, typeBinding, 2);
    }
    return typeBinding;
  }

  public LocalVariableBinding findVariable(char[] variable)
  {
    return null;
  }

  public Binding getBinding(char[] name, int mask, InvocationSite invocationSite, boolean needResolve)
  {
    CompilationUnitScope unitScope = compilationUnitScope();
    LookupEnvironment env = unitScope.environment;
    try {
      env.missingClassFileLocation = invocationSite;
      Binding binding = null;
      FieldBinding problemField = null;
      Object localObject2;
      if ((mask & 0x3) != 0) {
        boolean insideStaticContext = false;
        boolean insideConstructorCall = false;
        boolean insideTypeAnnotation = false;

        FieldBinding foundField = null;

        ProblemFieldBinding foundInsideProblem = null;

        Scope scope = this;
        int depth = 0;
        int foundDepth = 0;
        ReferenceBinding foundActualReceiverType = null;
        while (true) {
          switch (scope.kind) {
          case 2:
            MethodScope methodScope = (MethodScope)scope;
            insideStaticContext |= methodScope.isStatic;
            insideConstructorCall |= methodScope.isConstructorCall;
            insideTypeAnnotation = methodScope.insideTypeAnnotation;
          case 1:
            LocalVariableBinding variableBinding = scope.findVariable(name);

            if (variableBinding != null) {
              if ((foundField != null) && (foundField.isValidBinding()))
                return new ProblemFieldBinding(
                  foundField, 
                  foundField.declaringClass, 
                  name, 
                  5);
              if (depth > 0)
                invocationSite.setDepth(depth);
              return variableBinding;
            }
            break;
          case 3:
            ClassScope classScope = (ClassScope)scope;
            ReferenceBinding receiverType = classScope.enclosingReceiverType();
            if (!insideTypeAnnotation) {
              FieldBinding fieldBinding = classScope.findField(receiverType, name, invocationSite, needResolve);

              if (fieldBinding != null) {
                if (fieldBinding.problemId() == 3) {
                  if ((foundField == null) || (foundField.problemId() == 2))
                  {
                    return fieldBinding;
                  }
                  return new ProblemFieldBinding(
                    foundField, 
                    foundField.declaringClass, 
                    name, 
                    5);
                }

                ProblemFieldBinding insideProblem = null;
                if (fieldBinding.isValidBinding()) {
                  if (!fieldBinding.isStatic()) {
                    if (insideConstructorCall)
                      insideProblem = 
                        new ProblemFieldBinding(
                        fieldBinding, 
                        fieldBinding.declaringClass, 
                        name, 
                        6);
                    else if (insideStaticContext) {
                      insideProblem = 
                        new ProblemFieldBinding(
                        fieldBinding, 
                        fieldBinding.declaringClass, 
                        name, 
                        7);
                    }
                  }
                  if ((receiverType == fieldBinding.declaringClass) || (compilerOptions().complianceLevel >= 3145728L))
                  {
                    if (foundField == null) {
                      if (depth > 0) {
                        invocationSite.setDepth(depth);
                        invocationSite.setActualReceiverType(receiverType);
                      }

                      return insideProblem == null ? fieldBinding : insideProblem;
                    }
                    if (foundField.isValidBinding())
                    {
                      if (foundField.declaringClass != fieldBinding.declaringClass)
                      {
                        return new ProblemFieldBinding(
                          foundField, 
                          foundField.declaringClass, 
                          name, 
                          5);
                      }
                    }
                  }
                }
                if ((foundField == null) || ((foundField.problemId() == 2) && (fieldBinding.problemId() != 2)))
                {
                  foundDepth = depth;
                  foundActualReceiverType = receiverType;
                  foundInsideProblem = insideProblem;
                  foundField = fieldBinding;
                }
              }
            }
            insideTypeAnnotation = false;
            depth++;
            insideStaticContext |= receiverType.isStatic();

            MethodScope enclosingMethodScope = scope.methodScope();
            insideConstructorCall = enclosingMethodScope == null ? false : enclosingMethodScope.isConstructorCall;
            break;
          case 4:
            break;
          }
          scope = scope.parent;
        }

        if (foundInsideProblem != null)
          return foundInsideProblem;
        if (foundField != null) {
          if (foundField.isValidBinding()) {
            if (foundDepth > 0) {
              invocationSite.setDepth(foundDepth);
              invocationSite.setActualReceiverType(foundActualReceiverType);
            }
            return foundField;
          }
          problemField = foundField;
          foundField = null;
        }

        if (compilerOptions().sourceLevel >= 3211264L)
        {
          unitScope.faultInImports();
          ImportBinding[] imports = unitScope.imports;
          if (imports != null)
          {
            int i = 0; for (int length = imports.length; i < length; i++) {
              ImportBinding importBinding = imports[i];
              if ((importBinding.isStatic()) && (!importBinding.onDemand) && 
                (CharOperation.equals(importBinding.compoundName[(importBinding.compoundName.length - 1)], name)) && 
                (unitScope.resolveSingleImport(importBinding) != null) && ((importBinding.resolvedImport instanceof FieldBinding))) {
                foundField = (FieldBinding)importBinding.resolvedImport;
                ImportReference importReference = importBinding.reference;
                if ((importReference != null) && (needResolve)) {
                  importReference.bits |= 2;
                }
                invocationSite.setActualReceiverType(foundField.declaringClass);
                if (foundField.isValidBinding()) {
                  return foundField;
                }
                if (problemField == null) {
                  problemField = foundField;
                }
              }

            }

            boolean foundInImport = false;
            int i = 0; for (int length = imports.length; i < length; i++) {
              ImportBinding importBinding = imports[i];
              if ((importBinding.isStatic()) && (importBinding.onDemand)) {
                Binding resolvedImport = importBinding.resolvedImport;
                if ((resolvedImport instanceof ReferenceBinding)) {
                  FieldBinding temp = findField((ReferenceBinding)resolvedImport, name, invocationSite, needResolve);
                  if (temp != null) {
                    if (!temp.isValidBinding()) {
                      if (problemField == null)
                        problemField = temp;
                    } else if ((temp.isStatic()) && 
                      (foundField != temp)) {
                      ImportReference importReference = importBinding.reference;
                      if ((importReference != null) && (needResolve)) {
                        importReference.bits |= 2;
                      }
                      if (foundInImport)
                      {
                        return new ProblemReferenceBinding(name, null, 3);
                      }foundField = temp;
                      foundInImport = true;
                    }
                  }
                }
              }
            }
            if (foundField != null) {
              invocationSite.setActualReceiverType(foundField.declaringClass);
              return foundField;
            }
          }
        }

      }

      if ((mask & 0x4) != 0) {
        if ((binding = getBaseType(name)) != null)
          return binding;
        binding = getTypeOrPackage(name, (mask & 0x10) == 0 ? 4 : 20);
        if ((binding.isValidBinding()) || (mask == 4))
          return binding;
      }
      else if ((mask & 0x10) != 0) {
        unitScope.recordSimpleReference(name);
        if ((binding = env.getTopLevelPackage(name)) != null)
          return binding;
      }
      if (problemField != null) return problemField;
      if ((binding != null) && (binding.problemId() != 1))
        return binding;
      return new ProblemBinding(name, enclosingSourceType(), 1);
    } catch (AbortCompilation e) {
      e.updateContext(invocationSite, referenceCompilationUnit().compilationResult);
      throw e;
    } finally {
      env.missingClassFileLocation = null;
    }
  }

  public MethodBinding getConstructor(ReferenceBinding receiverType, TypeBinding[] argumentTypes, InvocationSite invocationSite) {
    CompilationUnitScope unitScope = compilationUnitScope();
    LookupEnvironment env = unitScope.environment;
    try {
      env.missingClassFileLocation = invocationSite;
      unitScope.recordTypeReference(receiverType);
      unitScope.recordTypeReferences(argumentTypes);
      MethodBinding methodBinding = receiverType.getExactConstructor(argumentTypes);
      Object localObject2;
      if ((methodBinding != null) && (methodBinding.canBeSeenBy(invocationSite, this)))
      {
        if (invocationSite.genericTypeArguments() != null)
          methodBinding = computeCompatibleMethod(methodBinding, argumentTypes, invocationSite);
        return methodBinding;
      }
      MethodBinding[] methods = receiverType.getMethods(TypeIds.INIT);
      if (methods == Binding.NO_METHODS) {
        return new ProblemMethodBinding(
          TypeIds.INIT, 
          argumentTypes, 
          1);
      }
      MethodBinding[] compatible = new MethodBinding[methods.length];
      int compatibleIndex = 0;
      MethodBinding problemMethod = null;
      int i = 0; for (int length = methods.length; i < length; i++) {
        MethodBinding compatibleMethod = computeCompatibleMethod(methods[i], argumentTypes, invocationSite);
        if (compatibleMethod != null) {
          if (compatibleMethod.isValidBinding())
            compatible[(compatibleIndex++)] = compatibleMethod;
          else if (problemMethod == null)
            problemMethod = compatibleMethod;
        }
      }
      if (compatibleIndex == 0) {
        if (problemMethod == null)
          return new ProblemMethodBinding(TypeIds.INIT, argumentTypes, 1);
        return problemMethod;
      }

      MethodBinding[] visible = new MethodBinding[compatibleIndex];
      int visibleIndex = 0;
      for (int i = 0; i < compatibleIndex; i++) {
        MethodBinding method = compatible[i];
        if (method.canBeSeenBy(invocationSite, this))
          visible[(visibleIndex++)] = method;
      }
      if (visibleIndex == 1) return visible[0];
      if (visibleIndex == 0) {
        return new ProblemMethodBinding(
          compatible[0], 
          TypeIds.INIT, 
          compatible[0].parameters, 
          2);
      }
      return mostSpecificMethodBinding(visible, visibleIndex, argumentTypes, invocationSite, receiverType);
    } catch (AbortCompilation e) {
      e.updateContext(invocationSite, referenceCompilationUnit().compilationResult);
      throw e;
    } finally {
      env.missingClassFileLocation = null;
    }
  }

  public final PackageBinding getCurrentPackage() {
    Scope unitScope = this;
    Scope scope;
    while ((scope = unitScope.parent) != null)
    {
      Scope scope;
      unitScope = scope;
    }return ((CompilationUnitScope)unitScope).fPackage;
  }

  public int getDeclarationModifiers()
  {
    switch (this.kind) {
    case 1:
    case 2:
      MethodScope methodScope = methodScope();
      if (!methodScope.isInsideInitializer())
      {
        MethodBinding context = ((AbstractMethodDeclaration)methodScope.referenceContext).binding;
        if (context != null)
          return context.modifiers;
      } else {
        SourceTypeBinding type = ((BlockScope)this).referenceType().binding;

        if (methodScope.initializedField != null)
          return methodScope.initializedField.modifiers;
        if (type != null)
          return type.modifiers;
      }
      break;
    case 3:
      ReferenceBinding context = ((ClassScope)this).referenceType().binding;
      if (context != null)
        return context.modifiers;
      break;
    }
    return -1;
  }

  public FieldBinding getField(TypeBinding receiverType, char[] fieldName, InvocationSite invocationSite) {
    LookupEnvironment env = environment();
    try {
      env.missingClassFileLocation = invocationSite;
      FieldBinding field = findField(receiverType, fieldName, invocationSite, true);
      Object localObject2;
      if (field != null) return field;

      return new ProblemFieldBinding(
        (receiverType instanceof ReferenceBinding) ? (ReferenceBinding)receiverType : null, 
        fieldName, 
        1);
    } catch (AbortCompilation e) {
      e.updateContext(invocationSite, referenceCompilationUnit().compilationResult);
      throw e;
    } finally {
      env.missingClassFileLocation = null;
    }
  }

  public MethodBinding getImplicitMethod(char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite)
  {
    boolean insideStaticContext = false;
    boolean insideConstructorCall = false;
    boolean insideTypeAnnotation = false;
    MethodBinding foundMethod = null;
    MethodBinding foundProblem = null;
    boolean foundProblemVisible = false;
    Scope scope = this;
    int depth = 0;
    CompilerOptions options;
    boolean inheritedHasPrecedence = (options = compilerOptions()).complianceLevel >= 3145728L;
    while (true)
    {
      switch (scope.kind) {
      case 2:
        MethodScope methodScope = (MethodScope)scope;
        insideStaticContext |= methodScope.isStatic;
        insideConstructorCall |= methodScope.isConstructorCall;
        insideTypeAnnotation = methodScope.insideTypeAnnotation;
        break;
      case 3:
        ClassScope classScope = (ClassScope)scope;
        ReferenceBinding receiverType = classScope.enclosingReceiverType();
        if (!insideTypeAnnotation)
        {
          MethodBinding methodBinding = classScope.findExactMethod(receiverType, selector, argumentTypes, invocationSite);
          if (methodBinding == null)
            methodBinding = classScope.findMethod(receiverType, selector, argumentTypes, invocationSite);
          if (methodBinding != null) {
            if (foundMethod == null) {
              if (methodBinding.isValidBinding()) {
                if ((!methodBinding.isStatic()) && ((insideConstructorCall) || (insideStaticContext))) {
                  if ((foundProblem != null) && (foundProblem.problemId() != 2))
                    return foundProblem;
                  return new ProblemMethodBinding(
                    methodBinding, 
                    methodBinding.selector, 
                    methodBinding.parameters, 
                    insideConstructorCall ? 
                    6 : 
                    7);
                }
                if ((inheritedHasPrecedence) || 
                  (receiverType == methodBinding.declaringClass) || 
                  (receiverType.getMethods(selector) != Binding.NO_METHODS))
                {
                  if (foundProblemVisible) {
                    return foundProblem;
                  }
                  if (depth > 0) {
                    invocationSite.setDepth(depth);
                    invocationSite.setActualReceiverType(receiverType);
                  }
                  return methodBinding;
                }

                if ((foundProblem == null) || (foundProblem.problemId() == 2)) {
                  if (foundProblem != null) foundProblem = null;

                  if (depth > 0) {
                    invocationSite.setDepth(depth);
                    invocationSite.setActualReceiverType(receiverType);
                  }
                  foundMethod = methodBinding;
                }
              } else {
                if ((methodBinding.problemId() != 2) && (methodBinding.problemId() != 1))
                  return methodBinding;
                if (foundProblem == null) {
                  foundProblem = methodBinding;
                }
                if ((!foundProblemVisible) && (methodBinding.problemId() == 1)) {
                  MethodBinding closestMatch = ((ProblemMethodBinding)methodBinding).closestMatch;
                  if ((closestMatch != null) && (closestMatch.canBeSeenBy(receiverType, invocationSite, this))) {
                    foundProblem = methodBinding;
                    foundProblemVisible = true;
                  }
                }
              }
            }
            else if ((methodBinding.problemId() == 3) || (
              (foundMethod.declaringClass != methodBinding.declaringClass) && (
              (receiverType == methodBinding.declaringClass) || (receiverType.getMethods(selector) != Binding.NO_METHODS))))
            {
              return new ProblemMethodBinding(
                methodBinding, 
                selector, 
                argumentTypes, 
                5);
            }
          }
        }
        insideTypeAnnotation = false;
        depth++;
        insideStaticContext |= receiverType.isStatic();

        MethodScope enclosingMethodScope = scope.methodScope();
        insideConstructorCall = enclosingMethodScope == null ? false : enclosingMethodScope.isConstructorCall;
        break;
      case 4:
        break;
      }
      scope = scope.parent;
    }

    if ((insideStaticContext) && (options.sourceLevel >= 3211264L)) {
      if (foundProblem != null) {
        if ((foundProblem.declaringClass != null) && (foundProblem.declaringClass.id == 1))
          return foundProblem;
        if ((foundProblem.problemId() == 1) && (foundProblemVisible)) {
          return foundProblem;
        }

      }

      CompilationUnitScope unitScope = (CompilationUnitScope)scope;
      unitScope.faultInImports();
      ImportBinding[] imports = unitScope.imports;
      if (imports != null) {
        ObjectVector visible = null;
        boolean skipOnDemand = false;
        int i = 0; for (int length = imports.length; i < length; i++) {
          ImportBinding importBinding = imports[i];
          if (importBinding.isStatic()) {
            Binding resolvedImport = importBinding.resolvedImport;
            MethodBinding possible = null;
            if (importBinding.onDemand) {
              if ((!skipOnDemand) && ((resolvedImport instanceof ReferenceBinding)))
              {
                possible = findMethod((ReferenceBinding)resolvedImport, selector, argumentTypes, invocationSite);
              }
            } else if ((resolvedImport instanceof MethodBinding)) {
              MethodBinding staticMethod = (MethodBinding)resolvedImport;
              if (CharOperation.equals(staticMethod.selector, selector))
              {
                possible = findMethod(staticMethod.declaringClass, selector, argumentTypes, invocationSite);
              } } else if ((resolvedImport instanceof FieldBinding))
            {
              FieldBinding staticField = (FieldBinding)resolvedImport;
              if (CharOperation.equals(staticField.name, selector))
              {
                char[][] importName = importBinding.reference.tokens;
                TypeBinding referencedType = getType(importName, importName.length - 1);
                if (referencedType != null)
                {
                  possible = findMethod((ReferenceBinding)referencedType, selector, argumentTypes, invocationSite);
                }
              }
            }
            if ((possible != null) && (possible != foundProblem)) {
              if (!possible.isValidBinding()) {
                if (foundProblem == null)
                  foundProblem = possible;
              } else if (possible.isStatic()) {
                MethodBinding compatibleMethod = computeCompatibleMethod(possible, argumentTypes, invocationSite);
                if (compatibleMethod != null) {
                  if (compatibleMethod.isValidBinding()) {
                    if (compatibleMethod.canBeSeenBy(unitScope.fPackage)) {
                      if ((visible == null) || (!visible.contains(compatibleMethod))) {
                        ImportReference importReference = importBinding.reference;
                        if (importReference != null) {
                          importReference.bits |= 2;
                        }
                        if ((!skipOnDemand) && (!importBinding.onDemand)) {
                          visible = null;
                          skipOnDemand = true;
                        }
                        if (visible == null)
                          visible = new ObjectVector(3);
                        visible.add(compatibleMethod);
                      }
                    } else if (foundProblem == null)
                      foundProblem = new ProblemMethodBinding(compatibleMethod, selector, compatibleMethod.parameters, 2);
                  }
                  else if (foundProblem == null)
                    foundProblem = compatibleMethod;
                }
                else if (foundProblem == null) {
                  foundProblem = new ProblemMethodBinding(possible, selector, argumentTypes, 1);
                }
              }
            }
          }
        }
        if (visible != null) {
          MethodBinding[] temp = new MethodBinding[visible.size];
          visible.copyInto(temp);
          foundMethod = mostSpecificMethodBinding(temp, temp.length, argumentTypes, invocationSite, null);
        }
      }
    }

    if (foundMethod != null) {
      invocationSite.setActualReceiverType(foundMethod.declaringClass);
      return foundMethod;
    }
    if (foundProblem != null) {
      return foundProblem;
    }
    return new ProblemMethodBinding(selector, argumentTypes, 1);
  }

  public final ReferenceBinding getJavaIoSerializable() {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(JAVA_IO_SERIALIZABLE);
    return unitScope.environment.getResolvedType(JAVA_IO_SERIALIZABLE, this);
  }

  public final ReferenceBinding getJavaLangAnnotationAnnotation() {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(JAVA_LANG_ANNOTATION_ANNOTATION);
    return unitScope.environment.getResolvedType(JAVA_LANG_ANNOTATION_ANNOTATION, this);
  }

  public final ReferenceBinding getJavaLangAssertionError() {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(JAVA_LANG_ASSERTIONERROR);
    return unitScope.environment.getResolvedType(JAVA_LANG_ASSERTIONERROR, this);
  }

  public final ReferenceBinding getJavaLangClass() {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(JAVA_LANG_CLASS);
    return unitScope.environment.getResolvedType(JAVA_LANG_CLASS, this);
  }

  public final ReferenceBinding getJavaLangCloneable() {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(JAVA_LANG_CLONEABLE);
    return unitScope.environment.getResolvedType(JAVA_LANG_CLONEABLE, this);
  }
  public final ReferenceBinding getJavaLangEnum() {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(JAVA_LANG_ENUM);
    return unitScope.environment.getResolvedType(JAVA_LANG_ENUM, this);
  }

  public final ReferenceBinding getJavaLangIterable() {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(JAVA_LANG_ITERABLE);
    return unitScope.environment.getResolvedType(JAVA_LANG_ITERABLE, this);
  }
  public final ReferenceBinding getJavaLangObject() {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(JAVA_LANG_OBJECT);
    return unitScope.environment.getResolvedType(JAVA_LANG_OBJECT, this);
  }

  public final ReferenceBinding getJavaLangString() {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(JAVA_LANG_STRING);
    return unitScope.environment.getResolvedType(JAVA_LANG_STRING, this);
  }

  public final ReferenceBinding getJavaLangThrowable() {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(JAVA_LANG_THROWABLE);
    return unitScope.environment.getResolvedType(JAVA_LANG_THROWABLE, this);
  }
  public final ReferenceBinding getJavaUtilIterator() {
    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(JAVA_UTIL_ITERATOR);
    return unitScope.environment.getResolvedType(JAVA_UTIL_ITERATOR, this);
  }

  public final ReferenceBinding getMemberType(char[] typeName, ReferenceBinding enclosingType)
  {
    ReferenceBinding memberType = findMemberType(typeName, enclosingType);
    if (memberType != null) return memberType;
    return new ProblemReferenceBinding(typeName, null, 1);
  }

  public MethodBinding getMethod(TypeBinding receiverType, char[] selector, TypeBinding[] argumentTypes, InvocationSite invocationSite) {
    CompilationUnitScope unitScope = compilationUnitScope();
    LookupEnvironment env = unitScope.environment;
    try {
      env.missingClassFileLocation = invocationSite;
      Object localObject2;
      switch (receiverType.kind()) {
      case 132:
        return new ProblemMethodBinding(selector, argumentTypes, 1);
      case 68:
        unitScope.recordTypeReference(receiverType);
        return findMethodForArray((ArrayBinding)receiverType, selector, argumentTypes, invocationSite);
      }
      unitScope.recordTypeReference(receiverType);

      ReferenceBinding currentType = (ReferenceBinding)receiverType;
      if (!currentType.canBeSeenBy(this)) {
        return new ProblemMethodBinding(selector, argumentTypes, 8);
      }

      MethodBinding methodBinding = findExactMethod(currentType, selector, argumentTypes, invocationSite);
      if (methodBinding != null) return methodBinding;

      methodBinding = findMethod(currentType, selector, argumentTypes, invocationSite);
      if (methodBinding == null)
        return new ProblemMethodBinding(selector, argumentTypes, 1);
      if (!methodBinding.isValidBinding()) {
        return methodBinding;
      }

      if ((receiverType.id != 1) && 
        (argumentTypes == Binding.NO_PARAMETERS) && 
        (CharOperation.equals(selector, GETCLASS)) && 
        (methodBinding.returnType.isParameterizedType())) {
        return ParameterizedMethodBinding.instantiateGetClass(receiverType, methodBinding, this);
      }
      return methodBinding;
    } catch (AbortCompilation e) {
      e.updateContext(invocationSite, referenceCompilationUnit().compilationResult);
      throw e;
    } finally {
      env.missingClassFileLocation = null;
    }
  }

  public final Binding getPackage(char[][] compoundName)
  {
    compilationUnitScope().recordQualifiedReference(compoundName);
    Binding binding = getTypeOrPackage(compoundName[0], 20);
    if (binding == null)
      return new ProblemReferenceBinding(compoundName[0], null, 1);
    if (!binding.isValidBinding()) {
      return binding;
    }
    if (!(binding instanceof PackageBinding)) return null;

    int currentIndex = 1;
    PackageBinding packageBinding = (PackageBinding)binding;
    while (currentIndex < compoundName.length) {
      binding = packageBinding.getTypeOrPackage(compoundName[(currentIndex++)]);
      if (binding == null)
        return new ProblemReferenceBinding(
          CharOperation.subarray(compoundName, 0, currentIndex), 
          null, 
          1);
      if (!binding.isValidBinding())
        return new ProblemReferenceBinding(
          CharOperation.subarray(compoundName, 0, currentIndex), 
          (binding instanceof ReferenceBinding) ? ((ReferenceBinding)binding).closestMatch() : null, 
          binding.problemId());
      if (!(binding instanceof PackageBinding))
        return packageBinding;
      packageBinding = (PackageBinding)binding;
    }
    return new ProblemReferenceBinding(compoundName, null, 1);
  }

  public final TypeBinding getType(char[] name)
  {
    TypeBinding binding = getBaseType(name);
    if (binding != null) return binding;
    return (ReferenceBinding)getTypeOrPackage(name, 4);
  }

  public final TypeBinding getType(char[] name, PackageBinding packageBinding)
  {
    if (packageBinding == null) {
      return getType(name);
    }
    Binding binding = packageBinding.getTypeOrPackage(name);
    if (binding == null)
      return new ProblemReferenceBinding(
        CharOperation.arrayConcat(packageBinding.compoundName, name), 
        null, 
        1);
    if (!binding.isValidBinding()) {
      return new ProblemReferenceBinding(
        CharOperation.arrayConcat(packageBinding.compoundName, name), 
        (binding instanceof ReferenceBinding) ? ((ReferenceBinding)binding).closestMatch() : null, 
        binding.problemId());
    }
    ReferenceBinding typeBinding = (ReferenceBinding)binding;
    if (!typeBinding.canBeSeenBy(this))
      return new ProblemReferenceBinding(
        CharOperation.arrayConcat(packageBinding.compoundName, name), 
        typeBinding, 
        2);
    return typeBinding;
  }

  public final TypeBinding getType(char[][] compoundName, int typeNameLength)
  {
    if (typeNameLength == 1)
    {
      TypeBinding binding = getBaseType(compoundName[0]);
      if (binding != null) return binding;
    }

    CompilationUnitScope unitScope = compilationUnitScope();
    unitScope.recordQualifiedReference(compoundName);
    Binding binding = 
      getTypeOrPackage(compoundName[0], typeNameLength == 1 ? 4 : 20);
    if (binding == null)
      return new ProblemReferenceBinding(compoundName[0], null, 1);
    if (!binding.isValidBinding()) {
      return (ReferenceBinding)binding;
    }
    int currentIndex = 1;
    boolean checkVisibility = false;
    if ((binding instanceof PackageBinding)) {
      PackageBinding packageBinding = (PackageBinding)binding;
      while (currentIndex < typeNameLength) {
        binding = packageBinding.getTypeOrPackage(compoundName[(currentIndex++)]);
        if (binding == null)
          return new ProblemReferenceBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            null, 
            1);
        if (!binding.isValidBinding())
          return new ProblemReferenceBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            (binding instanceof ReferenceBinding) ? ((ReferenceBinding)binding).closestMatch() : null, 
            binding.problemId());
        if (!(binding instanceof PackageBinding))
          break;
        packageBinding = (PackageBinding)binding;
      }
      if ((binding instanceof PackageBinding))
        return new ProblemReferenceBinding(
          CharOperation.subarray(compoundName, 0, currentIndex), 
          null, 
          1);
      checkVisibility = true;
    }

    ReferenceBinding typeBinding = (ReferenceBinding)binding;
    unitScope.recordTypeReference(typeBinding);
    if ((checkVisibility) && 
      (!typeBinding.canBeSeenBy(this))) {
      return new ProblemReferenceBinding(
        CharOperation.subarray(compoundName, 0, currentIndex), 
        typeBinding, 
        2);
    }
    while (currentIndex < typeNameLength) {
      typeBinding = getMemberType(compoundName[(currentIndex++)], typeBinding);
      if (!typeBinding.isValidBinding()) {
        if ((typeBinding instanceof ProblemReferenceBinding)) {
          ProblemReferenceBinding problemBinding = (ProblemReferenceBinding)typeBinding;
          return new ProblemReferenceBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            problemBinding.closestMatch(), 
            typeBinding.problemId());
        }
        return new ProblemReferenceBinding(
          CharOperation.subarray(compoundName, 0, currentIndex), 
          ((ReferenceBinding)binding).closestMatch(), 
          typeBinding.problemId());
      }
    }
    return typeBinding;
  }

  final Binding getTypeOrPackage(char[] name, int mask)
  {
    Scope scope = this;
    ReferenceBinding foundType = null;
    boolean insideStaticContext = false;
    boolean insideTypeAnnotation = false;
    if ((mask & 0x4) == 0) {
      Scope next = scope;
      while ((next = scope.parent) != null)
        scope = next;
    } else {
      while (true) {
        switch (scope.kind) {
        case 2:
          MethodScope methodScope = (MethodScope)scope;
          AbstractMethodDeclaration methodDecl = methodScope.referenceMethod();
          if ((methodDecl != null) && (methodDecl.binding != null)) {
            TypeVariableBinding typeVariable = methodDecl.binding.getTypeVariable(name);
            if (typeVariable != null) return typeVariable;
          }
          insideStaticContext |= methodScope.isStatic;
          insideTypeAnnotation = methodScope.insideTypeAnnotation;
        case 1:
          ReferenceBinding localType = ((BlockScope)scope).findLocalType(name);
          if (localType != null) {
            if ((foundType != null) && (foundType != localType))
              return new ProblemReferenceBinding(name, foundType, 5);
            return localType;
          }
          break;
        case 3:
          SourceTypeBinding sourceType = ((ClassScope)scope).referenceContext.binding;
          if ((scope == this) && ((sourceType.tagBits & 0x40000) == 0L))
          {
            TypeVariableBinding typeVariable = sourceType.getTypeVariable(name);
            if (typeVariable != null)
              return typeVariable;
            if (CharOperation.equals(name, sourceType.sourceName))
              return sourceType;
            insideStaticContext |= sourceType.isStatic();
          }
          else
          {
            if (!insideTypeAnnotation)
            {
              ReferenceBinding memberType = findMemberType(name, sourceType);
              if (memberType != null) {
                if (memberType.problemId() == 3) {
                  if ((foundType == null) || (foundType.problemId() == 2))
                  {
                    return memberType;
                  }
                  return new ProblemReferenceBinding(name, foundType, 5);
                }
                if ((memberType.isValidBinding()) && (
                  (sourceType == memberType.enclosingType()) || (compilerOptions().complianceLevel >= 3145728L))) {
                  if ((insideStaticContext) && (!memberType.isStatic()) && (sourceType.isGenericType())) {
                    return new ProblemReferenceBinding(name, memberType, 7);
                  }

                  if (foundType == null) {
                    return memberType;
                  }
                  if ((foundType.isValidBinding()) && (foundType != memberType)) {
                    return new ProblemReferenceBinding(name, foundType, 5);
                  }
                }
                if ((foundType == null) || ((foundType.problemId() == 2) && (memberType.problemId() != 2)))
                {
                  foundType = memberType;
                }
              }
            }
            TypeVariableBinding typeVariable = sourceType.getTypeVariable(name);
            if (typeVariable != null) {
              if (insideStaticContext)
                return new ProblemReferenceBinding(name, typeVariable, 7);
              return typeVariable;
            }
            insideStaticContext |= sourceType.isStatic();
            insideTypeAnnotation = false;
            if (CharOperation.equals(sourceType.sourceName, name)) {
              if ((foundType != null) && (foundType != sourceType) && (foundType.problemId() != 2))
                return new ProblemReferenceBinding(name, foundType, 5);
              return sourceType;
            }
          }
          break;
        case 4:
          break;
        }
        scope = scope.parent;
      }

      if ((foundType != null) && (foundType.problemId() != 2)) {
        return foundType;
      }
    }

    CompilationUnitScope unitScope = (CompilationUnitScope)scope;
    HashtableOfObject typeOrPackageCache = unitScope.typeOrPackageCache;
    if (typeOrPackageCache != null) {
      Binding binding = (Binding)typeOrPackageCache.get(name);
      if (binding != null) {
        if ((binding instanceof ImportBinding)) {
          ImportReference importReference = ((ImportBinding)binding).reference;
          if (importReference != null) {
            importReference.bits |= 2;
          }
          if ((binding instanceof ImportConflictBinding))
            typeOrPackageCache.put(name, binding = ((ImportConflictBinding)binding).conflictingTypeBinding);
          else
            typeOrPackageCache.put(name, binding = ((ImportBinding)binding).resolvedImport);
        }
        if ((mask & 0x4) != 0) {
          if ((foundType != null) && (foundType.problemId() != 2) && (binding.problemId() != 3))
            return foundType;
          if ((binding instanceof ReferenceBinding))
            return binding;
        }
        if (((mask & 0x10) != 0) && ((binding instanceof PackageBinding))) {
          return binding;
        }
      }
    }

    if ((mask & 0x4) != 0) {
      ImportBinding[] imports = unitScope.imports;
      if ((imports != null) && (typeOrPackageCache == null)) {
        int i = 0; for (int length = imports.length; i < length; i++) {
          ImportBinding importBinding = imports[i];
          if ((!importBinding.onDemand) && 
            (CharOperation.equals(importBinding.compoundName[(importBinding.compoundName.length - 1)], name))) {
            Binding resolvedImport = unitScope.resolveSingleImport(importBinding);
            if (resolvedImport != null) {
              if ((resolvedImport instanceof MethodBinding)) { resolvedImport = getType(importBinding.compoundName, importBinding.compoundName.length);
                if (!resolvedImport.isValidBinding());
              }
              else if ((resolvedImport instanceof TypeBinding)) {
                ImportReference importReference = importBinding.reference;
                if (importReference != null)
                  importReference.bits |= 2;
                return resolvedImport;
              }
            }
          }
        }

      }

      PackageBinding currentPackage = unitScope.fPackage;
      unitScope.recordReference(currentPackage.compoundName, name);
      Binding binding = currentPackage.getTypeOrPackage(name);
      if ((binding instanceof ReferenceBinding)) {
        if (typeOrPackageCache != null)
          typeOrPackageCache.put(name, binding);
        return binding;
      }

      if (imports != null) {
        boolean foundInImport = false;
        ReferenceBinding type = null;
        int i = 0; for (int length = imports.length; i < length; i++) {
          ImportBinding someImport = imports[i];
          if (someImport.onDemand) {
            Binding resolvedImport = someImport.resolvedImport;
            ReferenceBinding temp = null;
            if ((resolvedImport instanceof PackageBinding)) {
              temp = findType(name, (PackageBinding)resolvedImport, currentPackage);
            } else if (someImport.isStatic()) {
              temp = findMemberType(name, (ReferenceBinding)resolvedImport);
              if ((temp != null) && (!temp.isStatic()))
                temp = null;
            } else {
              temp = findDirectMemberType(name, (ReferenceBinding)resolvedImport);
            }
            if ((temp != type) && (temp != null)) {
              if (temp.isValidBinding()) {
                ImportReference importReference = someImport.reference;
                if (importReference != null) {
                  importReference.bits |= 2;
                }
                if (foundInImport)
                {
                  temp = new ProblemReferenceBinding(name, type, 3);
                  if (typeOrPackageCache != null)
                    typeOrPackageCache.put(name, temp);
                  return temp;
                }
                type = temp;
                foundInImport = true;
              } else if (foundType == null) {
                foundType = temp;
              }
            }
          }
        }
        if (type != null) {
          if (typeOrPackageCache != null)
            typeOrPackageCache.put(name, type);
          return type;
        }
      }
    }

    unitScope.recordSimpleReference(name);
    if ((mask & 0x10) != 0) {
      PackageBinding packageBinding = unitScope.environment.getTopLevelPackage(name);
      if (packageBinding != null) {
        if (typeOrPackageCache != null)
          typeOrPackageCache.put(name, packageBinding);
        return packageBinding;
      }

    }

    if (foundType == null) {
      foundType = new ProblemReferenceBinding(name, null, 1);
      if ((typeOrPackageCache != null) && ((mask & 0x10) != 0))
        typeOrPackageCache.put(name, foundType);
    }
    return foundType;
  }

  public final Binding getTypeOrPackage(char[][] compoundName)
  {
    int nameLength = compoundName.length;
    if (nameLength == 1) {
      TypeBinding binding = getBaseType(compoundName[0]);
      if (binding != null) return binding;
    }
    Binding binding = getTypeOrPackage(compoundName[0], 20);
    if (!binding.isValidBinding()) return binding;

    int currentIndex = 1;
    boolean checkVisibility = false;
    if ((binding instanceof PackageBinding)) {
      PackageBinding packageBinding = (PackageBinding)binding;

      while (currentIndex < nameLength) {
        binding = packageBinding.getTypeOrPackage(compoundName[(currentIndex++)]);
        if (binding == null)
          return new ProblemReferenceBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            null, 
            1);
        if (!binding.isValidBinding())
          return new ProblemReferenceBinding(
            CharOperation.subarray(compoundName, 0, currentIndex), 
            (binding instanceof ReferenceBinding) ? ((ReferenceBinding)binding).closestMatch() : null, 
            binding.problemId());
        if (!(binding instanceof PackageBinding))
          break;
        packageBinding = (PackageBinding)binding;
      }
      if ((binding instanceof PackageBinding)) return binding;
      checkVisibility = true;
    }

    ReferenceBinding typeBinding = (ReferenceBinding)binding;
    ReferenceBinding qualifiedType = (ReferenceBinding)environment().convertToRawType(typeBinding);

    if ((checkVisibility) && 
      (!typeBinding.canBeSeenBy(this))) {
      return new ProblemReferenceBinding(
        CharOperation.subarray(compoundName, 0, currentIndex), 
        typeBinding, 
        2);
    }
    while (currentIndex < nameLength) {
      typeBinding = getMemberType(compoundName[(currentIndex++)], typeBinding);

      if (!typeBinding.isValidBinding()) {
        return new ProblemReferenceBinding(
          CharOperation.subarray(compoundName, 0, currentIndex), 
          ((ReferenceBinding)binding).closestMatch(), 
          typeBinding.problemId());
      }
      if (typeBinding.isGenericType())
        qualifiedType = environment().createRawType(typeBinding, qualifiedType);
      else {
        qualifiedType = (qualifiedType != null) && ((qualifiedType.isRawType()) || (qualifiedType.isParameterizedType())) ? 
          environment().createParameterizedType(typeBinding, null, qualifiedType) : 
          typeBinding;
      }
    }
    return qualifiedType;
  }

  protected boolean hasErasedCandidatesCollisions(TypeBinding one, TypeBinding two, Map invocations, ReferenceBinding type, ASTNode typeRef) {
    invocations.clear();
    TypeBinding[] mecs = minimalErasedCandidates(new TypeBinding[] { one, two }, invocations);
    if (mecs != null) {
      int k = 0; for (int max = mecs.length; k < max; k++) {
        TypeBinding mec = mecs[k];
        if (mec != null) {
          Object value = invocations.get(mec);
          if ((value instanceof TypeBinding[])) {
            TypeBinding[] invalidInvocations = (TypeBinding[])value;
            problemReporter().superinterfacesCollide(invalidInvocations[0].erasure(), typeRef, invalidInvocations[0], invalidInvocations[1]);
            type.tagBits |= 131072L;
            return true;
          }
        }
      }
    }
    return false;
  }

  public CaseStatement innermostSwitchCase()
  {
    Scope scope = this;
    do {
      if ((scope instanceof BlockScope))
        return ((BlockScope)scope).enclosingCase;
      scope = scope.parent;
    }while (scope != null);
    return null;
  }

  protected boolean isAcceptableMethod(MethodBinding one, MethodBinding two) {
    TypeBinding[] oneParams = one.parameters;
    TypeBinding[] twoParams = two.parameters;
    int oneParamsLength = oneParams.length;
    int twoParamsLength = twoParams.length;
    if (oneParamsLength == twoParamsLength) {
      for (int i = 0; i < oneParamsLength; i++) {
        TypeBinding oneParam = oneParams[i];
        TypeBinding twoParam = twoParams[i];
        if (oneParam == twoParam) {
          if (oneParam.leafComponentType().isRawType())
          {
            if ((oneParam == one.original().parameters[i]) && (oneParam != two.original().parameters[i])) {
              return false;
            }
          }
        }
        else if (oneParam.isCompatibleWith(twoParam)) {
          if (oneParam.leafComponentType().isRawType())
          {
            if (oneParam.needsUncheckedConversion(two.declaringClass.isRawType() ? twoParam : two.original().parameters[i]))
              return false;
          }
        } else {
          if ((i == oneParamsLength - 1) && (one.isVarargs()) && (two.isVarargs())) {
            TypeBinding eType = ((ArrayBinding)twoParam).elementsType();
            if ((oneParam == eType) || (oneParam.isCompatibleWith(eType)))
              return true;
          }
          return false;
        }
      }
      return true;
    }

    if ((one.isVarargs()) && (two.isVarargs())) {
      if (oneParamsLength > twoParamsLength)
      {
        if (((ArrayBinding)twoParams[(twoParamsLength - 1)]).elementsType().id != 1) {
          return false;
        }
      }
      for (int i = (oneParamsLength > twoParamsLength ? twoParamsLength : oneParamsLength) - 2; i >= 0; i--)
        if ((oneParams[i] != twoParams[i]) && (!oneParams[i].isCompatibleWith(twoParams[i])))
          return false;
      if ((parameterCompatibilityLevel(one, twoParams) == -1) && 
        (parameterCompatibilityLevel(two, oneParams) == 2))
        return true;
    }
    return false;
  }

  public boolean isBoxingCompatibleWith(TypeBinding expressionType, TypeBinding targetType) {
    LookupEnvironment environment = environment();
    if ((environment.globalOptions.sourceLevel < 3211264L) || (expressionType.isBaseType() == targetType.isBaseType())) {
      return false;
    }

    TypeBinding convertedType = environment.computeBoxingType(expressionType);
    return (convertedType == targetType) || (convertedType.isCompatibleWith(targetType));
  }

  public final boolean isDefinedInField(FieldBinding field)
  {
    Scope scope = this;
    do {
      if ((scope instanceof MethodScope)) {
        MethodScope methodScope = (MethodScope)scope;
        if (methodScope.initializedField == field) return true;
      }
      scope = scope.parent;
    }while (scope != null);
    return false;
  }

  public final boolean isDefinedInMethod(MethodBinding method)
  {
    Scope scope = this;
    do {
      if ((scope instanceof MethodScope)) {
        ReferenceContext refContext = ((MethodScope)scope).referenceContext;
        if (((refContext instanceof AbstractMethodDeclaration)) && 
          (((AbstractMethodDeclaration)refContext).binding == method))
          return true;
      }
      scope = scope.parent;
    }while (scope != null);
    return false;
  }

  public final boolean isDefinedInSameUnit(ReferenceBinding type)
  {
    ReferenceBinding enclosingType = type;
    while ((type = enclosingType.enclosingType()) != null) {
      enclosingType = type;
    }

    Scope unitScope = this;
    Scope scope;
    while ((scope = unitScope.parent) != null)
    {
      Scope scope;
      unitScope = scope;
    }

    SourceTypeBinding[] topLevelTypes = ((CompilationUnitScope)unitScope).topLevelTypes;
    int i = topLevelTypes.length;
    do { if (topLevelTypes[i] == enclosingType)
        return true;
      i--; } while (i >= 0);

    return false;
  }

  public final boolean isDefinedInType(ReferenceBinding type)
  {
    Scope scope = this;
    do {
      if (((scope instanceof ClassScope)) && 
        (((ClassScope)scope).referenceContext.binding == type))
        return true;
      scope = scope.parent;
    }while (scope != null);
    return false;
  }

  public boolean isInsideCase(CaseStatement caseStatement)
  {
    Scope scope = this;
    do {
      switch (scope.kind) {
      case 1:
        if (((BlockScope)scope).enclosingCase == caseStatement)
          return true;
        break;
      }
      scope = scope.parent;
    }
    while (scope != null);
    return false;
  }

  public boolean isInsideDeprecatedCode() {
    switch (this.kind) {
    case 1:
    case 2:
      MethodScope methodScope = methodScope();
      if (!methodScope.isInsideInitializer())
      {
        MethodBinding context = ((AbstractMethodDeclaration)methodScope.referenceContext).binding;
        if ((context != null) && (context.isViewedAsDeprecated()))
          return true;
      } else {
        SourceTypeBinding type = ((BlockScope)this).referenceType().binding;

        if ((methodScope.initializedField != null) && (methodScope.initializedField.isViewedAsDeprecated()))
          return true;
        if (type != null) {
          type.initializeDeprecatedAnnotationTagBits();
          if (type.isViewedAsDeprecated())
            return true;
        }
      }
      break;
    case 3:
      ReferenceBinding context = ((ClassScope)this).referenceType().binding;
      if (context != null) {
        context.initializeDeprecatedAnnotationTagBits();
        if (context.isViewedAsDeprecated()) {
          return true;
        }
      }
      break;
    case 4:
      CompilationUnitDeclaration unit = referenceCompilationUnit();
      if ((unit.types != null) && (unit.types.length > 0)) {
        SourceTypeBinding type = unit.types[0].binding;
        if (type != null) {
          type.initializeDeprecatedAnnotationTagBits();
          if (type.isViewedAsDeprecated())
            return true; 
        }
      }
      break;
    }
    return false;
  }

  private TypeBinding leastContainingInvocation(TypeBinding mec, Object invocationData, List lubStack) {
    if (invocationData == null) return mec;
    if ((invocationData instanceof TypeBinding)) {
      return (TypeBinding)invocationData;
    }
    TypeBinding[] invocations = (TypeBinding[])invocationData;

    int dim = mec.dimensions();
    mec = mec.leafComponentType();

    int argLength = mec.typeVariables().length;
    if (argLength == 0) return mec;

    TypeBinding[] bestArguments = new TypeBinding[argLength];
    int i = 0; for (int length = invocations.length; i < length; i++) {
      TypeBinding invocation = invocations[i].leafComponentType();
      switch (invocation.kind()) {
      case 2052:
        TypeVariableBinding[] invocationVariables = invocation.typeVariables();
        for (int j = 0; j < argLength; j++) {
          TypeBinding bestArgument = leastContainingTypeArgument(bestArguments[j], invocationVariables[j], (ReferenceBinding)mec, j, lubStack);
          if (bestArgument == null) return null;
          bestArguments[j] = bestArgument;
        }
        break;
      case 260:
        ParameterizedTypeBinding parameterizedType = (ParameterizedTypeBinding)invocation;
        for (int j = 0; j < argLength; j++) {
          TypeBinding bestArgument = leastContainingTypeArgument(bestArguments[j], parameterizedType.arguments[j], (ReferenceBinding)mec, j, lubStack);
          if (bestArgument == null) return null;
          bestArguments[j] = bestArgument;
        }
        break;
      case 1028:
        return dim == 0 ? invocation : environment().createArrayType(invocation, dim);
      }
    }
    TypeBinding least = environment().createParameterizedType((ReferenceBinding)mec.erasure(), bestArguments, mec.enclosingType());
    return dim == 0 ? least : environment().createArrayType(least, dim);
  }

  private TypeBinding leastContainingTypeArgument(TypeBinding u, TypeBinding v, ReferenceBinding genericType, int rank, List lubStack)
  {
    if (u == null) return v;
    if (u == v) return u;
    if (v.isWildcard()) {
      WildcardBinding wildV = (WildcardBinding)v;
      if (u.isWildcard()) {
        WildcardBinding wildU = (WildcardBinding)u;
        switch (wildU.boundKind)
        {
        case 1:
          switch (wildV.boundKind)
          {
          case 1:
            TypeBinding lub = lowerUpperBound(new TypeBinding[] { wildU.bound, wildV.bound }, lubStack);
            if (lub == null) return null;

            if (lub == TypeBinding.INT) return environment().createWildcard(genericType, rank, null, null, 0);
            return environment().createWildcard(genericType, rank, lub, null, 1);
          case 2:
            if (wildU.bound == wildV.bound) return wildU.bound;
            return environment().createWildcard(genericType, rank, null, null, 0);
          }
          break;
        case 2:
          if (wildU.boundKind != 2) break;
          TypeBinding[] glb = greaterLowerBound(new TypeBinding[] { wildU.bound, wildV.bound });
          if (glb == null) return null;
          return environment().createWildcard(genericType, rank, glb[0], null, 2);
        default:
          break;
        }
      } else { switch (wildV.boundKind)
        {
        case 1:
          TypeBinding lub = lowerUpperBound(new TypeBinding[] { u, wildV.bound }, lubStack);
          if (lub == null) return null;

          if (lub == TypeBinding.INT) return environment().createWildcard(genericType, rank, null, null, 0);
          return environment().createWildcard(genericType, rank, lub, null, 1);
        case 2:
          TypeBinding[] glb = greaterLowerBound(new TypeBinding[] { u, wildV.bound });
          if (glb == null) return null;
          return environment().createWildcard(genericType, rank, glb[0], null, 2);
        case 0:
        } }
    }
    else if (u.isWildcard()) {
      WildcardBinding wildU = (WildcardBinding)u;
      switch (wildU.boundKind)
      {
      case 1:
        TypeBinding lub = lowerUpperBound(new TypeBinding[] { wildU.bound, v }, lubStack);
        if (lub == null) return null;

        if (lub == TypeBinding.INT) return environment().createWildcard(genericType, rank, null, null, 0);
        return environment().createWildcard(genericType, rank, lub, null, 1);
      case 2:
        TypeBinding[] glb = greaterLowerBound(new TypeBinding[] { wildU.bound, v });
        if (glb == null) return null;
        return environment().createWildcard(genericType, rank, glb[0], null, 2);
      case 0:
      }
    }
    TypeBinding lub = lowerUpperBound(new TypeBinding[] { u, v }, lubStack);
    if (lub == null) return null;

    if (lub == TypeBinding.INT) return environment().createWildcard(genericType, rank, null, null, 0);
    return environment().createWildcard(genericType, rank, lub, null, 1);
  }

  public TypeBinding lowerUpperBound(TypeBinding[] types)
  {
    int typeLength = types.length;
    if (typeLength == 1) {
      TypeBinding type = types[0];
      return type == null ? TypeBinding.VOID : type;
    }
    return lowerUpperBound(types, new ArrayList(1));
  }

  private TypeBinding lowerUpperBound(TypeBinding[] types, List lubStack)
  {
    int typeLength = types.length;
    if (typeLength == 1) {
      TypeBinding type = types[0];
      return type == null ? TypeBinding.VOID : type;
    }

    int stackLength = lubStack.size();
    for (int i = 0; i < stackLength; i++) {
      TypeBinding[] lubTypes = (TypeBinding[])lubStack.get(i);
      int lubTypeLength = lubTypes.length;
      if (lubTypeLength >= typeLength) {
        for (int j = 0; j < typeLength; j++) {
          TypeBinding type = types[j];
          if (type != null) {
            for (int k = 0; k < lubTypeLength; k++) {
              TypeBinding lubType = lubTypes[k];
              if ((lubType != null) && (
                (lubType == type) || (lubType.isEquivalentTo(type)))) break;
            }
            break;
          }
        }
        return TypeBinding.INT;
      }
    }
    lubStack.add(types);
    Map invocations = new HashMap(1);
    TypeBinding[] mecs = minimalErasedCandidates(types, invocations);
    if (mecs == null) return null;
    int length = mecs.length;
    if (length == 0) return TypeBinding.VOID;
    int count = 0;
    TypeBinding firstBound = null;
    int commonDim = -1;
    for (int i = 0; i < length; i++) {
      TypeBinding mec = mecs[i];
      if (mec != null) {
        mec = leastContainingInvocation(mec, invocations.get(mec), lubStack);
        if (mec == null) return null;
        int dim = mec.dimensions();
        if (commonDim == -1)
          commonDim = dim;
        else if (dim != commonDim) {
          return null;
        }
        if ((firstBound == null) && (!mec.leafComponentType().isInterface())) firstBound = mec.leafComponentType();
        mecs[(count++)] = mec;
      }
    }
    switch (count) { case 0:
      return TypeBinding.VOID;
    case 1:
      return mecs[0];
    case 2:
      if ((commonDim == 0 ? mecs[1].id : mecs[1].leafComponentType().id) == 1) return mecs[0];
      if ((commonDim == 0 ? mecs[0].id : mecs[0].leafComponentType().id) == 1) return mecs[1]; break;
    }
    TypeBinding[] otherBounds = new TypeBinding[count - 1];
    int rank = 0;
    for (int i = 0; i < count; i++) {
      TypeBinding mec = commonDim == 0 ? mecs[i] : mecs[i].leafComponentType();
      if (mec.isInterface()) {
        otherBounds[(rank++)] = mec;
      }
    }
    TypeBinding intersectionType = environment().createWildcard(null, 0, firstBound, otherBounds, 1);
    return commonDim == 0 ? intersectionType : environment().createArrayType(intersectionType, commonDim);
  }

  public final MethodScope methodScope() {
    Scope scope = this;
    do {
      if ((scope instanceof MethodScope))
        return (MethodScope)scope;
      scope = scope.parent;
    }while (scope != null);
    return null;
  }

  protected TypeBinding[] minimalErasedCandidates(TypeBinding[] types, Map allInvocations)
  {
    int length = types.length;
    int indexOfFirst = -1; int actualLength = 0;
    for (int i = 0; i < length; i++) {
      TypeBinding type = types[i];
      if (type != null) {
        if (type.isBaseType()) return null;
        if (indexOfFirst < 0) indexOfFirst = i;
        actualLength++;
      }
    }
    switch (actualLength) { case 0:
      return Binding.NO_TYPES;
    case 1:
      return types;
    }
    TypeBinding firstType = types[indexOfFirst];
    if (firstType.isBaseType()) return null;

    ArrayList typesToVisit = new ArrayList(5);

    int dim = firstType.dimensions();
    TypeBinding leafType = firstType.leafComponentType();
    TypeBinding firstErasure = (leafType.isTypeVariable()) || (leafType.isWildcard()) ? firstType : firstType.erasure();
    if (firstErasure != firstType) {
      allInvocations.put(firstErasure, firstType);
    }
    typesToVisit.add(firstType);
    int max = 1;

    for (int i = 0; i < max; i++) {
      TypeBinding typeToVisit = (TypeBinding)typesToVisit.get(i);
      dim = typeToVisit.dimensions();
      if (dim > 0) {
        leafType = typeToVisit.leafComponentType();
        switch (leafType.id) {
        case 1:
          if (dim > 1) {
            TypeBinding elementType = ((ArrayBinding)typeToVisit).elementsType();
            if (typesToVisit.contains(elementType)) continue;
            typesToVisit.add(elementType);
            max++;
          }
          break;
        case 2:
        case 3:
        case 4:
        case 5:
        case 7:
        case 8:
        case 9:
        case 10:
          TypeBinding superType = getJavaIoSerializable();
          if (!typesToVisit.contains(superType)) {
            typesToVisit.add(superType);
            max++;
          }
          superType = getJavaLangCloneable();
          if (!typesToVisit.contains(superType)) {
            typesToVisit.add(superType);
            max++;
          }
          superType = getJavaLangObject();
          if (typesToVisit.contains(superType)) continue;
          typesToVisit.add(superType);
          max++;

          break;
        case 6:
        }

        typeToVisit = leafType;
      }
      ReferenceBinding currentType = (ReferenceBinding)typeToVisit;
      if (currentType.isCapture()) {
        TypeBinding firstBound = ((CaptureBinding)currentType).firstBound;
        if ((firstBound != null) && (firstBound.isArrayType())) {
          TypeBinding superType = dim == 0 ? firstBound : environment().createArrayType(firstBound, dim);
          if (typesToVisit.contains(superType)) continue;
          typesToVisit.add(superType);
          max++;
          TypeBinding superTypeErasure = (firstBound.isTypeVariable()) || (firstBound.isWildcard()) ? superType : superType.erasure();
          if (superTypeErasure == superType) continue;
          allInvocations.put(superTypeErasure, superType);

          continue;
        }
      }

      ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
      if (itsInterfaces != null) {
        int j = 0; for (int count = itsInterfaces.length; j < count; j++) {
          TypeBinding itsInterface = itsInterfaces[j];
          TypeBinding superType = dim == 0 ? itsInterface : environment().createArrayType(itsInterface, dim);
          if (!typesToVisit.contains(superType)) {
            typesToVisit.add(superType);
            max++;
            TypeBinding superTypeErasure = (itsInterface.isTypeVariable()) || (itsInterface.isWildcard()) ? superType : superType.erasure();
            if (superTypeErasure != superType) {
              allInvocations.put(superTypeErasure, superType);
            }
          }
        }
      }
      TypeBinding itsSuperclass = currentType.superclass();
      if (itsSuperclass != null) {
        TypeBinding superType = dim == 0 ? itsSuperclass : environment().createArrayType(itsSuperclass, dim);
        if (!typesToVisit.contains(superType)) {
          typesToVisit.add(superType);
          max++;
          TypeBinding superTypeErasure = (itsSuperclass.isTypeVariable()) || (itsSuperclass.isWildcard()) ? superType : superType.erasure();
          if (superTypeErasure != superType) {
            allInvocations.put(superTypeErasure, superType);
          }
        }
      }
    }
    int superLength = typesToVisit.size();
    TypeBinding[] erasedSuperTypes = new TypeBinding[superLength];
    int rank = 0;
    for (Iterator iter = typesToVisit.iterator(); iter.hasNext(); ) {
      TypeBinding type = (TypeBinding)iter.next();
      leafType = type.leafComponentType();
      erasedSuperTypes[(rank++)] = ((leafType.isTypeVariable()) || (leafType.isWildcard()) ? type : type.erasure());
    }

    int remaining = superLength;
    for (int i = indexOfFirst + 1; i < length; i++) {
      TypeBinding otherType = types[i];
      if (otherType != null)
        if (otherType.isArrayType())
          for (int j = 0; j < superLength; j++) {
            TypeBinding erasedSuperType = erasedSuperTypes[j];
            if ((erasedSuperType != null) && (erasedSuperType != otherType))
            {
              TypeBinding match;
              if ((match = otherType.findSuperTypeWithSameErasure(erasedSuperType)) == null) {
                erasedSuperTypes[j] = null;
                remaining--; if (remaining == 0) return null;
              }
              else
              {
                Object invocationData = allInvocations.get(erasedSuperType);
                if (invocationData == null) {
                  allInvocations.put(erasedSuperType, match);
                } else if ((invocationData instanceof TypeBinding)) {
                  if (match != invocationData)
                  {
                    TypeBinding[] someInvocations = { (TypeBinding)invocationData, match };
                    allInvocations.put(erasedSuperType, someInvocations);
                  }
                } else {
                  TypeBinding[] someInvocations = (TypeBinding[])invocationData;

                  int invocLength = someInvocations.length;
                  int k = 0;
                  while (someInvocations[k] != match)
                  {
                    k++; if (k >= invocLength)
                    {
                      System.arraycopy(someInvocations, 0, someInvocations = new TypeBinding[invocLength + 1], 0, invocLength);
                      allInvocations.put(erasedSuperType, someInvocations);
                      someInvocations[invocLength] = match; } 
                  }
                }
              }
            }
          }
        else for (int j = 0; j < superLength; j++) {
            TypeBinding erasedSuperType = erasedSuperTypes[j];
            if (erasedSuperType != null)
            {
              TypeBinding match;
              TypeBinding match;
              if ((erasedSuperType == otherType) || ((erasedSuperType.id == 1) && (otherType.isInterface()))) {
                match = erasedSuperType;
              }
              else
              {
                TypeBinding match;
                if (erasedSuperType.isArrayType())
                  match = null;
                else {
                  match = otherType.findSuperTypeWithSameErasure(erasedSuperType);
                }
                if (match == null) {
                  erasedSuperTypes[j] = null;
                  remaining--; if (remaining != 0) continue; return null;
                }

              }

              Object invocationData = allInvocations.get(erasedSuperType);
              if (invocationData == null) {
                allInvocations.put(erasedSuperType, match);
              } else if ((invocationData instanceof TypeBinding)) {
                if (match != invocationData)
                {
                  TypeBinding[] someInvocations = { (TypeBinding)invocationData, match };
                  allInvocations.put(erasedSuperType, someInvocations);
                }
              } else {
                TypeBinding[] someInvocations = (TypeBinding[])invocationData;

                int invocLength = someInvocations.length;
                int k = 0;
                while (someInvocations[k] != match)
                {
                  k++; if (k >= invocLength)
                  {
                    System.arraycopy(someInvocations, 0, someInvocations = new TypeBinding[invocLength + 1], 0, invocLength);
                    allInvocations.put(erasedSuperType, someInvocations);
                    someInvocations[invocLength] = match;
                  }
                }
              }
            }
          } 
    }
    if (remaining > 1) {
      for (int i = 0; i < superLength; i++) {
        TypeBinding erasedSuperType = erasedSuperTypes[i];
        if (erasedSuperType != null) {
          for (int j = 0; j < superLength; j++)
            if (i != j) {
              TypeBinding otherType = erasedSuperTypes[j];
              if (otherType != null)
                if ((erasedSuperType instanceof ReferenceBinding)) {
                  if (((otherType.id != 1) || (!erasedSuperType.isInterface())) && 
                    (erasedSuperType.findSuperTypeWithSameErasure(otherType) != null)) {
                    erasedSuperTypes[j] = null;
                    remaining--;
                  }
                } else if ((erasedSuperType.isArrayType()) && 
                  ((!otherType.isArrayType()) || 
                  (otherType.leafComponentType().id != 1) || 
                  (otherType.dimensions() != erasedSuperType.dimensions()) || 
                  (!erasedSuperType.leafComponentType().isInterface())) && 
                  (erasedSuperType.findSuperTypeWithSameErasure(otherType) != null)) {
                  erasedSuperTypes[j] = null;
                  remaining--;
                }
            }
        }
      }
    }
    return erasedSuperTypes;
  }

  protected final MethodBinding mostSpecificClassMethodBinding(MethodBinding[] visible, int visibleSize, InvocationSite invocationSite)
  {
    MethodBinding previous = null;
    for (int i = 0; i < visibleSize; i++) {
      MethodBinding method = visible[i];
      if ((previous != null) && (method.declaringClass != previous.declaringClass)) {
        break;
      }
      if (!method.isStatic()) previous = method;
      int j = 0;
      while ((i == j) || 
        (visible[j].areParametersCompatibleWith(method.parameters)))
      {
        j++; if (j >= visibleSize)
        {
          compilationUnitScope().recordTypeReferences(method.thrownExceptions);
          return method;
        }
      }
    }
    return new ProblemMethodBinding(visible[0], visible[0].selector, visible[0].parameters, 3);
  }

  protected final MethodBinding mostSpecificInterfaceMethodBinding(MethodBinding[] visible, int visibleSize, InvocationSite invocationSite)
  {
    for (int i = 0; i < visibleSize; i++) {
      MethodBinding method = visible[i];
      int j = 0;
      while ((i == j) || 
        (visible[j].areParametersCompatibleWith(method.parameters)))
      {
        j++; if (j >= visibleSize)
        {
          compilationUnitScope().recordTypeReferences(method.thrownExceptions);
          return method;
        }
      }
    }
    return new ProblemMethodBinding(visible[0], visible[0].selector, visible[0].parameters, 3);
  }

  protected final MethodBinding mostSpecificMethodBinding(MethodBinding[] visible, int visibleSize, TypeBinding[] argumentTypes, InvocationSite invocationSite, ReferenceBinding receiverType)
  {
    int[] compatibilityLevels = new int[visibleSize];
    for (int i = 0; i < visibleSize; i++) {
      compatibilityLevels[i] = parameterCompatibilityLevel(visible[i], argumentTypes);
    }
    boolean useTiebreakMethod = invocationSite.genericTypeArguments() == null;
    MethodBinding[] moreSpecific = new MethodBinding[visibleSize];
    int count = 0;
    int level = 0; for (int max = 2; level <= max; level++) {
      for (int i = 0; i < visibleSize; i++)
        if (compatibilityLevels[i] == level) {
          max = level;
          MethodBinding current = visible[i];
          MethodBinding original = current.original();
          MethodBinding tiebreakMethod = useTiebreakMethod ? current.tiebreakMethod() : current;
          for (int j = 0; j < visibleSize; j++)
            if ((i != j) && (compatibilityLevels[j] == level)) {
              MethodBinding next = visible[j];
              if (original == next.original())
              {
                compatibilityLevels[j] = -1;
              }
              else
              {
                MethodBinding methodToTest = next;
                if ((next instanceof ParameterizedGenericMethodBinding)) {
                  ParameterizedGenericMethodBinding pNext = (ParameterizedGenericMethodBinding)next;
                  if ((!pNext.isRaw) || (pNext.isStatic()))
                  {
                    methodToTest = pNext.originalMethod;
                  }
                }
                MethodBinding acceptable = computeCompatibleMethod(methodToTest, tiebreakMethod.parameters, invocationSite);

                if ((acceptable == null) || (!acceptable.isValidBinding()))
                  break;
                if (!isAcceptableMethod(tiebreakMethod, acceptable)) {
                  break;
                }
                if ((current.isBridge()) && (!next.isBridge()) && 
                  (tiebreakMethod.areParametersEqual(acceptable))) break;
              }
            }
          moreSpecific[i] = current;
          count++;
        }
    }
    if (count == 1) {
      for (int i = 0; i < visibleSize; i++)
        if (moreSpecific[i] != null) {
          compilationUnitScope().recordTypeReferences(visible[i].thrownExceptions);
          return visible[i];
        }
    }
    else if (count == 0) {
      return new ProblemMethodBinding(visible[0], visible[0].selector, visible[0].parameters, 3);
    }

    if (receiverType != null)
      receiverType = (receiverType instanceof CaptureBinding) ? receiverType : (ReferenceBinding)receiverType.erasure();
    for (int i = 0; i < visibleSize; i++) {
      MethodBinding current = moreSpecific[i];
      if (current != null) {
        ReferenceBinding[] mostSpecificExceptions = (ReferenceBinding[])null;
        SimpleSet possibleMethods = null;
        MethodBinding original = current.original();
        for (int j = 0; j < visibleSize; j++) {
          MethodBinding next = moreSpecific[j];
          if ((next != null) && (i != j)) {
            MethodBinding original2 = next.original();
            if (original.declaringClass == original2.declaringClass) {
              break label1305;
            }
            if (!original.isAbstract()) {
              if (!original2.isAbstract())
              {
                TypeBinding superType = original.declaringClass.findSuperTypeWithSameErasure(original2.declaringClass.erasure());
                if (superType == null)
                  break;
                if ((current.hasSubstitutedParameters()) || (original.typeVariables != Binding.NO_TYPE_VARIABLES)) {
                  if (original2.declaringClass != superType)
                  {
                    MethodBinding[] superMethods = ((ReferenceBinding)superType).getMethods(original2.selector);
                    int m = 0; for (int l = superMethods.length; m < l; m++) {
                      if (superMethods[m].original() == original2) {
                        original2 = superMethods[m];
                        break;
                      }
                    }
                  }
                  if (!environment().methodVerifier().doesMethodOverride(original, original2))
                    break; 
                }
              } } else if (receiverType != null) {
              TypeBinding superType = receiverType.findSuperTypeWithSameErasure(original.declaringClass.erasure());
              if ((original.declaringClass != superType) && ((superType instanceof ReferenceBinding)))
              {
                MethodBinding[] superMethods = ((ReferenceBinding)superType).getMethods(original.selector);
                int m = 0; for (int l = superMethods.length; m < l; m++) {
                  if (superMethods[m].original() == original) {
                    original = superMethods[m];
                    break;
                  }
                }
              }
              superType = receiverType.findSuperTypeWithSameErasure(original2.declaringClass.erasure());
              if ((original2.declaringClass != superType) && ((superType instanceof ReferenceBinding)))
              {
                MethodBinding[] superMethods = ((ReferenceBinding)superType).getMethods(original2.selector);
                int m = 0; for (int l = superMethods.length; m < l; m++) {
                  if (superMethods[m].original() == original2) {
                    original2 = superMethods[m];
                    break;
                  }
                }
              }
              if (original.typeVariables != Binding.NO_TYPE_VARIABLES)
                original2 = original.computeSubstitutedMethod(original2, environment());
              if ((original2 == null) || (!original.areParameterErasuresEqual(original2)))
                break;
              if ((!original.returnType.isCompatibleWith(original2.returnType)) && 
                (!original.returnType.erasure().isCompatibleWith(original2.returnType.erasure())))
              {
                break;
              }
              if (original.thrownExceptions != original2.thrownExceptions) {
                if (mostSpecificExceptions == null)
                  mostSpecificExceptions = original.thrownExceptions;
                if (possibleMethods == null)
                  possibleMethods = new SimpleSet(3);
                int mostSpecificLength = mostSpecificExceptions.length;
                int original2Length = original2.thrownExceptions.length;
                SimpleSet temp = new SimpleSet(mostSpecificLength);
                for (int t = 0; t < mostSpecificLength; t++) {
                  ReferenceBinding exception = mostSpecificExceptions[t];
                  for (int s = 0; s < original2Length; s++) {
                    if (exception.isCompatibleWith(original2.thrownExceptions[s])) {
                      possibleMethods.add(current);
                      temp.add(exception);
                      break;
                    }if (original2.thrownExceptions[s].isCompatibleWith(exception)) {
                      possibleMethods.add(next);
                      temp.add(original2.thrownExceptions[s]);
                      break;
                    }
                  }
                }
                mostSpecificExceptions = temp.elementSize == 0 ? Binding.NO_EXCEPTIONS : new ReferenceBinding[temp.elementSize];
                temp.asArray(mostSpecificExceptions);
              }
            }
          }
        }
        if (mostSpecificExceptions != null) {
          Object[] values = possibleMethods.values;
          int exceptionLength = mostSpecificExceptions.length;
          int p = 0; label1283: for (int vLength = values.length; p < vLength; p++) {
            MethodBinding possible = (MethodBinding)values[p];
            if (possible != null) {
              ReferenceBinding[] itsExceptions = possible.thrownExceptions;
              if (itsExceptions.length == exceptionLength) {
                for (int e = 0; e < exceptionLength; e++) {
                  ReferenceBinding exception = itsExceptions[e];
                  int f = 0;
                  while (exception != mostSpecificExceptions[f])
                  {
                    f++; if (f >= exceptionLength)
                    {
                      break label1283;
                    }
                  }
                }
                return possible;
              }

            }

          }

        }

        return current;
      }

    }

    label1305: return new ProblemMethodBinding(visible[0], visible[0].selector, visible[0].parameters, 3);
  }

  public final ClassScope outerMostClassScope() {
    ClassScope lastClassScope = null;
    Scope scope = this;
    do {
      if ((scope instanceof ClassScope))
        lastClassScope = (ClassScope)scope;
      scope = scope.parent;
    }while (scope != null);
    return lastClassScope;
  }

  public final MethodScope outerMostMethodScope() {
    MethodScope lastMethodScope = null;
    Scope scope = this;
    do {
      if ((scope instanceof MethodScope))
        lastMethodScope = (MethodScope)scope;
      scope = scope.parent;
    }while (scope != null);
    return lastMethodScope;
  }

  public int parameterCompatibilityLevel(MethodBinding method, TypeBinding[] arguments) {
    TypeBinding[] parameters = method.parameters;
    int paramLength = parameters.length;
    int argLength = arguments.length;

    if (compilerOptions().sourceLevel < 3211264L) {
      if (paramLength != argLength)
        return -1;
      for (int i = 0; i < argLength; i++) {
        TypeBinding param = parameters[i];
        TypeBinding arg = arguments[i];
        if ((arg != param) && (!arg.isCompatibleWith(param)))
          return -1;
      }
      return 0;
    }

    int level = 0;
    int lastIndex = argLength;
    LookupEnvironment env = environment();
    if (method.isVarargs()) {
      lastIndex = paramLength - 1;
      if (paramLength == argLength) {
        TypeBinding param = parameters[lastIndex];
        TypeBinding arg = arguments[lastIndex];
        if (param != arg) {
          level = parameterCompatibilityLevel(arg, param, env);
          if (level == -1)
          {
            param = ((ArrayBinding)param).elementsType();
            if (parameterCompatibilityLevel(arg, param, env) == -1)
              return -1;
            level = 2;
          }
        }
      } else {
        if (paramLength < argLength) {
          TypeBinding param = ((ArrayBinding)parameters[lastIndex]).elementsType();
          for (int i = lastIndex; i < argLength; i++) {
            TypeBinding arg = arguments[i];
            if ((param != arg) && (parameterCompatibilityLevel(arg, param, env) == -1))
              return -1;
          }
        } else if (lastIndex != argLength) {
          return -1;
        }
        level = 2;
      }
    } else if (paramLength != argLength) {
      return -1;
    }

    for (int i = 0; i < lastIndex; i++) {
      TypeBinding param = parameters[i];
      TypeBinding arg = arguments[i];
      if (arg != param) {
        int newLevel = parameterCompatibilityLevel(arg, param, env);
        if (newLevel == -1)
          return -1;
        if (newLevel > level)
          level = newLevel;
      }
    }
    return level;
  }

  private int parameterCompatibilityLevel(TypeBinding arg, TypeBinding param, LookupEnvironment env)
  {
    if (arg.isCompatibleWith(param))
      return 0;
    if (arg.isBaseType() != param.isBaseType()) {
      TypeBinding convertedType = env.computeBoxingType(arg);
      if ((convertedType == param) || (convertedType.isCompatibleWith(param)))
        return 1;
    }
    return -1;
  }

  public abstract ProblemReporter problemReporter();

  public final CompilationUnitDeclaration referenceCompilationUnit() {
    Scope unitScope = this;
    Scope scope;
    while ((scope = unitScope.parent) != null)
    {
      Scope scope;
      unitScope = scope;
    }return ((CompilationUnitScope)unitScope).referenceContext;
  }

  public ReferenceContext referenceContext()
  {
    Scope current = this;
    do
      switch (current.kind) {
      case 2:
        return ((MethodScope)current).referenceContext;
      case 3:
        return ((ClassScope)current).referenceContext;
      case 4:
        return ((CompilationUnitScope)current).referenceContext;
      }
    while ((current = current.parent) != null);

    return null;
  }

  int startIndex()
  {
    return 0;
  }
}