package org.eclipse.jdt.internal.compiler.parser;

import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.ArrayQualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.ArrayTypeReference;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.FieldDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Statement;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;

public class RecoveredField extends RecoveredElement
{
  public FieldDeclaration fieldDeclaration;
  boolean alreadyCompletedFieldInitialization;
  public RecoveredType[] anonymousTypes;
  public int anonymousTypeCount;

  public RecoveredField(FieldDeclaration fieldDeclaration, RecoveredElement parent, int bracketBalance)
  {
    this(fieldDeclaration, parent, bracketBalance, null);
  }
  public RecoveredField(FieldDeclaration fieldDeclaration, RecoveredElement parent, int bracketBalance, Parser parser) {
    super(parent, bracketBalance, parser);
    this.fieldDeclaration = fieldDeclaration;
    this.alreadyCompletedFieldInitialization = (fieldDeclaration.initialization != null);
  }

  public RecoveredElement add(Statement statement, int bracketBalanceValue)
  {
    if ((this.alreadyCompletedFieldInitialization) || (!(statement instanceof Expression))) {
      return super.add(statement, bracketBalanceValue);
    }
    this.alreadyCompletedFieldInitialization = true;
    this.fieldDeclaration.initialization = ((Expression)statement);
    this.fieldDeclaration.declarationSourceEnd = statement.sourceEnd;
    this.fieldDeclaration.declarationEnd = statement.sourceEnd;
    return this;
  }

  public RecoveredElement add(TypeDeclaration typeDeclaration, int bracketBalanceValue)
  {
    if ((this.alreadyCompletedFieldInitialization) || 
      ((typeDeclaration.bits & 0x200) == 0) || (
      (this.fieldDeclaration.declarationSourceEnd != 0) && (typeDeclaration.sourceStart > this.fieldDeclaration.declarationSourceEnd))) {
      return super.add(typeDeclaration, bracketBalanceValue);
    }

    if (this.anonymousTypes == null) {
      this.anonymousTypes = new RecoveredType[5];
      this.anonymousTypeCount = 0;
    }
    else if (this.anonymousTypeCount == this.anonymousTypes.length) {
      System.arraycopy(
        this.anonymousTypes, 
        0, 
        this.anonymousTypes = new RecoveredType[2 * this.anonymousTypeCount], 
        0, 
        this.anonymousTypeCount);
    }

    RecoveredType element = new RecoveredType(typeDeclaration, this, bracketBalanceValue);
    this.anonymousTypes[(this.anonymousTypeCount++)] = element;
    return element;
  }

  public ASTNode parseTree()
  {
    return this.fieldDeclaration;
  }

  public int sourceEnd()
  {
    return this.fieldDeclaration.declarationSourceEnd;
  }
  public String toString(int tab) {
    StringBuffer buffer = new StringBuffer(tabString(tab));
    buffer.append("Recovered field:\n");
    this.fieldDeclaration.print(tab + 1, buffer);
    if (this.anonymousTypes != null) {
      for (int i = 0; i < this.anonymousTypeCount; i++) {
        buffer.append("\n");
        buffer.append(this.anonymousTypes[i].toString(tab + 1));
      }
    }
    return buffer.toString();
  }

  public FieldDeclaration updatedFieldDeclaration() {
    if (this.anonymousTypes != null) {
      if (this.fieldDeclaration.initialization == null) {
        for (int i = 0; i < this.anonymousTypeCount; i++) {
          RecoveredType recoveredType = this.anonymousTypes[i];
          TypeDeclaration typeDeclaration = recoveredType.typeDeclaration;
          if (typeDeclaration.declarationSourceEnd == 0) {
            typeDeclaration.declarationSourceEnd = this.fieldDeclaration.declarationSourceEnd;
            typeDeclaration.bodyEnd = this.fieldDeclaration.declarationSourceEnd;
          }
          if (recoveredType.preserveContent) {
            TypeDeclaration anonymousType = recoveredType.updatedTypeDeclaration();
            this.fieldDeclaration.initialization = anonymousType.allocation;
            if (this.fieldDeclaration.declarationSourceEnd == 0) {
              int end = anonymousType.declarationSourceEnd;
              this.fieldDeclaration.declarationSourceEnd = end;
              this.fieldDeclaration.declarationEnd = end;
            }
          }
        }
        if (this.anonymousTypeCount > 0) this.fieldDeclaration.bits |= 2; 
      }
      else if (this.fieldDeclaration.getKind() == 3)
      {
        for (int i = 0; i < this.anonymousTypeCount; i++) {
          RecoveredType recoveredType = this.anonymousTypes[i];
          TypeDeclaration typeDeclaration = recoveredType.typeDeclaration;
          if (typeDeclaration.declarationSourceEnd == 0) {
            typeDeclaration.declarationSourceEnd = this.fieldDeclaration.declarationSourceEnd;
            typeDeclaration.bodyEnd = this.fieldDeclaration.declarationSourceEnd;
          }
          recoveredType.updatedTypeDeclaration();
        }
      }
    }
    return this.fieldDeclaration;
  }

  public RecoveredElement updateOnClosingBrace(int braceStart, int braceEnd)
  {
    if (this.bracketBalance > 0) {
      this.bracketBalance -= 1;
      if (this.bracketBalance == 0) {
        if (this.fieldDeclaration.getKind() == 3) {
          updateSourceEndIfNecessary(braceEnd - 1);
          return this.parent;
        }
        this.alreadyCompletedFieldInitialization = true;
      }

      return this;
    }if (this.bracketBalance == 0) {
      this.alreadyCompletedFieldInitialization = true;
      updateSourceEndIfNecessary(braceEnd - 1);
    }
    if (this.parent != null) {
      return this.parent.updateOnClosingBrace(braceStart, braceEnd);
    }
    return this;
  }

  public RecoveredElement updateOnOpeningBrace(int braceStart, int braceEnd)
  {
    if ((this.fieldDeclaration.declarationSourceEnd == 0) && 
      (((this.fieldDeclaration.type instanceof ArrayTypeReference)) || ((this.fieldDeclaration.type instanceof ArrayQualifiedTypeReference))) && 
      (!this.alreadyCompletedFieldInitialization)) {
      this.bracketBalance += 1;
      return null;
    }
    if ((this.fieldDeclaration.declarationSourceEnd == 0) && 
      (this.fieldDeclaration.getKind() == 3)) {
      this.bracketBalance += 1;
      return null;
    }

    updateSourceEndIfNecessary(braceStart - 1, braceEnd - 1);
    return this.parent.updateOnOpeningBrace(braceStart, braceEnd);
  }
  public void updateParseTree() {
    updatedFieldDeclaration();
  }

  public void updateSourceEndIfNecessary(int bodyStart, int bodyEnd)
  {
    if (this.fieldDeclaration.declarationSourceEnd == 0) {
      this.fieldDeclaration.declarationSourceEnd = bodyEnd;
      this.fieldDeclaration.declarationEnd = bodyEnd;
    }
  }
}