package org.eclipse.jdt.internal.compiler.parser;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.Block;
import org.eclipse.jdt.internal.compiler.ast.FieldDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Initializer;
import org.eclipse.jdt.internal.compiler.ast.LocalDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Statement;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.lookup.BaseTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;

public class RecoveredInitializer extends RecoveredField
  implements TerminalTokens
{
  public RecoveredType[] localTypes;
  public int localTypeCount;
  public RecoveredBlock initializerBody;

  public RecoveredInitializer(FieldDeclaration fieldDeclaration, RecoveredElement parent, int bracketBalance)
  {
    this(fieldDeclaration, parent, bracketBalance, null);
  }
  public RecoveredInitializer(FieldDeclaration fieldDeclaration, RecoveredElement parent, int bracketBalance, Parser parser) {
    super(fieldDeclaration, parent, bracketBalance, parser);
    this.foundOpeningBrace = true;
  }

  public RecoveredElement add(Block nestedBlockDeclaration, int bracketBalanceValue)
  {
    if ((this.fieldDeclaration.declarationSourceEnd > 0) && 
      (nestedBlockDeclaration.sourceStart > this.fieldDeclaration.declarationSourceEnd)) {
      if (this.parent == null) return this;
      return this.parent.add(nestedBlockDeclaration, bracketBalanceValue);
    }

    if (!this.foundOpeningBrace) {
      this.foundOpeningBrace = true;
      this.bracketBalance += 1;
    }
    this.initializerBody = new RecoveredBlock(nestedBlockDeclaration, this, bracketBalanceValue);
    if (nestedBlockDeclaration.sourceEnd == 0) return this.initializerBody;
    return this;
  }

  public RecoveredElement add(FieldDeclaration newFieldDeclaration, int bracketBalanceValue)
  {
    char[][] fieldTypeName;
    if (((newFieldDeclaration.modifiers & 0xFFFFFFEF) != 0) || 
      (newFieldDeclaration.type == null) || (
      ((fieldTypeName = newFieldDeclaration.type.getTypeName()).length == 1) && 
      (CharOperation.equals(fieldTypeName[0], TypeBinding.VOID.sourceName())))) {
      if (this.parent == null) return this;
      updateSourceEndIfNecessary(previousAvailableLineEnd(newFieldDeclaration.declarationSourceStart - 1));
      return this.parent.add(newFieldDeclaration, bracketBalanceValue);
    }
    char[][] fieldTypeName;
    if ((this.fieldDeclaration.declarationSourceEnd > 0) && 
      (newFieldDeclaration.declarationSourceStart > this.fieldDeclaration.declarationSourceEnd)) {
      if (this.parent == null) return this;
      return this.parent.add(newFieldDeclaration, bracketBalanceValue);
    }

    return this;
  }

  public RecoveredElement add(LocalDeclaration localDeclaration, int bracketBalanceValue)
  {
    if ((this.fieldDeclaration.declarationSourceEnd != 0) && 
      (localDeclaration.declarationSourceStart > this.fieldDeclaration.declarationSourceEnd)) {
      if (this.parent == null) return this;
      return this.parent.add(localDeclaration, bracketBalanceValue);
    }

    Block block = new Block(0);
    block.sourceStart = ((Initializer)this.fieldDeclaration).sourceStart;
    RecoveredElement element = add(block, 1);
    return element.add(localDeclaration, bracketBalanceValue);
  }

  public RecoveredElement add(Statement statement, int bracketBalanceValue)
  {
    if ((this.fieldDeclaration.declarationSourceEnd != 0) && 
      (statement.sourceStart > this.fieldDeclaration.declarationSourceEnd)) {
      if (this.parent == null) return this;
      return this.parent.add(statement, bracketBalanceValue);
    }

    Block block = new Block(0);
    block.sourceStart = ((Initializer)this.fieldDeclaration).sourceStart;
    RecoveredElement element = add(block, 1);
    return element.add(statement, bracketBalanceValue);
  }

  public RecoveredElement add(TypeDeclaration typeDeclaration, int bracketBalanceValue)
  {
    if ((this.fieldDeclaration.declarationSourceEnd != 0) && 
      (typeDeclaration.declarationSourceStart > this.fieldDeclaration.declarationSourceEnd)) {
      if (this.parent == null) return this;
      return this.parent.add(typeDeclaration, bracketBalanceValue);
    }
    if ((typeDeclaration.bits & 0x100) != 0)
    {
      Block block = new Block(0);
      block.sourceStart = ((Initializer)this.fieldDeclaration).sourceStart;
      RecoveredElement element = add(block, 1);
      return element.add(typeDeclaration, bracketBalanceValue);
    }
    if (this.localTypes == null) {
      this.localTypes = new RecoveredType[5];
      this.localTypeCount = 0;
    }
    else if (this.localTypeCount == this.localTypes.length) {
      System.arraycopy(
        this.localTypes, 
        0, 
        this.localTypes = new RecoveredType[2 * this.localTypeCount], 
        0, 
        this.localTypeCount);
    }

    RecoveredType element = new RecoveredType(typeDeclaration, this, bracketBalanceValue);
    this.localTypes[(this.localTypeCount++)] = element;

    if (!this.foundOpeningBrace) {
      this.foundOpeningBrace = true;
      this.bracketBalance += 1;
    }
    return element;
  }
  public String toString(int tab) {
    StringBuffer result = new StringBuffer(tabString(tab));
    result.append("Recovered initializer:\n");
    this.fieldDeclaration.print(tab + 1, result);
    if (this.initializerBody != null) {
      result.append("\n");
      result.append(this.initializerBody.toString(tab + 1));
    }
    return result.toString();
  }

  public FieldDeclaration updatedFieldDeclaration() {
    if (this.initializerBody != null) {
      Block block = this.initializerBody.updatedBlock();
      if (block != null) {
        ((Initializer)this.fieldDeclaration).block = block;
      }
      if (this.localTypeCount > 0) this.fieldDeclaration.bits |= 2;
    }

    if (this.fieldDeclaration.sourceEnd == 0) {
      this.fieldDeclaration.sourceEnd = this.fieldDeclaration.declarationSourceEnd;
    }
    return this.fieldDeclaration;
  }

  public RecoveredElement updateOnClosingBrace(int braceStart, int braceEnd)
  {
    if ((--this.bracketBalance <= 0) && (this.parent != null)) {
      updateSourceEndIfNecessary(braceStart, braceEnd);
      return this.parent;
    }
    return this;
  }

  public RecoveredElement updateOnOpeningBrace(int braceStart, int braceEnd)
  {
    this.bracketBalance += 1;
    return this;
  }

  public void updateSourceEndIfNecessary(int braceStart, int braceEnd)
  {
    if (this.fieldDeclaration.declarationSourceEnd == 0) {
      Initializer initializer = (Initializer)this.fieldDeclaration;
      if (parser().rBraceSuccessorStart >= braceEnd) {
        if (initializer.bodyStart < parser().rBraceEnd)
          initializer.declarationSourceEnd = parser().rBraceEnd;
        else {
          initializer.declarationSourceEnd = initializer.bodyStart;
        }
        if (initializer.bodyStart < parser().rBraceStart)
          initializer.bodyEnd = parser().rBraceStart;
        else
          initializer.bodyEnd = initializer.bodyStart;
      }
      else {
        initializer.declarationSourceEnd = braceEnd;
        initializer.bodyEnd = (braceStart - 1);
      }
      if (initializer.block != null)
        initializer.block.sourceEnd = initializer.declarationSourceEnd;
    }
  }
}