package org.eclipse.jdt.internal.compiler.parser;

import java.util.ArrayList;
import java.util.HashMap;
import org.eclipse.jdt.core.IImportDeclaration;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Annotation;
import org.eclipse.jdt.internal.compiler.ast.AnnotationMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.ast.ArrayInitializer;
import org.eclipse.jdt.internal.compiler.ast.ArrayQualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.ArrayTypeReference;
import org.eclipse.jdt.internal.compiler.ast.Block;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ConstructorDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.FieldDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ImportReference;
import org.eclipse.jdt.internal.compiler.ast.Initializer;
import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ParameterizedQualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.ParameterizedSingleTypeReference;
import org.eclipse.jdt.internal.compiler.ast.QualifiedAllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.QualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.SingleTypeReference;
import org.eclipse.jdt.internal.compiler.ast.Statement;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeParameter;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.ast.Wildcard;
import org.eclipse.jdt.internal.compiler.env.ISourceImport;
import org.eclipse.jdt.internal.compiler.env.ISourceType;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.BaseTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.core.CompilationUnitElementInfo;
import org.eclipse.jdt.internal.core.ImportDeclaration;
import org.eclipse.jdt.internal.core.InitializerElementInfo;
import org.eclipse.jdt.internal.core.JavaElement;
import org.eclipse.jdt.internal.core.PackageFragment;
import org.eclipse.jdt.internal.core.SourceAnnotationMethodInfo;
import org.eclipse.jdt.internal.core.SourceField;
import org.eclipse.jdt.internal.core.SourceFieldElementInfo;
import org.eclipse.jdt.internal.core.SourceMethod;
import org.eclipse.jdt.internal.core.SourceMethodElementInfo;
import org.eclipse.jdt.internal.core.SourceType;
import org.eclipse.jdt.internal.core.SourceTypeElementInfo;
import org.eclipse.jdt.internal.core.util.Util;

public class SourceTypeConverter
{
  public static final int FIELD = 1;
  public static final int CONSTRUCTOR = 2;
  public static final int METHOD = 4;
  public static final int MEMBER_TYPE = 8;
  public static final int FIELD_INITIALIZATION = 16;
  public static final int FIELD_AND_METHOD = 7;
  public static final int LOCAL_TYPE = 32;
  public static final int NONE = 0;
  private int flags;
  private CompilationUnitDeclaration unit;
  private Parser parser;
  private ProblemReporter problemReporter;
  private org.eclipse.jdt.internal.compiler.env.ICompilationUnit cu;
  private char[] source;
  private HashMap annotationPositions;
  private boolean has1_5Compliance;
  int namePos;

  private SourceTypeConverter(int flags, ProblemReporter problemReporter)
  {
    this.flags = flags;
    this.problemReporter = problemReporter;
    this.has1_5Compliance = (problemReporter.options.complianceLevel >= 3211264L);
  }

  public static CompilationUnitDeclaration buildCompilationUnit(ISourceType[] sourceTypes, int flags, ProblemReporter problemReporter, CompilationResult compilationResult)
  {
    SourceTypeConverter converter = new SourceTypeConverter(flags, problemReporter);
    try {
      return converter.convert(sourceTypes, compilationResult); } catch (JavaModelException localJavaModelException) {
    }
    return null;
  }

  private CompilationUnitDeclaration convert(ISourceType[] sourceTypes, CompilationResult compilationResult)
    throws JavaModelException
  {
    this.unit = new CompilationUnitDeclaration(this.problemReporter, compilationResult, 0);

    if (sourceTypes.length == 0) return this.unit;
    SourceTypeElementInfo topLevelTypeInfo = (SourceTypeElementInfo)sourceTypes[0];
    org.eclipse.jdt.core.ICompilationUnit cuHandle = topLevelTypeInfo.getHandle().getCompilationUnit();
    this.cu = ((org.eclipse.jdt.internal.compiler.env.ICompilationUnit)cuHandle);
    this.annotationPositions = ((CompilationUnitElementInfo)((JavaElement)this.cu).getElementInfo()).annotationPositions;

    if ((this.has1_5Compliance) && (this.annotationPositions != null) && (this.annotationPositions.size() > 10))
    {
      return new Parser(this.problemReporter, true).dietParse(this.cu, compilationResult);
    }

    int start = topLevelTypeInfo.getNameSourceStart();
    int end = topLevelTypeInfo.getNameSourceEnd();

    String[] packageName = ((PackageFragment)cuHandle.getParent()).names;
    if (packageName.length > 0)
    {
      this.unit.currentPackage = 
        createImportReference(packageName, start, end, false, 0);
    }IImportDeclaration[] importDeclarations = topLevelTypeInfo.getHandle().getCompilationUnit().getImports();
    int importCount = importDeclarations.length;
    this.unit.imports = new ImportReference[importCount];
    for (int i = 0; i < importCount; i++) {
      ImportDeclaration importDeclaration = (ImportDeclaration)importDeclarations[i];
      ISourceImport sourceImport = (ISourceImport)importDeclaration.getElementInfo();
      String nameWithoutStar = importDeclaration.getNameWithoutStar();
      this.unit.imports[i] = 
        createImportReference(Util.splitOn('.', nameWithoutStar, 0, nameWithoutStar.length()), 
        sourceImport.getDeclarationSourceStart(), 
        sourceImport.getDeclarationSourceEnd(), 
        importDeclaration.isOnDemand(), 
        sourceImport.getModifiers());
    }
    try
    {
      int typeCount = sourceTypes.length;
      TypeDeclaration[] types = new TypeDeclaration[typeCount];

      for (int i = 0; i < typeCount; i++) {
        SourceTypeElementInfo typeInfo = (SourceTypeElementInfo)sourceTypes[i];
        types[i] = convert((SourceType)typeInfo.getHandle(), compilationResult);
      }
      this.unit.types = types;
      return this.unit; } catch (AnonymousMemberFound localAnonymousMemberFound) {
    }
    return new Parser(this.problemReporter, true).parse(this.cu, compilationResult);
  }

  private void addIdentifiers(String typeSignature, int start, int endExclusive, int identCount, ArrayList fragments)
  {
    if (identCount == 1)
    {
      char[] identifier;
      typeSignature.getChars(start, endExclusive, identifier = new char[endExclusive - start], 0);
      fragments.add(identifier);
    } else {
      fragments.add(extractIdentifiers(typeSignature, start, endExclusive - 1, identCount));
    }
  }

  private Initializer convert(InitializerElementInfo initializerInfo, CompilationResult compilationResult)
    throws JavaModelException
  {
    Block block = new Block(0);
    Initializer initializer = new Initializer(block, 0);

    int start = initializerInfo.getDeclarationSourceStart();
    int end = initializerInfo.getDeclarationSourceEnd();

    initializer.sourceStart = (initializer.declarationSourceStart = start);
    initializer.sourceEnd = (initializer.declarationSourceEnd = end);
    initializer.modifiers = initializerInfo.getModifiers();

    IJavaElement[] children = initializerInfo.getChildren();
    int typesLength = children.length;
    if (typesLength > 0) {
      Statement[] statements = new Statement[typesLength];
      for (int i = 0; i < typesLength; i++) {
        SourceType type = (SourceType)children[i];
        TypeDeclaration localType = convert(type, compilationResult);
        if ((localType.bits & 0x200) != 0) {
          QualifiedAllocationExpression expression = new QualifiedAllocationExpression(localType);
          expression.type = localType.superclass;
          localType.superclass = null;
          localType.superInterfaces = null;
          localType.allocation = expression;
          statements[i] = expression;
        } else {
          statements[i] = localType;
        }
      }
      block.statements = statements;
    }

    return initializer;
  }

  private FieldDeclaration convert(SourceField fieldHandle, TypeDeclaration type, CompilationResult compilationResult)
    throws JavaModelException
  {
    SourceFieldElementInfo fieldInfo = (SourceFieldElementInfo)fieldHandle.getElementInfo();
    FieldDeclaration field = new FieldDeclaration();

    int start = fieldInfo.getNameSourceStart();
    int end = fieldInfo.getNameSourceEnd();

    field.name = fieldHandle.getElementName().toCharArray();
    field.sourceStart = start;
    field.sourceEnd = end;
    field.declarationSourceStart = fieldInfo.getDeclarationSourceStart();
    field.declarationSourceEnd = fieldInfo.getDeclarationSourceEnd();
    int modifiers = fieldInfo.getModifiers();
    boolean isEnumConstant = (modifiers & 0x4000) != 0;
    if (isEnumConstant) {
      field.modifiers = (modifiers & 0xFFFFBFFF);
    } else {
      field.modifiers = modifiers;
      field.type = createTypeReference(fieldInfo.getTypeName(), start, end);
    }

    if (this.has1_5Compliance)
    {
      field.annotations = convertAnnotations(fieldHandle);
    }

    if ((this.flags & 0x10) != 0) {
      char[] initializationSource = fieldInfo.getInitializationSource();
      if (initializationSource != null) {
        if (this.parser == null) {
          this.parser = new Parser(this.problemReporter, true);
        }
        this.parser.parse(field, type, this.unit, initializationSource);
      }

    }

    if ((this.flags & 0x20) != 0) {
      IJavaElement[] children = fieldInfo.getChildren();
      int childrenLength = children.length;
      if (childrenLength == 1) {
        field.initialization = convert(children[0], isEnumConstant ? field : null, compilationResult);
      } else if (childrenLength > 1) {
        ArrayInitializer initializer = new ArrayInitializer();
        field.initialization = initializer;
        Expression[] expressions = new Expression[childrenLength];
        initializer.expressions = expressions;
        for (int i = 0; i < childrenLength; i++) {
          expressions[i] = convert(children[i], isEnumConstant ? field : null, compilationResult);
        }
      }
    }
    return field;
  }

  private QualifiedAllocationExpression convert(IJavaElement localType, FieldDeclaration enumConstant, CompilationResult compilationResult) throws JavaModelException {
    TypeDeclaration anonymousLocalTypeDeclaration = convert((SourceType)localType, compilationResult);
    QualifiedAllocationExpression expression = new QualifiedAllocationExpression(anonymousLocalTypeDeclaration);
    expression.type = anonymousLocalTypeDeclaration.superclass;
    anonymousLocalTypeDeclaration.superclass = null;
    anonymousLocalTypeDeclaration.superInterfaces = null;
    anonymousLocalTypeDeclaration.allocation = expression;
    if (enumConstant != null) {
      anonymousLocalTypeDeclaration.modifiers &= -16385;
      expression.enumConstant = enumConstant;
      expression.type = null;
    }
    return expression;
  }

  private AbstractMethodDeclaration convert(SourceMethod methodHandle, SourceMethodElementInfo methodInfo, CompilationResult compilationResult)
    throws JavaModelException
  {
    int start = methodInfo.getNameSourceStart();
    int end = methodInfo.getNameSourceEnd();

    TypeParameter[] typeParams = (TypeParameter[])null;
    if (this.has1_5Compliance)
    {
      char[][] typeParameterNames = methodInfo.getTypeParameterNames();
      if (typeParameterNames != null) {
        int parameterCount = typeParameterNames.length;
        if (parameterCount > 0) {
          char[][][] typeParameterBounds = methodInfo.getTypeParameterBounds();
          typeParams = new TypeParameter[parameterCount];
          for (int i = 0; i < parameterCount; i++) {
            typeParams[i] = createTypeParameter(typeParameterNames[i], typeParameterBounds[i], start, end);
          }
        }
      }
    }

    int modifiers = methodInfo.getModifiers();
    AbstractMethodDeclaration method;
    if (methodInfo.isConstructor()) {
      ConstructorDeclaration decl = new ConstructorDeclaration(compilationResult);
      decl.bits &= -129;
      AbstractMethodDeclaration method = decl;
      decl.typeParameters = typeParams;
    }
    else
    {
      MethodDeclaration decl;
      MethodDeclaration decl;
      if (methodInfo.isAnnotationMethod()) {
        AnnotationMethodDeclaration annotationMethodDeclaration = new AnnotationMethodDeclaration(compilationResult);

        SourceAnnotationMethodInfo annotationMethodInfo = (SourceAnnotationMethodInfo)methodInfo;
        boolean hasDefaultValue = (annotationMethodInfo.defaultValueStart != -1) || (annotationMethodInfo.defaultValueEnd != -1);
        if (((this.flags & 0x10) != 0) && 
          (hasDefaultValue)) {
          char[] defaultValueSource = CharOperation.subarray(getSource(), annotationMethodInfo.defaultValueStart, annotationMethodInfo.defaultValueEnd + 1);
          if (defaultValueSource != null) {
            Expression expression = parseMemberValue(defaultValueSource);
            if (expression != null)
              annotationMethodDeclaration.defaultValue = expression;
          }
          else
          {
            hasDefaultValue = false;
          }
        }

        if (hasDefaultValue)
          modifiers |= 131072;
        decl = annotationMethodDeclaration;
      } else {
        decl = new MethodDeclaration(compilationResult);
      }

      decl.returnType = createTypeReference(methodInfo.getReturnTypeName(), start, end);

      decl.typeParameters = typeParams;

      method = decl;
    }
    method.selector = methodHandle.getElementName().toCharArray();
    boolean isVarargs = (modifiers & 0x80) != 0;
    method.modifiers = (modifiers & 0xFFFFFF7F);
    method.sourceStart = start;
    method.sourceEnd = end;
    method.declarationSourceStart = methodInfo.getDeclarationSourceStart();
    method.declarationSourceEnd = methodInfo.getDeclarationSourceEnd();

    if (this.has1_5Compliance)
    {
      method.annotations = convertAnnotations(methodHandle);
    }

    String[] argumentTypeSignatures = methodHandle.getParameterTypes();
    char[][] argumentNames = methodInfo.getArgumentNames();
    int argumentCount = argumentTypeSignatures == null ? 0 : argumentTypeSignatures.length;
    if (argumentCount > 0) {
      long position = (start << 32) + end;
      method.arguments = new Argument[argumentCount];
      for (int i = 0; i < argumentCount; i++) {
        TypeReference typeReference = createTypeReference(argumentTypeSignatures[i], start, end);
        if ((isVarargs) && (i == argumentCount - 1)) {
          typeReference.bits |= 16384;
        }
        method.arguments[i] = 
          new Argument(
          argumentNames[i], 
          position, 
          typeReference, 
          0);
      }

    }

    char[][] exceptionTypeNames = methodInfo.getExceptionTypeNames();
    int exceptionCount = exceptionTypeNames == null ? 0 : exceptionTypeNames.length;
    if (exceptionCount > 0) {
      method.thrownExceptions = new TypeReference[exceptionCount];
      for (int i = 0; i < exceptionCount; i++) {
        method.thrownExceptions[i] = 
          createTypeReference(exceptionTypeNames[i], start, end);
      }

    }

    if ((this.flags & 0x20) != 0) {
      IJavaElement[] children = methodInfo.getChildren();
      int typesLength = children.length;
      if (typesLength != 0) {
        Statement[] statements = new Statement[typesLength];
        for (int i = 0; i < typesLength; i++) {
          SourceType type = (SourceType)children[i];
          TypeDeclaration localType = convert(type, compilationResult);
          if ((localType.bits & 0x200) != 0) {
            QualifiedAllocationExpression expression = new QualifiedAllocationExpression(localType);
            expression.type = localType.superclass;
            localType.superclass = null;
            localType.superInterfaces = null;
            localType.allocation = expression;
            statements[i] = expression;
          } else {
            statements[i] = localType;
          }
        }
        method.statements = statements;
      }
    }

    return method;
  }

  private TypeDeclaration convert(SourceType typeHandle, CompilationResult compilationResult)
    throws JavaModelException
  {
    SourceTypeElementInfo typeInfo = (SourceTypeElementInfo)typeHandle.getElementInfo();
    if (typeInfo.isAnonymousMember()) {
      throw new AnonymousMemberFound();
    }
    TypeDeclaration type = new TypeDeclaration(compilationResult);
    if (typeInfo.getEnclosingType() == null) {
      if (typeHandle.isAnonymous()) {
        type.name = CharOperation.NO_CHAR;
        type.bits |= 768;
      }
      else if (typeHandle.isLocal()) {
        type.bits |= 256;
      }
    }
    else {
      type.bits |= 1024;
    }
    if ((type.bits & 0x200) == 0) {
      type.name = typeInfo.getName();
    }
    type.name = typeInfo.getName();
    int start;
    type.sourceStart = (start = typeInfo.getNameSourceStart());
    int end;
    type.sourceEnd = (end = typeInfo.getNameSourceEnd());
    type.modifiers = typeInfo.getModifiers();
    type.declarationSourceStart = typeInfo.getDeclarationSourceStart();
    type.declarationSourceEnd = typeInfo.getDeclarationSourceEnd();
    type.bodyEnd = type.declarationSourceEnd;

    if (this.has1_5Compliance)
    {
      type.annotations = convertAnnotations(typeHandle);

      char[][] typeParameterNames = typeInfo.getTypeParameterNames();
      if (typeParameterNames.length > 0) {
        int parameterCount = typeParameterNames.length;
        char[][][] typeParameterBounds = typeInfo.getTypeParameterBounds();
        type.typeParameters = new TypeParameter[parameterCount];
        for (int i = 0; i < parameterCount; i++) {
          type.typeParameters[i] = createTypeParameter(typeParameterNames[i], typeParameterBounds[i], start, end);
        }
      }

    }

    if (typeInfo.getSuperclassName() != null) {
      type.superclass = createTypeReference(typeInfo.getSuperclassName(), start, end);
      type.superclass.bits |= 16;
    }
    char[][] interfaceNames = typeInfo.getInterfaceNames();
    int interfaceCount = interfaceNames == null ? 0 : interfaceNames.length;
    if (interfaceCount > 0) {
      type.superInterfaces = new TypeReference[interfaceCount];
      for (int i = 0; i < interfaceCount; i++) {
        type.superInterfaces[i] = createTypeReference(interfaceNames[i], start, end);
        type.superInterfaces[i].bits |= 16;
      }
    }

    if ((this.flags & 0x8) != 0) {
      SourceType[] sourceMemberTypes = typeInfo.getMemberTypeHandles();
      int sourceMemberTypeCount = sourceMemberTypes.length;
      type.memberTypes = new TypeDeclaration[sourceMemberTypeCount];
      for (int i = 0; i < sourceMemberTypeCount; i++) {
        type.memberTypes[i] = convert(sourceMemberTypes[i], compilationResult);
      }

    }

    InitializerElementInfo[] initializers = (InitializerElementInfo[])null;
    int initializerCount = 0;
    if ((this.flags & 0x20) != 0) {
      initializers = typeInfo.getInitializers();
      initializerCount = initializers.length;
    }
    SourceField[] sourceFields = (SourceField[])null;
    int sourceFieldCount = 0;
    if ((this.flags & 0x1) != 0) {
      sourceFields = typeInfo.getFieldHandles();
      sourceFieldCount = sourceFields.length;
    }
    int length = initializerCount + sourceFieldCount;
    if (length > 0) {
      type.fields = new FieldDeclaration[length];
      for (int i = 0; i < initializerCount; i++) {
        type.fields[i] = convert(initializers[i], compilationResult);
      }
      int index = 0;
      for (int i = initializerCount; i < length; i++) {
        type.fields[i] = convert(sourceFields[(index++)], type, compilationResult);
      }

    }

    boolean needConstructor = (this.flags & 0x2) != 0;
    boolean needMethod = (this.flags & 0x4) != 0;
    if ((needConstructor) || (needMethod))
    {
      SourceMethod[] sourceMethods = typeInfo.getMethodHandles();
      int sourceMethodCount = sourceMethods.length;

      int extraConstructor = 0;
      int methodCount = 0;
      int kind = TypeDeclaration.kind(type.modifiers);
      boolean isAbstract = (kind == 2) || (kind == 4);
      if (!isAbstract) {
        extraConstructor = needConstructor ? 1 : 0;
        for (int i = 0; i < sourceMethodCount; i++)
          if (sourceMethods[i].isConstructor()) {
            if (needConstructor) {
              extraConstructor = 0;
              methodCount++;
            }
          } else if (needMethod)
            methodCount++;
      }
      else
      {
        methodCount = needMethod ? sourceMethodCount : 0;
      }
      type.methods = new AbstractMethodDeclaration[methodCount + extraConstructor];
      if (extraConstructor != 0) {
        type.methods[0] = type.createDefaultConstructor(false, false);
      }
      int index = 0;
      boolean hasAbstractMethods = false;
      for (int i = 0; i < sourceMethodCount; i++) {
        SourceMethod sourceMethod = sourceMethods[i];
        SourceMethodElementInfo methodInfo = (SourceMethodElementInfo)sourceMethod.getElementInfo();
        boolean isConstructor = methodInfo.isConstructor();
        if ((methodInfo.getModifiers() & 0x400) != 0) {
          hasAbstractMethods = true;
        }
        if (((isConstructor) && (needConstructor)) || ((!isConstructor) && (needMethod))) {
          AbstractMethodDeclaration method = convert(sourceMethod, methodInfo, compilationResult);
          if ((isAbstract) || (method.isAbstract())) {
            method.modifiers |= 16777216;
          }
          type.methods[(extraConstructor + index++)] = method;
        }
      }
      if (hasAbstractMethods) type.bits |= 2048;
    }

    return type;
  }

  private Annotation[] convertAnnotations(JavaElement element) {
    if (this.annotationPositions == null) return null;
    char[] cuSource = getSource();
    long[] positions = (long[])this.annotationPositions.get(element);
    if (positions == null) return null;
    int length = positions.length;
    Annotation[] annotations = new Annotation[length];
    int recordedAnnotations = 0;
    for (int i = 0; i < length; i++) {
      long position = positions[i];
      int start = (int)(position >>> 32);
      int end = (int)position;
      char[] annotationSource = CharOperation.subarray(cuSource, start, end + 1);
      if (annotationSource != null) {
        Expression expression = parseMemberValue(annotationSource);

        if ((expression instanceof Annotation)) {
          annotations[(recordedAnnotations++)] = ((Annotation)expression);
        }
      }
    }
    if (length != recordedAnnotations)
    {
      System.arraycopy(annotations, 0, annotations = new Annotation[recordedAnnotations], 0, recordedAnnotations);
    }
    return annotations;
  }

  private ImportReference createImportReference(String[] importName, int start, int end, boolean onDemand, int modifiers)
  {
    int length = importName.length;
    long[] positions = new long[length];
    long position = (start << 32) + end;
    char[][] qImportName = new char[length][];
    for (int i = 0; i < length; i++) {
      qImportName[i] = importName[i].toCharArray();
      positions[i] = position;
    }
    return new ImportReference(
      qImportName, 
      positions, 
      onDemand, 
      modifiers);
  }

  private TypeParameter createTypeParameter(char[] typeParameterName, char[][] typeParameterBounds, int start, int end)
  {
    TypeParameter parameter = new TypeParameter();
    parameter.name = typeParameterName;
    parameter.sourceStart = start;
    parameter.sourceEnd = end;
    if (typeParameterBounds != null) {
      int length = typeParameterBounds.length;
      if (length > 0) {
        parameter.type = createTypeReference(typeParameterBounds[0], start, end);
        if (length > 1) {
          parameter.bounds = new TypeReference[length - 1];
          for (int i = 1; i < length; i++) {
            TypeReference bound = createTypeReference(typeParameterBounds[i], start, end);
            bound.bits |= 16;
            parameter.bounds[(i - 1)] = bound;
          }
        }
      }
    }
    return parameter;
  }

  private TypeReference createTypeReference(char[] typeName, int start, int end)
  {
    int length = typeName.length;
    this.namePos = 0;
    return decodeType(typeName, length, start, end);
  }

  private TypeReference createTypeReference(String typeSignature, int start, int end)
  {
    int length = typeSignature.length();
    this.namePos = 0;
    return decodeType(typeSignature, length, start, end);
  }

  private TypeReference decodeType(String typeSignature, int length, int start, int end) {
    int identCount = 1;
    int dim = 0;
    int nameFragmentStart = this.namePos; int nameFragmentEnd = -1;
    boolean nameStarted = false;
    ArrayList fragments = null;
    while (this.namePos < length) {
      char currentChar = typeSignature.charAt(this.namePos);
      switch (currentChar) {
      case 'Z':
        if (!nameStarted) {
          this.namePos += 1;
          if (dim == 0) {
            return new SingleTypeReference(TypeBinding.BOOLEAN.simpleName, (start << 32) + end);
          }
          return new ArrayTypeReference(TypeBinding.BOOLEAN.simpleName, dim, (start << 32) + end);
        }
        break;
      case 'B':
        if (!nameStarted) {
          this.namePos += 1;
          if (dim == 0) {
            return new SingleTypeReference(TypeBinding.BYTE.simpleName, (start << 32) + end);
          }
          return new ArrayTypeReference(TypeBinding.BYTE.simpleName, dim, (start << 32) + end);
        }
        break;
      case 'C':
        if (!nameStarted) {
          this.namePos += 1;
          if (dim == 0) {
            return new SingleTypeReference(TypeBinding.CHAR.simpleName, (start << 32) + end);
          }
          return new ArrayTypeReference(TypeBinding.CHAR.simpleName, dim, (start << 32) + end);
        }
        break;
      case 'D':
        if (!nameStarted) {
          this.namePos += 1;
          if (dim == 0) {
            return new SingleTypeReference(TypeBinding.DOUBLE.simpleName, (start << 32) + end);
          }
          return new ArrayTypeReference(TypeBinding.DOUBLE.simpleName, dim, (start << 32) + end);
        }
        break;
      case 'F':
        if (!nameStarted) {
          this.namePos += 1;
          if (dim == 0) {
            return new SingleTypeReference(TypeBinding.FLOAT.simpleName, (start << 32) + end);
          }
          return new ArrayTypeReference(TypeBinding.FLOAT.simpleName, dim, (start << 32) + end);
        }
        break;
      case 'I':
        if (!nameStarted) {
          this.namePos += 1;
          if (dim == 0) {
            return new SingleTypeReference(TypeBinding.INT.simpleName, (start << 32) + end);
          }
          return new ArrayTypeReference(TypeBinding.INT.simpleName, dim, (start << 32) + end);
        }
        break;
      case 'J':
        if (!nameStarted) {
          this.namePos += 1;
          if (dim == 0) {
            return new SingleTypeReference(TypeBinding.LONG.simpleName, (start << 32) + end);
          }
          return new ArrayTypeReference(TypeBinding.LONG.simpleName, dim, (start << 32) + end);
        }
        break;
      case 'S':
        if (!nameStarted) {
          this.namePos += 1;
          if (dim == 0) {
            return new SingleTypeReference(TypeBinding.SHORT.simpleName, (start << 32) + end);
          }
          return new ArrayTypeReference(TypeBinding.SHORT.simpleName, dim, (start << 32) + end);
        }
        break;
      case 'V':
        if (!nameStarted) {
          this.namePos += 1;
          new SingleTypeReference(TypeBinding.VOID.simpleName, (start << 32) + end);
        }
        break;
      case 'L':
      case 'Q':
        if (!nameStarted) {
          nameFragmentStart = this.namePos + 1;
          nameStarted = true;
        }
        break;
      case '*':
        this.namePos += 1;
        Wildcard result = new Wildcard(0);
        result.sourceStart = start;
        result.sourceEnd = end;
        return result;
      case '+':
        this.namePos += 1;
        Wildcard result = new Wildcard(1);
        result.bound = decodeType(typeSignature, length, start, end);
        result.sourceStart = start;
        result.sourceEnd = end;
        return result;
      case '-':
        this.namePos += 1;
        Wildcard result = new Wildcard(2);
        result.bound = decodeType(typeSignature, length, start, end);
        result.sourceStart = start;
        result.sourceEnd = end;
        return result;
      case '[':
        dim++;
        break;
      case ';':
      case '>':
        nameFragmentEnd = this.namePos - 1;
        this.namePos += 1;
        break;
      case '$':
      case '.':
        if (!nameStarted) {
          nameFragmentStart = this.namePos + 1;
          nameStarted = true;
        } else if (this.namePos > nameFragmentStart) {
          identCount++;
        }break;
      case '<':
        nameFragmentEnd = this.namePos - 1;

        if (!this.has1_5Compliance)
          break label1108;
        if (fragments == null) fragments = new ArrayList(2);
        addIdentifiers(typeSignature, nameFragmentStart, nameFragmentEnd + 1, identCount, fragments);
        this.namePos += 1;
        TypeReference[] arguments = decodeTypeArguments(typeSignature, length, start, end);
        fragments.add(arguments);
        identCount = 1;
        nameStarted = false;
      }

      this.namePos += 1;
    }
    label1108: if (fragments == null)
    {
      if (identCount == 1) {
        if (dim == 0) {
          char[] nameFragment = new char[nameFragmentEnd - nameFragmentStart + 1];
          typeSignature.getChars(nameFragmentStart, nameFragmentEnd + 1, nameFragment, 0);
          return new SingleTypeReference(nameFragment, (start << 32) + end);
        }
        char[] nameFragment = new char[nameFragmentEnd - nameFragmentStart + 1];
        typeSignature.getChars(nameFragmentStart, nameFragmentEnd + 1, nameFragment, 0);
        return new ArrayTypeReference(nameFragment, dim, (start << 32) + end);
      }

      long[] positions = new long[identCount];
      long pos = (start << 32) + end;
      for (int i = 0; i < identCount; i++) {
        positions[i] = pos;
      }
      char[][] identifiers = extractIdentifiers(typeSignature, nameFragmentStart, nameFragmentEnd, identCount);
      if (dim == 0) {
        return new QualifiedTypeReference(identifiers, positions);
      }
      return new ArrayQualifiedTypeReference(identifiers, dim, positions);
    }

    if (nameStarted) {
      addIdentifiers(typeSignature, nameFragmentStart, nameFragmentEnd + 1, identCount, fragments);
    }
    int fragmentLength = fragments.size();
    if (fragmentLength == 2) {
      Object firstFragment = fragments.get(0);
      if ((firstFragment instanceof char[]))
      {
        return new ParameterizedSingleTypeReference((char[])firstFragment, (TypeReference[])fragments.get(1), dim, (start << 32) + end);
      }
    }

    identCount = 0;
    for (int i = 0; i < fragmentLength; i++) {
      Object element = fragments.get(i);
      if ((element instanceof char[][]))
        identCount += ((char[][])element).length;
      else if ((element instanceof char[]))
        identCount++;
    }
    char[][] tokens = new char[identCount][];
    TypeReference[][] arguments = new TypeReference[identCount][];
    int index = 0;
    for (int i = 0; i < fragmentLength; i++) {
      Object element = fragments.get(i);
      if ((element instanceof char[][])) {
        char[][] fragmentTokens = (char[][])element;
        int fragmentTokenLength = fragmentTokens.length;
        System.arraycopy(fragmentTokens, 0, tokens, index, fragmentTokenLength);
        index += fragmentTokenLength;
      } else if ((element instanceof char[])) {
        tokens[(index++)] = ((char[])element);
      } else {
        arguments[(index - 1)] = ((TypeReference[])element);
      }
    }
    long[] positions = new long[identCount];
    long pos = (start << 32) + end;
    for (int i = 0; i < identCount; i++) {
      positions[i] = pos;
    }
    return new ParameterizedQualifiedTypeReference(tokens, arguments, dim, positions);
  }

  private TypeReference decodeType(char[] typeName, int length, int start, int end)
  {
    int identCount = 1;
    int dim = 0;
    int nameFragmentStart = this.namePos; int nameFragmentEnd = -1;
    ArrayList fragments = null;
    while (this.namePos < length) {
      char currentChar = typeName[this.namePos];
      switch (currentChar) {
      case '?':
        this.namePos += 1;
        while (typeName[this.namePos] == ' ') this.namePos += 1;
        int max;
        int ahead;
        switch (typeName[this.namePos])
        {
        case 's':
          max = TypeConstants.WILDCARD_SUPER.length - 1;
          ahead = 1; break;
        case 'e':
          while (typeName[(this.namePos + ahead)] == TypeConstants.WILDCARD_SUPER[(ahead + 1)])
          {
            ahead++; if (ahead >= max)
            {
              this.namePos += max;
              Wildcard result = new Wildcard(2);
              result.bound = decodeType(typeName, length, start, end);
              result.sourceStart = start;
              result.sourceEnd = end;
              return result;

              int max = TypeConstants.WILDCARD_EXTENDS.length - 1;
              int ahead = 1;
              while (typeName[(this.namePos + ahead)] == TypeConstants.WILDCARD_EXTENDS[(ahead + 1)])
              {
                ahead++; if (ahead >= max)
                {
                  this.namePos += max;
                  Wildcard result = new Wildcard(1);
                  result.bound = decodeType(typeName, length, start, end);
                  result.sourceStart = start;
                  result.sourceEnd = end;
                  return result;
                }
              }
            }
          }
        }
        Wildcard result = new Wildcard(0);
        result.sourceStart = start;
        result.sourceEnd = end;
        return result;
      case '[':
        if (dim == 0) nameFragmentEnd = this.namePos - 1;
        dim++;
        break;
      case ']':
        break;
      case ',':
      case '>':
        break;
      case '.':
        if (nameFragmentStart < 0) nameFragmentStart = this.namePos + 1;
        identCount++;
        break;
      case '<':
        if (!this.has1_5Compliance)
          break label541;
        if (fragments == null) fragments = new ArrayList(2);
        nameFragmentEnd = this.namePos - 1;
        char[][] identifiers = CharOperation.splitOn('.', typeName, nameFragmentStart, this.namePos);
        fragments.add(identifiers);
        this.namePos += 1;
        TypeReference[] arguments = decodeTypeArguments(typeName, length, start, end);
        fragments.add(arguments);
        identCount = 0;
        nameFragmentStart = -1;
        nameFragmentEnd = -1;
      }

      this.namePos += 1;
    }
    label541: if (nameFragmentEnd < 0) nameFragmentEnd = this.namePos - 1;
    if (fragments == null)
    {
      if (identCount == 1) {
        if (dim == 0)
        {
          char[] nameFragment;
          if ((nameFragmentStart != 0) || (nameFragmentEnd >= 0)) {
            int nameFragmentLength = nameFragmentEnd - nameFragmentStart + 1;
            char[] nameFragment;
            System.arraycopy(typeName, nameFragmentStart, nameFragment = new char[nameFragmentLength], 0, nameFragmentLength);
          } else {
            nameFragment = typeName;
          }
          return new SingleTypeReference(nameFragment, (start << 32) + end);
        }
        int nameFragmentLength = nameFragmentEnd - nameFragmentStart + 1;
        char[] nameFragment = new char[nameFragmentLength];
        System.arraycopy(typeName, nameFragmentStart, nameFragment, 0, nameFragmentLength);
        return new ArrayTypeReference(nameFragment, dim, (start << 32) + end);
      }

      long[] positions = new long[identCount];
      long pos = (start << 32) + end;
      for (int i = 0; i < identCount; i++) {
        positions[i] = pos;
      }
      char[][] identifiers = CharOperation.splitOn('.', typeName, nameFragmentStart, nameFragmentEnd + 1);
      if (dim == 0) {
        return new QualifiedTypeReference(identifiers, positions);
      }
      return new ArrayQualifiedTypeReference(identifiers, dim, positions);
    }

    if ((nameFragmentStart > 0) && (nameFragmentStart < length)) {
      char[][] identifiers = CharOperation.splitOn('.', typeName, nameFragmentStart, nameFragmentEnd + 1);
      fragments.add(identifiers);
    }
    int fragmentLength = fragments.size();
    if (fragmentLength == 2) {
      char[][] firstFragment = (char[][])fragments.get(0);
      if (firstFragment.length == 1)
      {
        return new ParameterizedSingleTypeReference(firstFragment[0], (TypeReference[])fragments.get(1), dim, (start << 32) + end);
      }
    }

    identCount = 0;
    for (int i = 0; i < fragmentLength; i++) {
      Object element = fragments.get(i);
      if ((element instanceof char[][])) {
        identCount += ((char[][])element).length;
      }
    }
    char[][] tokens = new char[identCount][];
    TypeReference[][] arguments = new TypeReference[identCount][];
    int index = 0;
    for (int i = 0; i < fragmentLength; i++) {
      Object element = fragments.get(i);
      if ((element instanceof char[][])) {
        char[][] fragmentTokens = (char[][])element;
        int fragmentTokenLength = fragmentTokens.length;
        System.arraycopy(fragmentTokens, 0, tokens, index, fragmentTokenLength);
        index += fragmentTokenLength;
      } else {
        arguments[(index - 1)] = ((TypeReference[])element);
      }
    }
    long[] positions = new long[identCount];
    long pos = (start << 32) + end;
    for (int i = 0; i < identCount; i++) {
      positions[i] = pos;
    }
    return new ParameterizedQualifiedTypeReference(tokens, arguments, dim, positions);
  }

  private TypeReference[] decodeTypeArguments(char[] typeName, int length, int start, int end)
  {
    ArrayList argumentList = new ArrayList(1);
    int count = 0;
    while (this.namePos < length) {
      TypeReference argument = decodeType(typeName, length, start, end);
      count++;
      argumentList.add(argument);
      if ((this.namePos >= length) || 
        (typeName[this.namePos] == '>')) {
        break;
      }
      this.namePos += 1;
    }
    TypeReference[] typeArguments = new TypeReference[count];
    argumentList.toArray(typeArguments);
    return typeArguments;
  }

  private TypeReference[] decodeTypeArguments(String typeSignature, int length, int start, int end) {
    ArrayList argumentList = new ArrayList(1);
    int count = 0;
    while (this.namePos < length) {
      TypeReference argument = decodeType(typeSignature, length, start, end);
      count++;
      argumentList.add(argument);
      if ((this.namePos >= length) || 
        (typeSignature.charAt(this.namePos) == '>')) {
        break;
      }
    }
    TypeReference[] typeArguments = new TypeReference[count];
    argumentList.toArray(typeArguments);
    return typeArguments;
  }

  private char[][] extractIdentifiers(String typeSignature, int start, int endInclusive, int identCount) {
    char[][] result = new char[identCount][];
    int charIndex = start;
    int i = 0;
    while (charIndex < endInclusive)
      if (typeSignature.charAt(charIndex) == '.') {
        typeSignature.getChars(start, charIndex, result[(i++)] =  = new char[charIndex - start], 0);
        charIndex++; start = charIndex;
      } else {
        charIndex++;
      }
    typeSignature.getChars(start, charIndex + 1, result[(i++)] =  = new char[charIndex - start + 1], 0);
    return result;
  }

  private char[] getSource() {
    if (this.source == null)
      this.source = this.cu.getContents();
    return this.source;
  }

  private Expression parseMemberValue(char[] memberValue)
  {
    if (this.parser == null) {
      this.parser = new Parser(this.problemReporter, true);
    }
    return this.parser.parseMemberValue(memberValue, 0, memberValue.length, this.unit);
  }

  static class AnonymousMemberFound extends RuntimeException
  {
    private static final long serialVersionUID = 1L;
  }
}